\3D.XPL		JUL-28-99
\3D Modeling Program by Loren Blaney
\
\REVISIONS:
\FEB-06-96, Original
\JUL-28-99, Modified for XPLP.
\
\A right-hand coordinate system is used:
\
\			       |   Z  (into screen)
\			       |  /
\			       | /
\			       |/
\		       --------+-------- X
\			      /|
\			     / |
\			    /  |
\			       |
\			       Y
\

include	CODES;

int	VCurX, VCurY;	\Character cursor coordinates in pixels
int	VLineX, VLineY;	\Line coordinates

def	VWidth = 640,	\VESA display dimensions in pixels
	VHeight = 480,
	VPoints = 11;	\Character point size (height in pixels) (see font table)

def	IntSize= 4;	\Number of bytes in an integer (2 or 4)

ext	GetTime,	\Get 32-bits of system timer counts
	HLine,		\Draw horizontal line
	VShow(ScreenBuf, GranShift); \Display 640x480x8 image at ScreenBuf

char	ScreenBuf;	\Array holding copy of screen

int	Cpureg,		\Address of CPU register array (from Getreg)
	DataSeg,	\Our data segment address
	GranShift,	\Bits to shift to compensate for VESA granularity
	HaveMouse,	\A mouse driver is installed and mouse works
	Lambert,	\Counter: Lambert lighting vs. colored facets vs. hilights
	OrigVidMode,	\Original video mode (to be restored upon exit)
	Solid,		\Flag: do solid filled polygons as opposed to lines
	II;		\Scratch for Main

real	Ambient;	\Ambient light level

def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0A, FF=$0C,	\Control chars
	CR=$0D, EOF=$1A, Esc=$1B, Sp=$20, Ctrl=$40;

def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\Scan codes

def	Pi = 3.14159265358979323846;

\-------------------------------------------------------------------------------

func real Rad(X);		\Convert degrees to radians
real	X;
return X /57.2957795130823;



func	WExtend(N);		\Sign extend 16-bit word to 32-bit double word
int	N;
return if N & $8000 then N ! $FFFF0000 else N & $0000FFFF;



func	ILn2(N);	\Integer log to base 2
int	N;		\ (this truncates, e.g: ILn2(31)=4)
int	B, I;
begin
B:= $8000;
for I:= -15, 0 do
	begin
	if N & B then return -I;
	B:= B>>1;
	end;
return -1;		\error
end;	\ILn2



proc	WaitVB;		 	\Wait for beginning of next vertical blank
int	SR;
begin
SR:= Peek($40, $63) + $306;	\Port address of status register
if (SR & $39A) # $39A then return;
while Pin(SR, 0) & $08 do;	\Wait for no vertical blank
repeat until Pin(SR, 0) & $08;	\Wait for vertical blank
end;	\WaitVB;



proc	Beep;			\A not-too-obnoxious beep
begin
Sound(false, 1, 1000);
Sound(true, 1, 3000);
end;	\Beep



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(Unused arguments need not be passed)
begin
Cpureg:= Getreg;
Cpureg(0):= AX;
Cpureg(1):= BX;
Cpureg(2):= CX;
Cpureg(3):= DX;
Cpureg(6):= BP;
Cpureg(9):= DS;
Cpureg(11):= ES;
Softint(Int);
return WExtend(Cpureg(0));	\return AX register
end;	\CallInt



func	GetVid;			\Returns the current video mode
return CallInt($10, $0F00) & $FF; \Function $0F



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl-C, and no cursor.
SC:= CallInt($16, $0000);	\Function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\Return non-ascii chars as negative scan code
return Ch;
end;	\GetKey



proc	WaitKey;		\Wait for a keystroke
begin
Openi(1);
repeat until ChkKey;
end;	\WaitKey



proc	SetTimerMode(M);	\Set 8254 timer mode
int	M;
begin
\00110100
Pout($30+M<<1, $43, 0);
Pout($FF, $40, 0);
Pout($FF, $40, 0);
end;	\SetTimerMode

\--------------------------- MOUSE ROUTINES ---------------------------

func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\Pointer is set to center of screen but is hidden
CallInt($21, $3533);		\Make sure mouse vector ($33) points to something
HaveMouse:= false;
if WExtend(Cpureg(1))=0 & WExtend(Cpureg(11))=0 then return false;
HaveMouse:= CallInt($33, $0000); \Reset mouse and get status
if HaveMouse then HaveMouse:= true; \(Beware of 'not' operator in 32-bit XPL)
return HaveMouse;		\Return 'false' if failure
end;	\OpenMouse



proc	ShowMouse(On);		\Turn mouse pointer on or off
\The video mode should be set before calling this routine.
\This counts the number of times the pointer is turned on or off. If the
\ pointer is turned on twice, it must be turned off twice before it
\ actually goes off. The pointer should be turned off before drawing over
\ it and before your program exits. Setting the video mode will also turn
\ off the pointer.
int	On;	\Flag: True = pointer on; False = pointer off
if HaveMouse then
	CallInt($33, if On then $0001 else $0002);



func	GetMousePosition(N);	\Return position of specified mouse coordinate
int	N;	\0 = X coordinate; 1 = Y coordinate
\For video modes $0-$E and $13 the maximum coordinates are 639x199, minus
\ the size of the pointer. For modes $F-$12 the coordinates are the same as
\ the pixels. For 80-column text modes divide the mouse coordinates by 8 to
\ get the character cursor position.
begin
if ~HaveMouse then return 0;
CallInt($33, $0003);
return if N then WExtend(Cpureg(3)) else WExtend(Cpureg(2));
end;	\GetMousePosition



func	GetMouseButton(N);	\Return 'true' if specified mouse button is down
int	N;	\Button number: 0 = left; 1 = right (or middle)
begin
if ~HaveMouse then return false;
CallInt($33, $0003);
return if N then (Cpureg(1)&2)=2 else (Cpureg(1)&1)=1;
end;	\GetMouseButton



proc	MoveMouse(X, Y);	\Move mouse pointer to X,Y
int	X, Y;
if HaveMouse then
	CallInt($33, $0004, 0, X, Y);

\-------------------------------------------------------------------------------

proc	VPoint(X, Y, C);	\Plot a point in ScreenBuf
int	X, Y, C;		\Coordinates, color
begin
if X < 0 then return;		\clip
if X >= VWidth then return;
if Y < 0 then return;
if Y >= VHeight then return;

ScreenBuf(Y*VWidth+X):= C;
end;	\VPoint



proc	VLine(X1, Y1, C);	\Draw line from VLineX,VLineY to X1,Y1
\This plots lines from both ends. Lines with an odd number of pixels will
\ have their middle pixel plotted twice. This would be a problem if XORed
\ lines were used.
\WARNING: This does not clip lines and assumes that lines do not extend beyond
\ ScreenBuf.
int	X1, Y1, C;	\Coordinates of end-of-line, Color (0-255)
int	DX, DY,		\Delta X, Delta Y
	XI, YI,		\Plotting increments along X and Y axes
	S, SD,		\Step (along X or Y), Step Diagonal (along X and Y)
	D,		\Decision variable (move straight or diagonally?)
	E, NE,		\East, Northeast (straight or diagonal)
	I;
char	P, Q;		\Pointers to ends of lines, Q plots from opposite end
begin
DX:= X1 - VLineX;
if DX >= 0 then
	XI:= 1
else	[DX:= -DX;   XI:= -1];

DY:= Y1 - VLineY;
if DY >= 0 then
	YI:= VWidth
else	[DY:= -DY;   YI:= -VWidth];

SD:= XI + YI;
if DX >= DY then
	S:= XI
else	begin
	S:= YI;
	I:= DX;   DX:= DY;   DY:= I;
	end;

D:= 2*DY - DX;					\Set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

P:= ScreenBuf + VLineY*VWidth + VLineX;		\Set up pointers
Q:= ScreenBuf + Y1*VWidth + X1;

P(0):= C;			\Plot initial point
Q(0):= C;
for I:= 1, DX>>1 do
	begin
	if D < 0 then
		begin		\Move straight
		D:= D + E;
		P:= P + S;
		Q:= Q - S;
		end
	else	begin		\Move diagonally
		D:= D + NE;
		P:= P + SD;
		Q:= Q - SD;
		end;
	P(0):= C;		\Plot point
	Q(0):= C;
	end;

VLineX:= X1;   VLineY:= Y1;	\End of current line is start of next line
end;	\VLine



proc	VMove(X, Y);		\Move pen to the beginning of a line
int	X, Y;
begin
VLineX:= X;
VLineY:= Y;
end;	\VMove

\-------------------------------------------------------------------------------

proc	VPoly(XX, YY, A, C);	\Draw a (filled) polygon
int	XX, YY,	\Location to plot polygon (coordinates of first point)
	A,	\One-dimensional array of X,Y coordinates with first value
		\ giving the number of points (X,Y pairs)
	C;	\Fill color
int	N, BASE, P, MEMPTR, \MEM0\;
int	X0, Y0, X1, Y1;
int	I, J,
	Y,
	X00, Y00,	\Coordinates (absolute) of first vertex in polygon
	MIN, MAX,	\Maximum and minimum (absolute) Y coordinate of polygon
	HEIGHT;		\Height of polygon in pixels
def	NIL= $80000000;
def	SS=12;		\Scale shift factor (one pixel error when slope = 4096)


	proc	PLINE(X0, Y0, X1, Y1);
	\Build table containing coordinates of X points along line. This is a
	\ linked list header for each horizontal line through the polygon.
	int	X0, Y0, X1, Y1;
	int	X, Y, DX, DY, M, T, Q;
	begin
	if Y0 > Y1 then				\Swap ends
		begin
		T:= Y0;   Y0:= Y1;   Y1:= T;
		T:= X0;   X0:= X1;   X1:= T;
		end;

	DY:= Y1 - Y0;
	if DY = 0 then return;
	DX:= X1 - X0;
	M:= DX<<SS /DY;				\Scaled slope

	X:= X0<<SS;
	for Y:= Y0-MIN, Y1-MIN-1 do
		begin
		T:= X>>SS;			\Unscale X
		P:= BASE + Y*IntSize;
		loop	begin
			case of P(0)=NIL, P(0,1)>T:
				begin		\Insert item at front of list
				Q:= MEMPTR;
				MEMPTR:= MEMPTR + 8;	\2*IntSize;
				Q(0):= P(0);
				Q(1):= T;
				P(0):= Q;
				quit;
				end
			other	P:= P(0);	\Follow pointers
			end;
		X:= X +M;
		end;
	end;	\PLINE


begin	\VPoly
\Find maximum and minimum (absolute) Y coordinate
N:= A(0);				\Get number of coordinates
MIN:= $7FFFFFFF;   MAX:= $80000000;
I:= 2;					\Index of first Y coordinate in A
for J:= 1, N do
	begin
	Y:= YY +A(I);
	if Y > MAX then MAX:= Y;
	if Y < MIN then MIN:= Y;
	I:= I +2;			\Next Y coordinate
	end;
HEIGHT:= MAX - MIN + 1;

\Set up array of list headers for each horizontal scan line in the polygon.
\ The list will contain the X coordinates (lowest value first) of the edges.
BASE:= Reserve(HEIGHT*IntSize);
for Y:= 0, HEIGHT-1 do BASE(Y):= NIL;

MEMPTR:= RESERVE(10000);		\Set up our own data space (or remove nested proc)
\MEM0:= MEMPTR;

I:= 1;					\Index of first coordinate in A
X0:= XX +A(I);   X00:= X0;   I:= I +1;	\Save first vertex of polygon in X00,Y00
Y0:= YY +A(I);   Y00:= Y0;   I:= I +1;
for J:= 2, N do				\Get rest of vertices
	begin
	X1:= XX +A(I);   I:= I +1;
	Y1:= YY +A(I);   I:= I +1;
	Pline(X0, Y0, X1, Y1);		\Draw edge of polygon between vertices
	X0:= X1;   Y0:= Y1;
	end;
Pline(X0, Y0, X00, Y00);		\Make sure the polygon is closed

for Y:= 0, HEIGHT-1 do			\Draw horizontal lines to fill polygon
	begin
	P:= BASE + Y*IntSize;
	loop	begin
		P:= P(0);
		if P = NIL then quit;
		X0:= P(1);
		P:= P(0);
		if P = NIL then quit;
		if X0#P(1) then HLine(ScreenBuf, X0, MIN+Y, P(1)-1, C);
		end;
	end;
end;	\VPoly

\-------------------------------------------------------------------------------

proc	VCrlf;			\Move cursor to the start of a new line
begin
VCurX:= 0;
VCurY:= VCurY + VPoints;
if VCurY+VPoints >= VHeight then VCurY:= 0;
end;	\VCrlf



proc	VCursor(X, Y);		\Set cursor coordinates (in character cells)
int	X, Y;
begin
VCurX:= X * 8;
VCurY:= Y * VPoints;
end;	\VCursor



proc	VChout(Ch, C);		\Plot character at current cursor location
int	Ch, C;	\character, color
\Also inputs: ScreenBuf, VPoints, VWidth, VHeight
int	I, J,	\indices
	B,	\byte from font table
	M,	\bit mask
	FT,
	;
char	A;
begin	\VChout
if Ch = $0D then [VCurX:= 0;   return];
if Ch = $0A then
	[VCurY:= VCurY + VPoints;
	if VCurY+VPoints >= VHeight then VCurY:= 0;
	return];

include	FONTTBL;

A:= ScreenBuf + VCurX + VCurY*VWidth;
for J:= 0, VPoints-1 do
	begin
	B:= FT(Ch*VPoints+J);			\Get byte from font table
	M:= $80;
	for I:= 0, 8-1 do
		begin				\Background is transparent
		if M & B then A(I):= C;
		M:= M >> 1;
		end;
	A:= A + VWidth;
	end;

VCurX:= VCurX + 8;				\Advance cursor
if VCurX+8 >= VWidth then VCrlf;
end;	\VChout



proc	VText(Str, C);		\Plot a string of characters at current cursor
char	Str;
int	C;
int	I, Ch;
begin
for I:= 0, 30000 do
	begin
	Ch:= Str(I);
	if Ch >= $80 then [Ch:= Ch&$7f;   I:= 30000];
	VChout(Ch, C);
	end;
end;	\VText



proc	VIntout(N, C);		\Plot the ascii representation of an integer
int	N, C;
int	Ch;
begin
Openo(8);
Intout(8, N);
Chout(8, ^,);
Openi(8);
loop	begin
	Ch:= Chin(8);
	if Ch = ^, then quit;
	VChout(Ch, C);
	end;
end;	\VIntout



proc	VRlout(N, C);		\Plot the ascii representation of a real
real	N;
int	C;
int	Ch;
begin
Openo(8);
Rlout(8, N);
Chout(8, ^,);
Openi(8);
loop	begin
	Ch:= Chin(8);
	if Ch = ^, then quit;
	VChout(Ch, C);
	end;
end;	\VRlout

\-------------------------------------------------------------------------------

proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
def	X, Y, Z;
begin
V3(X):= V1(Y)*V2(Z) - V1(Z)*V2(Y);
V3(Y):= V1(Z)*V2(X) - V1(X)*V2(Z);
V3(Z):= V1(X)*V2(Y) - V1(Y)*V2(X);
end;	\CrossProd



func real DotProd(V1, V2);	\Return the dot product of two 3D vectors
real	V1, V2;			\V1  V2
def	X, Y, Z;
return V1(X)*V2(X) + V1(Y)*V2(Y) + V1(Z)*V2(Z);



proc	Unitize(V);		\Convert a 3D vector to its unit vector
real	V;	\Vector to convert
real	M;	\Magnitude of vector
def	X, Y, Z;
begin
M:= Sqrt( V(X)*V(X) + V(Y)*V(Y) + V(Z)*V(Z) );
V(X):= V(X) / M;
V(Y):= V(Y) / M;
V(Z):= V(Z) / M;
end;	\Unitize



proc    Rotate(V, R, P, W); \3D rotate vector V
real    V,              \3D vector is 3-element array with X, Y and Z components
	R, P, W;        \Roll, Pitch and yaW (radians)
real    X, Y, Z, SW, SP, SR, CW, CP, CR, T;
begin
X:= V(0);   Y:= V(1);   Z:= V(2);	\get vector components
SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
\                               
\                     1   0   0 
\ [X Y Z]:= [X Y Z] * 0   CR  SR
\                     0  -SR  CR
\                               
T:= Y*CR - Z*SR;
Z:= Z*CR + Y*SR;
Y:= T;

\Rotate about Y axis (Pitch):
\                               
\                     CP  0  -SP
\ [X Y Z]:= [X Y Z] * 0   1   0 
\                     SP  0   CP
\                               
T:= X*CP + Z*SP;
Z:= Z*CP - X*SP;
X:= T;

\Rotate about Z axis (yaW):
\                                
\                      CW  SW  0 
\ [X Y Z]:= [X Y Z] * -SW  CW  0 
\                      0   0   1 
\                                
T:= X*CW - Y*SW;
Y:= Y*CW + X*SW;
X:= T;

V(0):= X;   V(1):= Y;   V(2):= Z;       \return values
end;    \Rotate

\===============================================================================

proc	LineTest;		\Draw lots of lines at various angles
def	XMax=640, YMax=480;
int	X, Y;
begin
for X:= 0, XMax-1 do [VMove(X, 0);   VLine(XMax-1-X, YMax-1, X&$0FF);   X:= X +0];
for Y:= 0, YMax-1 do [VMove(0, Y);   VLine(XMax-1, YMax-1-Y, Y&$0FF);   Y:= Y +0];
end;	\LineTest

\===============================================================================

proc	MoveTorus;	\Display and move a torus
def	Max=20;		\Maximum values for Sectors and Sections
real	TorusR1,	\Distance from the center of the hole in the middle
			\ to the center of the eatable tube-shaped part.
	TorusR2,	\The distance from the center of the eatable tube-
			\ shaped part to the frosting on top.
	TorusX, TorusY, TorusZ,	\Arrays of coordinates of vertices (2-dim)
	TorusNX, TorusNY, TorusNZ, \Arrays of coordinates of normals (2-dim)
	TNX, TNY, TNZ;
real	Ref1, Ref2, Ref3,	\Arrays: 3 unit vectors forming reference frame
	Zoom,		\Zoom factor
	;
int	Sectors,	\Slice of torus (tube)
	Sections,	\Slice of a sector (patch)
	Facets,		\Number of facets
	ITorusX, ITorusY, ITorusZ,	\Arrays of coordinates of vertices (2-dim)
	ITorusZZ,	\Sorted Z coordinates (1-dimensional array)
	ITorusIJ,	\Array accompanying ITorusZZ containing I and J indices
	II;		\Scratch

int	Time, Time0;
real	FPS;

\-------------------------------------------------------------------------------

proc	DrawTorus;	\Draw the torus
int	I, J, K,	\Indices
	Color,
	PA;		\Array: Polygon corners for a facet
real	Light,		\Light vector. Unit vector specifies direction of light
	N,		\Normal vector
	DP, DP2, DP4;
def	CX=640/2, CY=480/2;	\Coordinates of center of the screen
def	X, Y, Z;
\Inputs: Sectors, Sections, ITorusX, ITorusY, Facets, ITorusIJ
begin
\DRAW POLYGONS:
if Solid then
begin
PA:= Reserve(IntSize + 4*2*IntSize);	\Space for a rectangular polygon
PA(0):= 4;				\4 corners

N:= Rlres(3);
Light:= [0.5, 1.0, 1.0];		\Light direction
Unitize(Light);

\Draw facets starting with the back-most one (the one with the greatest Z)
for K:= 0, Facets-1 do
	begin
	I:= ITorusIJ(K);			\extract indices
	J:= I>>8;
	I:= I & $00FF;

	PA(1):= ITorusX(I, J);		PA(2):= ITorusY(I, J);
	PA(3):= ITorusX(I+1, J);	PA(4):= ITorusY(I+1, J);
	PA(5):= ITorusX(I+1, J+1);	PA(6):= ITorusY(I+1, J+1);
	PA(7):= ITorusX(I, J+1);	PA(8):= ITorusY(I, J+1);

	\This determines if a facet is visible. The surface normal of a facet
	\ is the cross product of two vectors determined by adjacent edges. If
	\ the dot product of this normal and the viewing direction [0,0,1] is
	\ negative, the facet is visible. This eliminates about half the polygons.
	if (PA(3)-PA(1)) * (PA(8)-PA(2)) < (PA(7)-PA(1)) * (PA(4)-PA(2)) then
		begin
		if Lambert \#0\ then
			begin
			\VN
			\[0,0,1]N
			\if TNZ(I,J) < 0 then \\visible

			N(X):= TNX(I,J);
			N(Y):= TNY(I,J);
			N(Z):= TNZ(I,J);

			case Lambert of
			  1:	begin
			  	Color:= Fix((189.0-Ambient) * -DotProd(N, Light) +
					Ambient) + 64;		\189=3*64-1-2
				end;
			  2:	begin
				DP:= -DotProd(N, Light);
				DP2:= DP * DP;
				DP4:= DP2 * DP2;
				Color:= Fix((100.0-Ambient/2.0) * DP +
					(189.0 - 100.0) * DP4*DP4*DP4*DP4 +
					Ambient/2.0) + 64;
				end
			other	[];

			if Color < 64 then Color:= 64;	\fix some rounding glitches
			end
		else	Color:= Rem((I+J+J)/63)+1;

		VPoly(CX, CY, PA, Color);
		end;
	end;
end
else	begin
\DRAW LINES:
Color:= 3;
for J:= 0, Sections-1 do
	begin
	VMove(ITorusX(0,J)+CX, ITorusY(0,J)+CY);
	for I:= 1, Sectors do
		VLine(ITorusX(I,J)+CX, ITorusY(I,J)+CY, Color);
	end;

for I:= 0, Sectors-1 do
	begin
	VMove(ITorusX(I,0)+CX, ITorusY(I,0)+CY);
	for J:= 1, Sections do
		VLine(ITorusX(I,J)+CX, ITorusY(I,J)+CY, Color);
	end;
end;
end;	\DrawTorus

\-------------------------------------------------------------------------------

proc	Sort(A, IA, N);		\Sort arrays into descending Z coordinates
int	A,			\Address of array of Z coordinates
	IA,			\Address of array of I and J coordinates
	N;			\Number of elements in array (size)
int	J, Gap, I, JG, T;
begin
Gap:= N /2;
while Gap > 0 do
	begin
	for I:= Gap, N-1 do
		begin
		J:= I -Gap;
		loop	begin
			if J < 0 then quit;
			JG:= J +Gap;
			if A(J) >= A(JG) then quit;
			T:= A(J);    A(J):= A(JG);    A(JG):= T;   \Swap
			T:= IA(J);  IA(J):= IA(JG);  IA(JG):= T;
			J:= J -Gap;
			end;
		end;
		Gap:= Gap /2;
	end;
end;	\Sort



proc	DepthSort;
\Find the Z coordinate (*2) of the center of each facet. Save result in
\ array ITorusZZ. Array ITorusIJ saves I & J coordinates for corresponding
\ facet. Z increases going into the screen.
\Inputs: Sectors, Sections, ITorusZ
\Outputs: ITorusZZ, ITorusIJ, Facets
int	I, J;
begin
Facets:= 0;
for I:= 0, Sectors-1 do
    for J:= 0, Sections-1 do
	begin
	ITorusZZ(Facets):= ITorusZ(I,J) + ITorusZ(I+1,J+1);
	ITorusIJ(Facets):= I + J<<8;
	Facets:= Facets +1;
	end;
Sort(ITorusZZ, ITorusIJ, Facets);
end;	\DepthSort

\-------------------------------------------------------------------------------

proc	RotateTorus;
\Inputs: Sectors, Sections, TorusX, TorusY, TorusZ
\Outputs: ITorusX, ITorusY, ITorusZ
int	I, J,
	DX, DY, DZ;		\Translational motion
real	AngX, AngY, AngZ,	\Rotation angle about indicated axis (radians)
	Rxx, Rxy, Rxz,		\Rotation matrix
	Ryx, Ryy, Ryz,
	Rzx, Rzy, Rzz;
begin
if GetMouseButton(0) then
	begin
	CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Z axis
	AngZ:= Float(WExtend(Cpureg(2))) / 500.0;	\get angle (in radians) from mouse
	\Mouse Y motion zooms
	Zoom:= Zoom + Float(-WExtend(Cpureg(3))) / 2500.0;
	AngX:= 0.0;
	AngY:= 0.0;
	end
else	begin
	CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Y axis
	AngY:= Float(-WExtend(Cpureg(2))) / 500.0;
	\Mouse Y motion rotates object about X axis
	AngX:= Float(WExtend(Cpureg(3))) / 500.0;
	AngZ:= 0.0;
	end;

\STEP 1:
\Rotate reference frame according to changes in mouse position.
\Independently rotating P1 and P2 can cause them to drift relative to
\ each other, but the amount of drift is insignificant because reals are
\ calculated to 16-place accuracy.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2

\STEP 2:
\Rotate each vertex in the object so that it aligns with the reference frame:
\Translation is done at the same time for efficiency.
DX:= 0;   DY:= 0;   DZ:= 0;

\Replace subscripted variables for efficiency:
Rxx:= Ref1(0);   Rxy:= Ref1(1);   Rxz:= Ref1(2);
Ryx:= Ref2(0);   Ryy:= Ref2(1);   Ryz:= Ref2(2);
Rzx:= Ref3(0);   Rzy:= Ref3(1);   Rzz:= Ref3(2);

\This multiplies each point at its initial position (Po) by the rotation
\ matrix to get the rotated position.
\	                                                    
\	                                  Rxx  Rxy  Rxz 
\	 Px  Py  Pz  =  Pxo  Pyo  Pzo  *  Ryx  Ryy  Ryz 
\	                                  Rzx  Rzy  Rzz 
\	                                                    
\
for I:= 0, Sectors do
  for J:= 0, Sections do
	begin
	ITorusX(I,J):= Fix((TorusX(I,J) *Rxx + TorusY(I,J) *Ryx + TorusZ(I,J) *Rzx)*Zoom) + DX;
	ITorusY(I,J):= Fix((TorusX(I,J) *Rxy + TorusY(I,J) *Ryy + TorusZ(I,J) *Rzy)*Zoom) + DY;
	ITorusZ(I,J):= Fix( TorusX(I,J) *Rxz + TorusY(I,J) *Ryz + TorusZ(I,J) *Rzz) + DZ;

	if Lambert \#0\ then
	  if I<Sectors & J<Sections then
		begin
		TNX(I,J):= TorusNX(I,J) *Rxx + TorusNY(I,J) *Ryx + TorusNZ(I,J) *Rzx;
		TNY(I,J):= TorusNX(I,J) *Rxy + TorusNY(I,J) *Ryy + TorusNZ(I,J) *Rzy;
		TNZ(I,J):= TorusNX(I,J) *Rxz + TorusNY(I,J) *Ryz + TorusNZ(I,J) *Rzz;
		end;
	end;
end;	\RotateTorus

\-------------------------------------------------------------------------------

proc	CalcTorus;
\Find coordinates of vertices of facets that approximate a torus.
\Chop the surface of the torus up into polygons. To simplify later
\ calculations, the last coordinates duplicate the first coordinates.
\ Torus is initially in the X,Y plane (like a donut floating in hot grease).
\Inputs: Sectors, Sections, TorusR1, TorusR2
\Outputs: TorusX, TorusY, TorusZ, TorusNX, TorusNY, TorusNZ
int	I, J;
real	A,	\Angular increments about hole in center
	B,	\Angular increments about center of tube
	SA, CA, SB, CB;	\Sines and Cosines of A and B
real	V1, V2, V3;
def	Pi2 = Pi*2.0;
def	X, Y, Z;
begin
V1:= Rlres(3);
V2:= Rlres(3);
V3:= Rlres(3);

for I:= 0, Sectors do
	begin
	A:= Float(I) / Float(Sectors) * Pi2;
	SA:= Sin(A);   CA:= Cos(A);
	for J:= 0, Sections do
		begin
		B:= Float(J) / Float(Sections) * Pi2;
		SB:= Sin(B);   CB:= Cos(B);
		TorusX(I,J):= (TorusR1 + TorusR2*CB) * CA;
		TorusY(I,J):= (TorusR1 + TorusR2*CB) * SA;
		TorusZ(I,J):= TorusR2 * SB;
		end;
	end;

\Find the normal vectors for each facet. A normal vector is a vector that
\ is one unit long and is perpendicular to the facet. It points outward,
\ away from the inside of the donut (the inside of a facet is the side that
\ can never be seen no matter how the donut is rotated.
for I:= 0, Sectors-1 do
	for J:= 0, Sections-1 do
		begin
		V1(X):= TorusX(I+1,J) - TorusX(I,J);
		V2(X):= TorusX(I,J+1) - TorusX(I,J);
		V1(Y):= TorusY(I+1,J) - TorusY(I,J);
		V2(Y):= TorusY(I,J+1) - TorusY(I,J);
		V1(Z):= TorusZ(I+1,J) - TorusZ(I,J);
		V2(Z):= TorusZ(I,J+1) - TorusZ(I,J);
		CrossProd(V1, V2, V3);
		Unitize(V3);
		TorusNX(I,J):= V3(X);
		TorusNY(I,J):= V3(Y);
		TorusNZ(I,J):= V3(Z);
		end;
end;	\CalcTorus

\-------------------------------------------------------------------------------

begin	\MoveTorus
TorusX:= Rlres(Max);
for II:= 0, Max-1 do
	TorusX(II):= Rlres(Max);
TorusY:= Rlres(Max);
for II:= 0, Max-1 do
	TorusY(II):= Rlres(Max);
TorusZ:= Rlres(Max);
for II:= 0, Max-1 do
	TorusZ(II):= Rlres(Max);

TorusNX:= Rlres(Max);
for II:= 0, Max-1 do
	TorusNX(II):= Rlres(Max);
TorusNY:= Rlres(Max);
for II:= 0, Max-1 do
	TorusNY(II):= Rlres(Max);
TorusNZ:= Rlres(Max);
for II:= 0, Max-1 do
	TorusNZ(II):= Rlres(Max);

TNX:= Rlres(Max);
for II:= 0, Max-1 do
	TNX(II):= Rlres(Max);
TNY:= Rlres(Max);
for II:= 0, Max-1 do
	TNY(II):= Rlres(Max);
TNZ:= Rlres(Max);
for II:= 0, Max-1 do
	TNZ(II):= Rlres(Max);

ITorusX:= Reserve(Max*IntSize);
for II:= 0, Max-1 do
	ITorusX(II):= Reserve(Max*IntSize);
ITorusY:= Reserve(Max*IntSize);
for II:= 0, Max-1 do
	ITorusY(II):= Reserve(Max*IntSize);
ITorusZ:= Reserve(Max*IntSize);
for II:= 0, Max-1 do
	ITorusZ(II):= Reserve(Max*IntSize);

ITorusZZ:= Reserve(Max*Max*IntSize);
ITorusIJ:= Reserve(Max*Max*IntSize);

Ref1:= [1.0, 0.0, 0.0];		\unit vectors for reference frame
Ref2:= [0.0, 1.0, 0.0];		\Rotate procedure modifies these arrays
Ref3:= Rlres(3);		\calculated from Ref1 & Ref2 using cross product
Zoom:= 1.0;

TorusR1:= 100.0;
TorusR2:= 60.0;
Sectors:= 13;
Sections:= 13;

CalcTorus;

VCursor(0, 0);
VText("FPS:



 [ ] [ ]
 Sectors


 [ ] [ ]
Segments


  [S]
Solid/Wire


  [L]
Lighting


 [+] [-]
 Ambient


[Enter]
 Reset


 [Esc]
 Exit
", $7);

\Arrows
VCursor(2,4);   VChout($1B,$7);
VCursor(6,4);   VChout($1A,$7);
VCursor(2,8);   VChout($18,$7);
VCursor(6,8);   VChout($19,$7);


\BEWARE OF Time0 = Time   ******  
Format(2,1);
Time:= GetTime;
loop	begin
	Time0:= Time;
	Time:= GetTime;

	FPS:= 1.19318E6 / Float(Time-Time0);

	VPoly(5*8, 0, [4, 0,0, 100,0, 100,12, 0,12], 0);	\clear background for FPS
	VCursor(5, 0);
	VRlout(FPS, $E);

	RotateTorus;
	DepthSort;
	VPoly(320, 240, [4, -239,-239, 239,-239, 239,239, -239,239], 0);  \clear background
	DrawTorus;

	\WaitVB;	DEBUG
	VShow(ScreenBuf,GranShift);

	if Chkkey then
		case GetKey of
		  ^S,^s: Solid:= ~Solid;
		  ^L,^l: [Lambert:= Lambert +1;   if Lambert > 2 then Lambert:= 0];
		  ^-:	[Ambient:= Ambient - 3.0;
		  	if Ambient<0.0 then Ambient:= 0.0];
		  ^+:	[Ambient:= Ambient + 3.0;
		  	if Ambient > 189.0 then Ambient:= 189.0];
		  -LtArrow:
			if Sectors > 3 then [Sectors:= Sectors -1; CalcTorus];
		  -RtArrow:
			if Sectors < Max-1 then [Sectors:= Sectors +1; CalcTorus];
		  -UpArrow:
			if Sections < Max-1 then [Sections:= Sections +1; CalcTorus];
		  -DnArrow:
			if Sections > 3 then [Sections:= Sections -1; CalcTorus];
		  CR:	[Ref1(0):= 1.0;   Ref1(1):= 0.0;   Ref1(2):= 0.0;
			 Ref2(0):= 0.0;   Ref2(1):= 1.0;   Ref2(2):= 0.0;
			Zoom:= 1.0];
		  Esc:	quit
		other	Beep;
	end;
end;	\MoveTorus

\===============================================================================

proc	SetColors;
int	I, J, R, G, B;
char	Pal;
char	Palette;	\256 color registers. Each color is 3 bytes (RGB).
begin
Pal:= Malloc(256*3/16);	\256 color registers. Each color is 3 bytes (RGB).

Palette:= (Pal - DataSeg) << 4;	\Convert absolute address to dseg offset addr

I:= 0;
R:= 0;   G:= 0;   B:= 0;
for J:= 0, 63 do
	begin
	Palette(I):= R;
	Palette(I+1):= G;
	Palette(I+2):= B;
	I:= I +3;
	R:= R +1;

	Palette(I):= R;
	Palette(I+1):= G;
	Palette(I+2):= B;
	I:= I +3;
	G:= G + 1;

	Palette(I):= R;
	Palette(I+1):= G;
	Palette(I+2):= B;
	I:= I +3;
	B:= B + 1;
	end;

\WARNING: The last two color registers are invalid

\Set block of 3*64 color registers starting with register 64
CallInt($10, $1012, 64, 3*64, 0, 0, 0, Pal);
Release(Pal);
end;	\SetColors



proc	XSetColors;
int	I, J, R, G, B;
char	Pal;
begin
Pal:= Malloc(256*3/16);	\256 color registers. Each color is 3 bytes (RGB).

I:= 0;
R:= 0;   G:= 0;   B:= 0;
for J:= 0, 63 do
	begin
	Poke(Pal, I, R);
	Poke(Pal, I+1, G);
	Poke(Pal, I+2, B);
	I:= I +3;
	R:= R +1;

	Poke(Pal, I, R);
	Poke(Pal, I+1, G);
	Poke(Pal, I+2, B);
	I:= I +3;
	G:= G + 1;

	Poke(Pal, I, R);
	Poke(Pal, I+1, G);
	Poke(Pal, I+2, B);
	I:= I +3;
	B:= B + 1;
	end;

\WARNING: The last two color registers are invalid

\Set block of 3*64 color registers starting with register 64
CallInt($10, $1012, 64, 3*64, 0, 0, 0, Pal);
Release(Pal);
end;	\SetColors



func	Granularity(Mode);
\Granularity is the smallest increment (in KB) of SVGA video RAM that can
\ be mapped into segment $A000. Normally, it is 64 KB, but for some cards
\ it is less. To specify the second 64K page of video RAM for access the
\ "page" (used by VESA function $4F05) must be 1*64/Granularity. For
\ example, if granularity is 64, the second page is 1. If granularity is 4,
\ the second page is 1*64/4 = 16.
int	Mode;		\VESA mode (e.g: $101)
char	VESAInfo;	\SVGA mode information returned by function 1 (=256 bytes)
int	T;
begin
VESAInfo:= Malloc(256/16);

Cpureg(0\AX\):= $4F01;
Cpureg(2\CX\):= Mode;
Cpureg(4\DI\):= 0;		\offset
Cpureg(11\ES\):= VESAInfo;	\segment
Softint($10);

T:= Peek(VESAInfo, 4)  +  Peek(VESAInfo, 5) << 8;
Release(VESAInfo);
return T;
end;	\Granularity

\-------------------------------------------------------------------------------

proc	SetVesaMode(M);		\Set SVGA mode M
int	M;
begin
CallInt($10, $4F02, M);
if (Cpureg(0) & $FF00) # 0 then
	[Text(0,"Error setting VESA mode: "); Hexout(0, Cpureg(0)); exit];
end;	\SetVesaMode

\===============================================================================

begin	\Main
CpuReg:= Getreg;
DataSeg:= Cpureg(12) & $FFFF;
if ~OpenMouse then [Text(0, "Mouse not detected");   exit];

\Start Screen Buffer on a paragraph boundary
II:= Reserve(0);
II:= Reserve((16-II)&$f);
ScreenBuf:= Reserve(640*480);
for II:= 0, 640*480-1 do ScreenBuf(II):= 0;

\Trapc(true);			\\Ignore Ctrl+C if struck on keyboard	DEBUG


Solid:= true;
Lambert:= 2;
Ambient:= 90.0;
OrigVidMode:= GetVid;

SetVesaMode($101);		\Set VESA mode 101: 640x480x8

GranShift:= ILn2(64/Granularity($101));

VCursor(0,0);

VLineX:= 0;   VLineY:= 0;
SetColors;
SetTimerMode(2);

MoveTorus;

SetVid(OrigVidMode);		\Restore original video mode (clears screen)
end;	\Main
