\CLIPPER.XPL	12-7-83
\WINDOW CLIPPING ALGORITHM BY LOREN BLANEY.
\REF: FUNDAMENTALS OF INTERACTIVE COMPUTER GRAPHICS, J.D.FOLEY
\ & A. VAN DAM, P149.

code	\COMPLETE SET, STANDARD XPLV5D INTRINSICS
ABS=0,		RAN=1,		REM=2,		RESERVE=3,
SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
ABORT=16,	TRAP=17,	FREE=18,	RERUN=19,
GETHP=20,	SETHP=21,	GETERR=22,	CURSOR=23,
SCAN=24,	SETRUN=25,	CHAIN=28,	OPENF=29,
WRITE=30,	READ=31,	SETTXT=33,	SETHI=34,
SETMIX=35,	SETLO=36,	BUTTON=37,	PADDLE=38,
SOUND=39,	CLEAR=40,	POINT=41,	LINE=42,
MOVE=43,	SCREEN=44,	BLOCK=45,	FIX=50;
code real	RLRES=46,	RLIN=47,	RLOUT=48,
FLOAT=49,	RLABS=51,	FORMAT=52,	SQRT=53,
LN=54,		EXP=55,		SIN=56,		ATAN2=57,
MOD=58,		LOG=59,		COS=60,		TAN=61,
ASIN=62,	ACOS=63;


real	XMIN,YMIN,XMAX,YMAX;	\WINDOW DIMENSIONS
int	I;


proc	DRAW(X1,Y1,X2,Y2);
real	X1,Y1,X2,Y2;
def	ON=$80;
begin
MOVE(FIX(X1),FIX(Y1));
LINE(FIX(X2),FIX(Y2),ON);
end;	\DRAW



proc	CLIPPER(X1,Y1,X2,Y2);
\DRAW THE PORTION OF THE LINE SEGMENT THAT APPEARS IN THE WINDOW
\ (COHEN-SUTHERLAND CLIPPING ALGORITHM).
real	X1,Y1,X2,Y2;
int	CODE1, CODE2;	\4-BIT BOOLEAN ARRAY
real	TR;

	func	OUTCODE(X,Y);
	real	X,Y;
	int	CODE;
	begin
	CODE:=if Y>YMAX then 1
	else if Y<YMIN then 2
	else 0;
	if X>XMAX then CODE:=CODE!4
	else if X<XMIN then CODE:=CODE!8;
	return CODE;
	end;	\OUTCODE

begin	\CLIPPER
loop	begin
	CODE1:=OUTCODE(X1,Y1); CODE2:=OUTCODE(X2,Y2);
	if CODE1 & CODE2 then quit;		\TRIVIALLY REJECT
	if (CODE1!CODE2) = 0 then [DRAW(X1,Y1,X2,Y2); quit];	\TRIVIALLY ACCEPT

	\SUBDIVIDE THE LINE SINCE AT MOST ONE ENDPOINT IS INSIDE THE WINDOW

	\FIRST, IF POINT X1,Y1 IS INSIDE THE WINDOW, EXCHANGE ENDPOINTS
	\ (AND THEIR OUTCODES) TO GUARANTEE THAT POINT X1,Y1 IS OUTSIDE
	\ THE WINDOW.
	if CODE1=0 then
		begin		\EXCHANGE
		TR:=X1; X1:=X2; X2:=TR;
		TR:=Y1; Y1:=Y2; Y2:=TR;
		CODE1:=CODE2;
		end;
	\NOW PERFORM A SUBDIVISION, MOVE POINT X1,Y1 TO THE INTERSECTION
	\ POINT; USE THE FORMULAS:	Y = Y1 + SLOPE*(X - X1)
	\				X = X1 + (1/SLOPE)*(Y - Y1)
	case of
	    CODE1&1:		\DIVIDE LINE AT TOP OF WINDOW
		[X1:=X1 + (X2 - X1)*(YMAX - Y1)/(Y2 - Y1);
		Y1:=YMAX];
	    CODE1&2:		\DIVIDE LINE AT BOTTOM OF WINDOW
		[X1:=X1 + (X2 - X1)*(YMIN - Y1)/(Y2 - Y1);
		Y1:=YMIN];
	    CODE1&4:		\DIVIDE LINE AT RIGHT EDGE OF WINDOW
		[Y1:=Y1 + (Y2 - Y1)*(XMAX - X1)/(X2 - X1);
		X1:=XMAX];
	    CODE1&8:		\DIVIDE LINE AT LEFT EDGE OF WINDOW
		[Y1:=Y1 + (Y2 - Y1)*(XMIN - X1)/(X2 - X1);
		X1:=XMIN]
	else	[];
	end;
end;	\CLIPPER



begin	\MAIN
XMIN:=70.; XMAX:=210.;
YMIN:=48.; YMAX:=144.;
SETHI; CLEAR;
CLIPPER(71.,49.,211.,145.);
CLIPPER(1.,96.,2000.,96.);
CLIPPER(68.,200.,500.,-300.);
I:=CHIN(0);
end;	\MAIN
