\CROSSREF.XPLV4D				2/10/81
\ CROSS-REFERENCE GENERATOR FOR XPL PROGRAMS
\ BY LOREN BLANEY
\THIS PROGRAM DEMONSTRATES THE USE OF BINARY SEARCH TREES
\ AND LINKED LISTS.
\ REF:  WITRH, N., "ALGORITHMS + DATA STRUCTURES = PROGRAMS,"
\ (NEW JERSEY: PRENTICE-HALL, 1976), PP. 206-210.

code	RESERVE=3	CHIN=7		CHOUT=8
	CRLF=9		INTOUT=11	TEXT=12
	OPENI=13	OPENO=14	SPACE=18;

def	TV=1, PR=2, FILE=3;		\I-O STUFF
def	CR=$0D, FF=$0C, TAB=$09, EOF=$1A;

def	IDLEN=6		\NUMBER OF CHARS IN AN IDENTIFIER
	MAXREF=8	\MAXIMUM LINE NUMBERS PER LINE
	NULL=0;		\EMPTY ENTRY

def	\TREEENTRY\	\COMPONENTS OF A TREE ENTRY
	FIRST,		\POINTER TO FIRST ENTRY IN LINKED LIST
	LAST,		\POINTER TO LAST ENTRY IN LINKED LIST
	LEFT,		\POINTER TO LEFT BRANCH OF TREE
	RIGHT,		\POINTER TO RIGHT BRANCH OF TREE
	KEY;		\IDENTIFIER CHARACTERS (MUST BE LAST)
def	\LISTENTRY\	\COMPONENTS OF A LINKED-LIST ENTRY
	LNO,		\LINE NUMBER OF IDENTIFIER
	NEXT;		\POINTER TO NEXT ENTRY IN THE LIST

addr	IDENT;		\IDENTIFIER CHARACTER ARRAY
int	ROOT;		\START OF THE SEARCH TREE
int	LEN,		\INDEX INTO "IDENT" ARRAY
	CHAR,		\THE CURRENT CHARACTER READ BY "GETCH"
	LINENO;		\CURRENT LINE NO. OF THE LISTING
int	TOPMEM		\TOP OF USABLE MEMORY (HEAP) SPACE
	RAM;		\POINTER TO ALLOCATED HEAP SPACE


proc	GETCH;	\GET NEXT CHARACTER AND PRINT IT
\ ALSO FILTERS OUT META CHARACTERS
begin
CHAR:=CHIN(FILE); CHOUT(PR,CHAR);
if CHAR=^^ then
	[CHAR:=CHIN(FILE); CHOUT(PR,CHAR);
	CHAR:=CHIN(FILE); CHOUT(PR,CHAR)];
end;


proc	ALLOCATE; int AMOUNT;
\"RESERVES" MEMORY, BUT UNLIKE THE RESERVE INTRINSIC, THIS
\ PROCEDURE DOESN'T RETURN THE SPACE TO THE HEAP MEMORY POOL
\ WHEN THE CALLING PROCEDURE RETURNS.  IN THIS RESPECT,
\ ALLOCATE BEHAVES LIKE THE PASCAL INTRINSIC "NEW."
int	TEMP;
begin
TEMP:=RAM;
RAM:=RAM + AMOUNT;	\RESERVE BYTES
if RAM>TOPMEM then
	[CRLF(PR); TEXT(PR,"OUT OF MEMORY--FILE IS TOO BIG");
	CRLF(PR); exit];
return TEMP;
end;


proc	SEARCH(ADDRTREEENTRY);
\SEARCH THE TREE; IF THE IDENTIFIER IS NOT PRESENT THEN
\ INSERT IT; OTHERWISE, APPEND THE LINE NUMBER TO THE
\ IDENTIFIER'S LINKED LIST.
int	ADDRTREEENTRY;	\ADDRESS OF POINTER TO THE TREE ENTRY
int	TREEENTRY,	\POINTER TO THE TREE ENTRY
	LISTENTRY,	\POINTER TO THE LINKED-LIST ENTRY
	N;		\SCRATCH
addr	ID;	\IDENTIFIER CHARACTER STRING ENTRY IN TREE
begin
TREEENTRY:=ADDRTREEENTRY(0);
if TREEENTRY=NULL then	\KEY IS NOT IN TREE SO INSERT IT.
	[TREEENTRY:=ALLOCATE(IDLEN+8); LISTENTRY:=ALLOCATE(4);
	ID:=TREEENTRY + 2*KEY;	\POINT "ID" TO IDENTIFIER
	for N:=0,IDLEN-1 do ID(N):=IDENT(N);
	TREEENTRY(LEFT):=NULL; TREEENTRY(RIGHT):=NULL;
	TREEENTRY(FIRST):=LISTENTRY; TREEENTRY(LAST):=LISTENTRY;
	LISTENTRY(LNO):=LINENO; LISTENTRY(NEXT):=NULL;
	ADDRTREEENTRY(0):=TREEENTRY;	\LINK UP NEW ENTRY
	return];
ID:=TREEENTRY + 2*KEY;		\POINT "ID" TO IDENTIFIER
loop	[for N:=0,IDLEN-1 do if IDENT(N)#ID(N) then quit;
	LISTENTRY:=ALLOCATE(4);		\IDENTIFIER FOUND
	LISTENTRY(LNO):=LINENO;		\INSERT REFERENCE NO.
	LISTENTRY(NEXT):=NULL;
	TREEENTRY(LAST,NEXT):=LISTENTRY;\LINK NEW ENTRY IN LIST
	TREEENTRY(LAST):=LISTENTRY;  \KEEP TRACK OF LAST ENTRY
	return];
if IDENT(N)<ID(N) then SEARCH(TREEENTRY + 2*LEFT)
	\PASS THE ADDRESS OF THE POINTER FOR THE LEFT BRANCH
else \IDENT(N)>ID(N)\ SEARCH(TREEENTRY + 2*RIGHT);
end;	\SEARCH


proc	PRINTENTRY; int TREEENTRY;
\PRINT THE IDENTIFIER NAME FOLLOWED BY ALL OF ITS LINE NUMBER
\ REFERENCES.  I.E. PRINT OUT ONE TREE ENTRY.
int	REFCNT,LISTENTRY,N;
begin
TEXT(PR,TREEENTRY + 2*KEY);	\PRINT IDENTIFIER NAME
\PRINT THE LINE NO. REFERENCES BY FOLLOWING THE LIST LINKAGES
LISTENTRY:=TREEENTRY(FIRST); REFCNT:=0;
repeat	if REFCNT=MAXREF then [CRLF(PR); REFCNT:=0];
	REFCNT:=REFCNT+1;
	CHOUT(PR,TAB); INTOUT(PR,LISTENTRY(LNO));
	LISTENTRY:=LISTENTRY(NEXT);
until LISTENTRY=NULL;
CRLF(PR);
end;	\PRINTENTRY


proc	PRINTTREE; int TREEENTRY;
\PRINTS THE ENTIRE TREE IN (ALPHABETICAL) ORDER.  I.E. PRINT
\ THE CROSS-REFERENCE LISTING
begin
if TREEENTRY#NULL then
	[PRINTTREE(TREEENTRY(LEFT));
	PRINTENTRY(TREEENTRY);
	PRINTTREE(TREEENTRY(RIGHT))];
end;	\PRINTTREE


begin	\MAIN PROCEDURE
IDENT:=RESERVE(IDLEN);
TEXT(TV,"CROSS REFERENCE VER 1.1"); CRLF(TV);
RAM:=RESERVE(SPACE-200);  \RESERVE MEMORY FOR TREE & LISTS
TOPMEM:=RESERVE(0)-10;	\POINTER TO TOP OF USABLE MEMORY
ROOT:=NULL; LINENO:=0;
OPENO(PR); OPENI(FILE);	\INITIALIZE DEVICES
loop	begin
	LINENO:=LINENO+1; INTOUT(PR,LINENO); CHOUT(PR,TAB);
	GETCH; if CHAR=EOF then quit;
	repeat	begin		\SCAN NON-EMPTY LINE
		if CHAR>=^A & CHAR<=^Z then
			begin		\IDENTIFIER FOUND
			LEN:=0;
			repeat	if LEN<IDLEN then
					[IDENT(LEN):=CHAR;
					LEN:=LEN+1];
				GETCH;
			until (CHAR<^A ! CHAR>^Z) & (CHAR<^0 !
				CHAR>^9);
			\FILL OUT "IDENT" WITH SPACES
			for LEN:=LEN,IDLEN-1 do IDENT(LEN):=^ ;
			IDENT(IDLEN-1):=IDENT(IDLEN-1) ! $80;
			SEARCH(addr ROOT);
			end
		else	begin	   \SKIP STRINGS, COMMENTS,
			case CHAR of	\ RESERVED WORDS, ETC.
			^":	[repeat GETCH until CHAR=^";
				GETCH];
			^\:	[repeat GETCH
				until CHAR=^\ ! CHAR=CR;
				if CHAR=^\ then GETCH];
			^':	[repeat GETCH until CHAR=^';
				GETCH];
			^$:	repeat GETCH
				until (CHAR<^0 ! CHAR>^9) &
					(CHAR<^A ! CHAR>^F)
			else	GETCH;
			end;
		if CHAR=EOF then quit;
		end;
	until CHAR=CR;
	end;
CHOUT(PR,FF); PRINTTREE(ROOT);
end;	\OF ALL
