\EXE2ROM.XPL	MAY-07-94
\ by Loren Blaney
\
\This program converts an EXE file to an executable memory image. It is
\ similar to Microsoft's EXE2BIN utility, but this one handles files
\ larger than 64K. It is intended to make ROM images, but it also makes
\ .COM files. This outputs a .ROM file.
\
\Usage: EXE2ROM infile [outfile]
\
\The entire .EXE file must be read in because the relocation pointer
\ table is not always in increasing order. This, combined with the
\ requirement that the executable code can be more than 64K, means that
\ 32-bit XPL must be used.

inc	\CXPL\CODESI;

int	CpuReg,		\address of CPU registers
	PspSeg,		\address of PSP segment (for file I/O)
	DataSeg;	\address of data segment (heap & stack)

def	SP=$20, CR=$0D;

\Global variables for DOS I/O routines:
int	InHandle,	\input file handle (= -1 if no input file)
	OutHandle;	\output file handle (= -1 if no output file)
addr	InFileName,	\input file name (all file names are zero terminated)
	OutFileName,	\output file name
	TmpFileName,	\temporary output file name (.$$$)
	BakFileName;	\backup output file name (.BAK)
def	NameSize=80;	\maximum number of chars in a file name (incl path)



func	IntLen(N);		\Return number of digits in N, a decimal integer
int	N;			\Adds 1 for a minus sign if N is negative
int	I;
for I:= if N>=0 then 1 else 2, 20 do
	[N:= N /10;   if N = 0 then return I];



proc	Justout(Dev, N, Places, Fill);	\Output a right-justified integer
\Always outputs correct value of N regardless of Places
int	Dev,	\Output device number
	N,	\16-bit integer (negative numbers should be filled with spaces)
	Places,	\Size of field in characters (right-justifies)
	Fill;	\Character to fill background of field. Usually space (Sp) or ^0.
int	I;	\ If not a space then beware of negative numbers.
begin
for I:= 1, Places-IntLen(N) do Chout(Dev, Fill);
Intout(Dev, N);
end;	\Justout

\========================== DOS I/O ROUTINES ==========================

func	DosOpen(InExt, OutExt);	\Open MS-DOS files. Returns: 0 if no errors;
\ 1 if error opening input file; 2 if error opening output file.
addr	InExt, OutExt;	\input and output extensions
int	Inx;		\index into CmdLine
addr	CmdLine;	\command line tail


	proc	CopyName(From, To, Ext);
	\Copy file name and replace any existing extension with "Ext".
	\If "Ext" = "???" then use extension from "From".
	addr	From, To,	\strings are 0 terminated
		Ext;		\extension
	int	I, J, Ch;
	begin
	I:= 0;
	loop	begin				\copy file name
		Ch:= From(I);
		if Ch=^. ! Ch=0 then quit;
		To(I):= Ch;
		if I < NameSize-1 then I:= I +1;
		end;
	To(I):= ^.;				\copy extension
	if I < NameSize-1 then I:= I +1;
	for J:= 0, 2 do
		begin
		Ch:= Ext(J) & $7F;
		To(I):= if Ch=^? then From(I) else Ch;
		if I < NameSize-1 then I:= I +1;
		end;
	To(I):= 0;				\terminate string
	end;	\CopyName



	proc	GetName(Name);
	\Read file name from command line and return it in "Name".
	\The command line is terminated by a CR.
	addr	Name;
	int	I, Ch;


		proc	GetCh;	\Get character, stripping out switches
		begin
		loop	begin
			Ch:= CmdLine(Inx);
			Inx:= Inx +1;
			if Ch # ^/ then quit;
			Ch:= CmdLine(Inx);	\ignore any one-char switches
			Inx:= Inx +1;
			if Ch = CR then quit;	\handle wise guys
			end;
		end;	\GetCh


	begin	\GetName
	repeat GetCh until Ch # SP;		\skip any leading spaces
	I:= 0;
	loop	begin
		case Ch of SP, CR, ^,: Ch:= 0 other; \terminate Name with 0
		Name(I):= Ch;
		if Ch = 0 then quit;
		if I < NameSize-1 then I:= I +1;
		GetCh;
		end;
	end;	\GetName


begin	\DosOpen
CmdLine:= Reserve($7F);			\get command tail from PSP
Blit(PspSeg, $81, DataSeg, CmdLine, $7F);
CmdLine($7E):= CR;			\make sure it's terminated with a CR

InHandle:= -1;   OutHandle:= -1;	\set to illegal values

Inx:= 0;
if InExt(0) # SP then
	begin				\get the input file name
	GetName(InFileName);
	CopyName(InFileName, InFileName, InExt);

	Trap(false);
	InHandle:= Fopen(InFileName, 0);
	Trap(true);
	if Geterr # 0 then		\file or path does not exist,
		[InHandle:= -1;   return 1]; \ SHARE conflict, etc.
	Fset(InHandle, ^I);
	end;

if OutExt(0) # SP then
	begin				\get the output file name
	if Inx > 0 then Inx:= Inx -1;	\backup in case of CR
	GetName(OutFileName);
	if OutFileName(0) =0 then	\if no outfile then use the infile
		CopyName(InFileName, OutFileName, OutExt)
	else	CopyName(OutFileName, OutFileName, OutExt);

	CopyName(OutFileName, TmpFileName, "$$$");
	CopyName(OutFileName, BakFileName, "BAK");

	Trap(false);
	OutHandle:= Fopen(TmpFileName, 1);
	Trap(true);
	if Geterr # 0 then		\path does not exist, file already exists
		[OutHandle:= -1;   return 2]; \ root directory full, etc.
	Fset(OutHandle, ^O);
	end;

return 0;				\no errors
end;	\DosOpen

\----------------------------------------------------------------------

func	DosClose;	\Close MS-DOS files. Returns: 0 if no errors;
\ 1 if error closing input file; 2 if error closing output file.
begin
if InHandle # -1 then
	begin
	Trap(false);
	Fclose(InHandle);
	Trap(true);
	if Geterr # 0 then return 1;
	end;

if OutHandle # -1 then
	begin
	Trap(false);
	Fclose(OutHandle);

	CpuReg(\AX\0):= $4100;		\DEL OutFileName.BAK
	CpuReg(\DS\9):= DataSeg;
	CpuReg(\DX\3):= BakFileName;
	Softint($21);			\(.BAK file might not exist)

	CpuReg(\AX\0):= $5600;		\REN OutFileName.EXT OutFileName.BAK
	CpuReg(\DS\9):= DataSeg;
	CpuReg(\DX\3):= OutFileName;
	CpuReg(\ES\11):= DataSeg;
	CpuReg(\DI\4):= BakFileName;
	Softint($21);

	CpuReg(\AX\0):= $5600;		\REN OutFileName.$$$ OutFileName.EXT
	CpuReg(\DS\9):= DataSeg;
	CpuReg(\DX\3):= TmpFileName;
	CpuReg(\ES\11):= DataSeg;
	CpuReg(\DI\4):= OutFileName;
	Softint($21);

	Trap(true);
	if Geterr # 0 then return 2;	\disk write error when flushing buffers
	end;

return 0;				\no errors
end;	\DosClose

\======================================================================

proc	EXE2ROM;	\Convert EXE file to executable memory image
char	Header,		\Array: EXE file header
	RelocTable,	\Array: Relocation pointer table, 4-byte entries
			\ (offset, segment) relative offset from start of
			\ code to a word that must have the base segment
			\ address added to it.
	Code;		\Array: Executable code
int	LastPageSize,	\Number of bytes in final 512-byte page of EXE file
	FilePages,	\Number of 512-byte pages in EXE file
	RelocItems,	\Number of relocation items
	HeadSize,	\Total number of bytes in all header info
	RelocOffset,	\Offset from start of EXE file to RelocTable
	CodeSize,	\Number of bytes of executable code
	BaseSeg,	\Segment address where code is executed from
	I, J,		\Scratch
	Off,		\Offset		\
	Seg,		\Segment	 |- relative to load segment
	Loc;		\Location	/
begin
Openi(3);   Openo(3);

Header:= Reserve($1c);
for I:= 0, $1b do			\Read in fixed length portion of header
	Header(I):= Chin(3);

if Header(0)#^M ! Header(1)#^Z then	\Verify format 
	begin
	Text(0, "Invalid EXE file signature");
	exit;
	end;

LastPageSize:= Header(2) + Header(3)<<8;
FilePages:= Header(4) + Header(5)<<8;
RelocItems:= Header(6) + Header(7)<<8;
HeadSize:= (Header(8) + Header(9)<<8) * 16;
RelocOffset:= Header($18) + Header($19)<<8;
CodeSize:= FilePages*512 - 512 + LastPageSize - HeadSize;

Text(0, "File size: ");   Justout(0, HeadSize+CodeSize, 8, SP);
Text(0, "  (hex ");   Hexout(0, HeadSize+CodeSize);   Chout(0, ^));   Crlf(0);
Text(0, "Code size: ");   Justout(0, CodeSize, 8, SP);
Text(0, "  (hex ");   Hexout(0, CodeSize);   Chout(0, ^));   Crlf(0);

if RelocItems > 0 then
	begin
	RelocTable:= Reserve(RelocItems*4);
	for I:= $1c, RelocOffset-1 do J:= Chin(3);	\Skip to relocation table

	for I:= 0, RelocItems*4-1 do			\Read in relocation table
		RelocTable(I):= Chin(3);

	Crlf(0);
	Text(0, "Fix-ups needed - base segment (hex): ");
	BaseSeg:= Hexin(0);
	end;

Openi(3);				\Skip entire header to get to code
for I:= 0, HeadSize-1 do J:= Chin(3);

Code:= Reserve(CodeSize);
for I:= 0, CodeSize-1 do		\Read in code
	Code(I):= Chin(3);

I:= 0;					\Relocate segment references in code
for J:= 0, RelocItems-1 do
	begin
	Off:= RelocTable(I) + RelocTable(I+1)<<8;
	Seg:= RelocTable(I+2) + RelocTable(I+3)<<8;
	Loc:= Seg <<4 + Off;
	I:= I +4;
	Seg:= Code(Loc) + Code(Loc+1)<<8 + BaseSeg;
	Code(Loc):= Seg;
	Code(Loc+1):= Seg>>8;
	end;

for I:= 0, CodeSize-1 do		\Output fixed-up code
	Chout(3, Code(I));
Close(3);
end;	\EXE2ROM

\======================================================================

begin	\Main
CpuReg:= Getreg;
PspSeg:= CpuReg(11);
DataSeg:= CpuReg(12);

InFileName:= Reserve(NameSize);
OutFileName:= Reserve(NameSize);
TmpFileName:= Reserve(NameSize);
BakFileName:= Reserve(NameSize);

Text(0, "
EXE File Locator Version 1.0
Copyright (c) 1994 Loren Blaney. All rights reserved.


THIS IS AN UNREGISTERED COPY.


");

case DosOpen("EXE", "ROM") of
  1:	[Text(0, "Cannot open input file");  exit];
  2:	[Text(0, "Cannot open output file");  exit]
other;

EXE2ROM;

case DosClose of
  1:	Text(0, "Cannot close input file");
  2:	Text(0, "Cannot close output file")
other;
end;	\Main
