\FERRIS.XPL	DEC-06-92
\Hi-res, flicker-free animated Ferris wheel by Loren Blaney.
\This demonstrates page flipping in VGA mode $12 (640x480x16).

\Revision History:
\JUN-23-88, Original on Amiga (640x480)
\APR-89,    Ported to the Mac II
\DEC-06-92, Ported to the PC

include C:\CXPL\CODESI;	\Intrinsic routines

ext	VLINE, VMOVE, VCLEAR, VDRAW, VSHOW;	\External assembly routines

def	INTSIZE= 4;	\Number of bytes in an integer

int	CX, CY, CZ,
	X, Y, Z,	\Array: (S,P)
	XA0, YA0, ZA0,
	XA1, YA1, ZA1,
	S,		\Spoke
	P,		\Point
	RY, RX, RZ,
	RX2, RY2, RZ2,
	RXA0, RYA0, RZA0,
	RX2A0, RY2A0, RZ2A0,
	RXA1, RYA1, RZA1,
	RX2A1, RY2A1, RZ2A1,
	SPEEDX,
	SPEEDY,
	SPEEDZ,
	PAGE;

\----------------------------------------------------------------------

proc	SETPALREG(N, C); \Set palette register
int	N,	\Register to set
	C;	\Color value (0..$3F)
int	CPUREG;
begin
CPUREG:= GETREG;
CPUREG(0):= $1000;
CPUREG(1):= C<<8 ! N;
SOFTINT($10);		\Call BIOS
end;	\SETPALREG



proc	SETBORDER(C);	\Set border color
int	C;	\Color value (0..$3F)
int	CPUREG;
begin
CPUREG:= GETREG;
CPUREG(0):= $1001;
CPUREG(1):= C<<8;
SOFTINT($10);
end;	\SETBORDER



proc	WAITVB;		\Wait for the beginning of the next vertical blank
begin
while PIN($3DA, 0) & $08 do;		\Wait for no vertical blank
repeat until PIN($3DA, 0) & $08;	\Wait for vertical blank
end;	\WAITVB;

\----------------------------------------------------------------------

proc	DRAW;		\Draw the Ferris wheel
def	COLOR=$0005, COLOR2=$7F0F, COLOR3=$CC0A, COLOR4=$D80F;
begin
\Circumfrence:
VMOVE(X(0,0), Y(0,0));
for S:= 1, 11 do
	VLINE(X(S,0), Y(S,0), COLOR);
VLINE(X(0,0), Y(0,0), COLOR);

VMOVE(X(0,1), Y(0,1));
for S:= 1, 11 do
	VLINE(X(S,1), Y(S,1), COLOR);
VLINE(X(0,1), Y(0,1), COLOR);

\Radial:
for S:= 0, 11 do
	begin
	VMOVE(XA0, YA0);
	VLINE(X(S,0), Y(S,0), COLOR);
	VMOVE(XA1, YA1);
	VLINE(X(S,1), Y(S,1), COLOR);
	end;

\Axel:
VMOVE(XA0, YA0);
VLINE(XA1, YA1, COLOR4);

\Cross members:
for S:= 0, 11 do
	begin
	VMOVE(X(S,2), Y(S,2));
	VLINE(X(S,4), Y(S,4), COLOR);
	end;

\Braces:
for S:= 0, 11 do
	begin
	VMOVE(XA1, YA1);
	VLINE(X(S,4), Y(S,4), COLOR2);
	S:= S +1;
	VMOVE(XA0, YA0);
	VLINE(X(S,2), Y(S,2), COLOR2);
	end;

\More braces:
VMOVE(X(11,2), Y(11,2));
for S:= 0, 11 do
	begin
	VLINE(X(S,4), Y(S,4), COLOR3);
	S:= S +1;
	VLINE(X(S,2), Y(S,2), COLOR3);
	end;
end;	\DRAW

\----------------------------------------------------------------------

proc	ROTATE;		\Rotate the coordinates of the Ferris wheel
begin
for S:= 0, 11 do
    for P:= 0, 4 do
	begin
	\Rotate about X axis:
	Y(S,P):= Y(S,P) + (Z(S,P) + RZ(S,P)) /CX;	RZ(S,P):= REM(0);
	Z(S,P):= Z(S,P) - (Y(S,P) + RY(S,P)) /CX;	RY(S,P):= REM(0);

	\Rotate about Y axis:
	X(S,P):= X(S,P) + (Z(S,P) + RZ2(S,P)) /CY;	RZ2(S,P):= REM(0);
	Z(S,P):= Z(S,P) - (X(S,P) + RX(S,P)) /CY;	RX(S,P):= REM(0);

	\Rotate about Z axis:
	X(S,P):= X(S,P) + (Y(S,P) + RY2(S,P)) /CZ;	RY2(S,P):= REM(0);
	Y(S,P):= Y(S,P) - (X(S,P) + RX2(S,P)) /CZ;	RX2(S,P):= REM(0);
	end;

YA0:= YA0 + (ZA0 +RZA0) /CX;	RZA0:= REM(0);
ZA0:= ZA0 - (YA0 +RYA0) /CX;	RYA0:= REM(0);
XA0:= XA0 + (ZA0 +RZ2A0) /CY;	RZ2A0:= REM(0);
ZA0:= ZA0 - (XA0 +RXA0) /CY;	RXA0:= REM(0);

XA0:= XA0 + (YA0 +RY2A0) /CZ;	RY2A0:= REM(0);
YA0:= YA0 - (XA0 +RX2A0) /CZ;	RX2A0:= REM(0);

YA1:= YA1 + (ZA1 +RZA1) /CX;	RZA1:= REM(0);
ZA1:= ZA1 - (YA1 +RYA1) /CX;	RYA1:= REM(0);
XA1:= XA1 + (ZA1 +RZ2A1) /CY;	RZ2A1:= REM(0);
ZA1:= ZA1 - (XA1 +RXA1) /CY;	RXA1:= REM(0);

XA1:= XA1 + (YA1 +RY2A1) /CZ;	RY2A1:= REM(0);
YA1:= YA1 - (XA1 +RX2A1) /CZ;	RX2A1:= REM(0);
end;	\ROTATE

\----------------------------------------------------------------------

func	GETKEY;		\Get character from keyboard
int	CPUREG, CH;	\This is a low-level routine with no echo,
begin			\ no ctrl-C, and no cursor.
CPUREG:= GETREG;
CPUREG(0):= 0;		\Function $00
SOFTINT($16);		\Call BIOS routine
CH:= CPUREG(0) & $FF;
if CH = 0 then
	begin			\Handle special keys
	case CPUREG(0)>>8 of	\Arrows:
	  72:	CH:= $0B;	\ up
	  80:	CH:= $0A;	\ down
	  75:	CH:= $08;	\ left
	  77:	CH:= $15	\ right
	other	CH:= -(CPUREG(0)>>8);
	end;
return CH;
end;	\GETKEY



proc	DOKEY;		\Process keyboard command
int	CH, SPDX;
begin
CH:= GETKEY;

if CH= $1B then [SETVID($03);   exit];		\Esc

if CH= $13 then					\Ctrl-S
	[repeat until CHKKEY;   CH:= CHIN(1)];

if CH= $20 then					\Spacebar
	[SPEEDX:= 0;   SPEEDY:= 0;   SPEEDZ:= 0];

if CH= $0B then SPEEDX:= SPEEDX -1;		\Up arrow
if SPEEDX > 12 then SPEEDX:= 12;
if CH= $0A then SPEEDX:= SPEEDX +1;		\Down arrow
if SPEEDX < -12 then SPEEDX:= -12;
SPDX:= 8 << ABS(SPEEDX);
if SPEEDX < 0 then SPDX:= -SPDX;
CX:= $1000<<4 /SPDX;

if CH= $15 then SPEEDY:= SPEEDY +1;		\ -->
if SPEEDY > 12 then SPEEDY:= 12;
if CH= $08 then SPEEDY:= SPEEDY -1;		\ <--
if SPEEDY < -12 then SPEEDY:= -12;
SPDX:= 8 << ABS(SPEEDY);
if SPEEDY < 0 then SPDX:= -SPDX;
CY:= $1000<<4 /SPDX;

if CH=^> ! CH=^. then SPEEDZ:= SPEEDZ -1;	\ >
if SPEEDZ > 12 then SPEEDZ:= 12;
if CH=^< ! CH=^, then SPEEDZ:= SPEEDZ +1;	\ <
if SPEEDZ < -12 then SPEEDZ:= -12;
SPDX:= 8 << ABS(SPEEDZ);
if SPEEDZ < 0 then SPDX:= -SPDX;
CZ:= $1000<<4 /SPDX;
end;	\DOKEY

\----------------------------------------------------------------------

proc	SETWHEEL;	\Define the coordinates of the Ferris wheel
def	A= 1000, B= 700, C=2800, D= 1500, E= 1000;
\Scaled constants: M= SQRT(3)/2, N= 1/2, U= 1
def	M= 8660, N= 5000, U= 10000;
int	F, G, TX, TY;
begin
TX:= [0, N, M, U, M, N, 0, -N, -M, -U, -M, -N];
TY:= [U, M, N, 0, -N, -M, -U, -M, -N, 0, N, M];

F:= (B*D -A*B +A*C)/C;

\Set up 5 points on each spoke (S)
for S:= 0, 11 do
	begin
	X(S,0):= -C * TX(S);	Y(S,0):= -C * TY(S);	Z(S,0):= -A;
	X(S,1):= -C * TX(S);	Y(S,1):= -C * TY(S);	Z(S,1):= A;
	X(S,2):= (B-C) * TX(S);	Y(S,2):= (B-C) * TY(S);	Z(S,2):= F;
	X(S,3):= (B-C) * TX(S);	Y(S,3):= (B-C) * TY(S);	Z(S,3):= 0;
	X(S,4):= (B-C) * TX(S);	Y(S,4):= (B-C) * TY(S);	Z(S,4):= -F;
	end;

for S:= 0, 11 do			\Divide out the TX/TY scale factor
    for P:= 0, 4 do
	[X(S,P):= X(S,P) /U;	Y(S,P):= Y(S,P) /U];

\Set up the axle:
XA0:= 0;	YA0:= 0;	ZA0:= -D;
XA1:= 0;	YA1:= 0;	ZA1:= D;
end;	\SETWHEEL

\----------------------------------------------------------------------

proc	SETCOLORS;	\Set up the palette registers
int	TBL, J;
def	BLUE=$08, WHITE=$07, GREEN=$02, RED=$04;
begin
TBL:=	[BLUE,	WHITE,	GREEN,	RED,
	WHITE,	0,	0,	WHITE,
	GREEN,	0,	0,	0,
	RED,	0,	0,	0];
for J:= 0, $F do SETPALREG(J, TBL(J));
SETBORDER(BLUE);
end;	\SETCOLORS

\----------------------------------------------------------------------

begin	\MAIN
SETVID($12);
SETCOLORS;

CHOUT(0, $0C);
TEXT(0, "


                                  FERRIS



   G. W. Ferris, an American engineer, designed an amusement ride for the

   1893 World's Columbian Exposition in Chicago. This first Ferris wheel

   could carry over 2,000 riders 250 feet into the air.



   Use the arrow keys and the ^"< >^" keys to rotate about the X, Y and Z axes.

   Press the spacebar to stop, and the ^"Esc^" key to exit.

   Beware of centrifugal force.



                                               -- Loren Blaney





   Press any key to start...");

if CHIN(1) then;

X:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do X(S):= RESERVE(5 *INTSIZE);
Y:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do Y(S):= RESERVE(5 *INTSIZE);
Z:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do Z(S):= RESERVE(5 *INTSIZE);

RX:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do RX(S):= RESERVE(5 *INTSIZE);
RY:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do RY(S):= RESERVE(5 *INTSIZE);
RZ:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do RZ(S):= RESERVE(5 *INTSIZE);

RX2:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do RX2(S):= RESERVE(5 *INTSIZE);
RY2:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do RY2(S):= RESERVE(5 *INTSIZE);
RZ2:= RESERVE(12 *INTSIZE);
for S:= 0, 11 do RZ2(S):= RESERVE(5 *INTSIZE);

for S:= 0, 11 do
    for P:= 0, 4 do
	begin
	RY(S,P):= 0;
	RX(S,P):= 0;
	RZ(S,P):= 0;
	RX2(S,P):= 0;
	RY2(S,P):= 0;
	RZ2(S,P):= 0;
	end;

RXA0:= 0;	RYA0:= 0;	RZA0:= 0;
RX2A0:= 0;	RY2A0:= 0;	RZ2A0:= 0;

RXA1:= 0;	RYA1:= 0;	RZA1:= 0;
RX2A1:= 0;	RY2A1:= 0;	RZ2A1:= 0;

SETWHEEL;

CX:= $1000<<4;	CY:= $1000<<4;	CZ:= $1000<<4;
SPEEDX:= 0;	SPEEDY:= 0;	SPEEDZ:= 0;

PAGE:= 0;
loop	begin
	VCLEAR(PAGE);		\Erase page
	VDRAW(PAGE);		\Set page to draw on
	DRAW;			\Draw wheel
	ROTATE;			\Rotate wheel
	if CHKKEY then DOKEY;	\Handle keyboard commands
	WAITVB;			\Wait for vertical blank
	VSHOW(PAGE);		\Show page with wheel drawn in new position
	PAGE:= (PAGE +1) & $01;	\Flip to other page (0 or 1)
	end;			\Loop until Esc key is pressed
end;	\MAIN
