\FLT.XPL	DEC-27-92
\Bee Blaster Flight Simulator
\by Loren Blaney
\
\Revisions:
\JUL-07-88, Amiga version
\DEC-27-92, Ported to PC
\
\NOTES:
\This program requires a mouse.
\
\This program draws images as seen by the pilot. Thus the hive appears to
\ be moving. A right-hand coordinate system is used.
\
\			           Z
\			       |  /
\			       | /
\			       |/
\			-------+-------> X
\			      /|
\			     / |
\			    /  |
\			       V
\			       Y
\
\ X increases to the right, Y increases downward, and Z increases into the
\ screen.
\
\			-- SIDE VIEW --
\
\		|			|
\		|			|
\		|			|
\		|			|
\		|			|
\		|			|
\	pilot  <O-----------------------+-----------------------> Z
\	(0,0)	|			|
\		|<------ ZSCREEN ------>|
\		|       (8 inches)	|
\		|			|
\		|			|
\		V		     monitor
\		Y		     screen
\
\
\Distances are scaled to 10 counts per foot.

include C:\CXPL\CODESI;	\Intrinsic routines

ext	VLINE, VMOVE, VCLEAR, VDRAW, VSHOW;	\External assembly routines

def	INTSIZE= 4;			\Number of bytes in an integer
def	BEL= $07, FF=$0C;		\ASCII control characters

def	INERTIAX= 8, INERTIAZ= 8;	\Joystick response (10 = no dampening)

\To simplify the perspective calculation, the distance to the screen
\ as expressed in pixels is chosen to be a power of two (ZN).
\	DIST = 2**ZN /768 = Distance from viewer to screen in feet
\	ZSCREEN = DIST *10 *$4000 = DIST *163840 = Scaled distance to screen
\	PPX = PX /PZ *1.33 *768		Perspective X in pixels
\					(There are 768 pixels/foot)
\	DIST   ZSCREEN   ZN
\	0.33    54613     8	(causes noticable distortion)
\	0.66   109227     9
\	1.33   218453    10
\	2.66   436907    11
\
def	\ZSCREEN= 109227,\ ZN= 9;
int	ZSCREEN;

int	CPUREG,
	REM0, REM1, REM2, REM3, REM4, REM5,	\Arrays
	II,
	MOUSEY,		\Mouse position (mouse counts)
	PTRX, PTRY,	\Mouse pointer position (affected by GAIN)
	GAINX, GAINY,	\Mouse pointer gain, scaled by 100
	THROTTLE,	\Throttle position (0 to 199)
	PX0, PY0, PZ0,	\Arrays of points (unrotated, initial hive positions)
	PX, PY, PZ,	\Arrays of points (rotated, translated)
	PPX, PPY,	\Arrays of points (rotated, translated, with perspective)
	REFPX, REFPY, REFPZ, \Arrays: 3 hive reference points, used to make
			\ unit vectors, scaled by $4000
	LBASE, PBASE,	\Line and point base index for each junction in PX etc.
	ANIMRATE,	\For animation rate display (debug)
	SPINX, SPINY, SPINZ,	\Rate of rotation about pilot's X, Y, and Z axes
	PITCHRATE,	\Pitch rate (commanded elevator position)
	FORWARD,	\Boolean: Commanded to travel in the forward direction
	SPEED,		\Actual speed: delta Z per iteration
	OLDSPEED,	\Previous iteration speed: delta Z per iteration
	BANKX,		\BANKX = COS(bank angle) *$4000,
	BANKY,		\BANKY = SIN(bank angle) *$4000,
	;
addr	PU,		\Array: point is connected to a line (i.e. is used)
	LMOVE, LBEGIN, LEND, LCOLOR,
	;		\Line arrays: MOVE required, 1st point, 2nd point, color

\----------------------------------------------------------------------

proc	SETPALREG(N, C); \Set palette register
int	N,	\Register to set
	C;	\Color value (0..$3F)
begin
CPUREG(0):= $1000;
CPUREG(1):= C<<8 ! N;
SOFTINT($10);		\Call BIOS
end;	\SETPALREG



proc	SETBORDER(C);	\Set border color
int	C;	\Color value (0..$3F)
begin
CPUREG(0):= $1001;
CPUREG(1):= C<<8;
SOFTINT($10);
end;	\SETBORDER



proc	SETCOLORS;	\Set up the palette registers
int	TBL, J;
def	BLACK=$00, DBLUE=$08, BLUE=$01, WHITE=$07, GREEN=$02, RED=$04, LRED=$24;
begin
\TBL:=	[DBLUE,	WHITE,	GREEN,	RED,
\	WHITE,	0,	0,	WHITE,
\	GREEN,	0,	0,	0,
\	RED,	0,	0,	0];
TBL:=	[BLACK,	BLUE,	LRED,	BLUE,
	BLUE,	0,	0,	WHITE,
	LRED,	0,	0,	0,
	BLUE,	0,	0,	0];
for J:= 0, $F do SETPALREG(J, TBL(J));
\SETBORDER(DBLUE);
end;	\SETCOLORS



proc	SHOCUR(FLAG);
int	FLAG;		\True = cursor on; False = cursor off
begin			\Requires: CPUREG:= GETREG;
CPUREG(0):= $0100;
CPUREG(2):= if FLAG then $0607 else $2000;
SOFTINT($10);		\Call BIOS routine
end;	\SHOCUR



func	GETVID;		\Returns the current video mode
begin			\Requires: CPUREG:= GETREG;
CPUREG(0):= $0F00;	\AX function $0F
SOFTINT($10);		\Call BIOS routine
return CPUREG(0) & $FF;
end;	\GETVID



proc	WAITVB;		\Wait for the beginning of the next vertical blank
begin
while PIN($3DA, 0) & $08 do;		\Wait for no vertical blank
repeat until PIN($3DA, 0) & $08;	\Wait for vertical blank
end;	\WAITVB;



func	GETKEY;		\Get character from keyboard
int	CH;		\This is a low-level routine with no echo,
begin			\ no ctrl-C, and no cursor.
CPUREG(0):= 0;		\Function $00
SOFTINT($16);		\Call BIOS routine
CH:= CPUREG(0) & $FF;
if CH = 0 then
	begin			\Handle special keys
	case CPUREG(0)>>8 of	\Arrows:
	  72:	CH:= $0B;	\ up
	  80:	CH:= $0A;	\ down
	  75:	CH:= $08;	\ left
	  77:	CH:= $15	\ right
	other	CH:= -(CPUREG(0)>>8);
	end;
return CH;
end;	\GETKEY

\----------------------------------------------------------------------

func	BUTTON(N);	\Returns 'true' if specified button is pressed
int	N;		\0 = left; 1 = right
int	M;
begin			\Requires: CPUREG:= GETREG
CPUREG(0):= $0003;
SOFTINT($33);		\Call BIOS routine
M:= CPUREG(1);
return if N then (M&2)=2 else (M&1)=1;
end;	\BUTTON



func	MOUSE(N);
int	N;		\0 = X coordinate; 1 = Y coordinate
begin			\Requires: CPUREG:= GETREG
CPUREG(0):= $0003;
SOFTINT($33);		\Call BIOS routine
return if N then CPUREG(3) else CPUREG(2);
end;	\MOUSE



func	OPENMOUSE;	\Reset mouse and get status
int	FLAG;
begin			\Requires: CPUREG:= GETREG;
CPUREG(0):= $0000;
SOFTINT($33);		\Call BIOS routine
FLAG:= CPUREG(0);	\False if no mouse

if FLAG then
	begin		\Show mouse pointer
	CPUREG(0):= $0001;
	SOFTINT($33);	\Call BIOS routine
	end;
return FLAG;		\Return false if no mouse
end;	\OPENMOUSE



proc	DOTHROTTLE;	\Get the throttle position, return it in THROTTLE
begin
\MOUSEY:= MOUSE(1);
\THROTTLE:= -MOUSEY *GAINY /100;
THROTTLE:= 20;

\Clip pointer position to text screen dimensions:
if THROTTLE > 199 then [THROTTLE:= 199;   MOUSEY:= -19900 /GAINY]
else if THROTTLE < 0 then [THROTTLE:= 0;   MOUSEY:= 0];

SPEED:= (THROTTLE *THROTTLE) >> 5;
if not FORWARD then SPEED:= -SPEED;

\SPEED:= SPEED + THRUST;	realistic equation
SPEED:= (OLDSPEED<<4 + SPEED) /17;
OLDSPEED:= SPEED;
end;	\DOTHROTTLE



proc	OPENTHROTTLE(Y0, G);	\Initialize throttle position
int	Y0,		\Initial throttle position
	G;		\Gain factors for mouse movement
begin
\OPENMOUSE;
GAINY:= G;
MOUSEY:= -Y0 *100 /GAINY;
DOTHROTTLE;
end;	\OPENTHROTTLE

\----------------------------------------------------------------------

func	RATIO(B, AA, BB);
\Return the value of A in the ratio A/B = AA/BB.
\WARNING: This will not work if B > BB.
int	B, AA, BB;
int	A1, B1, NEG;
begin
NEG:= false;
if B < 0 then [B:= -B;   NEG:= not NEG];
if AA < 0 then [AA:= -AA;   NEG:= not NEG];
if BB < 0 then [BB:= -BB;   NEG:= not NEG];
A1:=0;   B1:= 0;
loop	begin
	AA:= AA >> 1;
	if AA = 0 then quit;
	BB:= BB >> 1;
	if BB = 0 then quit;
	if B1+BB <= B then [B1:= B1 +BB;   A1:= A1 +AA];
	end;

return if NEG then -A1 else A1;
end;	\RATIO

\----------------------------------------------------------------------

proc	DRAW;		\Draw the objects on the graphics screen
\Inputs: BANKX, BANKY, PPX, PPY, PX, PY, PZ, LMOVE, LBEGIN, LEND, LBASE.
\Outputs: ANIMRATE.
int	ZEND, ZBEGIN,
	BX, BY,
	PX_, PY_,
	A, B,		\Tempories for screen intersection calculation
	I, J,
	L,
	Z,
	;
begin
VMOVE(ANIMRATE-320, 199);		\Display the rate indicator
VLINE(ANIMRATE-320, 199, $05);
ANIMRATE:= ANIMRATE +1;
if ANIMRATE >= 640 then ANIMRATE:= 0;

\Display artificial horizon and aiming point:
BX:= BANKX/$040;   BY:= BANKY/$040;
VMOVE(-BX, -BY);   VLINE(BX, BY, $7F05);	\(dotted line)
BX:= BANKX/$100;   BY:= BANKY/$100;
VMOVE(BY, -BX);   VLINE(0, 0, $7F05);

\Display throttle:

VMOVE(-319, 199);   VLINE(-319, 199-THROTTLE, $C005);	\(dotted line)

for L:= 0, LBASE-1 do			\For all lines...
	begin
	if LMOVE(L) then		\Must move to start of line
		begin
		I:= LBEGIN(L);		\Get index to point
		ZBEGIN:= PZ(I);
		if ZBEGIN >= ZSCREEN then VMOVE(PPX(I), PPY(I));
		end;
	J:= LEND(L);			\Get index to end point
	ZEND:= PZ(J);

	if ZBEGIN >= ZSCREEN then
		begin
		if ZEND >= ZSCREEN then		\Draw normal line
			VLINE(PPX(J), PPY(J), LCOLOR(L))
		else	begin			\ZEND < ZSCREEN
			\Calculate point of intersection with screen (PX_,PY_)
			I:= LBEGIN(L);		\Get index to beginning point
			A:= (ZSCREEN - PZ(J));	\Shift to avoid overflow
			B:= (PZ(I) - PZ(J));	\ A & B are always positive
			PX_:= PX(J) + RATIO( A, (PX(I) - PX(J)), B );
			PY_:= PY(J) + RATIO( A, (PY(I) - PY(J)), B );

			Z:= ZSCREEN >> ZN;
			PX_:= PX_ /Z;
			PY_:= PY_ /Z;

			VLINE(PX_, PY_, LCOLOR(L));
			end;
		end
	else	begin			\ZBEGIN < ZSCREEN
		if ZEND >= ZSCREEN then
			begin
			\Calculate point of intersection with screen (PX_,PY_)
			I:= LBEGIN(L);		\Get index to beginning point
			A:= (ZSCREEN - PZ(I));	\Shift to avoid overflow
			B:= (PZ(J) - PZ(I));	\ A & B are always positive
			PX_:= PX(I) + RATIO( A, (PX(J) - PX(I)), B );
			PY_:= PY(I) + RATIO( A, (PY(J) - PY(I)), B );

			Z:= ZSCREEN >> ZN;
			PX_:= PX_ /Z;
			PY_:= PY_ /Z;

			VMOVE(PX_, PY_);		\Move to intersection point
			VLINE(PPX(J), PPY(J), LCOLOR(L));	\Normal end point
			end;
		end;
	ZBEGIN:= ZEND;
	end;
end;	\DRAW

\----------------------------------------------------------------------

proc	PERSPECTIVE;
\Use similar triangles to make close points farther apart.
\Inputs: PX, PY, PZ, PBASE.
\Outputs: PPX, PPY.
int	I, Z;
begin
for I:= 0, PBASE-1 do			\For all points...
    if PU(I) then
	begin
	Z:= PZ(I);
	if Z >= ZSCREEN then	\(Points behind screen are handled by DRAW)
		begin
		Z:= Z >> ZN;
		PPX(I):= PX(I) /Z;	\(Scale factors divide out)
		PPY(I):= PY(I) /Z;
		end;
	end;
end;	\PERSPECTIVE

\----------------------------------------------------------------------

proc	ROTATE;		\Rotate and translate the points
int	I, J,
	DX, DY, DZ,	\Deltas
	DX1, DY1, DZ1,	\Deltas
	DX2, DY2, DZ2,	\Deltas
	DX3, DY3, DZ3,	\Deltas
	RX, RY, RZ,	\Temporaries for rotate
	UXX, UXY, UXZ,	\X unit vector
	UYX, UYY, UYZ,	\Y unit vector
	UZX, UZY, UZZ, 	\Z unit vector
	D2,
	;
int	K, L;



proc	ROT(PX, PY, PZ, PX0, PY0, PZ0, UXX, UYX, UZX, UXY, UYY, UZY, UXZ, UYZ, UZZ, N);
\Rotate object about 3 axes
int	PX, PY, PZ,		\Point position after being rotated (output)
	PX0, PY0, PZ0,		\Point at initial position
	UXX, UYX, UZX, UXY, UYY, UZY, UXZ, UYZ, UZZ,	\Rotation matrix
	DX, DY, DZ,		\Translation
	N;			\Number of points to rotate

\This multiplies each point at its initial position (Po) by the rotation
\ matrix to get the rotated position:
\	                                                    
\	                                  Uxx  Uxy  Uxz 
\	 Px  Py  Pz  =  Pxo  Pyo  Pzo  *  Uyx  Uyy  Uyz 
\	                                  Uzx  Uzy  Uzz 
\	                                                    
\
int	K;
begin
for K:= 0, N-1 do
	PX(K):= PX0(K) *UXX + PY0(K) *UYX + PZ0(K) *UZX + DX;
for K:= 0, N-1 do
	PY(K):= PX0(K) *UXY + PY0(K) *UYY + PZ0(K) *UZY + DY;
for K:= 0, N-1 do
	PZ(K):= PX0(K) *UXZ + PY0(K) *UYZ + PZ0(K) *UZZ + DZ;
end;	\ROT



begin	\ROTATE
\Rotate reference frame about the origin:
for J:= 0, 2 do
	begin
	RX:= REFPX(J);
	RY:= REFPY(J);
	RZ:= REFPZ(J);

	\Rotate about X axis:
	RY:= RY + (RZ *SPINX + REM0(J)) /$1000<<4;   REM0(J):= REM;
	RZ:= RZ - (RY *SPINX + REM1(J)) /$1000<<4;   REM1(J):= REM;
	\Rotate about Y axis:
	RX:= RX + (RZ *SPINY + REM2(J)) /$1000<<4;   REM2(J):= REM;
	RZ:= RZ - (RX *SPINY + REM3(J)) /$1000<<4;   REM3(J):= REM;
	\Rotate about Z axis:
	RY:= RY + (RX *SPINZ + REM4(J)) /$1000<<4;   REM4(J):= REM;
	RX:= RX - (RY *SPINZ + REM5(J)) /$1000<<4;   REM5(J):= REM;

	REFPX(J):= RX;
	REFPY(J):= RY;
	REFPZ(J):= RZ - SPEED;		\Also translate
	end;				\Forward speed moves objects closer

\Because these 3 points are independantly rotated, errors can accumulate and
\ distort the reference frame. To prevent this, mathematical cross braces are
\ used to maintain the correct distances between the points.

DX:= REFPX(1) - REFPX(0);   DY:= REFPY(1) - REFPY(0);   DZ:= REFPZ(1) - REFPZ(0);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $0FFF<<16+$0004 then			\$0FFF0004 = $3FFE **2
	begin				\Lengthen line
	REFPX(1):= REFPX(1) + DX /$1000;
	REFPY(1):= REFPY(1) + DY /$1000;
	REFPZ(1):= REFPZ(1) + DZ /$1000;
	end;
if D2 > $1001<<16+$0004 then			\$10010004 = 4002 **2
	begin				\Shorten line
	REFPX(1):= REFPX(1) - DX /$1000;
	REFPY(1):= REFPY(1) - DY /$1000;
	REFPZ(1):= REFPZ(1) - DZ /$1000;
	end;

DX:= REFPX(2) - REFPX(0);   DY:= REFPY(2) - REFPY(0);   DZ:= REFPZ(2) - REFPZ(0);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $0FFF<<16+$0004 then			\$0FFF0004 = $3FFE **2
	begin				\Lengthen line
	REFPX(2):= REFPX(2) + DX /$1000;
	REFPY(2):= REFPY(2) + DY /$1000;
	REFPZ(2):= REFPZ(2) + DZ /$1000;
	end;
if D2 > $1001<<16+$0004 then			\$10010004 = 4002 **2
	begin				\Shorten line
	REFPX(2):= REFPX(2) - DX /$1000;
	REFPY(2):= REFPY(2) - DY /$1000;
	REFPZ(2):= REFPZ(2) - DZ /$1000;
	end;

DX:= REFPX(2) - REFPX(1);   DY:= REFPY(2) - REFPY(1);   DZ:= REFPZ(2) - REFPZ(1);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $1FFD<<16+$8B01 then			\$1FFD8B01 = $5A7F **2
	begin				\Lengthen line
	REFPX(2):= REFPX(2) + DX /$1000;
	REFPY(2):= REFPY(2) + DY /$1000;
	REFPZ(2):= REFPZ(2) + DZ /$1000;
	end;
if D2 > $2001<<16+$C919 then			\$2001C919 = $5A85 **2
	begin				\Shorten line
	REFPX(2):= REFPX(2) - DX /$1000;
	REFPY(2):= REFPY(2) - DY /$1000;
	REFPZ(2):= REFPZ(2) - DZ /$1000;
	end;

UXX:= REFPX(1) - REFPX(0);		\UX = REFP(1) - REFP(0)
UXY:= REFPY(1) - REFPY(0);
UXZ:= REFPZ(1) - REFPZ(0);

UYX:= REFPX(2) - REFPX(0);		\UY = REFP(2) - REFP(0)
UYY:= REFPY(2) - REFPY(0);
UYZ:= REFPZ(2) - REFPZ(0);

\Find UZ using cross product:
UZX:= (UXY *UYZ - UXZ *UYY) /$4000;	\UZ = UX X UY
UZY:= (UXZ *UYX - UXX *UYZ) /$4000;
UZZ:= (UXX *UYY - UXY *UYX) /$4000;

\Rotate and translate all points to align with reference frame (unit vectors):
DX:= REFPX(0) << 14;	DY:= REFPY(0) << 14;	DZ:= REFPZ(0) << 14;

ROT(	PX, PY, PZ,		\Outputs
	PX0, PY0, PZ0,		\Inputs
	UXX, UYX, UZX,
	UXY, UYY, UZY,
	UXZ, UYZ, UZZ,
	DX, DY, DZ,
	42);

\Handle remaining junctions efficiently:
for I:= 1, 4 do
	begin			\Rotate reference point for each junction
	K:= I *42;
	PX(K):= PX0(K) *UXX + PY0(K) *UYX + PZ0(K) *UZX + DX;
	PY(K):= PX0(K) *UXY + PY0(K) *UYY + PZ0(K) *UZY + DY;
	PZ(K):= PX0(K) *UXZ + PY0(K) *UYZ + PZ0(K) *UZZ + DZ;

	RX:= PX(K) - PX(0);	RY:= PY(K) - PY(0);	RZ:= PZ(K) - PZ(0);
	J:= K;
	for L:= 1, 41 do	\Use offsets to determine remaining points
		begin
		K:= J + L;
		if PU(K) then
			begin
			PX(K):= PX(L) + RX;
			PY(K):= PY(L) + RY;
			PZ(K):= PZ(L) + RZ;
			end;
		end;
	end;

BANKY:= -UYX;
BANKX:= UYY;
\*** DEBUG ***
\BANKY:= BANKY + BANKX *SPINZ /$1000<<4;
\BANKX:= BANKX - BANKY *SPINZ /$1000<<4;
end;	\ROTATE

\----------------------------------------------------------------------

proc	MAKEJCT(X, Y, Z, COLOR);	\Set up junction geometry
int	X, Y, Z, COLOR;
int	I, J, TBL;
\Other inputs: PBASE, LBASE.
\Outputs: LINES, PX0, PY0, PZ0, LBASE, PBASE.
def	T= true, F= false;
def	DR= 500, D2R= 1000, DR2= 250,		\(L= 2000)
	DH= 433, D2H= 866, DH3= 144, D2H3= 288, D4H3= 576;
begin
\Set up initial (unrotated) point positions:
\	 X	 Y	 Z
TBL:= [	-DR,	 0,	-D2H,	\0
	 DR,	 0,	-D2H,	\1
	 D2R,	 0,	 0,	\2
	 DR,	 0,	 D2H,	\3
	-DR,	 0,	 D2H,	\4\	      \ 4                / 3
	-D2R,	 0,	 0,	\5\	       \                /
						\              /
	 0,	 0,	-D4H3,	\6\		 \_22_________/ 21
	 DR,	 0,	-D2H3,	\7		 /            \
	 DR,	 0,	 D2H3,	\8		/       | Z    \
	 0,	 0,	 D4H3,	\9	       /        |       \
	-DR,	 0,	 D2H3,	\10 5 ________/ 23    __|____ X \\_20______ 2
	-DR,	 0,	-D2H3,	\11\	      \         |        /
					       \                /
	-DR2,	-DH,	-DH,	\12\		\              /
	 DR2,	-DH,	-DH,	\13\		 \_18_________/ 19
	 DR,	-DH,	 0,	\14	         /            \
	 DR2,	-DH,	 DH,	\15             /              \
	-DR2,	-DH,	 DH,	\16            /                \
	-DR,	-DH,	 0,	\17         0 /                1 \

	-DR2,	 0,	-DH,	\18
	 DR2,	 0,	-DH,	\19
	 DR,	 0,	 0,	\20
	 DR2,	 0,	 DH,	\21
	-DR2,	 0,	 DH,	\22
	-DR,	 0,	 0,	\23

	-DR2,	 DH,	-DH,	\24
	 DR2,	 DH,	-DH,	\25
	 DR,	 DH,	 0,	\26
	 DR2,	 DH,	 DH,	\27
	-DR2,	 DH,	 DH,	\28
	-DR,	 DH,	 0,	\29

	 0,	-DH,	-D2H3,	\30
	 DR2,	-DH,	-DH3,	\31
	 DR2,	-DH,	 DH3,	\32
	 0,	-DH,	 D2H3,	\33
	-DR2,	-DH,	 DH3,	\34
	-DR2,	-DH,	-DH3,	\35

	 0,	 DH,	-D2H3,	\36
	 DR2,	 DH,	-DH3,	\37
	 DR2,	 DH,	 DH3,	\38
	 0,	 DH,	 D2H3,	\39
	-DR2,	 DH,	 DH3,	\40
	-DR2,	 DH,	-DH3 ];	\41

J:= 0;
for I:= PBASE, PBASE+41 do
	begin
	PX0(I):= TBL(J) + X;		J:= J +1;
	PY0(I):= TBL(J) + Y;		J:= J +1;
	PZ0(I):= TBL(J) + Z;		J:= J +1;
	end;


\Set up array for connecting points (LINES):
\ (MOVE required, 1st point, 2nd point)
TBL:= [	T,	12,	13,
	F,	13,	14,
	F,	14,	15,
	F,	15,	16,
	F,	16,	17,
	F,	17,	12,

	T,	24,	25,
	F,	25,	26,
	F,	26,	27,
	F,	27,	28,
	F,	28,	29,
	F,	29,	24,

	T,	12,	0,
	F,	0,	24,

	T,	13,	1,
	F,	1,	25,

	T,	14,	2,
	F,	2,	26,

	T,	15,	3,
	F,	3,	27,

	T,	16,	4,
	F,	4,	28,

	T,	17,	5,
	F,	5,	29 ];

J:= 0;
for I:= LBASE, LBASE+23 do
	begin
	LMOVE(I):= TBL(J);		J:= J +1;
	LBEGIN(I):= TBL(J) + PBASE;	J:= J +1;
	LEND(I):= TBL(J) + PBASE;	J:= J +1;
	LCOLOR(I):= COLOR;
	end;


PBASE:= PBASE + 42;
LBASE:= LBASE + 24;
end;	\MAKEJCT

\----------------------------------------------------------------------

proc	MAKEWORLD;	\Set up all the objects
int	X, Y, Z, I, J, K, TBL;
begin
\Initial position of reference frame:
X:= 0;		Y:= 0;		Z:= 8000;

REFPX(0):= X;		REFPY(0):= Y;		REFPZ(0):= Z;
REFPX(1):= $4000 +X;	REFPY(1):= Y;		REFPZ(1):= Z;
REFPX(2):= X;		REFPY(2):= $4000 +Y;	REFPZ(2):= Z;

LBASE:= 0;	PBASE:= 0;
for J:= 0, 42*5-1 do PU(J):= false;

MAKEJCT(0, 0, 0, $05);
MAKEJCT(0, 0, 4000, $05);	\(L = 2000)
MAKEJCT(0, 0, 8000, $05);
MAKEJCT(3464, 0, 6000, $05);
MAKEJCT(3464, 0, 2000, $01);

for J:= 0, 55 do
	begin
	LMOVE(J):= true;
	LCOLOR(J):= $05;
	end;

LBEGIN(0):= 35;			PU(35):= true;
LEND(0):= 16 + 42 *1;		PU(16+42):= true;
LBEGIN(1):= 32;			PU(32):= true;
LEND(1):= 31 + 42 *1;		PU(31+42):= true;
LBEGIN(2):= 8;			PU(8):= true;
LEND(2):= 7 + 42 *1;		PU(7+42):= true;
LBEGIN(3):= 38;			PU(38):= true;
LEND(3):= 37 + 42 *1;		PU(37+42):= true;
LBEGIN(4):= 41;			PU(41):= true;
LEND(4):= 28 + 42 *1;		PU(28+42):= true;
LBEGIN(5):= 11;			PU(11):= true;
LEND(5):= 4 + 42 *1;		PU(4+42):= true;

LBEGIN(6):= 16 + 42 *1;		PU(16+42):= true;
LEND(6):= 16 + 42 *2;		PU(16+84):= true;
LBEGIN(7):= 15 + 42 *1;		PU(15+42):= true;
LEND(7):= 13 + 42 *2;		PU(13+84):= true;
LBEGIN(8):= 3 + 42 *1;		PU(3+42):= true;
LEND(8):= 1 + 42 *2;		PU(1+84):= true;
LBEGIN(9):= 27 + 42 *1;		PU(27+42):= true;
LEND(9):= 25 + 42 *2;		PU(25+84):= true;
LBEGIN(10):= 28 + 42 *1;	PU(28+42):= true;
LEND(10):= 28 + 42 *2;	PU(28+84):= true;
LBEGIN(11):= 4 + 42 *1;		PU(4+42):= true;
LEND(11):= 4 + 42 *2;		PU(4+84):= true;

LBEGIN(12):= 15 + 42 *1;	PU(15+42):= true;
LEND(12):= 17 + 42 *3;	PU(17+126):= true;
LBEGIN(13):= 31 + 42 *1;	PU(31+42):= true;
LEND(13):= 12 + 42 *3;	PU(12+126):= true;
LBEGIN(14):= 7 + 42 *1;		PU(7+42):= true;
LEND(14):= 0 + 42 *3;		PU(0+126):= true;
LBEGIN(15):= 37 + 42 *1;	PU(37+42):= true;
LEND(15):= 24 + 42 *3;	PU(24+126):= true;
LBEGIN(16):= 27 + 42 *1;	PU(27+42):= true;
LEND(16):= 29 + 42 *3;	PU(29+126):= true;
LBEGIN(17):= 3 + 42 *1;		PU(3+42):= true;
LEND(17):= 5 + 42 *3;		PU(5+126):= true;

LBEGIN(18):= 16 + 42 *2;	PU(16+84):= true;
LEND(18):= 32 + 42 *3;	PU(32+126):= true;
LBEGIN(19):= 13 + 42 *2;	PU(13+84):= true;
LEND(19):= 17 + 42 *3;	PU(17+126):= true;
LBEGIN(20):= 1 + 42 *2;		PU(1+84):= true;
LEND(20):= 5 + 42 *3;		PU(5+126):= true;
LBEGIN(21):= 25 + 42 *2;	PU(25+84):= true;
LEND(21):= 29 + 42 *3;	PU(29+126):= true;
LBEGIN(22):= 28 + 42 *2;	PU(28+84):= true;
LEND(22):= 38 + 42 *3;	PU(38+126):= true;
LBEGIN(23):= 4 + 42 *2;		PU(4+84):= true;
LEND(23):= 8 + 42 *3;		PU(8+126):= true;

LBEGIN(24):= 12 + 42 *3;	PU(12+126):= true;
LEND(24):= 35 + 42 *4;	PU(35+168):= true;
LBEGIN(25):= 32 + 42 *3;	PU(32+126):= true;
LEND(25):= 32 + 42 *4;	PU(32+168):= true;
LBEGIN(26):= 8 + 42 *3;		PU(8+126):= true;
LEND(26):= 8 + 42 *4;		PU(8+168):= true;
LBEGIN(27):= 38 + 42 *3;	PU(38+126):= true;
LEND(27):= 38 + 42 *4;	PU(38+168):= true;
LBEGIN(28):= 24 + 42 *3;	PU(24+126):= true;
LEND(28):= 41 + 42 *4;	PU(41+168):= true;
LBEGIN(29):= 0 + 42 *3;		PU(0+126):= true;
LEND(29):= 11 + 42 *4;	PU(11+168):= true;

for J:= 31, 55 do
	LMOVE(J):= false;

LBEGIN(30):= 35;		PU(35):= true;
LEND(30):= 32;		PU(32):= true;
LEND(31):= 8;			PU(8):= true;
LEND(32):= 38;		PU(38):= true;
LEND(33):= 41;		PU(41):= true;
LEND(34):= 11;		PU(11):= true;
LEND(35):= 35;		PU(35):= true;

LMOVE(36):= true;
LBEGIN(36):= 31 + 42 *1;	PU(31+42):= true;
LEND(36):= 7 + 42 *1;		PU(7+42):= true;
LEND(37):= 37 + 42 *1;	PU(37+42):= true;

LMOVE(38):= true;
LBEGIN(38):= 15 + 42 *1;	PU(15+42):= true;
LEND(38):= 3 + 42 *1;		PU(3+42):= true;
LEND(39):= 27 + 42 *1;	PU(27+42):= true;

LMOVE(40):= true;
LBEGIN(40):= 12 + 42 *3;	PU(12+126):= true;
LEND(40):= 0 + 42 *3;		PU(0+126):= true;
LEND(41):= 24 + 42 *3;	PU(24+126):= true;

LMOVE(42):= true;
LBEGIN(42):= 13 + 42 *2;	PU(13+84):= true;
LEND(42):= 1 + 42 *2;		PU(1+84):= true;
LEND(43):= 25 + 42 *2;	PU(25+84):= true;

LMOVE(44):= true;
LBEGIN(44):= 16 + 42 *2;	PU(16+84):= true;
LEND(44):= 4 + 42 *2;		PU(4+84):= true;
LEND(45):= 28 + 42 *2;	PU(28+84):= true;

LMOVE(46):= true;
LBEGIN(46):= 32 + 42 *3;	PU(32+126):= true;
LEND(46):= 8 + 42 *3;		PU(8+126):= true;
LEND(47):= 38 + 42 *3;	PU(38+126):= true;

LMOVE(48):= true;
LBEGIN(48):= 17 + 42 *3;	PU(17+126):= true;
LEND(48):= 5 + 42 *3;		PU(5+126):= true;
LEND(49):= 29 + 42 *3;	PU(29+126):= true;

LMOVE(50):= true;
LBEGIN(50):= 35 + 42 *4;	PU(35+168):= true;
LEND(50):= 32 + 42 *4;	PU(32+168):= true;
LEND(51):= 8 + 42 *4;		PU(8+168):= true;
LEND(52):= 38 + 42 *4;	PU(38+168):= true;
LEND(53):= 41 + 42 *4;	PU(41+168):= true;
LEND(54):= 11 + 42 *4;	PU(11+168):= true;
LEND(55):= 35 + 42 *4;	PU(35+168):= true;

LBASE:= 56;
end;	\MAKEWORLD

\----------------------------------------------------------------------

proc	FLY;		\Animate objects as a function of the joystick
int	I, CH;
begin
BANKX:= $4000;	BANKY:= 0;		\*** DEBUG ***

PITCHRATE:= 0;	SPINX:= 0;	SPINY:= 0;	SPINZ:= 0;
FORWARD:= true;
OLDSPEED:= 32;
OPENTHROTTLE(32, 25);

loop	begin
	DOTHROTTLE;
	VDRAW(0);		\Set page to draw on
	WAITVB;
	VCLEAR(0);		\Erase page
	ROTATE;			\Give the blitter time to clear the screen
	DOTHROTTLE;
	PERSPECTIVE;
	DOTHROTTLE;
	DRAW;
	VSHOW(0);		\Note that viewing does not actually occur
				\ until after next vertical blank

	if CHKKEY then
		begin
		case GETKEY of
		  $1B:	quit;					\Esc
		  $12:	RESTART;				\CTRL-R
		  ^2, $0B:					\Back
			[PITCHRATE:= PITCHRATE +$100];
		  ^8, $0A:					\Forward
			[PITCHRATE:= PITCHRATE -$100];
		  ^4, $08:					\ <--
			[SPINZ:= SPINZ +$200];
		  ^6, $15:					\ -->
			[SPINZ:= SPINZ -$200];
		  ^., ^9:	FORWARD:= true;			\ >
		  ^,, ^3:	FORWARD:= false;		\ <
		  ^5, $0D, $20:
			[PITCHRATE:= 0; SPINZ:= 0]		\STOP
		other CHOUT(0, BEL);
		end;

	DOTHROTTLE;

	if BUTTON(0) then	\Left mouse button stops rotations
		[PITCHRATE:= 0; SPINZ:= 0];

	\Alieron control
	\if not KEYCON then SPINZ:= INERTIAZ *SPINZ /10;
	SPINZ:= SPINZ - BANKY /$20;		\Slow correction to level
	if SPINZ > $1000 then SPINZ:= $1000;
	if SPINZ < -$1000 then SPINZ:= -$1000;

	\Elevator control
	\if not KEYCON then PITCHRATE:= INERTIAX *PITCHRATE /10;

	\Humans can climb faster than they can dive (blackout versus redout)
	if PITCHRATE > $2000 then PITCHRATE:= $2000;
	if PITCHRATE < -$800 then PITCHRATE:= -$800;

	\The turn rate is the rate of turn about the hive's Y axis:
	\	turn rate = max rate * SIN(bank angle).
	\The rotation rates about the pilot's reference frame are:
	\	SPINY = turn rate * COS(bank angle)
	\	SPINX = turn rate * SIN(bank angle) + PITCHRATE
	\	BANKX = COS(bank angle) *$4000,
	\	BANKY = SIN(bank angle) *$4000,
	SPINY:= BANKY *BANKX /$2000<<4;		\Divide by scale and rate adj
	SPINX:= BANKY *BANKY /$2000<<4 + PITCHRATE;

	VDRAW(1);		\Set page to draw on
	WAITVB;			\Make sure PAGE 0 is actually being displayed
	VCLEAR(1);		\Erase page
	ROTATE;
	DOTHROTTLE;
	PERSPECTIVE;
	DOTHROTTLE;
	DRAW;
	VSHOW(1);
	end;
end;	\FLY

\----------------------------------------------------------------------

begin	\MAIN
CPUREG:= GETREG;
ZSCREEN:= 109*1000 + 227;

REM0:= RESERVE(3 *INTSIZE);	REM1:= RESERVE(3 *INTSIZE);
REM2:= RESERVE(3 *INTSIZE);	REM3:= RESERVE(3 *INTSIZE);
REM4:= RESERVE(3 *INTSIZE);	REM5:= RESERVE(3 *INTSIZE);

for II:= 0, 2 do
	begin
	REM0(II):= 0;
	REM1(II):= 0;
	REM2(II):= 0;
	REM3(II):= 0;
	REM4(II):= 0;
	REM5(II):= 0;
	end;

REFPX:= RESERVE(3 *INTSIZE);
REFPY:= RESERVE(3 *INTSIZE);
REFPZ:= RESERVE(3 *INTSIZE);

PX:= RESERVE(42 *5 *INTSIZE);
PY:= RESERVE(42 *5 *INTSIZE);
PZ:= RESERVE(42 *5 *INTSIZE);
PU:= RESERVE(42 *5);

PX0:= RESERVE(42 *5 *INTSIZE);
PY0:= RESERVE(42 *5 *INTSIZE);
PZ0:= RESERVE(42 *5 *INTSIZE);

PPX:= RESERVE(42 *5 *INTSIZE);
PPY:= RESERVE(42 *5 *INTSIZE);

LMOVE:= RESERVE(45 *5);
LBEGIN:= RESERVE(45 *5);
LEND:= RESERVE(45 *5);
LCOLOR:= RESERVE(45 *5);

OPENI(0);   OPENO(0);

CHOUT(0, FF);
TEXT(0, "-- FLIGHT SIMULATOR, V0.0x0 --
");
\if CHIN(1) then;

SETVID($12);
SETCOLORS;

MAKEWORLD;
ANIMRATE:= 0;
FLY;
SETVID($03);
end;	\MAIN
