\FLT20.XPL	JUL-07-88
\Flight simulator
\by Loren Blaney
\
\NOTES:
\This program requires a mouse in port 1 and a joystick in port 2 (although
\ the keypad may be used instead of the joystick).
\
\This program draws images as seen by the pilot. Thus the hive appears to
\ be moving. A right-hand coordinate system is used.
\
\			           Z
\			       |  /
\			       | /
\			       |/
\			-------+-------> X
\			      /|
\			     / |
\			    /  |
\			       V
\			       Y
\
\ X increases to the right, Y increases downward, and Z increases into the
\ screen.
\
\			-- SIDE VIEW --
\
\		|			|
\		|			|
\		|			|
\		|			|
\		|			|
\		|			|
\	pilot  <O-----------------------+-----------------------> Z
\	(0,0)	|			|
\		|<------ ZSCREEN ------>|
\		|	(0.66 ft)	|
\		|			|
\		|			|
\		V		     monitor
\		Y		     screen
\
\
\Distances are scaled to 10 counts per foot.

ext	ROT= $A000;	\Routine to rotate an array of points

code
ABS= 0,		RAN= 1,		REM= 2,		RESERVE= 3,
SWAP= 4,	EXTEND= 5,	RESTART= 6,	CHIN= 7,
CHOUT= 8,	CRLF= 9,	INTIN= 10,	INTOUT= 11,
TEXT= 12,	OPENI= 13,	OPENO= 14,	CLOSE= 15,
CURSOR= 23,	HEXIN= 26,	HEXOUT= 27,
CLEAR= 40,	POINT= 41,	LINE= 42,	MOVE= 43,
BITMAP= 108,	BITMAP2= 109,	VIEW= 110,	PALETTE= 111,
PEEK_W= 113,	POKE_W= 114,	SWAP_W= 117,
GETKEY= 125,	KEYHIT= 126;

def	INTSIZE= 4;	\Number of bytes in an integer
def	TV= 0, KB= 0;	\Device numbers
def	WIDTH= 640, HEIGHT= 480, DEPTH= 1;	\RASTER (bit map) dimensions
def	JOY0DAT= $DFF00A, JOY1DAT= $DFF00C;	\Chip register addresses

def	INERTIAX= 8, INERTIAZ= 8;	\Joystick response (10 = no dampening)

\To simplify the perspective calculation, the distance to the screen
\ as expressed in pixels is chosen to be a power of two (ZN).
\	DIST = 2**ZN /768 = Distance from viewer to screen in feet
\	ZSCREEN = DIST *10 *$4000 = DIST *163840 = Scaled distance to screen
\	PPX = PX /PZ *1.33 *768		Perspective X in pixels
\					(There are 768 pixels/foot)
\	DIST   ZSCREEN   ZN
\	0.33    54613     8	(causes noticable distortion)
\	0.66   109227     9
\	1.33   218453    10
\	2.66   436907    11
\
def	ZSCREEN= 109227, ZN= 9;

int	REM0, REM1, REM2, REM3, REM4, REM5,	\Arrays
	II,
	MOUSEX, MOUSEY,	\Mouse position (mouse counts)
	MOUSEX0, MOUSEY0,\Mouse position (OWNs for GETMOUSE)
	PTRX, PTRY,	\Mouse pointer position (affected by GAIN)
	GAINX, GAINY,	\Mouse pointer gain, scaled by 100
	THROTTLE,	\Throttle position (0 to 199)
	PX0, PY0, PZ0,	\Arrays of points (unrotated, initial hive positions)
	PX, PY, PZ,	\Arrays of points (rotated, translated)
	PPX, PPY,	\Arrays of points (rotated, translated, with perspective)
	REFPX, REFPY, REFPZ, \Arrays: 3 hive reference points, used to make
			\ unit vectors, scaled by $4000
	LBASE, PBASE,	\Line and point base index for each junction in PX etc.
	ANIMRATE,	\For animation rate display (debug)
	SPINX, SPINY, SPINZ,	\Rate of rotation about pilot's X, Y, and Z axes
	PITCHRATE,	\Pitch rate (commanded elevator position)
	FORWARD,	\Boolean: Commanded to travel in the forward direction
	SPEED,		\Actual speed: delta Z per iteration
	OLDSPEED,	\Previous iteration speed: delta Z per iteration
	BANKX,		\BANKX = COS(bank angle) *$4000,
	BANKY,		\BANKY = SIN(bank angle) *$4000,
	;
addr	PU,		\Array: point is connected to a line (i.e. is used)
	LMOVE, LBEGIN, LEND, LCOLOR,
			\Line arrays: MOVE required, 1st point, 2nd point, color
	RASTER0, RASTER1; \Graphics bit maps

\----------------------------------------------------------------------

func	BUTTON(N);	\Returns 'TRUE' if specified button is pressed.
int	N;
addr	ADDR;
def	CIAA= $BFE001;
begin
ADDR:= 0;
case N of
  0: return (ADDR(CIAA) & $40) = 0;	\Port 1, left button
  1: return false;			\Port 1, right button
  2: return (ADDR(CIAA) & $80) = 0;	\Port 2, left button
  3: return false			\Port 2, right button
other return false;
end;	\BUTTON



func	JOYSTICK(N);	\Returns 'TRUE' if specified direction is pressed.
reg int	N;
reg int	DATA;
def	JOY0DAT= $DFF00A, JOY1DAT= $DFF00C;	\Chip register addresses
\
\	    LEFT PORT		   RIGHT PORT
\
\		2			6
\		|			|
\	   3 ---+--- 1		   7 ---+--- 5
\		|			|
\		0			4
\
begin
if N & 4 then
	begin
	DATA:= PEEK_W(JOY1DAT);
	if N & 1 then
		begin
		if N & 2 then
			return (DATA & $0200) # 0	\3 <--
		else	return (DATA & $0002) # 0;	\1 -->
		end
	else	begin
		DATA:= DATA>>1 | DATA;
		if N & 2 then
			return (DATA & $0100) # 0	\2 Forward
		else	return (DATA & $0001) # 0;	\0 Back
		end;
	end
else	begin
	DATA:= PEEK_W(JOY0DAT);
	if N & 1 then
		begin
		if N & 2 then
			return (DATA & $0200) # 0	\3 <--
		else	return (DATA & $0002) # 0;	\1 -->
		end
	else	begin
		DATA:= DATA>>1 | DATA;
		if N & 2 then
			return (DATA & $0100) # 0	\2 Forward
		else	return (DATA & $0001) # 0;	\0 Back
		end;
	end;
end;	\JOYSTICK



proc	GETMOUSE;	\Return the current mouse position in MOUSEX, MOUSEY
int	X, Y;
begin
X:= PEEK_W(JOY0DAT);
Y:= SWAP(X);
MOUSEX:= MOUSEX + EXTEND(X - MOUSEX0);
MOUSEY:= MOUSEY + EXTEND(Y - MOUSEY0);
MOUSEX0:= X;
MOUSEY0:= Y;
end;	\GETMOUSE



proc	OPENMOUSE;	\Initialize mouse position to 0,0
begin
MOUSEX0:= PEEK_W(JOY0DAT);		\Read position from mouse
MOUSEY0:= SWAP(MOUSEX0);
MOUSEX:= 0;	MOUSEY:= 0;
end;



proc	DOTHROTTLE;	\Get the throttle position, return it in THROTTLE
begin
GETMOUSE;
THROTTLE:= -MOUSEY *GAINY /100;

\Clip pointer position to text screen dimensions:
if THROTTLE > 199 then [THROTTLE:= 199;   MOUSEY:= -19900 /GAINY]
else if THROTTLE < 0 then [THROTTLE:= 0;   MOUSEY:= 0];

SPEED:= (THROTTLE *THROTTLE) >> 5;
if not FORWARD then SPEED:= -SPEED;

\SPEED:= SPEED + THRUST;	realistic equation
SPEED:= (OLDSPEED<<4 + SPEED) /17;
OLDSPEED:= SPEED;
end;	\DOTHROTTLE



proc	OPENTHROTTLE(Y0, G);	\Initialize throttle position
int	Y0,		\Initial throttle position
	G;		\Gain factors for mouse movement
begin
OPENMOUSE;
GAINY:= G;
MOUSEY:= -Y0 *100 /GAINY;
DOTHROTTLE;
end;	\OPENTHROTTLE

\----------------------------------------------------------------------

proc	WAITVB;		\Wait for vertical blank
def	INTREQR= $DFF01E, INTREQ= $DFF09C;
begin
POKE_W(INTREQ, $0020);			\Clear vertical blank bit
repeat until PEEK_W(INTREQR) & $20;	\Wait for vertical blank
end;	\WAITVB

\----------------------------------------------------------------------

func	RATIO(B, AA, BB);
\Return the value of A in the ratio A/B = AA/BB.
\WARNING: This will not work if B > BB.
int	B, AA, BB;
int	A1, B1, NEG;
begin
NEG:= false;
if B < 0 then [B:= -B;   NEG:= not NEG];
if AA < 0 then [AA:= -AA;   NEG:= not NEG];
if BB < 0 then [BB:= -BB;   NEG:= not NEG];
A1:=0;   B1:= 0;
loop	begin
	AA:= AA >> 1;
	if AA = 0 then quit;
	BB:= BB >> 1;
	if BB = 0 then quit;
	if B1+BB <= B then [B1:= B1 +BB;   A1:= A1 +AA];
	end;

return if NEG then -A1 else A1;
end;	\RATIO

\----------------------------------------------------------------------

proc	DRAW;		\Draw the objects on the graphics screen
\Inputs: BANKX, BANKY, PPX, PPY, PX, PY, PZ, LMOVE, LBEGIN, LEND, LBASE.
\Outputs: ANIMRATE.
int	ZEND, ZBEGIN,
	BX, BY,
	PX_, PY_,
	A, B,		\Tempories for screen intersection calculation
	I, J,
	L,
	Z,
	;
begin
POINT(ANIMRATE-320,199,$01);		\Display the rate indicator
ANIMRATE:= ANIMRATE +1;
if ANIMRATE >= 640 then ANIMRATE:= 0;

\Display artificial horizon and aiming point:
BX:= BANKX/$040;   BY:= BANKY/$040;
MOVE(-BX, -BY);   LINE(BX, BY, $7F7F0001);	\(dotted line)
BX:= BANKX/$100;   BY:= BANKY/$100;
MOVE(BY, -BX);   LINE(0, 0, $7F7F0001);

\Display throttle:

MOVE(-319, 199);   LINE(-319, 199-THROTTLE, $C0000001);	\(dotted line)

for L:= 0, LBASE-1 do			\For all lines...
	begin
	if LMOVE(L) then		\Must move to start of line
		begin
		I:= LBEGIN(L);		\Get index to point
		ZBEGIN:= PZ(I);
		if ZBEGIN >= ZSCREEN then MOVE(PPX(I), PPY(I));
		end;
	J:= LEND(L);			\Get index to end point
	ZEND:= PZ(J);

	if ZBEGIN >= ZSCREEN then
		begin
		if ZEND >= ZSCREEN then		\Draw normal line
			LINE(PPX(J), PPY(J), LCOLOR(L))
		else	begin			\ZEND < ZSCREEN
			\Calculate point of intersection with screen (PX_,PY_)
			I:= LBEGIN(L);		\Get index to beginning point
			A:= (ZSCREEN - PZ(J));	\Shift to avoid overflow
			B:= (PZ(I) - PZ(J));	\ A & B are always positive
			PX_:= PX(J) + RATIO( A, (PX(I) - PX(J)), B );
			PY_:= PY(J) + RATIO( A, (PY(I) - PY(J)), B );

			Z:= ZSCREEN >> ZN;
			PX_:= PX_ /Z;
			PY_:= PY_ /Z;

			LINE(PX_, PY_, LCOLOR(L));
			end;
		end
	else	begin			\ZBEGIN < ZSCREEN
		if ZEND >= ZSCREEN then
			begin
			\Calculate point of intersection with screen (PX_,PY_)
			I:= LBEGIN(L);		\Get index to beginning point
			A:= (ZSCREEN - PZ(I));	\Shift to avoid overflow
			B:= (PZ(J) - PZ(I));	\ A & B are always positive
			PX_:= PX(I) + RATIO( A, (PX(J) - PX(I)), B );
			PY_:= PY(I) + RATIO( A, (PY(J) - PY(I)), B );

			Z:= ZSCREEN >> ZN;
			PX_:= PX_ /Z;
			PY_:= PY_ /Z;

			MOVE(PX_, PY_);		\Move to intersection point
			LINE(PPX(J), PPY(J), LCOLOR(L));	\Normal end point
			end;
		end;
	ZBEGIN:= ZEND;
	end;
end;	\DRAW

\----------------------------------------------------------------------

proc	PERSPECTIVE;
\Use similar triangles to make close points farther apart.
\Inputs: PX, PY, PZ, PBASE.
\Outputs: PPX, PPY.
reg int	I, Z;
begin
for I:= 0, PBASE-1 do			\For all points...
    if PU(I) then
	begin
	Z:= PZ(I);
	if Z >= ZSCREEN then	\(Points behind screen are handled by DRAW)
		begin
		Z:= Z >> ZN;
		PPX(I):= PX(I) /Z;	\(Scale factors divide out)
		PPY(I):= PY(I) /Z;
		end;
	end;
end;	\PERSPECTIVE

\----------------------------------------------------------------------

proc	ROTATE;		\Rotate and translate the points
int	I, J,
	DX, DY, DZ,	\Deltas
	DX1, DY1, DZ1,	\Deltas
	DX2, DY2, DZ2,	\Deltas
	DX3, DY3, DZ3,	\Deltas
	RX, RY, RZ,	\Temporaries for rotate
	UXX, UXY, UXZ,	\X unit vector
	UYX, UYY, UYZ,	\Y unit vector
	UZX, UZY, UZZ, 	\Z unit vector
	D2,
	;
reg int	K, L;
begin
\Rotate reference frame about the origin:
for J:= 0, 2 do
	begin
	RX:= REFPX(J);
	RY:= REFPY(J);
	RZ:= REFPZ(J);

	\Rotate about X axis:
	RY:= RY + (RZ *SPINX + REM0(J)) /$10000;   REM0(J):= REM;
	RZ:= RZ - (RY *SPINX + REM1(J)) /$10000;   REM1(J):= REM;
	\Rotate about Y axis:
	RX:= RX + (RZ *SPINY + REM2(J)) /$10000;   REM2(J):= REM;
	RZ:= RZ - (RX *SPINY + REM3(J)) /$10000;   REM3(J):= REM;
	\Rotate about Z axis:
	RY:= RY + (RX *SPINZ + REM4(J)) /$10000;   REM4(J):= REM;
	RX:= RX - (RY *SPINZ + REM5(J)) /$10000;   REM5(J):= REM;

	REFPX(J):= RX;
	REFPY(J):= RY;
	REFPZ(J):= RZ - SPEED;		\Also translate
	end;				\Forward speed moves objects closer

\Because these 3 points are independantly rotated, errors can accumulate and
\ distort the reference frame. To prevent this, mathematical cross braces are
\ used to maintain the correct distances between the points.

DX:= REFPX(1) - REFPX(0);   DY:= REFPY(1) - REFPY(0);   DZ:= REFPZ(1) - REFPZ(0);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $FFF0004 then			\$FFF0004 = $3FFE **2
	begin				\Lengthen line
	REFPX(1):= REFPX(1) + DX /$1000;
	REFPY(1):= REFPY(1) + DY /$1000;
	REFPZ(1):= REFPZ(1) + DZ /$1000;
	end;
if D2 > $10010004 then			\$10010004 = 4002 **2
	begin				\Shorten line
	REFPX(1):= REFPX(1) - DX /$1000;
	REFPY(1):= REFPY(1) - DY /$1000;
	REFPZ(1):= REFPZ(1) - DZ /$1000;
	end;

DX:= REFPX(2) - REFPX(0);   DY:= REFPY(2) - REFPY(0);   DZ:= REFPZ(2) - REFPZ(0);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $FFF0004 then			\$FFF0004 = $3FFE ^2
	begin				\Lengthen line
	REFPX(2):= REFPX(2) + DX /$1000;
	REFPY(2):= REFPY(2) + DY /$1000;
	REFPZ(2):= REFPZ(2) + DZ /$1000;
	end;
if D2 > $10010004 then			\$10010004 = 4002 **2
	begin				\Shorten line
	REFPX(2):= REFPX(2) - DX /$1000;
	REFPY(2):= REFPY(2) - DY /$1000;
	REFPZ(2):= REFPZ(2) - DZ /$1000;
	end;

DX:= REFPX(2) - REFPX(1);   DY:= REFPY(2) - REFPY(1);   DZ:= REFPZ(2) - REFPZ(1);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $1FFD8B01 then			\$1FFD8B01 = $5A7F ^2
	begin				\Lengthen line
	REFPX(2):= REFPX(2) + DX /$1000;
	REFPY(2):= REFPY(2) + DY /$1000;
	REFPZ(2):= REFPZ(2) + DZ /$1000;
	end;
if D2 > $2001C919 then			\$2001C919 = $5A85 ^2
	begin				\Shorten line
	REFPX(2):= REFPX(2) - DX /$1000;
	REFPY(2):= REFPY(2) - DY /$1000;
	REFPZ(2):= REFPZ(2) - DZ /$1000;
	end;

UXX:= REFPX(1) - REFPX(0);		\UX = REFP(1) - REFP(0)
UXY:= REFPY(1) - REFPY(0);
UXZ:= REFPZ(1) - REFPZ(0);

UYX:= REFPX(2) - REFPX(0);		\UY = REFP(2) - REFP(0)
UYY:= REFPY(2) - REFPY(0);
UYZ:= REFPZ(2) - REFPZ(0);

\Find UZ using cross product:
UZX:= (UXY *UYZ - UXZ *UYY) /$4000;	\UZ = UX X UY
UZY:= (UXZ *UYX - UXX *UYZ) /$4000;
UZZ:= (UXX *UYY - UXY *UYX) /$4000;

\Rotate and translate all points to align with reference frame (unit vectors):
DX:= REFPX(0) << 14;	DY:= REFPY(0) << 14;	DZ:= REFPZ(0) << 14;

ROT(PX, PY, PZ, PX0, PY0, PZ0,
	UXX, UYX, UZX, UXY, UYY, UZY, UXZ, UYZ, UZZ,
	DX, DY, DZ, 42);

\Handle remaining junctions efficiently:
for I:= 1, 4 do
	begin			\Rotate reference point for each junction
	K:= I *42;
	PX(K):= PX0(K) *UXX + PY0(K) *UYX + PZ0(K) *UZX + DX;
	PY(K):= PX0(K) *UXY + PY0(K) *UYY + PZ0(K) *UZY + DY;
	PZ(K):= PX0(K) *UXZ + PY0(K) *UYZ + PZ0(K) *UZZ + DZ;

	RX:= PX(K) - PX(0);	RY:= PY(K) - PY(0);	RZ:= PZ(K) - PZ(0);
	J:= K;
	for L:= 1, 41 do	\Use offsets to determine remaining points
		begin
		K:= J + L;
		if PU(K) then
			begin
			PX(K):= PX(L) + RX;
			PY(K):= PY(L) + RY;
			PZ(K):= PZ(L) + RZ;
			end;
		end;
	end;

BANKY:= -UYX;
BANKX:= UYY;
\*** DEBUG ***
\BANKY:= BANKY + BANKX *SPINZ /$10000;
\BANKX:= BANKX - BANKY *SPINZ /$10000;
end;	\ROTATE

\----------------------------------------------------------------------

proc	MAKEJCT(X, Y, Z, COLOR);	\Set up junction geometry
int	X, Y, Z, COLOR;
int	I, J, TBL;
\Other inputs: PBASE, LBASE.
\Outputs: LINES, PX0, PY0, PZ0, LBASE, PBASE.
def	T= true, F= false;
def	DR= 500, D2R= 1000, DR2= 250,		\(L= 2000)
	DH= 433, D2H= 866, DH3= 144, D2H3= 288, D4H3= 576;
begin
\Set up initial (unrotated) point positions:
\	 X	 Y	 Z
TBL:= [	-DR,	 0,	-D2H,	\0
	 DR,	 0,	-D2H,	\1
	 D2R,	 0,	 0,	\2
	 DR,	 0,	 D2H,	\3
	-DR,	 0,	 D2H,	\4\	      \ 4                / 3
	-D2R,	 0,	 0,	\5\	       \                /
						\              /
	 0,	 0,	-D4H3,	\6\		 \_22_________/ 21
	 DR,	 0,	-D2H3,	\7		 /            \
	 DR,	 0,	 D2H3,	\8		/       | Z    \
	 0,	 0,	 D4H3,	\9	       /        |       \
	-DR,	 0,	 D2H3,	\10 5 ________/ 23    __|____ X \\_20______ 2
	-DR,	 0,	-D2H3,	\11\	      \         |        /
					       \                /
	-DR2,	-DH,	-DH,	\12\		\              /
	 DR2,	-DH,	-DH,	\13\		 \_18_________/ 19
	 DR,	-DH,	 0,	\14	         /            \
	 DR2,	-DH,	 DH,	\15             /              \
	-DR2,	-DH,	 DH,	\16            /                \
	-DR,	-DH,	 0,	\17         0 /                1 \

	-DR2,	 0,	-DH,	\18
	 DR2,	 0,	-DH,	\19
	 DR,	 0,	 0,	\20
	 DR2,	 0,	 DH,	\21
	-DR2,	 0,	 DH,	\22
	-DR,	 0,	 0,	\23

	-DR2,	 DH,	-DH,	\24
	 DR2,	 DH,	-DH,	\25
	 DR,	 DH,	 0,	\26
	 DR2,	 DH,	 DH,	\27
	-DR2,	 DH,	 DH,	\28
	-DR,	 DH,	 0,	\29

	 0,	-DH,	-D2H3,	\30
	 DR2,	-DH,	-DH3,	\31
	 DR2,	-DH,	 DH3,	\32
	 0,	-DH,	 D2H3,	\33
	-DR2,	-DH,	 DH3,	\34
	-DR2,	-DH,	-DH3,	\35

	 0,	 DH,	-D2H3,	\36
	 DR2,	 DH,	-DH3,	\37
	 DR2,	 DH,	 DH3,	\38
	 0,	 DH,	 D2H3,	\39
	-DR2,	 DH,	 DH3,	\40
	-DR2,	 DH,	-DH3 ];	\41

J:= 0;
for I:= PBASE, PBASE+41 do
	begin
	PX0(I):= TBL(J) + X;		J:= J +1;
	PY0(I):= TBL(J) + Y;		J:= J +1;
	PZ0(I):= TBL(J) + Z;		J:= J +1;
	end;


\Set up array for connecting points (LINES):
\ (MOVE required, 1st point, 2nd point)
TBL:= [	T,	12,	13,
	F,	13,	14,
	F,	14,	15,
	F,	15,	16,
	F,	16,	17,
	F,	17,	12,

	T,	24,	25,
	F,	25,	26,
	F,	26,	27,
	F,	27,	28,
	F,	28,	29,
	F,	29,	24,

	T,	12,	0,
	F,	0,	24,

	T,	13,	1,
	F,	1,	25,

	T,	14,	2,
	F,	2,	26,

	T,	15,	3,
	F,	3,	27,

	T,	16,	4,
	F,	4,	28,

	T,	17,	5,
	F,	5,	29 ];

J:= 0;
for I:= LBASE, LBASE+23 do
	begin
	LMOVE(I):= TBL(J);		J:= J +1;
	LBEGIN(I):= TBL(J) + PBASE;	J:= J +1;
	LEND(I):= TBL(J) + PBASE;	J:= J +1;
	LCOLOR(I):= COLOR;
	end;


PBASE:= PBASE + 42;
LBASE:= LBASE + 24;
end;	\MAKEJCT

\----------------------------------------------------------------------

proc	MAKEWORLD;	\Set up all the objects
int	X, Y, Z, I, J, K, TBL;
begin
\Initial position of reference frame:
X:= 0;		Y:= 0;		Z:= 8000;

REFPX(0):= X;		REFPY(0):= Y;		REFPZ(0):= Z;
REFPX(1):= $4000 +X;	REFPY(1):= Y;		REFPZ(1):= Z;
REFPX(2):= X;		REFPY(2):= $4000 +Y;	REFPZ(2):= Z;

LBASE:= 0;	PBASE:= 0;
for J:= 0, 42*5-1 do PU(J):= false;

MAKEJCT(0, 0, 0, $01);
MAKEJCT(0, 0, 4000, $01);	\(L = 2000)
MAKEJCT(0, 0, 8000, $01);
MAKEJCT(3464, 0, 6000, $01);
MAKEJCT(3464, 0, 2000, $01);

for J:= 0, 55 do
	begin
	LMOVE(J):= true;
	LCOLOR(J):= $01;
	end;

LBEGIN(0):= 35;			PU(35):= true;
LEND(0):= 16 + 42 *1;		PU(16+42):= true;
LBEGIN(1):= 32;			PU(32):= true;
LEND(1):= 31 + 42 *1;		PU(31+42):= true;
LBEGIN(2):= 8;			PU(8):= true;
LEND(2):= 7 + 42 *1;		PU(7+42):= true;
LBEGIN(3):= 38;			PU(38):= true;
LEND(3):= 37 + 42 *1;		PU(37+42):= true;
LBEGIN(4):= 41;			PU(41):= true;
LEND(4):= 28 + 42 *1;		PU(28+42):= true;
LBEGIN(5):= 11;			PU(11):= true;
LEND(5):= 4 + 42 *1;		PU(4+42):= true;

LBEGIN(6):= 16 + 42 *1;		PU(16+42):= true;
LEND(6):= 16 + 42 *2;		PU(16+84):= true;
LBEGIN(7):= 15 + 42 *1;		PU(15+42):= true;
LEND(7):= 13 + 42 *2;		PU(13+84):= true;
LBEGIN(8):= 3 + 42 *1;		PU(3+42):= true;
LEND(8):= 1 + 42 *2;		PU(1+84):= true;
LBEGIN(9):= 27 + 42 *1;		PU(27+42):= true;
LEND(9):= 25 + 42 *2;		PU(25+84):= true;
LBEGIN(10):= 28 + 42 *1;	PU(28+42):= true;
LEND(10):= 28 + 42 *2;	PU(28+84):= true;
LBEGIN(11):= 4 + 42 *1;		PU(4+42):= true;
LEND(11):= 4 + 42 *2;		PU(4+84):= true;

LBEGIN(12):= 15 + 42 *1;	PU(15+42):= true;
LEND(12):= 17 + 42 *3;	PU(17+126):= true;
LBEGIN(13):= 31 + 42 *1;	PU(31+42):= true;
LEND(13):= 12 + 42 *3;	PU(12+126):= true;
LBEGIN(14):= 7 + 42 *1;		PU(7+42):= true;
LEND(14):= 0 + 42 *3;		PU(0+126):= true;
LBEGIN(15):= 37 + 42 *1;	PU(37+42):= true;
LEND(15):= 24 + 42 *3;	PU(24+126):= true;
LBEGIN(16):= 27 + 42 *1;	PU(27+42):= true;
LEND(16):= 29 + 42 *3;	PU(29+126):= true;
LBEGIN(17):= 3 + 42 *1;		PU(3+42):= true;
LEND(17):= 5 + 42 *3;		PU(5+126):= true;

LBEGIN(18):= 16 + 42 *2;	PU(16+84):= true;
LEND(18):= 32 + 42 *3;	PU(32+126):= true;
LBEGIN(19):= 13 + 42 *2;	PU(13+84):= true;
LEND(19):= 17 + 42 *3;	PU(17+126):= true;
LBEGIN(20):= 1 + 42 *2;		PU(1+84):= true;
LEND(20):= 5 + 42 *3;		PU(5+126):= true;
LBEGIN(21):= 25 + 42 *2;	PU(25+84):= true;
LEND(21):= 29 + 42 *3;	PU(29+126):= true;
LBEGIN(22):= 28 + 42 *2;	PU(28+84):= true;
LEND(22):= 38 + 42 *3;	PU(38+126):= true;
LBEGIN(23):= 4 + 42 *2;		PU(4+84):= true;
LEND(23):= 8 + 42 *3;		PU(8+126):= true;

LBEGIN(24):= 12 + 42 *3;	PU(12+126):= true;
LEND(24):= 35 + 42 *4;	PU(35+168):= true;
LBEGIN(25):= 32 + 42 *3;	PU(32+126):= true;
LEND(25):= 32 + 42 *4;	PU(32+168):= true;
LBEGIN(26):= 8 + 42 *3;		PU(8+126):= true;
LEND(26):= 8 + 42 *4;		PU(8+168):= true;
LBEGIN(27):= 38 + 42 *3;	PU(38+126):= true;
LEND(27):= 38 + 42 *4;	PU(38+168):= true;
LBEGIN(28):= 24 + 42 *3;	PU(24+126):= true;
LEND(28):= 41 + 42 *4;	PU(41+168):= true;
LBEGIN(29):= 0 + 42 *3;		PU(0+126):= true;
LEND(29):= 11 + 42 *4;	PU(11+168):= true;

for J:= 31, 55 do
	LMOVE(J):= false;

LBEGIN(30):= 35;		PU(35):= true;
LEND(30):= 32;		PU(32):= true;
LEND(31):= 8;			PU(8):= true;
LEND(32):= 38;		PU(38):= true;
LEND(33):= 41;		PU(41):= true;
LEND(34):= 11;		PU(11):= true;
LEND(35):= 35;		PU(35):= true;

LMOVE(36):= true;
LBEGIN(36):= 31 + 42 *1;	PU(31+42):= true;
LEND(36):= 7 + 42 *1;		PU(7+42):= true;
LEND(37):= 37 + 42 *1;	PU(37+42):= true;

LMOVE(38):= true;
LBEGIN(38):= 15 + 42 *1;	PU(15+42):= true;
LEND(38):= 3 + 42 *1;		PU(3+42):= true;
LEND(39):= 27 + 42 *1;	PU(27+42):= true;

LMOVE(40):= true;
LBEGIN(40):= 12 + 42 *3;	PU(12+126):= true;
LEND(40):= 0 + 42 *3;		PU(0+126):= true;
LEND(41):= 24 + 42 *3;	PU(24+126):= true;

LMOVE(42):= true;
LBEGIN(42):= 13 + 42 *2;	PU(13+84):= true;
LEND(42):= 1 + 42 *2;		PU(1+84):= true;
LEND(43):= 25 + 42 *2;	PU(25+84):= true;

LMOVE(44):= true;
LBEGIN(44):= 16 + 42 *2;	PU(16+84):= true;
LEND(44):= 4 + 42 *2;		PU(4+84):= true;
LEND(45):= 28 + 42 *2;	PU(28+84):= true;

LMOVE(46):= true;
LBEGIN(46):= 32 + 42 *3;	PU(32+126):= true;
LEND(46):= 8 + 42 *3;		PU(8+126):= true;
LEND(47):= 38 + 42 *3;	PU(38+126):= true;

LMOVE(48):= true;
LBEGIN(48):= 17 + 42 *3;	PU(17+126):= true;
LEND(48):= 5 + 42 *3;		PU(5+126):= true;
LEND(49):= 29 + 42 *3;	PU(29+126):= true;

LMOVE(50):= true;
LBEGIN(50):= 35 + 42 *4;	PU(35+168):= true;
LEND(50):= 32 + 42 *4;	PU(32+168):= true;
LEND(51):= 8 + 42 *4;		PU(8+168):= true;
LEND(52):= 38 + 42 *4;	PU(38+168):= true;
LEND(53):= 41 + 42 *4;	PU(41+168):= true;
LEND(54):= 11 + 42 *4;	PU(11+168):= true;
LEND(55):= 35 + 42 *4;	PU(35+168):= true;

LBASE:= 56;
end;	\MAKEWORLD

\----------------------------------------------------------------------

proc	FLY;		\Animate objects as a function of the joystick
int	I, CH,
	KEYCON;		\Flag: keyboard control
begin
KEYCON:= false;		\Assume joystick

BANKX:= $4000;	BANKY:= 0;		\*** DEBUG ***

PITCHRATE:= 0;	SPINX:= 0;	SPINY:= 0;	SPINZ:= 0;
FORWARD:= true;
OLDSPEED:= 32;
OPENTHROTTLE(32, 25);

loop	begin
	DOTHROTTLE;
	BITMAP(RASTER0, WIDTH, HEIGHT, DEPTH);
	WAITVB;
	CLEAR;
	ROTATE;			\Give the blitter time to clear the screen
	DOTHROTTLE;
	PERSPECTIVE;
	DOTHROTTLE;
	DRAW;
	VIEW(RASTER0, $9004);	\Note that viewing does not actually occur
				\ until after next vertical blank
	if KEYHIT then
		begin
		case GETKEY of
		  $03:	quit;					\CTRL-C
		  $13:	[repeat until KEYHIT; CH:= GETKEY];	\CTRL-S
		  $12:	RESTART;				\CTRL-R
		  ^2, $0B:					\Back
			[KEYCON:= true;   PITCHRATE:= PITCHRATE +$100];
		  ^8, $0A:					\Forward
			[KEYCON:= true;   PITCHRATE:= PITCHRATE -$100];
		  ^4, $08:					\ <--
			[KEYCON:= true;   SPINZ:= SPINZ +$200];
		  ^6, $15:					\ -->
			[KEYCON:= true;   SPINZ:= SPINZ -$200];
		  ^., ^9:	FORWARD:= true;			\ >
		  ^,, ^3:	FORWARD:= false;		\ <
		  ^5, $0D:
			[PITCHRATE:= 0; SPINZ:= 0]		\STOP
		other CHOUT(0,7);
		end;

	DOTHROTTLE;

	if BUTTON(0) then	\Left mouse button stops rotations
		[PITCHRATE:= 0; SPINZ:= 0];

	case of						\Alieron control
	  JOYSTICK(5):
		[KEYCON:= false;   SPINZ:= SPINZ -$200];	\ -->
	  JOYSTICK(7):
		[KEYCON:= false;   SPINZ:= SPINZ +$200]		\ <--
	other if not KEYCON then SPINZ:= INERTIAZ *SPINZ /10;
	SPINZ:= SPINZ - BANKY /$20;		\Slow correction to level
	if SPINZ > $1000 then SPINZ:= $1000;
	if SPINZ < -$1000 then SPINZ:= -$1000;

	case of						\Elevator control
	  JOYSTICK(4):
		[KEYCON:= false;   PITCHRATE:= PITCHRATE +$100]; \Back
	  JOYSTICK(6):
		[KEYCON:= false;   PITCHRATE:= PITCHRATE -$100]	 \Forward
	other if not KEYCON then PITCHRATE:= INERTIAX *PITCHRATE /10;

	\Humans can climb faster than they can dive (blackout versus redout)
	if PITCHRATE > $2000 then PITCHRATE:= $2000;
	if PITCHRATE < -$800 then PITCHRATE:= -$800;

	\The turn rate is the rate of turn about the hive's Y axis:
	\	turn rate = max rate * SIN(bank angle).
	\The rotation rates about the pilot's reference frame are:
	\	SPINY = turn rate * COS(bank angle)
	\	SPINX = turn rate * SIN(bank angle) + PITCHRATE
	\	BANKX = COS(bank angle) *$4000,
	\	BANKY = SIN(bank angle) *$4000,
	SPINY:= BANKY *BANKX /$20000;		\Divide by scale and rate adj
	SPINX:= BANKY *BANKY /$20000 + PITCHRATE;

	BITMAP(RASTER1, WIDTH, HEIGHT, DEPTH);
	WAITVB;			\Make sure RASTER0 is actually being displayed
	CLEAR;			\ before clearing RASTER1
	ROTATE;
	DOTHROTTLE;
	PERSPECTIVE;
	DOTHROTTLE;
	DRAW;
	VIEW(RASTER1, $9004);
	end;
end;	\FLY

\----------------------------------------------------------------------

begin	\MAIN
REM0:= RESERVE(3 *INTSIZE);
REM1:= RESERVE(3 *INTSIZE);
REM2:= RESERVE(3 *INTSIZE);
REM3:= RESERVE(3 *INTSIZE);
REM4:= RESERVE(3 *INTSIZE);
REM5:= RESERVE(3 *INTSIZE);
for II:= 0, 2 do
	begin
	REM0(II):= 0;
	REM1(II):= 0;
	REM2(II):= 0;
	REM3(II):= 0;
	REM4(II):= 0;
	REM5(II):= 0;
	end;

RASTER0:= RESERVE(WIDTH *HEIGHT /8 *DEPTH);
RASTER1:= RESERVE(WIDTH *HEIGHT /8 *DEPTH);

REFPX:= RESERVE(3 *INTSIZE);
REFPY:= RESERVE(3 *INTSIZE);
REFPZ:= RESERVE(3 *INTSIZE);

PX:= RESERVE(42 *5 *INTSIZE);
PY:= RESERVE(42 *5 *INTSIZE);
PZ:= RESERVE(42 *5 *INTSIZE);
PU:= RESERVE(42 *5);

PX0:= RESERVE(42 *5 *INTSIZE);
PY0:= RESERVE(42 *5 *INTSIZE);
PZ0:= RESERVE(42 *5 *INTSIZE);

PPX:= RESERVE(42 *5 *INTSIZE);
PPY:= RESERVE(42 *5 *INTSIZE);

LMOVE:= RESERVE(45 *5);
LBEGIN:= RESERVE(45 *5);
LEND:= RESERVE(45 *5);
LCOLOR:= RESERVE(45 *5);

OPENI(0); OPENO(0);

CHOUT(0,$0C);
TEXT(0, "-- FLIGHT SIMULATOR, V0.0x20 --
");

BITMAP2(0,0,320,240,false,false);

PALETTE(0, $0000);		\Black
PALETTE(1, $000F);		\Blue

MAKEWORLD;
ANIMRATE:= 0;
FLY;

CURSOR(0,23);
end;	\MAIN
