	page	59,130
;GETTIME.ASM	JUL-25-99
;Sub-Microsecond Precision Timer by Loren Blaney
;
;Revisions:
;JAN-09-96, Original
;JUL-25-99, Modify for XPLP


	.386
	.387
cseg	segment dword public use32 'code'
	assume	cs:cseg

;-------------------------------------------------------------------------------
;Each count of the system timer (timer 0 in the 5254 chip) is: 1 / 1.19318MHz
; = 838 ns. The BIOS tick occurs every 838 * 65535 = 54,919,168 ns, or
; 1,000,000,000 / 54,919,168 = 18.2 times per second.
;
;Example of how to use this routine:
; Start:= GetTime;
; for I:= 1, 1000 do Something;
; End:= GetTime;
; Seconds:= Float(End-Start) * 8.380965E-7;
; OperationsPerSecond:= 1.19318E6 / Float(End-Start);
;
;WARNING: This assumes that timer 0 is running in its normal range of 0-ffff;
; some versions of BASIC change this range.
;
;eax and edx are destroyed.
;
	public	GetTime
GetTime:push	bx		;save registers
	push	cx

	mov	ax,fs:046ch	;get BIOS tick count (1/18th second resolution)
	shl	eax,16		;shift into high 16 bits
	mov	bx,ax		;save a copy

	call	get8254		;read 8254 timer 0

	mov	cx,fs:046ch	;get BIOS tick count again
	cmp	bx,cx		;jump if there is no change, otherwise a carry-in
	je	gt10		; occurred and timer 0 must be reread to get the
	mov	ax,cx		; correct value.
	shl	eax,16		;shift BIOS count into high 16 bits
	call	get8254		;re-read 8254 timer 0 because of carry-in
gt10:
	xchg	al,ah		;untangle the timer 0 value
	neg	ax		;make it ascending

	pop	cx		;restore registers
	pop	bx

	pop	edx		;pop return address
	push	eax		;push time
	push	edx
	ret

;-------------------------------------------------------------------------------
;Read the system timer: timer 0 in the 8254 counter/timer chip.
;
;Ports $40-$42 access one of the three counters, and port $43 accesses the
; control register. The bits in the control register are:
;
;	7-6	Select counter (0-2)
;	5-4	Select write/read operation
;	3-1	Mode (0-5)
;	0	Use BCD
;
get8254:mov	al,00h		;latch counter 0
	out	43h,al
	jmp	short $+2	;delay for flaky ATs
	in	al,40h		;read counter 0 latch (low byte 1st)
	mov	ah,al
	jmp	short $+2	;delay
	in	al,40h		;read high byte
	ret

cseg	ends
	end
