\HANOID0.XPL	AUG-20-88
\This program solves the tower of Hanoi puzzle, which involves 3 needles
\ and a stack of disks of varying diameters arranged such that they form a
\ cone on the left needle. The problem is to move the stack to the right
\ needle. The disks must be moved one at a time, and a larger disk can
\ never be placed on top of a smaller disk. The needle in the center can be
\ used as a temporary place to put disks.
\
\This program uses recursion to solve the puzzle.
\
\REVISION HISTORY:
\NOV-03-81, Original, by Richard Willmore
\FEB-18-86, Modified for 32-bit XPL on the Stride, Loren Blaney.
\AUG-15-88, Added character graphics for the Wyse terminal on the Stride.
\AUG-20-88, Modified for Videx character graphics on the Apple.

\IBM PC XPL CODE SET FOR XPL V5.6I

code

ABS=0,		RAN=1,		REM=2,		RESERVE=3,
SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
ABORT=16,	TRAP=17,	FREE=18,	RERUN=19,
GETHP=20,	SETHP=21,	GETERR=22,	CURSOR=23,
FSET=24,	SETRUN=25,	HEXIN=26,	HEXOUT=27,
CHAIN=28,	FOPEN=29,	WRITE=30,	READ=31,
FCLOSE=32,	CHKKEY=33,	SOFTINT=34,	GETREG=35,
BLIT=36,

SOUND=39,	CLEAR=40,
POINT=41,	LINE=42,	MOVE=43,	READPIX=44,
SETVID=45;


def	BEL=$07, TAB= $09, FF= $0C, CR= $0D, EOF=$1A, SP= $20; \ASCII characters
def	DISKCH= $87, NEEDLCH= $8B, BASECH= $8F;
def	WX0= 10, WY0= 2,	\Window 1 location
	WX1= 51, WY1= 2,	\Window 2 location
	BX= 5, BY= 20;		\Base location

int	DISKS0,		\Total number of disks
	LEVEL,		\Level of recursion in MOVETOWER, determines RADIUS
	MOVECNT,	\Count of the number of disk moves
	TIME,		\Amount of time to delay
	VOLUME,		\Turns sound on (1) or off (0)
	;
addr	TOWERX, TOWERY	\Arrays: Coordinates of top of each of 3 towers
	;

\------------------------ I/O ROUTINES FOR APPLE ----------------------



func	GETKEY;		\ROUTINE TO READ A CHARACTER FROM THE THE APPLE KEYBOARD
int	CH;
addr	ADDR;
begin
repeat until CHKKEY;
CH:=CHIN(1);
return CH;
end;	\GETKEY



proc	CHOUTX(CH);
int	CH;
int	TBL;
begin
if CH >=$80 & CH <=$8F then		\CHANGE TO VIDEX CHARACTERS
	begin
	TBL:= [	$00C2, $00C0, $00DA, $00BF, $00C3, $00D9, $00B3, $00DB,
		$00C5, $00B4, $00C4, $00B1, $00CD, $00C1, $00BA, $00B2];
	CH:= TBL(CH -$80);
	end;
CHOUT(0, CH);
end;	\CHOUTX

\----------------------------------------------------------------------

proc	EXIT;		\Exit program in a tidy fashon
begin
CLEAR;
exit;
end;	\EXIT



proc	CTXT(X, Y, STR);
int	X, Y;
addr	STR;
begin
CURSOR(X, Y);
TEXT(0, STR);
end;	\CTXT



proc	BOX(X, Y, W, H);	\Draw a box
int	X, Y,		\Coordinates of upper left corner
	W, H;		\Width and height of box (|_| is 3 wide)
int	I, J;
begin
CURSOR(X, Y);   CHOUTX($82);		\Upper left corner
for I:= 1, W-2 do			\Top line
	CHOUTX($8A);
CHOUTX($83);				\Upper right corner

for J:= 1, H-2 do
	begin
	CURSOR(X, Y+J);   CHOUTX($86);		\Left side
	CURSOR(X+W-1, Y+J);   CHOUTX($86);	\Right side
	end;

CURSOR(X, Y+H-1);   CHOUTX($81);	\Lower left corner
for I:= 1, W-2 do			\Bottom line
	CHOUTX($8A);
CHOUTX($85);				\Lower right corner
end;	\BOX

\----------------------------------------------------------------------

proc	BINOUT(DEV, N, SIZE, FILL);
\Display N in binary on the specified device.
\This always displays the correct value of N regardless of the field size.
int	DEV,		\Output device number
	N,		\Unsigned, 15-bit integer
	SIZE,		\Size of field in characters (right-justifies number)
	FILL;		\Character to fill the background of the field
int	I,		\Index (bits are counted backwards)
	J,		\Index to first bit in the field
	MASK,		\Mask used to test a bit in N
	FLAG;		\Flag: Display zeros, used to suppress leading zeros
begin
MASK:= $4000;				\Initialize mask to first bit
J:= 15 -SIZE;
FLAG:= false;
for I:= 0, 13 do
	begin
	if MASK & N then		\Bit is set
		begin
		CHOUT(DEV, ^1);
		FLAG:= true;		\Don't suppress zeros anymore
		end
	else	begin
		if FLAG then CHOUT(DEV, ^0)
		else	begin
			if I >= J then		\Leading zero is within field size
			CHOUT(DEV, FILL);	\Replace it with background
			end;
		end;
	MASK:= MASK /2;			\Shift mask to next bit
	end;
CHOUT(DEV, if MASK & N then ^1 else ^0);	\Last bit is always displayed
end;	\BINOUT

\----------------------------------------------------------------------

proc	DELAY;		\Time delay set by keys 0 through 9
\Inputs: TIME
int	C, I, J;
begin
if CHKKEY then
	begin
	C:= GETKEY;
	if C=\CTRL-C\$03 then EXIT;
	if C=\CTRL-S\$13 then [repeat until CHKKEY;   I:= GETKEY];
	if C=SP then VOLUME:= not VOLUME &1;	\Toggle sound on and off
	if C>=^0 & C<=^9 then
		TIME:= C -^0;
	end;
for I:= 1, TIME do
	for J:= 1, 100 do;
end;	\DELAY

\======================================================================

proc	SHOWMOVECNT;		\Show move counter value
\Inputs: MOVECNT, DISKS0
begin
CURSOR(WX1+5, WY1+1);
if MOVECNT<10 then CHOUT(0, SP);
if MOVECNT<1000 then CHOUT(0, SP);
CHOUT(0, SP);			\Erase old number
INTOUT(0, MOVECNT);
CHOUT(0, SP);

CURSOR(WX1+8-(DISKS0/2), WY1+2);
BINOUT(0, MOVECNT, DISKS0, ^0);
end;	\SHOWMOVECNT



proc	SHOWDISK(X, Y, R, CH);	\Show (or erase) a disk
int	X, Y,	\Coordinates of center
	R,	\Radius
	CH;	\DISKCH or space
int	I;
begin
CURSOR(X-R, Y);
for I:= 1, R do CHOUTX(CH);
CURSOR(X+1, Y);
for I:= 1, R do CHOUTX(CH);
end;	\SHOWDISK

\----------------------------------------------------------------------

proc	MOVEDISK(TAKEOFF, PUTON);
\Move a disk from one needle (TAKEOFF) to another (PUTON)
int	TAKEOFF,	\Needle numbers (0..2)
	PUTON;
int	RADIUS, X, Y, DIR, ALT;
begin
RADIUS:= DISKS0 -(LEVEL-1);
ALT:= 17 -DISKS0;

\Raise disk up to top of needle:
\Move from TOWERY(TAKEOFF) up to ALT
X:= TOWERX(TAKEOFF);   Y:= TOWERY(TAKEOFF);
\Take a disk off the top:
TOWERY(TAKEOFF):= TOWERY(TAKEOFF) +1;		\Move down one
loop	begin
	if Y <= ALT then quit;
	DELAY;   DELAY;				\Double delay for vertical motion
	SHOWDISK(X, Y, RADIUS, SP);		\Blank out the disk
	Y:= Y -1;
	SHOWDISK(X, Y, RADIUS, DISKCH);		\Show disk in new position
	end;

\Move disk over to PUTON needle:
DIR:= if X < TOWERX(PUTON) then +1 else -1;	\Move right or left?
loop	begin
	if X = TOWERX(PUTON) then quit;
	DELAY;
	SHOWDISK(X, Y, RADIUS, SP);		\Blank out the disk
	X:= X +DIR;
	SHOWDISK(X, Y, RADIUS, DISKCH);		\Show disk in new position
	end;

TOWERY(PUTON):= TOWERY(PUTON) -1;		\Move up one
\Lower disk onto PUTON stack:
loop	begin
	if Y>= TOWERY(PUTON) then quit;
	DELAY;   DELAY;				\Double delay for vertical motion
	SHOWDISK(X, Y, RADIUS, SP);		\Blank out the disk
	Y:= Y +1;
	SHOWDISK(X, Y, RADIUS, DISKCH);		\Show disk in new position
	end;

SOUND(VOLUME, 2, 400*RADIUS);			\Clink

MOVECNT:= MOVECNT +1;				\Count this move
SHOWMOVECNT;
end;	\MOVEDISK

\----------------------------------------------------------------------

proc	MOVETOWER(DISKS, FROM, TO, USING);
\Move the tower from one needle to another
int	DISKS,		\Number of disks on the (sub)tower
	FROM,		\Needle numbers (0..2)
	TO,
	USING;		\Use this needle as a temporary place
begin
LEVEL:= LEVEL +1;			\Next level down
if DISKS >0 then
	begin					\At level 1 the needles are:
	MOVETOWER(DISKS-1, FROM, USING, TO);	\Move sub-tower from 0 to 1
	MOVEDISK(FROM, TO);			\move disk from 0 to 2
	MOVETOWER(DISKS-1, USING, TO, FROM);	\Move sub-tower from 1 to 2
	end;
LEVEL:= LEVEL -1;
end;	\MOVETOWER

\----------------------------------------------------------------------

proc	INITRUN;	\Initialize for one run of the program
int	I, J,
	ROW, RADIUS;
begin
LEVEL:= 0;   MOVECNT:= 0;

loop	begin
	CTXT(WX0+2, WY0+1, "  HOW MANY   ");	\Also blank "CONTROL SPEED"
	CTXT(WX0+2, WY0+2, "   DISKS?    ");	\ message (if any)
	CURSOR(WX0+12, WY0+2);
	DISKS0:= INTIN(0);
	if DISKS0 <= 0 then EXIT;
	if DISKS0 <= 10 then quit;
	CHOUT(0, BEL);
	end;

CTXT(WX0+2, WY0+1, "KEYS 0 THRU 9");
CTXT(WX0+2, WY0+2, "CONTROL SPEED");
CTXT(WX1+2, WY1+1, "             ");		\Blank move count (if any)
CTXT(WX1+2, WY1+2, "             ");
SHOWMOVECNT;

\Set the top of each tower:
TOWERY(0):= BY -DISKS0;			\Left tower is full
TOWERY(1):= BY;				\Other towers are empty
TOWERY(2):= BY;

\     |_									_|
\       ----------|---------- ----------|---------- ----------|----------
\_____==##########.##########.##########.##########.##########.##########==

for J:= 9, BY-4 do			\Erase old disks and needles if any
	begin
	CURSOR(BX, J);
	for I:= 1, 69 do
		CHOUT(0, SP);
	DELAY;
	end;
for J:= BY-3, BY-1 do			\Erase old disks if any
	begin
	CURSOR(BX, J);
	for I:= BX, TOWERX(0)-1 do
		CHOUT(0, SP);
	CHOUTX(NEEDLCH);
	for I:= TOWERX(0)+1, TOWERX(1)-1 do
		CHOUT(0, SP);
	CHOUTX(NEEDLCH);
	for I:= TOWERX(1)+1, TOWERX(2)-1 do
		CHOUT(0, SP);
	CHOUTX(NEEDLCH);
	for I:= 1, 12 do
		CHOUT(0, SP);
	DELAY;
	DELAY;
	end;

\Show the initial towers and needles:
ROW:= BY;				\From the bottom up
for I:= 1, DISKS0 do
	begin
	RADIUS:= DISKS0 -(I-1);
	ROW:= ROW -1;
	CURSOR(TOWERX(0), ROW);   CHOUTX(NEEDLCH);
	SHOWDISK(TOWERX(0), ROW, RADIUS, DISKCH);
	CURSOR(TOWERX(1), ROW);   CHOUTX(NEEDLCH);
	CURSOR(TOWERX(2), ROW);   CHOUTX(NEEDLCH);
	SOUND(VOLUME, 2, 400*RADIUS);
	DELAY;
	DELAY;
	end;
ROW:= ROW -1;
CURSOR(TOWERX(0), ROW);   CHOUTX(NEEDLCH);
CURSOR(TOWERX(1), ROW);   CHOUTX(NEEDLCH);
CURSOR(TOWERX(2), ROW);   CHOUTX(NEEDLCH);

for I:= 1, 10 do DELAY;
end;	\INITRUN

\======================================================================

proc	INITALL;	\Initialize the entire program
int	I, J,
	;
begin
TIME:= 5;				\Set time delay to average
VOLUME:= 1;				\Turn the sound on

CLEAR;
for I:= 1, 12 do			\Zoom box
	begin
	BOX(40-3*I, 12-I, 6*I, 2*I);
	I:= I +1;			\Speed it up
	end;
CLEAR;
BOX(0, 0, 79, 24);

CURSOR(0, 0);
for I:= 0, 78 do
	CHOUTX($8C);
CTXT(32, 0, "  H A N O I D  ");

BOX(WX0, WY0, 17, 4);
BOX(WX1, WY1, 17, 4);

for J:= 0, 1 do				\Draw base
	begin
	CURSOR(BX, BY+J);
	for I:= 1, 69 do
		CHOUTX(BASECH);
	end;

\Set the center of each tower:
TOWERX(0):= 17;
TOWERX(1):= 39;
TOWERX(2):= 61;

for J:= BY-3, BY-1 do
	begin
	CURSOR(TOWERX(0), J);   CHOUTX(NEEDLCH);
	CURSOR(TOWERX(1), J);   CHOUTX(NEEDLCH);
	CURSOR(TOWERX(2), J);   CHOUTX(NEEDLCH);
	end;
end;	\INITALL

\----------------------------------------------------------------------

begin	\MAIN
TOWERY:= RESERVE(3);		\For 3 needles
TOWERX:= RESERVE(3);
INITALL;
loop	begin
	INITRUN;			\Get number of DISKS, etc.
	MOVETOWER(DISKS0, 0, 2, 1);	\Move tower from needle 0 to 2
	end;
end;	\MAIN
