\HD.XPL		OCT-10-93
\A Simple Hex/Ascii Dump Utility by Loren Blaney
\32-bit XPL0 (2.0) is required for large arrays.

inc	C:\CXPL\CODESI;		\Include code definitions for intrinsic routines

def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0a, FF=$0c,	\Control chars
	CR=$0d, EOF=$1a, Esc=$1b, Sp=$20, Ctrl=$40;
def	Up=-$48, Dn=-$50, Lt=-$4b, Rt=-$4d, PgUp=-$49,		\-Scan codes
	PgDn=-$51, End=-$4f, Home=-$47, Del=-$53;

int	Cpureg;			\Address of CPU register array (from Getreg)
def	AX, BX, CX, DX, DI, SI, BP, CF, CS, DS, SS, ES;		\Getreg registers

def	TV=6;		\Device 6 is faster than device 0
def	Lines=16;	\Number of lines shown on the display

char	File;		\Array containing the input file
int	CurShape,	\Cursor shape
	FileInx,	\Index into File
	FileSize,	\Number of bytes in the file
	HexMode,	\Flag: Hex vs. Ascii mode
	ItemsPerLine,	\16 (Hex mode) or 64 (Ascii mode)
	MaxSize;	\Maximum size of file (bytes)

\--------------------------- INPUT ROUTINES ---------------------------

func	ShiftKey;		\Returns 'true' if a shift key is down
begin
Cpureg(AX):= $0200;
SoftInt($16);
return (Cpureg(AX)&$03) # 0;
end;	\ShiftKey



func	GetKey;			\Get character from keyboard (wait if necessary)
int	Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl-C, and no cursor.
Cpureg(AX):= 0;			\Function $00
SoftInt($16);			\Call BIOS routine
Ch:= Cpureg(AX) & $ff;
if Ch = 0 then			\Return non-ascii chars as negative scan code
	Ch:= -(Cpureg(AX)>>8);
return Ch;
end;	\GetKey



func	GetKeyX;		\Get character from keyboard including
int	Ch;			\ shifted UgUp/PgDn
begin
Ch:= GetKey;
if ShiftKey then
	case Ch of
	  PgUp:	Ch:= ^9;
	  PgDn:	Ch:= ^3
	other;
return Ch;
end;	\GetKeyX



func	GetCurShape;		\Get cursor shape
begin
Cpureg(AX):= $0300;
SoftInt($10);
return Cpureg(CX);
end;	\GetCurShape



proc	SetCurShape(Shape);	\Set cursor shape
\WARNING: $0607 does not work for monochrome mode 7. Use $0b0c instead.
\ $000d gives a solid block for all modes.
int	Shape;
begin
Cpureg(AX):= $0100;
Cpureg(CX):= Shape;
SoftInt($10);
end;	\SetCurShape



proc	ShowCursor(On);		\Turn flashing cursor off and on
int	On;	\Flag: True = cursor on; False = cursor off
int	CS;	\Own-type variable to save original cursor shape
begin
CS:= [$000d];	\Default is solid block
if On then SetCurShape(CS(0))
else	[CS(0):= GetCurShape;
	SetCurShape($2000)];
end;	\ShowCursor

\-------------------------- OUTPUT ROUTINES ---------------------------

proc	Spout(Dev, N);		\Output N spaces to specified device
int	Dev, N;
int	I;
for I:= 1, N do Chout(Dev, ^ );



proc	Hex2out(Dev, N);	\Output two hex digits (a byte)
int	Dev, N;
char	HexDigit;
begin
HexDigit:= "0123456789ABCDEF ";
Chout(Dev, HexDigit(N>>4 & $0f));
Chout(Dev, HexDigit(N & $0f));
end;	\Hex2out

\----------------------------------------------------------------------

proc	ShowAscii(P);	\Display a line of ascii starting at offset P
char	P;
int	I;
\  PPPPPP:    ................................................................
begin
Spout(TV, 2);
Hex2out(TV, P>>16);			\Display the offset (P)
Hex2out(TV, P>>8);
Hex2out(TV, P);
Text(TV, ":    ");

for I:= 0, 63 do			\Display 64 bytes of ascii
	begin
	if P+I < FileSize then 
		Chout(TV, if File(P+I)<$20 then ^. else File(P+I))
	else	Chout(TV, Sp);
	end;
Spout(TV, 2);				\Blank what is left behind by hex display
Crlf(TV);
end;	\ShowAscii



proc	ShowHex(P);	\Display a line of hex/ascii, starting at P
char	P;
char	L;
int	I;
\PPPPPP:  XX XX XX XX  XX XX XX XX  XX XX XX XX  XX XX XX XX    ................
begin
L:= Reserve(16);			\Line buffer

Hex2out(TV, P>>16);			\Display the offset (P)
Hex2out(TV, P>>8);
Hex2out(TV, P);
Text(TV, ":  ");
for I:= 0, 15 do			\Display 16 hex bytes
	begin
	if P+I < FileSize then 
		[L(I):= File(P+I);
		Hex2out(TV, L(I))]
	else	Spout(TV, 2);
	Chout(TV, Sp);
	if (I+P & 3) = 3 then Chout(TV, Sp);
	end;
Spout(TV, 2);

for I:= 0, 15 do			\Display the corresponding ascii
	begin
	if P+I < FileSize then 
		Chout(TV, if L(I)<$20 then ^. else L(I))
	else	Chout(TV, Sp);
	end;
Crlf(TV);
end;	\ShowHex



proc	ShowPage(P);	\Display a page of hex/ascii starting at pointer P
char	P;
int	J;
begin
Cursor(0, 6);
for J:= 0, Lines-1 do
	begin
	if P < FileSize then
		[if HexMode then ShowHex(P) else ShowAscii(P)]
	else	[Spout(TV, 79);   Crlf(TV)];
	if ChkKey then return;
	P:= P +ItemsPerLine;
	end;
end;	\ShowPage

\----------------------------------------------------------------------

proc	ReadFile;	\Read input file from disk
int	I;
begin
Trap(false);				\We will handle read beyond end of file
I:= 0;
loop	begin				\Read until EOF or array full
	File(I):= Chin(3);
	if Geterr then [I:= I -1;   quit];	\Back up over automatic Ctrl-Z
	I:= I +1;
	if I = MaxSize then quit;
	end;
Trap(true);
FileSize:= I;				\Mark end of file
end;	\ReadFile



func	OpenFile;	\Open the file on the command line for input
int	HandIn;				\File handle
char	CmdTail;			\Copy of command tail
begin
CmdTail:= Reserve($80);			\Get copy of command tail
Blit(Cpureg(11), $81, Cpureg(12), CmdTail, $7f);

Trap(false);				\We will handle any errors, thank you
HandIn:= FOpen(CmdTail, 0);		\Open file for input
Trap(true);
if Geterr then return false;		\Failed
FSet(HandIn, ^I);
Openi(3);
return true;				\Success
end;	\OpenFile

\----------------------------------------------------------------------

begin	\Main
Cpureg:= Getreg;			\Get address of CPU registers for SoftInt
MaxSize:= $40 <<12; \$40000\		\Reserve a big array to hold input file
File:= Reserve(MaxSize);

\Read in file
if not OpenFile then
	begin
	Text(0, "File not found
");	exit;
	end;
ReadFile;

ShowCursor(false);			\Turn off annoying flashing cursor

Chout(0, FF);
RawText(0,
"
  []			  H E X    D U M P    1 . 0			  []

  Arrows    (Shift) Page Up    (Shift) Page Down    Home    End    Tab    Esc 
 ");

HexMode:= true;
ItemsPerLine:= 16;
FileInx:= 0;
loop	begin
	ShowPage(FileInx);		\Display a page of hex starting at FileInx
	case GetKeyX of			\Move according to key command
	  Up:	FileInx:= FileInx -ItemsPerLine;
	  Dn:	FileInx:= FileInx +ItemsPerLine;
	  Lt:	FileInx:= FileInx -1;
	  Rt:	FileInx:= FileInx +1;
	  PgUp:	FileInx:= FileInx -Lines*ItemsPerLine;
	  PgDn:	FileInx:= FileInx +Lines*ItemsPerLine;
	  ^9:	FileInx:= FileInx -Lines*ItemsPerLine*16;
	  ^3:	FileInx:= FileInx +Lines*ItemsPerLine*16;
	  Home:	FileInx:= 0;
	  End:	FileInx:= FileSize -Lines*ItemsPerLine;
	  Tab:	[HexMode:= not HexMode;
		ItemsPerLine:= if HexMode then 16 else 64];
	  Esc, ^C-CTRL:	quit
	other	Chout(0, Bel);		\Error

	if FileInx > Filesize -Lines*ItemsPerLine then	\Check limits
		FileInx:= FileSize -Lines*ItemsPerLine;
	if FileInx < 0 then FileInx:= 0;
	end;

ShowCursor(true);			\Restore flashing cursor
end;	\Main
