\HUFFMAN.XPL		7-2-83
\HUFFMAN CODE GENERATOR
\BY LOREN BLANEY
\   THIS PROGRAM OUTPUTS THE HUFFMAN CODES (SPACE-OPTIMAL BINARY CODES
\BASED ON THE FREQUENCY OF A LETTER'S USE) FOR THE INPUT FILE.
\THIS PROGRAM DETERMINES WHAT THE CODES SHALL BE. IT IS THE JOB OF A
\SEPARATE PROGRAM (THE ENCODER) THAT TRANSLATES THE ASCII INTO THESE
\CODES.

code	REM=2 RESERVE=3, CHIN=7, CHOUT=8, CRLF=9 INTIN=10, INTOUT=11,
	TEXT=12, OPENI=13, OPENO=14, CLOSE=15, FREE=18;

def	TV=0, KB=0, FILE=3;	\I/O DEVICE NUMBERS
def	BS=$08, TAB=$09, LF=$0A, CR=$0D, EOF=$1A, SP=$20;
def	NIL=0;	\EMPTY NODE

def	\NODE\	\COMPONENTS OF A NODE
	LINK,	\POINTER TO NEXT NODE IN LINKED LIST (MUST BE 1ST, I.E. = 0)
	FREQ,	\FREQUENCY OF CODED CHARACTER USAGE
	LETTER,	\CHARACTER TO BE CODED
	LEFT,	\POINTER TO LEFT BRANCH OF THE TREE
	RIGHT;	\POINTER TO THE RIGHT BRANCH OF THE TREE

int	TOPMEM,	\TOP OF USABLE MEMORY
	RAM,	\POINTER TO ALLOCATED HEAP SPACE
	DEV,	\THE OUTPUT DEVICE NUMBER
	COUNTERS, \ARRAY OF LETTER FREQUENCIES
	LIST;	\HUFFMAN TREE (INITIALLY A LINKED LIST)
char	OUTBUF;	\SPECIAL CHARACTER OUTPUT BUFFER TO SHOW TREE
def	BUFMAX=40;	\LAST ELEMENT OF "OUTBUF"
int	BUFPTR;	\POINTER TO THE NEXT CHARACTER IN OUTPUT BUFFER

\----------------------------------------------------------------------

proc	COUNT_CHARS;
\COUNTS THE OCCURRENCES OF ALL CHARACTERS IN THE INPUT FILE.
\INPUT: INPUT FILE
\OUTPUT: COUNTERS ARRAY
int	I;
begin
for I:=0,$7F do COUNTERS(I):=0;		\INITIALIZE COUNTERS
repeat	I:=CHIN(FILE) & $7F;
	COUNTERS(I):=COUNTERS(I)+1;
until I=EOF;
end;	\COUNT_CHARS

\----------------------------------------------------------------------

func	ALLOCATE(AMOUNT);
\ALLOCATES THE NUMBER OF BYTES OF HEAP SPACE CONTAINED IN "AMOUNT".
\(UNLIKE THE RESERVE INTRINSIC THIS FUNCTION DOES NOT RELINQUISH
\ THE RESERVED HEAP SPACE WHEN IT RETURNS.)
int	AMOUNT;
int	TEMP;
begin
TEMP:=RAM;
RAM:=RAM + AMOUNT;	\RESERVE BYTES
if RAM>TOPMEM then
	[CRLF(0); TEXT(0,"OUT OF MEMORY"); exit];
return TEMP;	\THE ADDRESS OF THE ALLOCATED SPACE
end;	\ALLOCATE

\----------------------------------------------------------------------

proc	MAKE_LEAVES;	\CONVERT "COUNTERS" INTO A LINKED LIST
int	I,N;
begin
LIST:=NIL;
for I:=0,$7F do if COUNTERS(I)>0 then
	begin
	N:=ALLOCATE(10);	\MAKE A NEW ENTRY
	N(LINK):=LIST;		\INSERT IT INTO THE LIST
	LIST:=N;
	N(FREQ):=COUNTERS(I);
	N(LETTER):=I;
	N(LEFT):=NIL;
	N(RIGHT):=NIL;
	end;
end;	\MAKE_LEAVES

\----------------------------------------------------------------------

proc	MAKE_TREE;	\CREATE A HUFFMAN CODING TREE
int	SMALL, PNTR, NODE;



proc	SMALLEST;
\SCAN THE LINKED LIST FOR THE SMALLEST LETTER FREQUENCY.
\OUTPUTS: PNTR - POINTER TO THE NODE PRECEDING THE NODE WITH
\		 THE SMALLEST FREQUENCY.
\	  SMALL - THE VALUE OF THE SMALLEST FREQUENCY.
int	N;
begin
SMALL:=32767;		\INITIALIZE AT BIGGEST
N:=addr LIST;	\THE 1ST TIME:  N(LINK) = LIST
while N(LINK)#NIL do	\FOLLOW THE LINKAGES
	begin
	if N(0,FREQ)<SMALL then
		begin
		SMALL:=N(0,FREQ);
		PNTR:=N;
		end;
	N:=N(LINK);
	end;
end;	\SMALLEST



begin	\MAKE_TREE
while LIST(0)#NIL do	\I.E. MORE THAN ONE ITEM IN LIST
	begin
	NODE:=ALLOCATE(10);	\MAKE A NEW NODE
	SMALLEST;	\FIND THE SMALLEST FREQ IN THE LIST
	NODE(LEFT):=PNTR(0);	\POINT TO IT
	NODE(FREQ):=SMALL;
	PNTR(0):=PNTR(0,0);	\REMOVE SMALLEST NODE FROM LIST

	SMALLEST;	\DO SAME THING FOR THE RIGHT BRANCH
	NODE(RIGHT):=PNTR(0);
		\ NODE FREQ IS SUM OF TWO SMALLEST FREQS:
	NODE(FREQ):=NODE(FREQ) + SMALL;
	PNTR(0):=PNTR(0,0);
	NODE(LINK):=LIST;	\INSERT NEW NODE AT HEAD OF LIST
	LIST:=NODE;
	end;
end;	\MAKE_TREE

\----------------------------------------------------------------------

proc	BUFOUT(CH);	\OUTPUT CHAR USING SPECIAL LINE BUFFER
int	CH;
begin
	if CH=BS then
		[if BUFPTR>0 then BUFPTR:=BUFPTR-1]
	else	[OUTBUF(BUFPTR):=CH;
		BUFPTR:=BUFPTR+1];
end;	\BUFOUT

\----------------------------------------------------------------------

proc	DUMPBUF(CH);	\DUMP THE OUTPUT BUFFER TO THE OUTPUT DEVICE
\SHOWS THE HUFFMAN CODE, THE CHARACTER, AND THE NUMBER OF OCCURRENCES
int	CH;
int	I, CTRLTBL;
begin
INTOUT(DEV,CH); CHOUT(DEV,TAB);		\PUT OUT DECIMAL ASCII AND
INTOUT(DEV,BUFPTR); CHOUT(DEV,TAB);	\ BIT COUNT FOR OTHER PROGRAMS

CTRLTBL:=[
	"NUL", "SOH", "STX", "ETX", "EOT", "ENQ", "ACK", "BEL",
	"BS ", "TAB", "LF ", "VT ", "FF ", "CR ", "SO ", "SI ",
	"DLE", "DC1", "DC2", "DC3", "DC4", "NAK", "SYN", "ETB",
	"CAN", "EM ", "EOF", "ESC", "FS ", "GS ", "RS ", "US ",
	"SP "];
case of
  CH<=$20:	TEXT(DEV,CTRLTBL(CH));
  CH=$7F:	TEXT(DEV,"DEL")
else	[CHOUT(DEV,CH); TEXT(DEV,"  ")];
TEXT(DEV,"  "); INTOUT(DEV,COUNTERS(CH));
TEXT(DEV,"   "); CHOUT(DEV,TAB);

loop for I:=0,BUFMAX do
	[if I=BUFPTR then quit else CHOUT(DEV,OUTBUF(I));
	if REM(I/4)=3 then CHOUT(DEV,SP)];
CRLF(DEV);
BUFPTR:=BUFPTR+1;	\COMPENSATE FOR BS IN TREEOUT
end;	\DUMPBUF

\----------------------------------------------------------------------

proc	TREEOUT(PNTR);	\OUTPUT THE HUFFMAN CODES IN TREE FORM
int	PNTR;
begin	\TREEOUT
if PNTR(RIGHT)=NIL then DUMPBUF(PNTR(LETTER))
else	[BUFOUT(^0);
	TREEOUT(PNTR(RIGHT));
	BUFOUT(BS);
	BUFOUT(^1);
	TREEOUT(PNTR(LEFT))];
	BUFOUT(BS);
end;	\TREEOUT

\----------------------------------------------------------------------

begin	\MAIN
OUTBUF:=RESERVE(BUFMAX+1);
COUNTERS:=RESERVE(2*$80);	\FOR ALL ASCII CHARS
RAM:=RESERVE(FREE-500);		\RESERVE MEMORY FOR TREE AND LISTS
TOPMEM:=RESERVE(0) - 10;	\POINTER TO TOP OF USABLE MEMORY

TEXT(TV,"-- HUFFMAN CODE GENERATOR, VER 1.1 --
");
TEXT(TV,"OUTPUT DEVICE? ");
DEV:=INTIN(KB);
OPENO(DEV);

OPENI(FILE);
COUNT_CHARS;
COUNTERS(LF):=0; COUNTERS(EOF):=0;	\WE DON'T WANT THESE CHARACTRS
\COUNTERS(CR):=0;

MAKE_LEAVES;
MAKE_TREE;
BUFPTR:=0;	\INITIALIZE SPECIAL BUFFER POINTER
TREEOUT(LIST);
CLOSE(DEV);
end;	\MAIN
