	page	55,136
;IMAGE.ASM	NOV-20-93

;Image save and restore routines. Based on code in "Programmer's Guide to
; PC & PS/2 Video Systems" by Richard Wilton.

cseg	segment dword public 'code'
	assume cs:cseg

;----------------------------------------------------------------------
;Save an image with specified area into memory. Works for video modes
; 0eh-12h. (If mode 10h, EGA card must have at least 128K of video RAM.)
; GetImage(X0, Y0, X1, Y1, BufSeg);
;
;Save a mask for an image. A mask is required to remove the background
; before writing an image with PutImage. To remove the background, use the
; AND operation and to write the image use the OR operation. For example:
;	GetImage(X0, Y0, X1, Y1, ImageBuf);	\Get the image
;	GetMask(X0, Y0, X1, Y1, MaskBuf);	\Get a mask for the image
;	PutImage(X, Y, MaskBuf, 1);		\1 = AND operation
;	PutImage(X, Y, ImageBuf, 2);		\2 = OR operation
;
argx0	equ	word ptr [bp]+14	;coordinates of upper-left corner
argy0	equ	word ptr [bp]+12	; of rectangular area to save
argx1	equ	word ptr [bp]+10	;lower-right corner
argy1	equ	word ptr [bp]+8
bufSeg	equ	word ptr [bp]+6		;segment address of RAM location

pixRows	equ	word ptr [bp]-2		;number of pixel rows (= y1 -y0 +1)
pixRowLen equ	word ptr [bp]-4		;approx no. of bytes in a row for image

bytesPerRow	equ	80	;number of bytes in a row on the screen

	public	getImage
getImage:
	mov	ax,0005h	;5 = index = mode register, set read mode 0
	jmp	short l00

	public	getMask
getMask:
	mov	ax,0805h	;set read mode 1 (color compare reg = 0)
l00:	mov	dx,3ceh		;graphics controller registers
	out	dx,ax

	push	es		;save registers
	mov	bp,sp
	sub	sp,4		;establish stack frame
	push	ds
	push	si
	push	di

;compute dimensions of bit block
	mov	ax,argx1
	sub	ax,argx0
	mov	cx,0ff07h	;ch = unshifted bit mask
				;cl = AND mask for al
	and	cl,al		;cl = number of pixels in last byte of row
	xor	cl,7		;cl = number of bits to shift
	shl	ch,cl		;bit mask for last byte of row
	mov	cl,ch
	push	cx

	mov	cl,3		;convert width in pixels to width in bytes
	shr	ax,cl
	inc	ax
	push	ax

	mov	ax,argy1	;calculate number of pixel rows (pixRows)
	sub	ax,argy0
	inc	ax
	push	ax

;establish addressing
	mov	ax,argy0
	mov	bx,argx0
	call	pixAddr
	xor	cl,7		;number of bits to shift left
	push	es
	pop	ds
	mov	si,bx		;ds:si -> video memory

	mov	ax,bufSeg	;es:di -> buffer in RAM
	mov	es,ax
	sub	di,di

;build 5-byte bit block header
	pop	ax
	mov	pixRows,ax
	cld
	stosw			;byte 0-1 = number of pixel rows
	pop	ax
	mov	pixRowLen,ax
	stosw			;byte 2-3 = bytes per pixel row
	pop	ax
	mov	ch,al		;bit mask for last byte in row
	stosb			;byte 4 = bit mask for last byte

;set up graphics controller
	mov	dx,3ceh		;graphics controller registers
	mov	ax,0304h	;index = 4 = read map select register
				;set up for first bit plane
;copy from video memory to system RAM
l01:	out	dx,ax		;select next memory map to read
	push	ax
	push	pixRows
	push	si		;save offset of x0,y0

l02:	mov	bx,pixRowLen
	push	si		;save si at start of pixel row

l03:	lodsw			;al = next byte in video memory
				;ah = (next byte) + 1
	dec	si		;ds:si -> (next byte) +1
	rol	ax,cl		;al = next 4 pixels in row
	stosb			;copy to system RAM
	dec	bx		;loop across row
	jnz	l03

	and	es:[di]-1,ch	;mask last byte in row
	pop	si		;ds:si -> start of row
	add	si,bytesPerRow	;ds:si -> start of next row

	dec	pixRows
	jnz	l02		;loop down rows

	pop	si		;ds:si -> start of bit block
	pop	pixRows	;restore number of pixel rows
	pop	ax		;ah = last map read
				;al = read map select register number
	dec	ah
	jns	l01		;loop across bit plane

	mov	ax,0005h	;5 = index = mode register, set read mode 0
	mov	dx,3ceh		;graphics controller registers
	out	dx,ax

	pop	di		;restore registers
	pop	si
	pop	ds
	mov	sp,bp
	pop	es
	retf	10		;discard arguments and return

;----------------------------------------------------------------------
;Output an image to the screen. Works for video modes 0eh-12h. (If mode 10h,
; EGA card must have at least 128K of video RAM.)
; PutImage(X, Y, BufSeg, Op);
;
argx	equ	word ptr [bp]+12	;upper-left corner of image is put here
argy	equ	word ptr [bp]+10
bufSeg	equ	word ptr [bp]+8		;segment address of RAM containing image
op	equ	word ptr [bp]+6		;operation: 0=copy, 1=and, 2=or, 3=xor

pixRows	equ	word ptr [bp]-2
pixRowLen	equ	word ptr [bp]-4
rowCounter	equ	word ptr [bp]-6
startMask	equ	word ptr [bp]-8
endMaskL	equ	word ptr [bp]-10
endMaskR	equ	word ptr [bp]-12

bytesPerRow	equ	80	;logical width of video screen in bytes

	public	putImage
putImage:
	push	es		;save registers
	mov	bp,sp
	sub	sp,12		;establish stack frame
	push	ds
	push	si
	push	di

;establish addressing
	mov	ax,argy
	mov	bx,argx
	call	pixAddr		;es:bx -> byte offset of x,y
	inc	cl
	and	cl,7		;number of bits to shift left
	mov	di,bx		;es:di -> X,Y in video memory

	mov	ax,bufSeg	;ds:si -> buffer in system RAM
	mov	ds,ax
	sub	si,si

;get dimensions of bit block from header
	lodsw			;number of pixel rows
	mov	pixrows,ax
	lodsw			;bytes per pixel row
	mov	pixrowlen,ax
	lodsb			;bit mask for last byte in row
	mov	ch,al

;set up graphics controller registers
	mov	dx,3ceh		;graphics controller registers
	mov	ax,op		;get operation code (copy / and / or / xor)
	and	al,3		;limit to legal range
	xchg	ah,al		;shift bits into positions 4 and 3
	shl	ah,1
	shl	ah,1
	shl	ah,1
	mov	al,3		;3 = data rotate register
	out	dx,ax

	mov	ax,0805h	;5 = mode register, read mode 1 write mode 0
	out	dx,ax

	mov	ax,0007h	;7 = color do(nt) care, reads return 0FFh
	out	dx,ax

	mov	ax,0ff08h	;8 = bit mask
	out	dx,ax

	mov	dl,0c4h		;sequencer I/O port
	mov	ax,0802h	;2 = map mask register, 1000b

	cmp	cx,0ff00h	;if mask # ffh or bits to shift # 0 then jump
	jne	l15

;routine for byte-aligned bit blocks
	mov	cx,pixRowLen
l10:	out	dx,ax		;enable one bit plane for writes
	push	ax		;save map mask value
	push	di		;save video memory offset to X,Y
	mov	bx,pixRows

l11:	push	di		;save registers
	push	cx

l12:	lodsb			;al = next byte of pixels
	and	es:[di],al	;update bit plane
	inc	di
	loop	l12

	pop	cx
	pop	di
	add	di,bytesPerRow	;es:di -> next pixel row in video memory
	dec	bx
	jnz	l11		;loop down pixel rows

	pop	di		;es:di -> video memory offset to X,Y
	pop	ax		;ah = current map mask register value
	shr	ah,1		;ah = new map mask value
	jnz	l10
	jmp	lexit

;routine for non-aligned bit blocks
l15:	push	ax		;save map mask value
	mov	bx,000ffh	;bh = 00 = mask for first byte in row
	mov	al,ch		;al = mask for last byte in pixel row
	cbw			;ah = 0ffh = mask for last-1 bytte

	cmp	pixRowLen,1
	jne	l16		;jump if more than one byte per row

	mov	bl,ch
	mov	ah,ch		;ah = mask for last-1 byte
	sub	al,al		;al = 0 = mask for last byte

l16:	shl	ax,cl		;shift masks into position
	shl	bx,cl

	mov	bl,al		;save masks
	mov	al,8		;8 = bit mask register
	mov	endMaskL,ax
	mov	ah,bl
	mov	endMaskR,ax
	mov	ah,bh
	mov	startMask,ax

	mov	bx,pixRowLen
	pop	ax		;restore map mask register value

;set pixels row by row in the bit planes
l17:	out	dx,ax		;enable one bit plane for writes
	push	ax		;save map mask value
	push	di		;save video memory offset
	mov	dl,0ceh		;dx = 3ceh = graphics controller port

	mov	ax,pixRows	;init loop counter
	mov	rowCounter,ax

;set pixels in start of row in currently enabled bit plane
l18:	push	di		;save offset of start of pixel row
	push	si		;save offset of row in bit block
	push	bx		;save bytes per pixel row

	mov	ax,startMask
	out	dx,ax

	lodsw			;two bytes of pixels
	dec	si		;ds:si -> second byte of pixels
	test	cl,cl
	jnz	l19		;jump if not left-aligned

	dec	bx		;bytes per row -1
	jnz	l20		;jump if at least 2 bytes per row
	jmp	short l22	;jump if only one byte per row

l19:	rol	ax,cl		;ah = left part of first byte and
				; right part of second byte
				;al = right part of first byte and
				; left part of second byte
	and	es:[di],ah	;set pixels for left part of first byte
	inc	di
	dec	bx		;bx = bytes per row -2

l20:	push	ax		;save pixels
	mov	ax,0ff08h	;set bit mask register for succeeding bytes
	out	dx,ax
	pop	ax

	dec	bx
	jng	l22		;jump if only 1 or 2 bytes in pixel row

;set pixels in middle of row
l21:	and	es:[di],al	;set pixels in right part of current byte
	inc	di		; and left part of next byte

	lodsw			;next word of pixels
	dec	si
	rol	ax,cl		;ah = left part of next byte and
				; right part of next+1 byte
				;al = right part of next byte and
	dec	bx		; left part of next+1 byte
	jnz	l21		;loop across pixel row

;set pixels at end of row
l22:	mov	bx,ax		;bh = right part of last byte and
				; left part of last-1 byte
				;bl = left part of last byte and
				; right part of last-1 byte
	mov	ax,endMaskL	;ah = mask for last-1 byte; al = bit mask reg
	out	dx,ax		;set bit mask register
	and	es:[di],bl	;set pixels for last-1 byte

	mov	ax,endMaskR	;mask for last byte in pixel row
	out	dx,ax		;last byte in pixel row
	and	es:[di]+1,bh	;set pixels for last byte

	pop	bx		;bytes per pixel row
	pop	si
	add	si,bx		;ds:si -> next row in bit block
	pop	di
	add	di,bytesPerRow	;es:di -> next pixel row in video memory
	dec	rowCounter
	jnz	l18		;loop down pixel rows

	pop	di		;es:di -> video memory offset to X,Y
	pop	ax		;current map mask value
	mov	dl,0c4h		;3c4h = sequencer port
	shr	ah,1		;next map mask value
	jnz	l17

;restore graphics controller and sequencer registers to their default states
lexit:	mov	ax,0f02h	;default map mask value
	out	dx,ax

	mov	dl,0ceh		;dx = 3ceh = graphics controller registers
	mov	ax,0003h	;default data rotate/function select (copy)
	out	dx,ax

	mov	ax,0005h	;default mode value
	out	dx,ax

	mov	ax,0f07h	;default color compare value
	out	dx,ax

	mov	ax,0ff08h	;default bit mask value
	out	dx,ax

	pop	di		;restore registers
	pop	si
	pop	ds
	mov	sp,bp
	pop	es
	retf	8		;discard arguments and return

;----------------------------------------------------------------------
;Return offset in video memory for pixel at X,Y
; Inputs:
;	ax = Y coordinate
;	bx = X coordinate
; Outputs:
;	ah = bit mask
;	bx = byte offset in buffer
;	cl = number of bits to shift left
;	es = video buffer segment
;
bytesPerLine	equ	80	;number of bytes in one horizontal line
originOffset	equ	0	;byte offset of (0,0)
videoMemSeg	equ	0a000h	;base segment address of video memory

pixAddr:
	mov	cl,bl		;cl = low-order byte of X
	push	dx

	mov	dx,bytesPerLine	;ax = Y * bytesPerLine
	mul	dx

	pop	dx
	shr	bx,1		;bx = X / 8
	shr	bx,1
	shr	bx,1
	add	bx,ax
	add	bx,originOffset

	mov	ax,videoMemSeg
	mov	es,ax		;es:bx = byte address of pixel

	and	cl,7		;X & 7
	xor	cl,7		;number of bits to shift left
	mov	ah,1		;unshifted bit mask
	ret

;----------------------------------------------------------------------
;Returns the number of paragraphs (16 bytes) needed to store an image.
; For video modes 0dh-12h.
; ImageSize(X0, Y0, X1, Y1);
;
argx0	equ	word ptr [bp]+10	;coordinates of upper-left corner
argy0	equ	word ptr [bp]+8		; of rectangular area to save
argx1	equ	word ptr [bp]+6		;lower-right corner
argy1	equ	word ptr [bp]+4

	public	imageSize
imageSize:
	mov	bp,sp

; ( (X1-X0+1+7)/8 * (Y1-Y0+1) * 4 + 5 + 15 ) /16
; ( ((X1-X0)/8 +1) * (Y1-Y0+1) + 5) /4
	mov	ax,argx1	; ((X1-X0)/8 +1)
	sub	ax,argx0
	shr	ax,1
	shr	ax,1
	shr	ax,1
	inc	ax

	mov	dx,argy1	; * (Y1-Y0+1)
	sub	dx,argy0
	inc	dx
	mul	dx

	add	ax,5		; + 5) /4
	adc	dx,0
	shr	dx,1
	rcr	ax,1
	shr	dx,1
	rcr	ax,1

	mov	argx0,ax	;replace first argument with size
	retf	6		;discard rest of arguments and return

cseg	ends
	end
