\LIB2.XPL	JUN-17-99
\Library of Non-Standard Routines



proc	Hex3Out(Dev, N); \Output N as three ASCII hex digits
int	Dev, N;
begin
Hex1Out(Dev, N >>8);
Hex2Out(Dev, N);
end;	\Hex3Out



proc	SetBkgndColor(X, Y, C);	\Set background color for text
int	X, Y, C;	\Text coordinates (X,Y) and background color (C)
int	AC;		\Attribute (high byte) and character (low byte)
begin
\Set cursor position
CallInt($10, $0200, $0000, 0, Y<<8!X);

\Get attribute and character at cursor
AC:= CallInt($10, $0800, $0000);

\Change background color and write attribute back at cursor
CallInt($10,   $0900 ! AC&$00FF,   AC>>8 & $0F ! C<<4,   1);
end;	\SetBkgndColor



proc	Backlight(X0, Y0, X1, Y1, C); \Hilight but don't change foreground color
int	X0, Y0, X1, Y1, C;
int	X, Y;
begin
for Y:= Y0, Y1 do
    for X:= X0, X1 do
	SetBkgndColor(X, Y, C);
end;	\Backlight



proc	Beep;		\A not-too-obnoxious beep
begin
Sound(false, 1, 1000);	\Synchronize with system timer to make tone a
Sound(true, 1, 3000);	\ consistent duration and a consistent sound.
end;	\Beep



func	MAX3(A, B, C);		\Return the largest of the arguments
int	A, B, C;
return MAX(MAX(A, B), C);




func	ILn2(N);		\Integer log to base 2 (or bit number of MSB)
int	N;			\ (this truncates, e.g: ILn2(31)=4)
int	B, I;
begin
B:= $8000;
for I:= -15, 0 do
	begin
	if N & B then return -I;
	B:= B>>1;
	end;
return -1;			\error
end;	\ILn2



func	CLIP(X, L);		\Clip +/-X to +/-L
int	X, L;
begin
if X > L then X:= L
else if X < -L then X:= -L;
return X;
end;	\CLIP



func	KEYIN;			\Input a keystroke and handle special keys
int	KEY;
begin
SHOWCURSOR(true);			\Turn on flashing cursor
KEY:= Chin(1);
SHOWCURSOR(false);			\Turn off flashing cursor
if KEY = 0 then
	KEY:= -Chin(1);		\Get scan code for non-ASCII keys
return KEY;			\ and return it as a negative number
end;	\KEYIN



func	Hex1in;			\Read in one ASCII hex digit from disk file
int	Ch;			\ (for .HEX and 6502 .BIN files)
begin
Ch:= Chin(3);
if Ch>=^0 & Ch<=^9 then return Ch -^0;
if Ch>=^A & Ch<=^F then return Ch -^A +10;
return -1;			\Error--hex digit expected
end;	\Hex1in



func	Hex2in;			\Read in two ASCII hex digits from disk file
return Hex1in<<4 +Hex1in;



func	Hex4in;			\Read in four ASCII hex digits from disk file
return Hex2in<<8 +Hex2in;



proc	STRDIV(DIVISOR, POINT);		\POINT:= POINT /DIVISOR
int	DIVISOR; 
char	POINT;
int	DIGIT, PLACE, QUOTIENT, RESIDUE;
begin
DIGIT:= 0;   NOTDONE:= FALSE;
for PLACE:= 0, SIZE do
	begin
	DIGIT:= DIGIT + POINT(PLACE);
	QUOTIENT:= DIGIT /DIVISOR;
	RESIDUE:= REM(0);
	NOTDONE:= NOTDONE ! QUOTIENT ! RESIDUE;
	POINT(PLACE):= QUOTIENT;
	DIGIT:= 10 *RESIDUE;
	end;
end;	\STRDIV



func	ATOI2(STR, SIZE);	\Convert a string of ASCII digits to an integer
\Non-digit terminates number. Only works for positive integers.
addr	STR;
int	SIZE;
int	I, NUM, DIGIT, FOUND;
begin
NUM:= 0;
FOUND:= false;
for I:= 0, SIZE-1 do
	begin
	DIGIT:= (STR(I) & $7F) - ^0;
	if DIGIT>=0 & DIGIT<=9 then
		[NUM:= NUM*10 + DIGIT;   FOUND:= true]
	else if FOUND then return NUM;
	end;
return NUM;
end;	\ATOI2



func	ATOIN(STR, SIZE);	\Convert positive string to positive integer
\This ignores decimal points, commas, leading blanks, etc.
char	STR;
int	SIZE;
int	I, NUM, DIGIT;
begin
NUM:= 0;
for I:= 0, SIZE-1 do
	begin
	DIGIT:= (STR(I) & $7F) - ^0;
	if DIGIT>=0 & DIGIT<=9 then
		NUM:= NUM*10 + DIGIT
	end;
return NUM;
end;	\ATOIN



proc	ITOA(N, STR, SIZE);	\Integer to ASCII
\This produces a string containing the ASCII representation of an integer.
\ The output string (STR) is assumed to be at least SIZE bytes long. The
\ last byte does not have the MSB set. The number is right rather than left
\ justified. SIZE must be at least 1.
int	N;	\Input integer
char	STR;	\Output string
int	SIZE;	\Number of digits allowed for (size of STR)
int	I;
begin
for I:= 0, SIZE-1 do
	[N:= N /10;
	STR(SIZE-I-1):= REM(0) +^0];
if N >= 1 then [FATAL("ITOA OVERFLOW");  return];
I:= 0;
while I<SIZE & STR(I)=^0 do [STR(I):= SP;   I:= I +1];
if I = SIZE then STR(I-1):= ^0;
end;	\ITOA



proc	ITOA2(N, STR, X);
\This routine produces a string containing the ASCII representation of an
\ integer number. The string is assumed to be at least N bytes long. The last
\ byte does NOT have the high bit set. The number is right rather than left
\ justified. If the quantity will not fit then the first byte of the string is
\ set to "*" and the remainder of the string is undefined. N must be at least
\ one.
int	N;
addr	STR;
int	X;
int	I, NEG;
begin
NEG:= X < 0;
if NEG then X:= -X;
for I:= 0, N-1 do
	begin
	X:= X /10;
	STR(N-I-1):= REM(0) +^0;
	end;
if X >= 1 then [STR(0):= ^*;   return];
I:= 0;
while I<N & STR(I)=^0 do [STR(I):= SP;   I:= I+1];
if I = N then STR(I-1):= ^0;
if NEG then
	[if I >= 1 then STR(I-1):= ^- else STR(0):= ^*];
end;	\ITOA2



func	GETDEC;			\Read in a decimal string and return its value
\The first digit is already read in, the first non-digit character (terminator)
\ will be in Char
int	VAL;
begin
VAL:= CHAR -^0;
GETCH;
loop	begin
	if CHAR<^0 ! CHAR>^9 then quit;
	VAL:= VAL *10 + CHAR - ^0;
	GETCH;
	end;
return VAL;
end;	\GETDEC



func	GETHEX;			\Return the value of a hex string, e.g: $1AC4
int	VAL, I;
begin
GETCH;					\Skip the "$"
case of
  CHAR>=^0 & CHAR<=^9: VAL:= CHAR-^0;
  CHAR>=^A & CHAR<=^F: VAL:= CHAR-$37
other [ERROR("HEX DIGIT EXPECTED");   return 0];
loop	begin
	GETCH;
	case of
	  CHAR>=^0 & CHAR<=^9: I:= CHAR-^0;
	  CHAR>=^A & CHAR<=^F: I:= CHAR-$37
	other quit;
	if VAL > $FFF then ERROR("OVERFLOW");

	VAL:= VAL<<4 +I;
	end;
return VAL;
end;	\GETHEX



proc	NUMOUT(DEV, NUM);   \Output a positive integer; suppress leading zeros
int	DEV, NUM;
int	DIV, FLAG, DIGIT;
begin
DIV:= 1000000000;		\FOR 32-BIT XPL
FLAG:= false;
repeat	DIGIT:= NUM / DIV;
	NUM:= REM(0);
	FLAG:= FLAG ! (DIGIT # 0);
	if FLAG then CHOUT(DEV, DIGIT+^0);
	DIV:= DIV / 10;
until	DIV = 1;
CHOUT(DEV, NUM+^0);
end;	\NUMOUT



proc	NUMOUT(DEV, N);	\Output a 5-digit positive integer with leading zeros
int	DEV, N;
begin
if N < 10000 then CHOUT(DEV, ^0);
if N < 1000 then CHOUT(DEV, ^0);
if N < 100 then CHOUT(DEV, ^0);
if N < 10 then CHOUT(DEV, ^0);
INTOUT(DEV, N);
end;	\NUMOUT



proc	JUSTOUT(DEV, N, PLACES); \Output a right-justified integer
int	DEV, N, PLACES;
begin
FORMAT(PLACES, 0);
RLOUT(DEV, FLOAT(N));		\(No decimal point is displayed)
end;	\JUSTOUT



proc	Binout(Dev, N, Places, Fill);	\Output N in binary (right-justified)
\Always outputs correct value of N regardless of Places
int	Dev,	\Output device number
	N,	\16-bit unsigned integer
	Places,	\Size of field in characters (right-justifies)
	Fill;	\Character to fill background of field (Usually Sp or ^0)
int	I,	\Index (bits are counted backwards)
	J,	\Index to first bit in the field
	Mask,	\Mask used to test a bit in N
	Flag;	\Flag: Output zeros, used to suppress leading zeros
begin
Mask:= $8000;					\Initialize mask to first bit
J:= 16 -Places;
Flag:= false;
for I:= 0, 14 do
	begin
	if Mask & N then			\Bit is set
		begin
		Chout(Dev, ^1);
		Flag:= true;			\Don't suppress zeros anymore
		end
	else	begin
		if Flag then Chout(Dev, ^0)
		else	begin
			if I >= J then		\Leading zero within field size
			Chout(Dev, Fill);	\Replace it with background
			end;
		end;
	Mask:= Mask >> 1;			\Shift mask to next bit
	end;
Chout(Dev, if Mask & N then ^1 else ^0);	\Last bit is always sent out
end;	\Binout



proc	DATEOUT(DEV);		\Output current date. E.g: JUN-12-90
int	DEV;			\I/O device number
int	YEAR, MO, DAY, I;
addr	MOSTR;


	proc	NUMOUT(N);	\Output 2-digit number with leading 0
	int	N;
	begin
	if N < 10 then CHOUT(DEV, ^0);
	INTOUT(DEV, N);
	end;


begin	\DATEOUT
CPUREG(AX):= $2A00;
SOFTINT($21);
DAY:= CPUREG(DX) &$FF;			\[1..31]
MO:= SWAP(CPUREG(DX)) &$FF;		\[1..12]
YEAR:= CPUREG(CX);			\[1980..2099]

MOSTR:= "JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC ";
MO:= (MO-1) *3;
for I:= 0, 2 do CHOUT(DEV, MOSTR(MO+I));
CHOUT(DEV, ^-);
NUMOUT(DAY);
CHOUT(DEV, ^-);
YEAR:= REM(YEAR/100);
NUMOUT(YEAR);
end;	\DATEOUT



func real HMSIN(Dev);
\Input HH:MM:SS from the specified device, and return the equivalent
\ hours (a real number).
int	Dev;	\Device number
int	NEG,	\Flag: Minus sign seen
	CRFLAG,	\Flag: Carriage return seen
	N,	\Integer value after the decimal point (if any)
	DEM;	\Denominator for correcting fractional values
real	X;	\Value that is returned

	func	CHINX(DEV);
	int	DEV;
	int	CH;
	begin
	if CRFLAG then return ^0;
	CH:= CHIN(KB);
	if CH=CR then CRFLAG:= true;
\	if CH=ESC ! CH=^E then ESC_FLAG:= true;
	return CH;
	end;	\CHINX


	func	GETNUM;		\Get number from device, look for minus signs
	int	CH, NUM;
	begin
	loop	begin		\Ignore leading garbage (non-numerics)
		CH:= CHINX(DEV);
		if CH>=^0 & CH<=^9 ! CH=^- then quit;
		end;
	if CH=^- then [NEG:= true;   CH:= CHINX(DEV)];

	NUM:= CH -^0;
	loop	begin
		CH:= CHINX(DEV);
		if CH<^0 ! CH>^9 ! CRFLAG then quit;
		NUM:= NUM*10 + CH-^0;
		end;

	return NUM;
	end;	\GETNUM

begin	\HMSIN
NEG:= false;
CRFLAG:= false;
X:= FLOAT(GETNUM) + FLOAT(GETNUM)/60.0 + FLOAT(GETNUM)/3600.0;
N:= GETNUM;	\Deal with wise guys who put more than one digit after the D.P.
DEM:= 1;
while N/DEM>=10 do DEM:= DEM*10;
X:= X + FLOAT(N)/36000.0/FLOAT(DEM);
return if NEG then -X else X;
end;	\HMSIN



proc	HMSOut(Dev,HMS,TENTHS);	\Output HMS in the format: hour, min, sec
\Hour must be <100.
int	Dev;	\Output device number
real	HMS;	\Value to output
int	Tenths;	\Flag to display tenths of a second
char	A;
int	HH, MM, SS;

	func	Trunc(X);	\Truncate X to integer
	real	X;
	int	I;
	begin			\(Beware of numbers like 51.9999999)
	I:= Fix(X-.5);
	if I<0 then I:= 0;	\(Beware of 0.999999999)
	return I;
	end;	\Trunc


	proc	NumOut(N);
	\Output a positive 2-digit integer with leading zeros
	int	N;
	begin
	Chout(Dev, N/10 + ^0);
	Chout(Dev, REM(0) + ^0);
	end;	\NumOut

begin	\HMSOut
if HMS<0.0 then
	[Chout(Dev, ^-); HMS:= -HMS]
else Chout(Dev, ^ );

HMS:= HMS + (if TENTHS then 1.39E-5 else 1.39E-4);	\Round

A:= addr HMS;			\Make HMS a nice number that won't
A(4):= 0;			\ give rounding problems

HH:= TRUNC(HMS);
NUMOUT(HH);
CHOUT(DEV, ^ );

HMS:= (HMS - FLOAT(HH)) * 60.0;
MM:= TRUNC(HMS);
NUMOUT(MM);
CHOUT(DEV, ^ );

HMS:= (HMS - FLOAT(MM)) * 60.0;
SS:= TRUNC(HMS);
NUMOUT(SS);

if TENTHS then			\Show tenths of seconds
	begin
	CHOUT(DEV, ^.);
	HMS:= (HMS - FLOAT(SS)) * 10.0;
	CHOUT(DEV, TRUNC(HMS) +^0);
	end;
CHOUT(DEV, CR);
end;	\HMSOUT

\-------------------------- GRAPHICS ROUTINES -------------------------

proc	LOADBMP(FNAME, IMAGE);	\Read a .BMP image file into IMAGE
char	FNAME;
seg addr IMAGE;				\320x200 image
int	HAND, I, J, JJ, R, G, B;
begin
\Set up globally:
\IMAGE:= RESERVE(2);
\IMAGE(0):= MALLOC(320/16*200);

HAND:= FOPEN(FNAME, 0);			\Graphic image
FSET(HAND, ^I);
OPENI(3);

for I:= 0, $39 do J:= CHIN(3);		\Skip header
FIXPALETTE;
for I:= 1, $FF do			\Read in colors
	begin
	B:= CHIN(3) >>2;
	G:= CHIN(3) >>2;
	R:= CHIN(3) >>2;
	J:= CHIN(3);
	SETCOLREG(I, R, G, B);		\Set color register
	end;

\I:= 320 *200;				\\WARNING: this swaps left and right
\repeat	I:= I -1;			\\Read image data
\	IMAGE(0, I):= CHIN(3);
\until I = 0;

for J:= -199, 0 do			\Read image data
	begin
	JJ:= -J *320;
	for I:= 0, 319 do IMAGE(0, JJ+I):= CHIN(3);
	\CPUREG(AX):= $3F00;
	\CPUREG(BX):= HANDLE;
	\CPUREG(CX):= 320;
	\CPUREG(DX):= -J *320;
	\CPUREG(DS):= IMAGE(0);
	\SOFTINT($21);
	end;

FCLOSE(HAND);
end;	\LOADBMP



proc	LINE(X1, Y1, X2, Y2, TYPE);
\Draw the portion of the line segment that appears in the window
\ using the Cohen-Sutherland clipping algorithm.
\ Ref: Fundamentals of Interactive Computer Graphics, J.D.Foley
\ & A. Van Dam, p149.
int	X1, Y1, X2, Y2, TYPE;
int	CODE1, CODE2;		\4-bit boolean array
int	TR, IX, IY;


	func	OUTCODE(X, Y);
	int	X, Y;
	int	CODE;
	begin
	CODE:= if Y>YHI then 1
	else if Y<YLO then 2
	else 0;
	if X > XHI then CODE:=CODE!4
	else if X < XLO then CODE:=CODE!8;
	return CODE;
	end;	\OUTCODE


	proc	XLAT(X, Y);	\Translate to hi-res coordinates
	int	X, Y;
	begin
	IX:= (X-XLO) *SCALE /XTICK;
	IY:= (if PLOT then 95 else 191) - (Y-YLO) *SCALE /YTICK;
	end;	\XLAT


begin	\LINE
if PLOT & ITOH then
	begin
	CHOUT(PR, ^M);   ITOHOUT(X1);   CHOUT(PR, ^,);   ITOHOUT(Y1);   CRLF(PR);
	CHOUT(PR, ^D);   ITOHOUT(X2);   CHOUT(PR, ^,);   ITOHOUT(Y2);   CRLF(PR);
	return;
	end;
loop	begin
	CODE1:= OUTCODE(X1, Y1);   CODE2:= OUTCODE(X2, Y2);
	if CODE1 & CODE2 then quit;		\Trivially reject
	if (CODE1 ! CODE2) = 0 then 		\Trivially accept
		begin
		XLAT(X1, Y1);
		if PLOT then [XPT:= IX;   YPT:= IY] else MOVE(IX, IY);
		XLAT(X2, Y2);
		if PLOT then PRLINE(IX, IY, TYPE) else LINE0(IX, IY, TYPE);
		quit;
		end;
	\Subdivide the line, since at most one end point is inside the window.
	\ First, if point X1,Y1 is inside the window, exchange end points
	\ (and their OUTCODEs) to guarantee that point X1,Y1 is outside
	\ the window.
	if CODE1 = 0 then			\Exchange
		begin
		TR:= X1;   X1:= X2;   X2:= TR;
		TR:= Y1;   Y1:= Y2;   Y2:= TR;
		CODE1:= CODE2;
		end;
	\Now perform a subdivision, move point X1,Y1 to the intersection
	\ point (X,Y); use the formulas:	Y = Y1 + SLOPE *(X - X1)
	\					X = X1 + (1/SLOPE) *(Y - Y1)
	case of
	    CODE1 & 1:			\Divide line at top of window
		[X1:= X1 + FIX(FLOAT(X2 - X1) *FLOAT(YHI - Y1) /FLOAT(Y2 - Y1));
		Y1:= YHI];
	    CODE1 & 2:			\Divide line at bottom of window
		[X1:= X1 + FIX(FLOAT(X2 - X1) *FLOAT(YLO - Y1) /FLOAT(Y2 - Y1));
		Y1:= YLO];
	    CODE1 & 4:			\Divide line at right edge of window
		[Y1:= Y1 + FIX(FLOAT(Y2 - Y1) *FLOAT(XHI - X1) /FLOAT(X2 - X1));
		X1:= XHI];
	    CODE1 & 8:			\Divide line at left edge of window
		[Y1:= Y1 + FIX(FLOAT(Y2 - Y1) *FLOAT(XLO - X1) /FLOAT(X2 - X1));
		X1:= XLO]
	other	[];
	end;
end;	\LINE



proc	Rectangle(X0, Y0, X1, Y1, C, F);	\Draw a rectangle
int	X0, Y0, X1, Y1, C, F;
int	I, J, T;
begin
if not F then				\If not filled then draw hollow rectangle
	begin
	MOVE(X0, Y0);
	LINE(X1, Y0, C);
	LINE(X1, Y1, C);
	LINE(X0, Y1, C);
	LINE(X0, Y0, C);
	return
	end;

\Else draw a filled rectangle:
if X1 < X0 then [T:=X1;   X1:=X0;   X0:=T];
if Y1 < Y0 then [T:=Y1;   Y1:=Y0;   Y0:=T];

if (Y1-Y0) > 8*(X1-X0) then		\If tall and skinny
	for I:= X0, X1 do		\Draw vertical lines
		[MOVE(I, Y0);   LINE(I, Y1, C)]
else	for J:= Y0, Y1 do		\Draw fast horizontal lines
		[MOVE(X0, J);   LINE(X1, J, C)];
end;	\Rectangle



proc	Circle(X0, Y0, Radius, Color, Segs);	\Draw a circle
int	X0, Y0,	\Coordinates of center
	Radius,	\Radius in C(pixels)
	Color,	\Line color
	Segs;
int	X, Y, E, U, V;

	proc	PlotOctants;
	begin						\Segment
	if Segs & $01 then Point(X0+Y, Y0+X, Color);	\ 0
	if Segs & $02 then Point(X0+X, Y0+Y, Color);	\ 1
	if Segs & $04 then Point(X0-X, Y0+Y, Color);	\ 2
	if Segs & $08 then Point(X0-Y, Y0+X, Color);	\ 3
	if Segs & $10 then Point(X0-Y, Y0-X, Color);	\ 4
	if Segs & $20 then Point(X0-X, Y0-Y, Color);	\ 5
	if Segs & $40 then Point(X0+X, Y0-Y, Color);	\ 6
	if Segs & $80 then Point(X0+Y, Y0-X, Color);	\ 7
	end;	\PlotOctants

begin	\Circle
X:= 0;
Y:= Radius;
U:= 1;
V:= 1 -Radius -Radius;
E:= 1 -Radius;
while X < Y do
	begin
	PlotOctants;
	if E < 0 then
		[U:= U +2;   V:= V +2;   E:= E +U]
	else	[U:= U +2;   V:= V +4;   E:= E +V;   Y:= Y -1];
	X:= X +1;
	end;
if X = Y then PlotOctants;
end;	\Circle



proc	Ellipse(X0, Y0, A, B, Color);	\Draw an ellipse
\This is limited to small ellipses (A=B=45) because A*A and B*B cannot exceed
\ 32767. This is not a problem with 32-bit XPL.
int	X0, Y0,		\Coordinates
	A, B,		\Width and height
	Color;
int	X, Y,
	T1, T2, T3, T4, T5, T6, T7, T8, T9,
	D1, D2;

	proc	Set4pts(X, Y, Xc, Yc, Color);
	int	X, Y, Xc, Yc, Color;
	begin
	Point(Xc+X, Yc+Y, Color);
	Point(Xc+X, Yc-Y, Color);
	Point(Xc-X, Yc+Y, Color);
	Point(Xc-X, Yc-Y, Color);
	end;	\Set4pts

begin	\Ellipse
X:= A;
Y:= 0;
T1:= A *A;
T2:= T1 +T1;
T3:= T2 +T2;
T4:= B *B;
T5:= T4 +T4;
T6:= T5 +T5;
T7:= A *T5;
T8:= T7 +T7;
T9:= 0;

D1:= T2 - T7 + T4/2;
D2:= T1/2 - T8 + T5;

\Do region 1 of quadrant
while D2 < 0 do
	begin
	Set4pts(X, Y, X0, Y0, Color);
	Y:= Y +1;
	T9:= T9 +T3;
	if D1 < 0 then
		begin
		D1:= D1 +T9 +T2;
		D2:= D2 +T9;
		end
	else	begin
		X:= X -1;
		T8:= T8 -T6;
		D1:= D1 +T9 +T2 -T8;
		D2:= D2 +T9 +T5 -T8;
		end;
	end;

\Do region 2 of quadrant
repeat	Set4pts(X, Y, X0, Y0, Color);
	X:= X -1;
	T8:= T8 -T6;
	if D2 < 0 then
		begin
		Y:= Y +1;
		T9:= T9 +T3;
		D2:= D2 +T5 +T9 -T8;
		end
	else	D2:= D2 +T5 -T8;
until	X < 0;
end;	\Ellipse

\----------------------- MISCELLANEOUS ROUTINES -----------------------

proc	Sort(A, N);		\Shell sort array in ascending order
int	A,			\Address of array
	N;			\Number of elements in array (size)
int	J, GAP, I, JG, T;
begin
Gap:= N /2;
while Gap > 0 do
	begin
	for I:= Gap, N-1 do
		begin
		J:= I -Gap;
		loop	begin
			if J < 0 then quit;
			JG:= J +Gap;
			if A(J) <= A(JG) then quit;
			T:= A(J);   A(J):= A(JG);   A(JG):= T;
			J:= J -Gap;
			end;
		end;
		Gap:= Gap /2;
	end;
end;	\Sort

\----------------------------------------------------------------------

proc	QUICKSORT;	\Quicksort SYMTBL into alphabetical order
\??? MULTIPLE LABELS ???
\Inputs: SYMTBL, SYMINX
\This routine is subtle and treacherous. You better understand it well
\ before dinging with it, or it will surly bomb.
int	TEMP, PVTLINE;
int	X, Y;
def	SYMONE=0\$9E\;	\Index of the first symbol beyond mnemonics, etc.



proc	SORT(L, R);
int	L, R;		\Left and right limits to SYMINX array
int	I, J;		\Indicies
begin
loop	begin
	I:= L;   J:= R;
	PVTLINE:= SYMINX((L+R) >>1);
	repeat	begin
		loop	begin
			X:= SYMINX(I);   Y:= PVTLINE;
			if X = Y then quit;
			while SYMTBL(X) = SYMTBL(Y) do
				[X:= X +SYMSIZE;   Y:= Y +SYMSIZE];
			if SYMTBL(X) > SYMTBL(Y) then quit;
			I:= I +1;
			end;
		loop	begin
			X:= PVTLINE;   Y:= SYMINX(J);
			if X = Y then quit;
			while SYMTBL(X) = SYMTBL(Y) do
				[X:= X +SYMSIZE;   Y:= Y +SYMSIZE];
			if SYMTBL(X) > SYMTBL(Y) then quit;
			J:= J -1;
			end;
		if I <= J then
			begin		\Swap elements
			TEMP:= SYMINX(I);
			SYMINX(I):= SYMINX(J);
			SYMINX(J):= TEMP;
			I:= I +1;
			J:= J -1;
			end;
		end;
	until I > J;

	if J-L < R-I then
		begin
		if L < J then SORT(L, J);
		if I < R then L:= I else quit;	\SORT(I, R)
		end
	else	begin
		if I < R then SORT(I, R);
		if L < J then R:= J else quit;	\SORT(L, J)
		end;
	end;
end;	\SORT



begin	\QUICKSORT
if SYMNUM > SYMONE then SORT(SYMONE, SYMNUM-1);
end;	\QUICKSORT

\----------------------------------------------------------------------

func	LOOKUP(CLASS);	\Lookup current identifier and return its index
\Other inputs: IDENT, HASH;
\Outputs: Index of identifier. Returns -1 if not found.
int	CLASS;		\Class of symbol (LAB, MN, RES)
int	I, K, INX;
begin
INX:= HASHTBL(HASH);
loop	begin
	if INX= -1 then quit;				\Not found
	I:= 0;   K:= INX;
	while IDENT(I)=SYMTBL(K) & I<SIGCHAR do
		[I:= I+1;   K:= K+SYMSIZE];
	if I=SIGCHAR & SYMCLASS(INX)=CLASS then quit;	\Found
	INX:= SYMNEXT(INX);		\Follow the linkage pointers
	end;
return INX;
end;	\LOOKUP



proc	INSERT(CLASS, VAL, TYPE);
\**** ???? reverse insertion order for maximum efficiency ????
\Insert the current identifier into the symbol table
\Other inputs: IDENT, HASH;
int	CLASS, VAL, TYPE;
int	I, INX, K, SPASS;
begin
INX:= LOOKUP(CLASS);
if INX # -1 then
	begin
	SPASS:= PASS;   PASS:= 3;	\Kludge to report error on pass 1
	ERROR("LABEL ALREADY USED");
	PASS:= SPASS;
	XREFON:= false;			\(XREF can't handle multiple labels)
	end;
if SYMNUM>=SYMSIZE then
	begin
	SPASS:= PASS;   PASS:= 3;	\Kludge to report error on pass 1
	ERROR("SYMBOL TABLE OVERFLOW");
	PASS:= SPASS;
	SYMNUM:= SYMSIZE-1;
	end;
K:= SYMNUM;
for I:= 0, SIGCHAR-1 do
	[SYMTBL(K):= IDENT(I);   K:= K+SYMSIZE];

SYMCLASS(SYMNUM):= CLASS;
SYMVAL(SYMNUM):= VAL;
SYMTYPE(SYMNUM):= TYPE;

SYMNEXT(SYMNUM):= HASHTBL(HASH);	\Link back
HASHTBL(HASH):= SYMNUM;

SYMNUM:= SYMNUM+1;
end;	\INSERT

\-----------------------------------------------------------------------

proc	COLUMNOUT(DEV, N);
int	DEV;
int	N;
int	H, B, B0, I;
begin
H:= (N+5) /6;			\Height of a column

for B0:= 0, H-1 do		\For each row
	begin
	B:= B0;			\Value at start of row
	for I:= 1, 6 do				\For 6 columns
		begin
		if B < N then			\If it exists
			begin
			INTOUT(DEV, B);
			CHOUT(DEV, 9);
			end;
		B:= B +H;			\Next column
		end;
	CRLF(DEV);
	end;
end;	\COLUMNOUT

\----------------------------------------------------------------------

proc	ProseNo(Dev, Num);	\Output the integer NUM in prose to device DEV
int	Dev, Num;
int	OneTbl, TenTbl, ThoTbl,
	ThoPwr, I, Quot;

	proc	Out999(N);
	\Output number in the range 0..999 (0 does nothing)
	int	N;
	int	Huns, Tens, Ones;
	begin
	Huns:= N /100;				\0..9
	N:= Rem(0);				\0..99
	Tens:= N /10;				\0..9
	Ones:= Rem(0);				\0..9

	if Huns # 0 then
		[Text(Dev, OneTbl(Huns));	\1..9
		Text(Dev, " hundred ")];

	if Tens >= 2 then
		begin
		Text(Dev, TenTbl(Tens));
		if Ones # 0 then
			[Chout(Dev, ^-);   Text(Dev, OneTbl(Ones))];
		end
	else	if N # 0 then Text(Dev, OneTbl(N));	\N = 1..19
	end;	\Out999

begin	\ProseNo
if Num = 0 then [Text(Dev, "zero");   return];

if Num < 0 then [Num:= -Num;   Text(Dev, "minus ")];

OneTbl:=[0, "one", "two", "three", "four",
	"five", "six", "seven", "eight", "nine",
	"ten", "eleven", "twelve", "thirteen", "fourteen",
	"fifteen", "sixteen", "seventeen", "eighteen", "nineteen"];

TenTbl:=[0, 0, "twenty", "thirty", "forty",
	"fifty", "sixty", "seventy", "eighty", "ninety"];

ThoTbl:=[" billion ", " million ", " thousand "];

\ThoPwr:= 1000000000;			\\32 BIT XPL
ThoPwr:= 1000;				\16 bit XPL
for I:= 0, 2 do				\(I:= 2, 2 if 16 bits)
	begin
	Quot:= Num /ThoPwr;
	Num:= Rem(0);
	if Quot # 0 then
		[Out999(Quot);
		Text(Dev, ThoTbl(I))];
	ThoPwr:= ThoPwr /1000;
	end;
Out999(NUM);
end;	\ProseNo



proc	BIGNUMOUT(X, Y, NUM, W, H);
\Output a right-justified number using the format: -XXX.XXX
\ Use large, seven-segment digits if W and H are not both zero.
int	X, Y;		\Initial cursor position
real	NUM,		\The number
int	W, H;		\Width and height of large digits
int	LZ, NEG, DIG, CH;


	proc	BIGDIGOUT(CH);	\Display a digit
	int	CH;
	int	JCTTBL, SEGTBL;
	addr	JCT, SEG;


		proc	HSEGOUT(N);	\Display a horizontal segment
		int	N;		\Segment number (0-6)
		int	I;
		for I:= 1, W do CHOUT(0, SEG(N));


		proc	VSEGOUT(N);	\Display a vertical segment
		int	N;		\Segment number (0-6)
		int	I;
		for I:= 1, H do
			begin
			CHOUT(0, SEG(N));
			CHOUT(0, LF);		\Move down
			CHOUT(0, BS);		\ and under
			end;


	begin	\BIGDIGOUT
	if H=0 & W=0 then
		[CHOUT(0, CH);   return];	\Display at normal size
	\Otherwise display a large, seven-segment digit

	\This table contains the graphic codes which describe the junctions
	\ for each digit.
	\	 01234567	DIGIT		ORDER OF THE JUNCTIONS
	JCTTBL:=[
		"ڿ  ",	\0			01
		"     ",	\1			   
		"ڿ  ",	\2			23
		"ڿ   ",	\3			   
		"   ",	\4			45
		"  ",	\5
		" ÿ  ",	\6
		"ڿ    ",	\7
		"ڿô  ",	\8
		"ڿ   ",	\9
		"        ",	\space
		"        "];	\minus sign

	\This table contains the graphic codes which describe the segments
	\ for each digit.
	\	 01234567	DIGIT		ORDER OF THE SEGMENTS
	SEGTBL:=[
		"ĳ  ",	\0			 0 
		"      ",	\1			1   2
		" ĳ  ",	\2			 3 
		"   ",	\3			4   5
		"    ",	\4			 6 
		"ĳ   ",	\5
		"  ĳ ",	\6
		"     ",	\7
		"ĳĳ ",	\8
		"ĳ   ",	\9
		"        ",	\space
		"       "];	\minus sign

	case CH of				\Convert ASCII to table index
	  SP:	CH:= 10;
	  ^-:	CH:= 11
	other CH:= CH -^0;

	if CH>=0 & CH <=11 then			\Decimal point just skips
		begin
		JCT:= JCTTBL(CH);		\Get string for digit
		SEG:= SEGTBL(CH);

		\Display a seven-segment digit
		CHOUT(0, JCT(0));	HSEGOUT(0);	CHOUT(0, JCT(1));
		CHOUT(0, LF);		CHOUT(0, BS);	VSEGOUT(2);
		CURSOR(X, Y+1);		VSEGOUT(1);
		CHOUT(0, JCT(2));	HSEGOUT(3);	CHOUT(0, JCT(3));
		CHOUT(0, LF);		CHOUT(0, BS);	VSEGOUT(5);
		CURSOR(X, Y+2+H);	VSEGOUT(4);
		CHOUT(0, JCT(4));	HSEGOUT(6);	CHOUT(0, JCT(5));
		end;
	X:= X +2 +W +3;				\Move to start of next digit
	CURSOR(X, Y);
	end;	\BIGDIGOUT


begin	\NUMOUT
OPENO(8);
if NUM >=0.0 then CHOUT(8, SP);
FORMAT(3, 3);
RLOUT(8, NUM);
CHOUT(8, EOF);

OPENI(8);
loop	begin
	CH:= CHIN(8);
	if CH=EOF then quit;
	BIGDIGOUT(CH);
	end;
end;	\NUMOUT



proc	DOTAB(COL);	\Move cursor to column COL on the monitor
int	COL;
int	I;
begin
I:= 0;
repeat	begin
	CHOUT(TV, ^ );
	CPUREG(0):= $0300;	\function $03
	CPUREG(1):= $0000;
	SOFTINT($10);
	I:= I +1;		\prevent potiential infinite loop
	end;
until	(CPUREG(3)&$00FF)>=COL ! I>=COL;
end;	\DOTAB



proc	OpenInFile(FN);			\Open file for input
char	FN;	\File Name string
int	H;	\Handle
begin
H:= FOpen(FN, 0);
FSet(H, ^i);				\Small buffer allows multiple files
OpenI(3);
end;	\OpenInFile



proc	OpenOutFile(FN);		\Open file for output
char	FN;	\File Name string
int	H;	\Handle
begin
H:= FOpen(FN, 1);
FSet(H, ^o);				\Small buffer allows multiple files
OpenO(3);				\Closed automatically when program exits
end;	\OpenOutFile
