\MANDEL.XPL	MAY-23-87
\Mandelbrot pattern generator
\By Loren Blaney

code
ABS=0,		RAN=1,		REM=2,		RESERVE=3,
SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
ABORT=16,	TRAP=17,	FREE=18,	RERUN=19,
GETHP=20,	SETHP=21,	GETERR=22,	CURSOR=23,
SCAN=24,	SETRUN=25,	HEXIN=26,	HEXOUT=27,
CHAIN=28,	OPENF=29,	WRITE=30,	READ=31,
TESTPT=32,	SETTXT=33,	SETHI=34,	SETMIX=35,
SETLO=36,	BUTTON=37,	PADDLE=38,	SOUND=39,
CLEAR=40,	POINT=41,	LINE=42,	MOVE=43,
BITMAP= 108,	BITMAP2= 109,	VIEW= 110,	PALETTE= 111,
GETKEY=125,	KEYHIT=126,	SHOCUR=127,	PEEK_W=113,
POKE_W=114;

addr	RASTER;
int	X, Y,			\Screen coordinates
	CORNERX, CORNERY,	\Coordinates of upper left hand corner
	II, JJ,		\Scratch
	COLORS,		\Array of palette values
	MAGNIFICATION,	\Magnification factor (power of 2)
	HIRES,		\Flag: 640x400 verses 320x200 display
	WIDTH,
	HEIGHT;



proc	SYNC;
def	INTREQR= $DFF01E, INTREQ= $DFF09C;
begin
POKE_W(INTREQ, $0020);			\Clear vertical blank bit
repeat until PEEK_W(INTREQR) & $20;	\Wait for vertical blank
end;	\SYNC



proc	MANDEL(X,Y);
int	X, Y;		\Screen coordinates of current point
int	CX, CY,		\Scaled and relocated coordinates
	COUNT,		\Iteration counter
	LIMIT,		\Size limit scaled and squared
	ASR;		\Correction factor to change LSR to ASR
reg int	ZXZX, ZYZY,	\Complex accumulator, squared values
	ZX, ZY;		\Complex accumulator

begin
LIMIT:= 1 <<(MAGNIFICATION+2);
ASR:= -1 <<(32-MAGNIFICATION);

CX:= CORNERX + X;
CY:= CORNERY + Y;

COUNT:= 0;
ZX:= 0; ZY:= 0;

\	Z and C are complex numbers
\	Z:= Z**2 + C
\	Z**2 = ZX**2 - ZY**2 + 2 *ZX *ZY
\	ZX' = ZX**2 - ZY**2 + CX
\	ZY' = 2 *ZX *ZY + CY
loop	begin
	ZXZX:= (ZX *ZX) >> MAGNIFICATION;
	ZYZY:= (ZY *ZY) >> MAGNIFICATION;

	if ZXZX + ZYZY > LIMIT then quit;	\If size of Z >2 then quit

	ZY:= ZY *ZX;
	if ZY < 0 then 
		ZY:= ZY >>MAGNIFICATION ! ASR
	else	ZY:= ZY >>MAGNIFICATION;

	ZY:= ZY + ZY + CY;
	ZX:= ZXZX - ZYZY + CX;

	COUNT:= COUNT +1;
	if COUNT > 64 then quit;	\Quit if it is in the Mandelbrot set
	end;

if COUNT < 64 then POINT(X, Y, COUNT);  \Else leave Mandelbrot point black
end;	\MANDEL



begin	\MAIN
WIDTH:= 640;			\Reserve maximum size
HEIGHT:= 400;
RASTER:= RESERVE(WIDTH *HEIGHT /8 *4);
loop	begin
	OPENI(0); OPENO(0);
	CHOUT(0,$0C);
	TEXT(0, "-- MANDELBROT PATTERN GENERATOR, V2.0x5 --

WARNING: This program will write over the RAM disk.

Do you want high resolution (N/Y)? ");
	HIRES:= (CHIN(0) ! $20) = ^y;
	TEXT(0,"
Magnification (power of two: 4-16)?	");
	MAGNIFICATION:= INTIN(0);
	TEXT(0,"
Magnified coordinates (X,Y)?		");

	CORNERX:= INTIN(0);
	CORNERY:= INTIN(0);

	WIDTH:= if HIRES then 640 else 320;
	HEIGHT:= if HIRES then 400 else 200;

	BITMAP(RASTER, WIDTH, HEIGHT, 4);
	BITMAP2(0, 0, 0, 0, false, true);
	CLEAR;

	COLORS:= [$0000, $000F, $00F0, $0F00];

	for II:= 0, 3 do
		PALETTE(II, COLORS(II));

	VIEW(RASTER, if HIRES then $A004 else $2000);	\SELECT 2 BIT PLANES
							\ TO MAINTAIN SPEED
	for Y:= 0, HEIGHT-1 do
		for X:= 0, WIDTH-1 do
			begin
			MANDEL(X,Y);
			if KEYHIT then [X:=1000; Y:=GETKEY; Y:=1000;];
			end;

	COLORS:= [$0000, $0260, $04C0,
		$00F0, $00C4, $0088,
		$004C, $000F, $040C,
		$0808, $0C04, $0F00,
		$0C40, $0880, $0CC0, $0FF0];

	for II:= 0, 15 do
		PALETTE(II, COLORS(II));

	VIEW(RASTER, if HIRES then $C004 else $4000);	\SELECT 4 BIT PLANES
	
	if GETKEY then;		\WAIT FOR KEYSTROKE

	JJ:= 0;
	loop	begin
		for II:= 1, 15 do
			[JJ:= JJ + 1;
			if JJ > 15 then JJ:= 0;
			PALETTE(II, COLORS(JJ));
			];
		JJ:= JJ +1;
		if JJ > 15 then JJ:= 0;
		JJ:= JJ +1;
		if JJ > 15 then JJ:= 0;
		if KEYHIT then quit;
		for II:= 0, 4000 do;		\Delay
		SYNC;
		end;
	end;
end;	\MAIN
