\MAZE.XPL

\16x16 maze is divided into square cells.
\Each cell is 64x64

\A mathematical coordinate system is used. Up is positive except when plotting.

int	WallH,		\Array
	WallV,
	VidMode,	\Initial video mode to be restored upon exit
	;

int	Px, Py;		\Player position
real	Pa;		\Player angle, direction player is facing

inc	\CXPL\STDLIB;



proc	Exit;		\Make a graceful exit
begin
SetVid(VidMode);	\Restore original video mode (clears screen)
exit;
end;	\Exit

\------------------------------------------------------------------------------

proc	CastRays(Px, Py, Pa);
\The descriptions are from the perspective of an aireal view looking down on the maze
int	Px, Py;		\Player position
real	Pa;		\Player angle, direction player is facing
int	Vx,
	Bx, By,		\Coordinates at base of cell
	Dx, Dy,		\Increments to move to next base line
	C, H;		\Distance, color, height
real	A		\Angle (0..360)
	TanA,
	Ix, Iy,		\Coordinates of intercepts
	DH2, DV2,	\Distance squared
	D;		\Distance
begin
Clear;
for Vx:= 0, 320-1 do
	begin
	A:= Pa - Float(Vx-160)/5.3333;
	while A > 360.0 do A:= A - 360.0;
	while A < 0.0 do A:= A + 360.0;

	TanA:= Tan(Rad(A));

	\Find intercepts (Ix,Iy):
	Bx:= Px & $FFC0;
	By:= Py & $FFC0;

	Dx:= -64;
	if A<90.0 ! A>270.0 then	\Moving to the right
		begin
		Bx:= Bx +64;
		Dx:= 64;
		end;

	Dy:= -64;
	if A>0.0  & A<180.0 then	\Moving upward
		begin
		By:= By +64;
		Dy:= 64;
		end;

	Ix:= Float(By-Py) / TanA + Float(Px);
	Iy:= Float(Bx-Px) * TanA + Float(Py);

	\Project ray until it hits a horizontal wall
	loop	begin
		if WallH(Fix(Ix-.5)>>6, By>>6) then quit;
		By:= By +Dy;
		Ix:= Ix + Float(Dy)/TanA;
		end;
	DH2:= (Ix-Float(Px))*(Ix-Float(Px)) + Float(By-Py)*Float(By-Py);

	\Project ray until it hits a vertical wall
	loop	begin
		if WallV(Bx>>6, Fix(Iy-.5)>>6) then quit;
		Bx:= Bx +Dx;
		Iy:= Iy + Float(Dx)*TanA;
		end;
	DV2:= Float(Bx-Px)*Float(Bx-Px) + (Iy-Float(Py))*(Iy-Float(Py));

	\Get distance and color of closer wall
	if DH2 <= DV2 then
		begin
		D:= Sqrt(DH2);
		C:= WallH(Fix(Ix-.5)>>6, By>>6);
		end
	else	begin
		D:= Sqrt(DV2);
		C:= WallV(Bx>>6, Fix(Iy-.5)>>6);
		end;

	\Draw vertical line representing slice of the wall
	H:= Fix(3200.0 /D);
	if H > 198 then H:= 198;
	Move(Vx, 100-H/2);
	Line(Vx, 100+H/2, C);
	end;
end;	\CastRays

\------------------------------------------------------------------------------

proc	SetUp;
begin

\    4	+---+---+---+---+
\	|               |
\    3	+---+           +
\	|   |           |
\    2	+---+   +---+   +
\	|       |   |   |
\    1	+---+   +   +   +
\	|   |   |   |   |
\    0	+---+---+---+---+
\	0   1   2   3   4


\Horizontal wall colors
\1 1 1 1
\2 0 0 0
\3 0 4 0
\3 0 0 0
\1 1 1 1

WallH:=
      [ [1, 2, 3, 3, 1],
	[1, 0, 0, 0, 1],
	[1, 0, 4, 0, 1],
	[1, 0, 0, 0, 1] ];

\Vertical wall colors
\1 2 4 4 1
\1 0 4 4 1
\1 3 0 0 1
\1 0 0 0 1

WallV:=
      [ [1, 1, 1, 1],
	[2, 0, 3, 0],
	[4, 4, 0, 0],
	[4, 4, 0, 0],
	[1, 1, 1, 1] ];
end;	\SetUp

\------------------------------------------------------------------------------

proc	StartUp;	\Start program with VGA and mouse
int	I, X, Y;
begin
I:= Equip;
if (I(2)&$08) = 0 then
	[Text(0, "

This program requires a VGA display.
");   exit];

Trapc(true);				\Ignore Ctrl+C if struck on keyboard
VidMode:= GetVid;
Setvid($13);				\320x200x256

\Set up mouse
if not OpenMouse then
	[Text(0, "This program requires a mouse.

Press Enter to continue...");
	WaitKey;   Exit];		\(Changing video modes erases screen)
MoveMouse(640/2, 480/2);		\Show mouse pointer in center of screen
end;	\StartUp

\==============================================================================

begin	\Main
Trap(false);
VidMode:= GetVid;
StartUp;
SetUp;

Px:= 1;   Py:= 96;   Pa:= 0.0;
loop	begin
	CastRays(Px, Py, Pa);
	case GetKey of
	  -RtArrow:
		Px:= Px +4;
	  -LfArrow:
		Px:= Px -4;
	  -UpArrow:
		Py:= Py +4;
	  -DnArrow:
		Py:= Py -4;
	  ^4:	Pa:= Pa +5.0;
	  ^6:	Pa:= Pa -5.0;
	  Esc:	quit
	other	[];
	end;
Exit;
end;	\Main
