\METRONOM.XPL		OCT-23-93
\Yer Basic XPL0 Metronome by Loren Blaney

inc	C:\CXPL\CODESI;		\Include code definitions for intrinsic routines

def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0a, FF=$0c,	\Control chars
	CR=$0d, EOF=$1a, Esc=$1b, Sp=$20, Ctrl=$40;
def	Up=-$48, Dn=-$50, Lt=-$4b, Rt=-$4d, PgUp=-$49,		\-Scan codes
	PgDn=-$51, End=-$4f, Home=-$47, Del=-$53;

int	Cpureg;			\Address of CPU register array (from Getreg)
def	AX, BX, CX, DX, DI, SI, BP, CF, CS, DS, SS, ES;		\Getreg registers

def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,	\Attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, LWhite; \ EGA palette

def	Pi = 3.14159265358979323846;


def	ArmLen = 150.0,	\Length of arm (pixels)
	Amp = 0.6;	\Amplitude of swing (radians)
int	Mode,		\Initial video mode
	DelX, DelY,	\Relative position of swinging end of arm
	ClickVol;	\Adjustment for click sound volume
real	Angle,		\Angle of arm relative to Y axis
	Angle0,		\Previous angle
	Time,		\Elapsed time
	Freq;		\Frequency in beats (clicks) per minute
\Constant to convert Freq to radians per vertical blank
def	Const= Pi /60.0 /60.0;



func	GetVid;			\Returns the current video mode
begin
Cpureg(AX):= $0f00;		\Function $0f
Softint($10);			\Call BIOS routine
return Cpureg(AX) & $ff;
end;	\GetVid



func	GetKey;			\Get character from keyboard (wait if necessary)
int	Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl-C, and no cursor.
Cpureg(AX):= 0;			\Function $00
SoftInt($16);			\Call BIOS routine
Ch:= Cpureg(AX) & $ff;
if Ch = 0 then			\Return non-ascii chars as negative scan code
	Ch:= -(Cpureg(AX)>>8);
return Ch;
end;	\GetKey



proc	WaitVB;			 \Wait for beginning of next vertical blank
begin				 \WARNING: even $3ba might not work for monochrome
while Pin($3da, 0) & $08 do;	 \Wait for no vertical blank
repeat until Pin($3da, 0) & $08; \Wait for vertical blank
end;	\WaitVB;



proc	Click;			\Make click sound
int	I;
begin
Pout(Pin($61, 0)!$03, $61, 0);	\Turn on speaker
for I:= 0, ClickVol do; 	\Delay for click noise
Pout(Pin($61, 0)&$fc, $61, 0);	\Turn off speaker
end;	\Click



proc	DrawPoly(N, A, C);	\Draw a polygon
int	N,	\Number of points in A
	A,	\Array: N pairs of X,Y coordinates
	C;	\Color
int	I;
begin
if N >= 1 then Move(A(0), A(1));
for I:= 2, N+N-1 do
	begin
	Line(A(I), A(I+1), C);
	I:= I +1;
	end;
end;	\DrawPoly



proc	DrawMetronome;		\Draw the body of the metronome
int	Shape;
begin
Shape:= [319,144, 294,156, 244,406, 395,406, 345,156, 320,144];
DrawPoly(6, Shape, Brown);
Move(260,326);   Line(379, 326, Brown);
Move(294,156);   Line(345, 156, Brown);
end;	\DrawMetronome



proc	DrawArm;		\Draw the swinging arm
def	X0=320, Y0=310;		\Pivot point
begin
Move(X0+DelX, Y0-DelY);
Line(X0, Y0, Yellow!$80);	\XOR color to draw and erase
end;	\DrawArm



proc	Calc;			 \Calculate position of arm
begin
Angle:= Amp * Sin(Freq*Time);	 \Simple harmonic motion
DelX:= Fix(ArmLen * Sin(Angle)); \Position of end of arm
DelY:= Fix(ArmLen * Cos(Angle));
end;	\Calc



proc	SwingArm;		\Swing the metronome arm

	proc	ShowFreq;	\Display the current frequency
	begin
	Format(3, 0);
	Cursor(38, 23);
	Rlout(0, Freq);
	Time:= 0.0;		\Restart time to make jump less annoying
	end;	\ShowFreq

begin
ClickVol:= 400;
Freq:= 120.0;			\Initial frequency
ShowFreq;
loop	begin
	Angle0:= Angle;		\Click when arm passes through angle = 0.0
	Calc;
	DrawArm;
	if Angle<0.0 & Angle0>=0.0 ! Angle>=0.0 & Angle0<0.0 then Click;
	Time:= Time + Const;

	if ChkKey then
		case GetKey of
		  Up:	[Freq:= Freq -1.0;   ShowFreq];
		  Dn:	[Freq:= Freq +1.0;   ShowFreq];
		  Lt:	[ClickVol:= ClickVol -100];
		  Rt:	[ClickVol:= ClickVol +100];
		  Esc:	quit
		other;

	WaitVB;
	DrawArm;		\Erase arm
	end;
end;	\SwingArm



begin	\Main
Cpureg:= Getreg;
Mode:= GetVid;
SetVid($12);			\VGA: 640x480x16

Cursor(15, 29);
Text(0, "Presto: |           Largo: |             Fine: Esc ");
Cursor(23, 29);   Chout(0, $19);	\Down arrow
Cursor(42, 29);   Chout(0, $18);	\Up arrow

DrawMetronome;
Sound(0, 1, 500);		\Set up speaker for click

SwingArm;

SetVid(Mode);
end;	\Main
