\MOLECULE.XPL	MAR-21-94
\This draws a 3-D perspective view of some molecules that are meant to be 
\ viewed using Sega-like 3-D goggles.
\ by Loren Blaney and Richard Ottosen
\Translated from a BASIC program by E. Kikland (see Byte, FEB 1985, p 253).

\REVISIONS:
\FEB-22-85, Original on Apple II
\FEB-23-93, PC version using 3-D Sega glasses for Rich and Bill.
\NOV-01-93, Modified output for Sega glasses (shift a copy of the bits left
\ two places).
\NOV-16-93, Show commands and files on screen with settings, ignore invalid
\ commands, make glasses clear when not doing 3-D.  --R. Ottosen
\MAR-15-94, Make scaling match for both 3-D images, color. --R. Ottosen/L.B.
\MAR-21-94, The Windows look with font, fix distortion caused by EGA, give
\ atoms consistent colors, intro screen selects with or w/o goggles.


\The data file contains the description of the molecule. Each atom is
\ specified by X, Y, Z coordinates, a size (atomic radius), and it's atomic
\ number. There is also a list of which atoms are bonded to which (atoms
\ are numbered starting with 0)
\
\The molecule is projected onto the X,Z plane, which corresponds to the
\ X,Y dimensions of the graphics screen:
\
\		  Z
	      \   |
		\ |
\ Eye -->    -----*----- Y
\		  | \
\		  |   \
\		        X

inc	C:\CXPL\CODESI;

def	AX, BX, CX, DX, DI, SI, BP, CF, CS, DS, SSEG, ES;	\GETREG registers
def	BS=$08, TAB=$09, FF=$0C, CR=$0D, EOF=$1A, ESC=$1B, SP=$20; \ASCII chars
def	UP=-72, DN=-80, LT=-75, RT=-77,
	PGUP=-73, PGDN=-81, END=-79, HOME=-71;		\Scan codes

def	HEIGHT=350, WIDTH=640;		\EGA Screen dimensions in pixels

\The colors are determined by MOLECULE.BMP, which is loaded at the start.
\	0	black
\	1	dark yellow
\	2	med yel
\	3	yellow
\	4	dark blue
\	5	med d. blue
\	6	med l. blue (gray)
\	7	light blue (white)
\	8	dark red
\	9	med d red
\	10	med l red
\	11	light red
\	12	dark green
\	13	med d green
\	14	med l green
\	15	light green

def	LEFT, RIGHT;	\Left and right views for 3D goggles

seg addr IMAGE;		\320x200 image of atom
char	TITLE;		\Array: Title of molecule
int	CPUREG,		\Array: CPU registers for BIOS calls
	DATASEG,	\Our data segment address
	DispMode,	\Display mode to return to at end of program
	Goggles,	\Boolean: Have goggles
	COLD,		\Boolean: True if this is a cold start
	CROSSEYE,	\Boolean: True to flip which eye is which
	LPT1PORT,	\Printer port address
	ATOMS,		\Total number of atoms in data file
	BONDS,		\Total number of bonds in data file
	BOND0, BOND1,	\Arrays: bond goes from BOND0(I) to BOND1(I)
	TYPE;		\Array: type of atom (Hydrogen, Oxygen, Carbon...)
real	XX, YY, ZZ, SS;	\Arrays: initial position and size for each atom
real	AZ, EL,		\Azimuth and elevation
	DIST,		\View distance
	SIZE,		\Scale factor for atom sizes
	DEPTH,		\Angle between eyeballs
	SCALE;		\Scaling factor used for both views

\Variables for GETFILE:
int	IFNDISP,	\Index to top of displayed file names
	IFNSEL;		\Index to selected (highlighted) file name

char	FontTable;	\8x14 font, for 256 characters
def	FontSize=3584;	\ = 256 * 14

\----------------------------------------------------------------------

proc	EXIT;			\Orderly exit from program
begin
Setvid(DispMode);		\Restore original display mode
exit;
end;	\EXIT



func	ToUpper(Ch);		\Convert character to uppercase
int	Ch;
return if Ch>=^a & Ch<=^z then Ch&$df else Ch;



func real RAD(X);	\Convert degrees to radians
def	PI=3.141592654;
real	X;
return X *PI /180.0;



func	GetVid;			\Returns the current video mode
begin
Cpureg(AX):= $0f00;		\Function $0f
Softint($10);			\Call BIOS routine
return Cpureg(AX) & $ff;
end;	\GetVid



func	GETKEY;			\Get character from keyboard
int	CH;			\This is a low-level routine with no echo,
begin				\ no ctrl-C, and no cursor.
CPUREG(AX):= 0;			\Function $00
SOFTINT($16);			\Call BIOS routine
CH:= CPUREG(AX) & $FF;
if CH = 0 then CH:= -(CPUREG(AX)>>8);
return CH;
end;	\GETKEY



func	ShiftKey;		\Returns 'true' if a shift key is down
begin
Cpureg(AX):= $0200;
SoftInt($16);
return (Cpureg(AX)&$03) # 0;
end;	\ShiftKey



proc	PLACE(X, Y, STR, VAL, UNITS);	\Cursor + Text + Rlout + Chout
int	X, Y, STR;
real	VAL;
int	UNITS;
begin
CURSOR(X, Y);
TEXT(6, STR);
RLOUT(6, VAL);
CHOUT(6, UNITS);
end;	\PLACE



proc	Ctxt(X, Y, Str);	\Display a string on device 0 at X, Y
int	X, Y;
char	Str;
begin
Cursor(X, Y);
Text(6, Str);
end;	\Ctxt



proc	HLine(X, Y, W, T);	\Draw a horizontal line using type T box chars
int	X, Y, W, T;		\ with left end at X, Y
int	C, I;			\The minimum width (W) is 2
begin
C:= if T&1\#0\ then ^ else ^;
Cursor(X, Y);
for I:= 0, W-1 do Chout(6, C);
end;	\HLine



proc	Box(X, Y, W, H, T);	\Draw a rectangle using type T box chars
int	X, Y, W, H, T;		\ with upper-left corner at X, Y
int	TBL, I;			\The minimum width (W) & height (H) is 2
char	S;			\Note this uses device 6 with attributes
begin
\	 012345   Box Character Type (T):
TBL:=  ["Ŀ ",	\0
	"͸Ծ ",	\1
	"ķӽ ",	\2
	"ͻȼ "];	\3
S:= TBL(T&3);				\(Mask for safety)

Cursor(X, Y);	Chout(6, S(0));		\Draw top line with brackets
for I:= 1, W-2 do Chout(6, S(1));
Chout(6, S(2));
for I:= 1, H-2 do			\Draw sides
	[Cursor(X, Y+I);   Chout(6, S(3));
	Cursor(X+W-1, Y+I);   Chout(6, S(3))];
Cursor(X, Y+H-1);   Chout(6, S(4));	\Draw bottom line with brackets
for I:= 1, W-2 do Chout(6, S(1));
Chout(6, S(5));
end;	\Box



proc	SETCOLREG(N, R, G, B);	\Set color register N
int	N,	\Register to set
	R, G, B;\Red, Green, and Blue [0..255]
begin
CPUREG(AX):= $1010;		\Function $10, subfunction $10
CPUREG(BX):= N;			\Select color register
CPUREG(CX):= G<<8 ! (B&$FF);	\Green<<8 blue
CPUREG(DX):= R<<8;		\Red<<8
SOFTINT($10);			\Call BIOS routine
end;	\SETCOLREG



proc	FIXPALETTE;	\Make palette registers correspond to DAC registers
int	N;
begin
for N:= 0, 15 do
	begin
	CPUREG(AX):= $1000;	\Function $10, subfunction $00
	CPUREG(BX):= N<<8!N;	\DAC and palette registers
	SOFTINT($10);		\Call BIOS routine
	end;
end;	\FIXPALETTE



proc	WAITVB;		\Wait for the beginning of the next vertical blank
begin
while PIN($3DA, 0) & $08 do;		\Wait for no vertical blank
repeat until PIN($3DA, 0) & $08;	\Wait for vertical blank
end;	\WAITVB;



proc	SetPage(P);		\Set video page for writing and viewing
int	P;
begin
Cpureg(AX):= $0500 ! P&$ff;
Softint($10);
end;	\SetPage



proc	RECTANGLE(X0, Y0, X1, Y1, C);	\Draw a solid rectangle
int	X0, Y0, X1, Y1, C;
int	I, J, T;
begin
if X1 < X0 then [T:=X1; X1:=X0; X0:=T];
if Y1 < Y0 then [T:=Y1; Y1:=Y0; Y0:=T];

if (Y1-Y0) > 8*(X1-X0) then		\If tall and skinny
	begin
	for I:= X0, X1 do		\Draw vertical lines
		[MOVE(I, Y0);   LINE(I, Y1, C)];
	end
else	begin
	for J:= Y0, Y1 do		\Draw fast horizontal lines
		[MOVE(X0, J);   LINE(X1, J, C)];
	end;
end;	\RECTANGLE



proc	SPHERE(X, Y, R, COL);	\Draw a sphere
int	X, Y,	\Screen coordinates of center
	R,	\Radius in pixels
	COL;	\Color (0..15)
int	RY, I, J, II, JJ, KK, C;
begin
if R = 0 then return;
RY:= FIX(FLOAT(R) * .73);	\Correct aspect ratio
COL:= COL -12;
for J:= -RY, RY do
	begin
	JJ:= (64 + J*64/RY) *320;
	KK:= Y +J;
	for I:= -R, R do
		begin
		C:= IMAGE(0, JJ + (319-64 - I*64/R));	\DARK GREEN = 12
		if C\#0\ then POINT(X+I, KK, C+COL);
		end;
	end;
end;	\SPHERE

\----------------------------------------------------------------------

func	COLORTYPE(A);		\Return color used by sphere for Atomic No.
int	A;
def	H=1, C=6, N=7, O=8, Si=14, P=15, S=16,
	Yellow=0, Blue=4, Red=8, Green=12;
begin
case A of		\Valence
\  H:	return Green;	\\ 1
  C,Si:	return Yellow;	\-4
  N,P:	return Blue;	\-3
  O,S:	return Red	\-2
other	return Green;	\Hydrogen and Miscellanea
end;	\COLORTYPE

\----------------------------------------------------------------------

func	STRCMP(A, B, SIZE);	\Compare string A to string B
\This returns:
\	-1 if A < B
\	 0 if A = B
\	+1 if A > B
\This provides a general string compare, for example:
\ if CMPSTR(A, B, SIZE) >= 0 then...	(if A >= B then...)
addr	A, B;		\Strings containing numbers to be compared
int	SIZE;		\Number of digits (number of bytes in A and B)
int	I;		\A & B must be right justified
begin
for I:= 0, SIZE-1 do
	if A(I) # B(I) then
		return if A(I) > B(I) then 1 else -1;
return 0;				\They're equal
end;	\STRCMP



proc	SORT(A, N);		\Sort array using the shell sort method
int	A,			\Address of array
	N;			\Number of elements in array (size)
int	J, GAP, I, JG, T;
begin
GAP:= N /2;
while GAP > 0 do
	begin
	for I:= GAP, N-1 do
		begin
		J:= I -GAP;
		loop	begin
			if J < 0 then quit;
			JG:= J +GAP;
			if STRCMP(A(J), A(JG), 8) <= 0 then quit;
			T:= A(J);   A(J):= A(JG);   A(JG):= T;
			J:= J -GAP;
			end;
		end;
		GAP:= GAP /2;
	end;
end;	\SORT

\======================================================================

func	GETFILE(TEMPLATE, WINX, WINY, WINWIDE, WINHIGH);
char	TEMPLATE;
int	WINX, WINY, WINWIDE, WINHIGH;
int	J, KEY,
	NAMES,		\Array of pointers to strings
	QMAXX;		\Index to last file + 1



proc	READDIR;		\Read a directory with specified template
int	I, P, CH;
char	DTA,N;
begin	
DTA:= RESERVE(43);		\Set up data for directory work
CPUREG(0):= $1A00;
CPUREG(3):= DTA;
CPUREG(9):= DATASEG;
SOFTINT($21);

CPUREG(0):= $4E00;		\Look up first file
CPUREG(2):= $FFF7;
CPUREG(9):= DATASEG;
CPUREG(3):= TEMPLATE;
SOFTINT($21);
QMAXX:= 0;
if (CPUREG(0) & $FF) # 0 then return;

loop	begin
	P:= 0;
	N:= NAMES(QMAXX);
	loop	begin
		CH:= DTA(30+P);
		if CH = ^. then quit;
		N(P):= CH;
		P:= P+1;
		end;
	for I:=P,7 do N(I):= SP;
	N(7):= N(7) ! $80;
	QMAXX:= QMAXX+1;

	CPUREG(0):= $4F00;		\Look up next file
	SOFTINT($21);
	if (CPUREG(0) & $FF) # 0 then quit;
	end;
end;	\READDIR



proc	SHOWFILES;		\Fill window with list of files
int	Q, I;
begin
Q:= IFNDISP;
for I:= 1, WINHIGH-2 do
	begin
	CURSOR(WINX+1, WINY+I);
	ATTRIB(if Q = IFNSEL then $47 else $70);	\Highlight selected file
	CHOUT(6, SP);
	TEXT(6, if Q < QMAXX then NAMES(Q) else "        "); \Display file name
	CHOUT(6, SP);
	Q:= Q+1;
	end;
end;	\SHOWFILES



proc	SCROLLBAR;		\Draw scroll bar with slider
int	J;
begin
ATTRIB($60);
CURSOR(WINX+WINWIDE-1, WINY+1);   CHOUT(6, $18);	 \up arrow
for J:= WINY+2, WINY+WINHIGH-3 do
	[CURSOR(WINX+WINWIDE-1, J);   CHOUT(6, ^)];
CURSOR(WINX+WINWIDE-1, WINY+WINHIGH-2);   CHOUT(6, $19); \down arrow

\Show highlighted slider
ATTRIB($06);
J:= if QMAXX > 1 then (IFNSEL * (WINHIGH-5) + (QMAXX-1)/2 ) / (QMAXX-1)
	else 0;
CURSOR(WINX+WINWIDE-1, J+WINY+2);
CHOUT(6, ^);
end;	\SCROLLBAR



func	SEARCH(CH, I);	\Search NAMES for file that starts with CH.
\Returns index of next name or current name if none.
int	CH, I;
char	STR;
begin
loop	begin
	I:= I + 1;
	if I >= QMAXX then I:= 0;	\loop back to start of list
	if I = IFNSEL then return I;	\back where we started

	STR:= NAMES(I);
	if CH = STR(0) then return I;
	end;
end;	\SEARCH



begin	\GETFILE
NAMES:= RESERVE(100*2);
for J:= 0, 99 do
	NAMES(J):= RESERVE(8);

ATTRIB($74);	\Blue on white
BOX(WINX, WINY, WINWIDE, WINHIGH, 3);

\Fix up TEMPLATE by terminating it with 0
TEMPLATE(5):= 0;
READDIR;
SORT(NAMES, QMAXX);

loop	begin
	SHOWFILES;
	SCROLLBAR;

	KEY:= GETKEY;
	case KEY of
	  UP:	begin
		IFNSEL:= IFNSEL -1;
		if IFNSEL < 0 then IFNSEL:= 0;
		if IFNSEL < IFNDISP then IFNDISP:= IFNSEL;
		end;
	  DN:	begin
		IFNSEL:= IFNSEL +1;
		if IFNSEL >= QMAXX then IFNSEL:= QMAXX-1;
		if IFNSEL >= IFNDISP+WINHIGH-2 then
			IFNDISP:= IFNSEL-(WINHIGH-2)+1;
		end;
	  HOME:	begin
		IFNSEL:= 0;
		IFNDISP:= 0;
		end;
	  END:	begin
		IFNSEL:= QMAXX-1;
		IFNDISP:= QMAXX-1 - (WINHIGH-2)+1;
		if IFNDISP < 0 then IFNDISP:= 0;
		end;
	  PGUP:	begin
		IFNDISP:= IFNDISP - (WINHIGH-2);
		if IFNDISP < 0 then IFNDISP:= 0;
		IFNSEL:= IFNDISP;
		end;
	  PGDN:	begin
		IFNDISP:= IFNDISP + (WINHIGH-2);
		if IFNDISP > QMAXX-1 - (WINHIGH-2)+1 then
			[IFNDISP:= QMAXX-1-(WINHIGH-2)+1;
			if IFNDISP < 0 then IFNDISP:= 0];
		IFNSEL:= IFNDISP + WINHIGH-2-1;
		if IFNSEL >= QMAXX then IFNSEL:= QMAXX-1;
		end;
	  CR:	return NAMES(IFNSEL);	\Return selected file name
	  ESC:	return -1
	other	begin
		IFNSEL:= SEARCH(ToUpper(KEY), IFNSEL);
		if IFNDISP>IFNSEL ! IFNDISP+(WINHIGH-2)<=IFNSEL then
			IFNDISP:= IFNSEL;
		if IFNDISP > QMAXX-1 - (WINHIGH-2)+1 then
			[IFNDISP:= QMAXX-1-(WINHIGH-2)+1;
			if IFNDISP < 0 then IFNDISP:= 0];
		end;
	end;
end;	\GETFILE

\======================================================================

proc	VIEW3D;			\Continuously flip goggles and video screen

	proc	EYE(P);
	int	P;
	begin
	SETPAGE(P);
	if CROSSEYE then POUT(if P\#0\ then $F5 else $FA, LPT1PORT, 0)
	else POUT(if P\#0\ then $FA else $F5, LPT1PORT, 0);
	end;	\EYE

begin	\VIEW3D
loop	begin
	WAITVB; EYE(0);
	WAITVB; EYE(1);
	if CHKKEY then quit;
	end;
end;	\VIEW3D

\======================================================================

proc	INITSCREEN;	\Draw initial screen
begin
RECTANGLE(0, 0, WIDTH-1, HEIGHT-1, 7);

RECTANGLE(0, 0, 3, HEIGHT-1, 4);
RECTANGLE(WIDTH-1, 0, WIDTH-4, HEIGHT-1, 4);
RECTANGLE(0, HEIGHT-3, WIDTH-1, HEIGHT-1, 4);
RECTANGLE(0, 0, WIDTH-1, 13, 4);

ATTRIB($47);	\White on blue
CURSOR(WIDTH/2/8-9, 0);   TEXT(6, "MOLECULE");
if ~COLD then [TEXT(6, ": ");  TEXT(6, TITLE)];
end;	\INITSCREEN



proc	ERASE;		\Erase any existing molecule - redraw entire screen
def	X0=2, Y0=2;
def	X1=3, Y1=20;
begin
INITSCREEN;

FORMAT(4, 1);
ATTRIB($70);	\Black on white
BOX(X0, Y0, 26, 13, 0);
CTXT(X0+1, Y0, " Commands ");
\Display left and right arrows:
 PLACE(X0+2, Y0+1, "     Azimuth:  ", AZ, ^);
CURSOR(X0+2, Y0+1);  CHOUT(6,$1B);  CHOUT(6,^ );  CHOUT(6,$1A);
 PLACE(X0+2, Y0+2, "   Elevation:", EL, ^);
 HLINE(X0+1, Y0+3, 24, 0);
 PLACE(X0+2, Y0+4, "D d  Distance: ", DIST, ^);
 PLACE(X0+2, Y0+5, "S s  Size:     ", SIZE, ^ );
 if ~Goggles then ATTRIB($76);
 PLACE(X0+2, Y0+6, "A a  3-D Angle:", DEPTH, ^);
 if ~Goggles then ATTRIB($70);
 HLINE(X0+1, Y0+7, 24, 0);
  CTXT(X0+2, Y0+8, "R      Reset View");
  if ~Goggles then ATTRIB($76);
  CTXT(X0+2, Y0+9, "I      Swap Eyes");
  if ~Goggles then ATTRIB($70);
  CTXT(X0+2, Y0+10,"Enter  New Molecule...");
  CTXT(X0+2, Y0+11,"Esc    Exit Program");

\Legend:
ATTRIB($70+COLORTYPE(1)+1);
CTXT(X1, Y1+0," Hydrogen  Miscellanea");
ATTRIB($70+COLORTYPE(6)+1+1);
CTXT(X1, Y1+1," Carbon    Silicon");
ATTRIB($70+COLORTYPE(7)+1);
CTXT(X1, Y1+2," Nitrogen  Phosphorus");
ATTRIB($70+COLORTYPE(8)+1);
CTXT(X1, Y1+3," Oxygen    Sulphur");
end;	\ERASE

\----------------------------------------------------------------------

proc	DRAWMOLE(VIEW);	\Draw the molecule
int	VIEW;		\Select which eye's view (LEFT/RIGHT)
int	IX, IY, IR, COL,
	IGAP, IT,
	I, J, JG;
real	X, Y, Z, S;	\ARRAYS: POSITION AND SIZE FOR EACH ATOM
real	XMIN, XMAX,
	YMIN, YMAX,
	ZMIN, ZMAX,
	XCEN, YCEN, ZCEN,
	SMAX, T,
	CAZ, SAZ, CEL, SEL,
	XA, YA, ZA;
def	WINCENX= 2*WIDTH/3-16,   WINCENY= (HEIGHT-10)/2 +10;
begin
X:= RLRES(100);   Y:= RLRES(100);   Z:= RLRES(100);   S:= RLRES(100);
COL:= RESERVE(100*2);

\Get a copy of the data and determine max and min limits:
XMIN:= 1E25;   XMAX:= -XMIN;   YMIN:= XMIN;   YMAX:= XMAX;
ZMIN:= XMIN;   ZMAX:= XMAX;
for I:= 1, ATOMS do
	begin
	X(I):= XX(I);   Y(I):= YY(I);   Z(I):= ZZ(I);   S(I):= SS(I);
	if X(I) > XMAX then XMAX:= X(I);   if X(I) < XMIN then XMIN:= X(I);
	if Y(I) > YMAX then YMAX:= Y(I);   if Y(I) < YMIN then YMIN:= Y(I);
	if Z(I) > ZMAX then ZMAX:= Z(I);   if Z(I) < ZMIN then ZMIN:= Z(I);
	COL(I):= COLORTYPE(TYPE(I));
	end;
\Rotate molecule around its center:
CAZ:= COS(RAD(AZ));   SAZ:= SIN(RAD(AZ));   CEL:= COS(RAD(EL));   SEL:= SIN(RAD(EL));
XCEN:= .5*(XMAX+XMIN);   YCEN:= .5*(YMAX+YMIN);   ZCEN:= .5*(ZMAX+ZMIN);
for I:= 1, ATOMS do
	begin
	X(I):= X(I)-XCEN;
	Y(I):= Y(I)-YCEN;
	Z(I):= Z(I)-ZCEN;

	\Polar rotation (about X axis)
	YA:= Y(I);
	ZA:= Z(I);
	Y(I):= CEL*YA + SEL*ZA;
	Z(I):= CEL*ZA - SEL*YA;

	\Azimuthal rotation (about Z axis)
	XA:= X(I);
	YA:= Y(I);
	X(I):= CAZ*XA + SAZ*YA;
	Y(I):= CAZ*YA - SAZ*XA;
	end;

\Do perspective projection and scale coordinates:
SMAX:= SCALE;

for I:= 1, ATOMS do
	begin
	YA:= 1.0/(DIST-Y(I));
	X(I):= X(I)*YA;   Z(I):= Z(I)*YA;   S(I):= S(I)*YA;

	if VIEW =LEFT then	\Compute scaling for left eye only
	  [if SCALE<RLABS(X(I)) then SCALE:= RLABS(X(I));  \Expand SCALE to
	  if SCALE<RLABS(Z(I)) then SCALE:= RLABS(Z(I));   \ largest dimension
	  ];
	if SMAX<S(I) then SMAX:= S(I);
	end;
if VIEW =LEFT then		\Compute scaling for left eye only
	SCALE:= FLOAT((HEIGHT-12)/2)/0.73 /(SCALE + SMAX*SIZE);

\Draw the bonds:
\WaitVB;
for I:= 1, BONDS do
	begin
	J:= BOND0(I);
	IX:= FIX(X(J)*SCALE) + WINCENX;
	IY:= FIX(-Z(J)*SCALE*0.73) + WINCENY;
	MOVE(IX, IY);
	J:= BOND1(I);
	IX:= FIX(X(J)*SCALE) + WINCENX;
	IY:= FIX(-Z(J)*SCALE*0.73) + WINCENY;
	LINE(IX, IY, $04);
	end;

if CHKKEY then return;

\Sort atoms by depth (Y axis) using Shell sort:
IGAP:= ATOMS /2;
while IGAP >= 1 do
	begin
	for I:= IGAP+1, ATOMS do
		begin
		J:= I -IGAP;		\synthesized decrementing FOR loop
		loop	begin
			if J < 1 then quit;
			JG:= J +IGAP;
			if Y(J) <= Y(JG) then quit;
			T:= X(J);   X(J):= X(JG);   X(JG):= T;
			T:= Y(J);   Y(J):= Y(JG);   Y(JG):= T;
			T:= Z(J);   Z(J):= Z(JG);   Z(JG):= T;
			T:= S(J);   S(J):= S(JG);   S(JG):= T;
			IT:= COL(J);   COL(J):= COL(JG);   COL(JG):= IT;
			J:= J -IGAP;
			end;
		end;
	IGAP:= IGAP /2;
	end;

\Plot atoms
for I:= 1, ATOMS do
	begin
	IX:= FIX(X(I)*SCALE) + WINCENX;
	IY:= FIX(-Z(I)*SCALE*0.73) + WINCENY;
	IR:= FIX(S(I)*SCALE*SIZE);
	SPHERE(IX, IY, IR, COL(I));
	if CHKKEY then return;
	end;
end;	\DRAWMOLE

\----------------------------------------------------------------------

proc	LOADFILE;	\Read atomic coordinates from data file
int	INHAND;
char	FILENAME;
char	STR, EXT;
int	I, J, CH;
begin
POUT($FF, LPT1PORT, 0);			\Make glasses clear
STR:= RESERVE(12);

FILENAME:= GETFILE("*.MOL ", 24, 11, 12, 12);
if FILENAME = -1 then EXIT;

EXT:= ".MOL";				\Default extension
I:= 0;
loop	begin
	CH:= FILENAME(I);
	if CH=SP then quit;
	if CH>=$80 then
		[CH:= CH & $7F;
		if CH=SP then quit;
		STR(I):= CH;
		TITLE(I):= CH;
		I:= I+1;
		quit];
	STR(I):= CH;
	TITLE(I):= CH;
	I:= I+1;
	end;
TITLE(I):= SP ! $80;
for J:= 0, 3 do
	begin
	STR(I):= EXT(J);
	I:= I+1;
	end;

COLD:= false;				\We now have a valid title

INHAND:= FOPEN(STR, 0);
FSET(INHAND, ^i);			\Set handle for device 3
OPENI(3);

ATOMS:= 0;
repeat	ATOMS:= ATOMS +1;
	XX(ATOMS):= RLIN(3) *7.69;	\Convert to Angstroms
	YY(ATOMS):= RLIN(3) *7.69;
	ZZ(ATOMS):= RLIN(3) *7.69;
	SS(ATOMS):= RLIN(3);		\Already in Angstroms
	TYPE(ATOMS):= INTIN(3);
until	CHIN(3)=^$;	\There better be a multiple of 5 numbers in the input
			\ file. " $" must immediately follow the last number.
BONDS:= 0;
repeat	BONDS:= BONDS +1;
	BOND0(BONDS):= INTIN(3);
	BOND1(BONDS):= INTIN(3);
until	CHIN(3)=^$;	\There better be a multiple of 4 numbers in input file

FCLOSE(INHAND);
end;	\LOADFILE

\----------------------------------------------------------------------

proc	RESET;		\Normal initial viewing parameters
begin
AZ:= 0.0;
EL:= 0.0;
DIST:= 30.0;
SIZE:=  0.5;
DEPTH:= if Goggles then 3.5 else 0.0;
end;	\RESET



proc	COMMAND;	\Get commands and draw molecule
int	REDRAW;
begin
RESET;
ERASE;
IFNDISP:= 0;
IFNSEL:= 0;
LOADFILE;
REDRAW:= true;
loop	begin
	POUT($FF, LPT1PORT, 0);			\Make glasses clear
	SETPAGE(0);
	if REDRAW then
		[POUT($FF, LPT1PORT, 0);	\Make glasses clear
		ERASE;
		SCALE:= -1E25;
		DRAWMOLE(LEFT)
		];

	if Goggles then
		begin
		AZ:= AZ -DEPTH;	\For stereo view
		SETPAGE(1);
		if REDRAW then
			[POUT($FF, LPT1PORT, 0);	\Make glasses clear
			ERASE;
			DRAWMOLE(RIGHT)
			];
		AZ:= AZ +DEPTH;	\Unadjust
		VIEW3D;
		end;

	REDRAW:= true;
	case GETKEY of
	  CR: LOADFILE;
	  RT: AZ:= AZ + (if ShiftKey then 0.5 else 5.0);
	  LT: AZ:= AZ - (if ShiftKey then 0.5 else 5.0);
	  DN: EL:= EL + (if ShiftKey then 0.5 else 5.0);
	  UP: EL:= EL - (if ShiftKey then 0.5 else 5.0);
	  ^D: DIST:= DIST +.5;
	  ^d: [DIST:= DIST -.5;   if DIST < 0.0 then DIST:= 0.0];
	  ^S: SIZE:= SIZE +.1;
	  ^s: [SIZE:= SIZE -.1;   if SIZE < 0.0 then SIZE:= 0.0];
	  ^A: if Goggles then DEPTH:= DEPTH +.5;
	  ^a: if Goggles then
	  	[DEPTH:= DEPTH -.5; if DEPTH < 0.0 then DEPTH:= 0.0];
	  ^I,^i: CROSSEYE:= not CROSSEYE;
	  ^R,^r: RESET;
	  ESC: quit
	other REDRAW:= false;
	end;
POUT($FF, LPT1PORT, 0);	\Make glasses clear (so you can program with 'em on)
end;	\COMMAND

\======================================================================

proc	GETIMAGE;	\Read a bit mapped image file (.BMP) into IMAGE
int	HAND, I, J, R, G, B;
begin
HAND:= FOPEN("MOLECULE.BMP", 0);	\Graphic image for nice fuzzy atom
FSET(HAND, ^I);
OPENI(3);

for I:= 0, $39 do J:= CHIN(3);		\Skip header
FIXPALETTE;
for I:= 1, $0F do			\Read in colors
	begin
	B:= CHIN(3) >>2;
	G:= CHIN(3) >>2;
	R:= CHIN(3) >>2;
	J:= CHIN(3);
	SETCOLREG(I, R, G, B);	\Set color register
	end;
for I:= $10, $FF do			\Skip rest of colors for speed (not used)
	begin
	B:= CHIN(3);
	G:= CHIN(3);
	R:= CHIN(3);
	J:= CHIN(3);
	end;

I:= 320 *200;				\WARNING: this swaps left and right
repeat	I:= I -1;			\Read image data
	IMAGE(0, I):= CHIN(3);
until I = 0;

FCLOSE(HAND);
end;	\GETIMAGE



proc	LoadFont;		\Load in a font table for EGA mode $10
int	Hand, I, Ch;
begin
HAND:= FOPEN("MOLECULE.F14", 0);	\Graphic image for nice fuzzy atom
FSET(HAND, ^I);
Openi(3);
for I:= 0, FontSize-1 do
	begin
	FontTable(I):= Chin(3);
	end;
FCLOSE(HAND);
end;	\LoadFont

\======================================================================

proc	INTRO;		\Display intro screen an find out if there are goggles
int	Ch;
begin
ATTRIB($07);
RAWTEXT(6, "                              / / /       / / /
                            /    /    /   / 
                                  /    /      / 
                                / /     /      / 
                                  /    /      / 
                            /      /    /     /
                             / / /     / / /
                                        

                     M O L E C U L E   V I E W E R   0.9

                      Loren Blaney  -  Richard Ottosen ");

ATTRIB($1F);
CURSOR(15,14); RAWTEXT(6," ͻ  ");
CURSOR(15,15); RAWTEXT(6,"                                              ");
CURSOR(15,16); RAWTEXT(6,"                                              ");
CURSOR(15,17); RAWTEXT(6,"                                              ");
CURSOR(15,18); RAWTEXT(6,"       Do you have 3-D goggles (Y\N)?         ");
CURSOR(15,19); RAWTEXT(6,"                                              ");
CURSOR(15,20); RAWTEXT(6,"                                              ");
CURSOR(15,21); RAWTEXT(6,"                                              ");
CURSOR(15,22); RAWTEXT(6," ͼ  ");

loop	begin
	CURSOR(54, 18); Ch:= Chin(6);
	if Ch=^n ! Ch=^N then [Goggles:= false; quit];
	if Ch=^y ! Ch=^Y then [Goggles:= true;  quit];
	if Ch=Esc then EXIT;
	Chout(0, 7);
	CURSOR(54, 18); Text(6, "   ");
	end;

Openi(0);
if not Goggles then
	begin
	CURSOR(18, 20);
	RAWTEXT(6,"(Come to a 6502 meeting and build a pair) ");
	for Ch:= 1, 46 do
		begin
		SOUND(0, 1, 1000);
		if Chkkey then Ch:= 32766;
		end;
	Openi(0);
	end;

Clear;
ATTRIB($07);
CURSOR(25,24); TEXT(6, "Loading graphic images...");
end;	\INTRO

\======================================================================

begin	\MAIN
COLD:= true;		\Flag cold start for displaying title
CPUREG:= GETREG;
DATASEG:= CPUREG(12);

FontTable:= Reserve(FontSize);

XX:= RLRES(100);  YY:= RLRES(100);  ZZ:= RLRES(100);  SS:= RLRES(100);
TYPE:= RESERVE(2*100);
BOND0:= RESERVE(2*100);   BOND1:= RESERVE(2*100);
IMAGE:= RESERVE(2);
IMAGE(0):= MALLOC(320/16*200);

TITLE:= RESERVE(9);

\Get printer port address from BIOS system parameters:
LPT1PORT:= PEEK($40, $08) ! SWAP(PEEK($40, $09));
CROSSEYE:= false;	\Start with eyes un-swapped

DispMode:= Getvid;

LoadFont;
Setvid($10);		\640x350x16
\Set up font so it can be displayed, i.e. set up vector 43
Cpureg(AX):= $1121;
Cpureg(BX):= $0002;	\25 rows
Cpureg(CX):= 14;	\bytes per char
Cpureg(ES):= DataSeg;
Cpureg(BP):= FontTable;
Softint($10);

INTRO;
GETIMAGE;
OPENI(0);
COMMAND;
EXIT;
end;	\MAIN
