\NODEPLT.XPL	OCT-21-86
\
\This program shows how to trace the lines connecting N nodes equally spaced
\ on the perimeter of a circle. Each node is connected to all the other nodes.
\ The algorithm used here is intended for a plotter and is the most efficient
\ method possible. That is, it retraces the minimum length of lines.
\
\For an odd number of nodes there are no retraced lines which is 100% efficient.
\ But for an even number of nodes the number of retraced lines in N/2 -1. Is
\ this the most efficient solution?
\
\A node can have either an even or odd number of lines connected to it. Tracing
\ a group of nodes where each is connected to the rest by an even number of
\ lines can be done without a single retrace since each pair of lines provides
\ an entrance and exit for each node. However, tracing a group of nodes each
\ with an odd number of lines cannot be done without some retraces.
\
\If a node has an odd number of lines, such as three, and you begin tracing
\ outside the node, then you must enter, exit and enter the node such that you
\ end the trace by entering the node. If the other nodes also have an odd number
\ of lines, then you must simultaneously enter each of these nodes when you
\ finish, which of course is impossible. Actually, you don't need to enter the
\ node where you started. This leaves N-2 nodes which have impossible endings.
\ Now consider retraces. A retraced line effectively adds a line to a pair of
\ nodes giving them each an even number of lines. The minimum number of retraced
\ lines which can resolve N-2 nodes with impossible endings in (N-2)/2 = N/2 -1.
\
\This is what the algorithm does. Furthermore, since all the retraced lines are
\ between adjacent nodes their total length is a minimum.

code	\Complete set, standard XPL V5.6 intrinsics
ABS=0,		RAN=1,		REM=2,		RESERVE=3,
SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
ABORT=16,	TRAP=17,	FREE=18,	RERUN=19,
GETHP=20,	SETHP=21,	GETERR=22,	CURSOR=23,
SCAN=24,	SETRUN=25,	HEXIN=26,	HEXOUT=27,
CHAIN=28,	OPENF=29,	WRITE=30,	READ=31,
TESTPT=32,	SETTXT=33,	SETHI=34,	SETMIX=35,
SETLO=36,	BUTTON=37,	PADDLE=38,	SOUND=39,
CLEAR=40,	POINT=41,	LINE=42,	MOVE=43,
SCREEN=44,	BLOCK=45,	FIX=50,
BACKUP=64,	HICHAR=65,	PEEK=66,	POKE=67;
code real	RLRES=46,	RLIN=47,	RLOUT=48,
FLOAT=49,	RLABS=51,	FORMAT=52,	SQRT=53,
LN=54,		EXP=55,		SIN=56,		ATAN2=57,
MOD=58,		LOG=59,		COS=60,		TAN=61,
ASIN=62,	ACOS=63;

def	INTSIZE= 2, DEV=0;
def	PI2= 6.2831853, SIZE=96.0;

int	I,		\Scratch
	DONE,		\Array: Indicates which lines have been done
	NODES,		\Total number of nodes (input)
	LINES,		\Total number of lines
	FROM,		\Node we're drawing (or moving) from
	TO,		\Node we're drawing (or moving) to
	DRAWS,		\Number of lines drawn so far
	REDRAWS,	\Number of lines redrawn (or retraced)
	X, Y;		\Array: Coordinates of nodes


begin	\Main
OPENO(DEV);
TEXT(0,"HOW MANY NODES? ");
NODES:= INTIN(0);
DONE:= RESERVE(NODES *INTSIZE);
for I:= 0, NODES-1 do DONE(I):= >ESERVE(NODES *INTSIZE);
\Initialize the DONE array, mark off nonsense lines where FROM = TO
for FROM:=0, NODES-1 do
    for TO:=0, NODES-1 do
	DONE(FROM,TO):= if FROM=TO then true else false;

X:= RESERVE(NODES *INTSIZE);
Y:= RESERVE(NODES *INTSIZE);

for I:=0, NODES-1 do
	begin
	X(I):= FIX(SIZE *SIN( FLOAT(I) *PI2/ FLOAT(NODES) )) + 140;
	Y(I):= 96 - FIX(SIZE *COS( FLOAT(I) *PI2/ FLOAT(NODES) ));
	end;

\Calculate the number of lines we need to draw
LINES:= NODES *(NODES-1) /2;

DRAWS:= 0;
REDRAWS:= 0;
FROM:= 0;
TO:= 0;

CLEAR;
SETHI;				\Get into graphics mode
MOVE(X(TO), Y(TO));

loop	begin
	I:=0;
	loop	begin
		TO:= TO +1;
		if TO = NODES then TO:= 0;
		I:= I +1;
		if not DONE(FROM,TO) then
			begin
			LINE(X(TO), Y(TO), $40);	\Draw a line
			DRAWS:= DRAWS +1;
			TEXT(DEV,"  DRAW FROM:	"); INTOUT(DEV, FROM);
			TEXT(DEV," TO:	"); INTOUT(DEV, TO); CRLF(DEV);

			DONE(TO,FROM):= true;		\Mark it done
			DONE(FROM,TO):= true;		\ (both directions)
			quit;
			end;
		if I > NODES then	\We're back at the initial node +1
			begin
			LINE(X(TO), Y(TO), $40);
			REDRAWS:= REDRAWS +1;
			TEXT(DEV,"REDRAW FROM:	"); INTOUT(DEV, FROM);
			TEXT(DEV," TO:	"); INTOUT(DEV, TO); CRLF(DEV);
			quit;
			end;
		end;

	if DRAWS >= LINES then quit;
	FROM:= TO;
	end;

TEXT(DEV,"NODES = "); INTOUT(DEV,NODES);
TEXT(DEV,"	LINES = "); INTOUT(DEV,DRAWS);
TEXT(DEV,"	REDRAWS = "); INTOUT(DEV,REDRAWS);
CRLF(DEV);

RESTART;
end;	\MAIN
