\Routines to convert numbers to strings and vice versa

\func	ATOI(STR, N);	CONVERT STRING TO INTEGER
\func	real ATOF(STR, N);	CONVERT STRING TO REAL
\proc	ITOA(N, STR, X);	CONVERT INTEGER TO STRING
\proc	FTOA(N, M, STR, X);	CONVERT REAL TO STRING
\proc	FLOUT(N, M, X);		PRINT A REAL, RIGHT JUSTIFY, SCREEN
\proc	FLOUTF(D, N, M, X);	PRINT REAL. LEFT JUSTIFY, DEVICE


func	ATOI(STR, N);		\Convert ASCII string to integer
addr	STR;
int	N;
int I, X, C;
begin
X:= 0;
for I:= 0, N-1 do
	begin
	C:= STR(I) & $7F;
	if C=^. then return -1;
	C:= C - ^0;
	if C>=0 & C<=9 then X:= X *10 + C;
	end;
return X;
end;	\ATOI



func real ATOF(STR, N);		\Convert ASCII string to floating point
addr	STR;
int	N;
int	I, C;
real	X, Y;
begin
X:= 0.0;   Y:= 0.0;
for I:= 0, N-1 do
	begin
	C:= STR(I) & $7F;
	if  C>=^0 & C<=^9  !  C=^. then
		begin
		if C=^. then 
			[if Y = 0.0 then Y:= 1.0]
		else	begin
			X:= X *10.0 + FLOAT(C-^0);
			Y:= Y *10.0;
			end;
		end;
	end;
if Y > 0.0 then X:= X /Y;
return X;
end;	\ATOF



proc	ITOA(N, STR, X);		\CONVERT INTEGER TO ASCII STRING
\This produces a string containing the ASCII representation of an
\ integer. The string is assumed to be at least N bytes long. The last byte
\ does NOT have the high bit set. The number is right rather than left
\ justified. If the quantity will not fit then the first byte of the string
\ is set to "*" and the remainder of the string is undefined. N must be at
\ least one.
int	N;
char	STR;
int	X;
int	I, NEG;
begin
NEG:= X < 0;
if NEG then X:= -X;
for I:= 0, N-1 do
	begin
	X:= X /10;
	STR(N-I-1):= REM(0) +^0;
	end;
if X >= 1 then [STR(0):= ^*;   return];
I:= 0;
while I<N & STR(I)=^0 do [STR(I):= ^ ;   I:= I +1];
if I = N then STR(I-1):= ^0;
if NEG then
	[if I >= 1 then STR(I-1):= ^- else STR(0):= ^*];
end;	\ITOA



proc	FTOA(N, M, STR, X);	\Convert floating point to ASCII string
\This produces a string containing the ASCII representation of a floating
\ point number. The string is assumed to be at least N+M+1 bytes long. It
\ produces N bytes before the decimal and M bytes after it. The last byte
\ does NOT have the high bit set. The number is right rather than left
\ justified. If the quantity will not fit then the first byte of the string
\ is set to "*" and the remainder of the string is undefined. The decimal
\ itself is always present. It is in the last byte if M=0 (which is legal).
\ N must be at least one. Exponential and other notations are not
\ supported.
int	N, M;
char	STR;
real	X;
int	I, NEG, DIGIT, LEN;
real	HOLD, EXPO, COMP, ROUND;


	proc	FILL;
	int	I;
	begin
	for I:= 0, LEN-1 do STR(I):= ^*;
	end;


begin
LEN:= if M = 0 then N else N +M +1;
NEG:= X < 0.0;
if NEG then X:= -X;
ROUND:= 0.5;
for I:= 0, M-1 do ROUND:= ROUND *0.1;
X:= X +ROUND;
HOLD:= X;
EXPO:= 0.0;   COMP:= 1.0;
for I:= 0, N-1 do
	begin
	DIGIT:= IFIX(MOD(X, 10.0) -0.5);
	EXPO:= EXPO + COMP *FLOAT(DIGIT);
	COMP:= COMP *10.0;
	STR(N-I-1):= DIGIT +^0;
	X:= X /10.0;
	end;
if X >= 1.0 then [FILL;   return];

I:= 0;
while I<N & STR(I)=^0 do [STR(I):= ^ ;   I:= I +1];
if I = N then [STR(I-1):= ^0;   I:= I -1];
if NEG then
	[if I <= 0 then [FILL;   return] else STR(I-1):= ^-];

X:= HOLD -EXPO;
STR(N):= ^.;
for I:= 1, M do
	begin
	X:= X *10.0;
	DIGIT:= IFIX(X-0.5);
	STR(N+I):= DIGIT +^0;
	X:= X -FLOAT(DIGIT);
	end;
end;	\FTOA



proc	FLOUT(N, M, X);
int	N, M;
real	X;
char	STR;
int	I, LEN;
begin
STR:= RESERVE(N+M+1);
FTOA(N, M, STR, X);
I:= 0;
while STR(I)=^  do I:= I +1;
LEN:= if M = 0 then N else N+M+1;
for I:= I, LEN-1 do CHOUT(0, STR(I));
end;	\FLOUT



proc	FLOUTF(D, N, M, X);
int	D, N, M;
real	X;
char	STR;
int	I, LEN;
begin
STR:= RESERVE(N+M+1);
FTOA(N, M, STR, X);
LEN:= if M = 0 then N else N+M+1;
for I:= 0, LEN-1 do CHOUT(D, STR(I));
end;	\FLOUTF
