\OUT.XPL	DEC-12-85
\PROGRAM TO REDIRECT BYTE-STREAM OUTPUT TO AN ALTERNATIVE DEVICE.

\WARNING! IF OUTPUT IS REDIRECTED TO THE DISK, AN OUTPUT BUFFER MUST BE
\ SPECIFIED. FOR EXAMPLE, IF THE PRINT UTILITY (PR) IS USED AND OUTPUT
\ HAS BEEN CHANGED FROM 2 TO 3, THEN PR.SAV MUST PROVIDE AN OUTPUT BUFFER.
\ ALSO, THE OUTPUT FILE MUST BE OPENED FROM APEX USING THE OPEN COMMAND.

code
ABS=0,		RAN=1,		REM=2,		RESERVE=3,
SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
TRAP=17,	WRITE=30,	READ=31;

def	BEL=$07;	\ASCII CHARACTER(S)
def	RESSIZ=48,	\SIZE OF RESCOD.SYS (BLOCKS)
	SYSDEV=$52,	\OFFSET INTO SYSTEM PAGE TO GET SYSTEM UNIT NUMBER
	LINIDX=$5A,	\COMMAND FILE INDEX
	EXECUT=$5D,	\COMMAND FILE FLAG
	;
int	UNIT,		\UNIT RESCOD.SYS IS ON
	LDEV,		\LOGICAL DEVICE NUMBER (FROM DEVICE)
	PDEV,		\PHYSICAL DEVICE NUMBER (TO DEVICE)
	I,		\INDEX
	PAGE,		\PAGE THAT HANDLER VECTORS START ON
	DEVTAB,		\DEVICE TABLE (IN SYSTEM PAGE)
	BLOCK,		\ARRAY: START AND END BLOCK
	;
addr	BUFFER,		\BUFFER USED TO GET PHYSICAL DEVICE VECTORS FROM RESCOD
	SYSPAG,		\SYSTEM PAGE STARTING ADDRESS
	BASE,		\BASE ADDRESS OF HANDLER FUNCTION CODE VECTORS
	;



proc	ERROR(STR);	\DISPLAY ERROR MESSAGE AND RESTART PROGRAM
addr	STR;
begin
CRLF(0);
TEXT(0,"NOPE - "); TEXT(0,STR); CRLF(0);
CHOUT(0,BEL);
CRLF(0);

SYSPAG(EXECUT):= $FF;		\ABORT ANY COMMAND FILE
SYSPAG(LINIDX):= $FF;
RESTART;
end;	\ERROR



proc	CHKPAGE(PAGE);	\"PAGE" MUST BE IN THE RANGE: $B1-$BF OR $E0-$FF
int	PAGE;
begin
if PAGE<$B1 then
	ERROR("HANDLER MUST BE IN RESCOD.SYS");
case PAGE/16 of
  $B,$E,$F:	\OK
other	ERROR("HANDLER MUST BE IN RESCOD.SYS");
end;	\CHKPAGE



begin	\MAIN
BUFFER:=RESERVE($200);	\(NOTE THAT HANDLER VECTORS MAY CROSS PAGE BOUNDARIES)
DEVTAB:=$BFC0;
SYSPAG:=$BF00;
BLOCK:= RESERVE(4);

TEXT(0,"-- REDIRECT OUTPUT, V1.8 --
");
loop	begin
	TEXT(0,"FROM, TO DEVICES: ");
	LDEV:=INTIN(0);
	PDEV:=INTIN(0);
	if LDEV>=0 & LDEV<=7 & PDEV>=0 & PDEV<=7 then quit;
	CHOUT(0,BEL);
	end;

\GET THE PHYSICAL DEVICE VECTORS FROM RESCOD.SYS ON DISK:
BASE:= DEVTAB(PDEV);
PAGE:= SWAP(BASE) & $FF;
CHKPAGE(PAGE);

\CONVERT PAGE IN MEMORY TO PAGE IN RESCOD.SYS:
case PAGE/16 of
  $B:	 PAGE:= PAGE -$B0;
  $E,$F: PAGE:= PAGE -$D0
other;

UNIT:= SYSPAG(SYSDEV);
READ(UNIT, 17+PAGE, BUFFER, 2);

\CHANGE THE LOGICAL DEVICE VECTORS:
\ FUNCTION CODES ARE 3-BYTE ENTRIES:
\	0	OPENI
\	3	OPENO
\	6	CHIN
\	9	CHOUT
\	12	CLOSE

I:= BASE & $FF;			\GET INDEX FOR PHYSICAL DEVICE IN THE BUFFER
BASE:= DEVTAB(LDEV);

CHKPAGE(BUFFER(I+5));		\CHECK FOR LEGAL VECTORS (HE MAY NOT HAVE THE
CHKPAGE(BUFFER(I+11));		\ SYSTEM DISK INSERTED)
CHKPAGE(BUFFER(I+14));

\LOGICAL DEVICE VECTOR := PHYSICAL DEVICE VECTOR
BASE(4):= BUFFER(I+4);		\OPENO
BASE(5):= BUFFER(I+5);
BASE(10):= BUFFER(I+10);	\CHOUT
BASE(11):= BUFFER(I+11);
BASE(13):= BUFFER(I+13);	\CLOSE
BASE(14):= BUFFER(I+14);

\MAKE SURE THE DEVICE IS OPEN. (REDIRECTED CONSOLE OUTPUT SELDOM OPENS ANYTHING.)
\THIS MIGHT GENERATE AN I2L ERROR FOR THE DISK, BUT WE'RE FINISHED ANYWAY.
OPENO(LDEV);
end;	\MAIN
