\PENTS3D.XPL	DEC-08-90
\Pentominoes Puzzle Solver in 3-D
\ by Lenny Boreal

include	\CXPL\CODESI;

def	BEL=$07, CR=$0D, FF=$0C, ESC=$1B, SP=$20, CTRL=$40;   \Ascii characters
def	BOXW=5, BOXH=4, BOXD=3;	\Box dimensions
int	PIECE,		\Array: Puzzle pieces and their orientations
	OMAX,		\Array: Maximum orientation subscript for each piece
	SOLVECNT,	\Number of solutions found so far
	TIME,		\Amount of time to delay (18ths of seconds)
	PAUSEFL,	\Flag: Wait for 'ENTER' keystroke after each solution
	SOUNDFL,	\Flag: Beep on each solution
	CELLEND;	\Last cell in BOX +1
real	COUNT;		\Number of tries (pieces inserted or removed)
addr	BOX,		\Array: Contains the pieces (i.e. the puzzle)
	AVAIL;		\Array: Flags indicating piece is not in BOX
int	CPUREG,		\Pointer to system info
	MODE0,		\Video mode when this program started
	ZX, ZY,		\Paralax caused by depth (Z) for X and Y axes
	ZS;		\Size of cubes


include	LIB;



proc	EXIT;		\Exit program in a tidy fashon
begin
SETVID(MODE0);		\Restore initial video mode
exit;
end;	\EXIT



proc	DELAY(T);	\Delay T 18ths of a second
int	T;
int	I;
begin
for I:= 1, T do
	begin
	SOUND(0, 1, 1);
	if CHKKEY then return;		\Abort delay if keystroke
	end;
end;	\DELAY

\----------------------------------------------------------------------

proc	SHOWBLK(CELL, COLOR);	\Display a block (cube)
int	CELL, COLOR;
int	X, Y, Z, X1, Y1;


	proc	CUBE(X0, Y0, C);
	int	X0, Y0, C;
	int	X1, Y1, X2, Y2;
	int	J;
	begin
	\C:= C ! $80;

	X1:= X0 -ZX;   Y1:= Y0 +ZY;
	RECTANGLE(X1-ZS, Y1-ZS, X1+ZS, Y1+ZS, C);

	\for J:= 1, 7 do
	\	RECTANGLE(X1-ZS+J, Y1-ZS+J, X1+ZS-J, Y1+ZS-J, C);

	X2:= X0 +ZX;   Y2:= Y0 -ZY;
	\RECTANGLE(X2-ZS, Y2-ZS, X2+ZS, Y2+ZS, C);

	MOVE(X2-ZS, Y2-ZS);   LINE(X2+ZS, Y2-ZS, C);
	LINE(X2+ZS, Y2+ZS, C);

	MOVE(X1-ZS, Y1-ZS);   LINE(X2-ZS, Y2-ZS, C);
	MOVE(X1+ZS, Y1-ZS);   LINE(X2+ZS, Y2-ZS, C);
	\MOVE(X1-ZS, Y1+ZS);   LINE(X2-ZS, Y2+ZS, C);
	MOVE(X1+ZS, Y1+ZS);   LINE(X2+ZS, Y2+ZS, C);
	end;	\CUBE


begin	\SHOWBLK
X:= CELL /20;
Y:= REM(0) /4;
Z:= REM(0);

X1:= X*80 +170 -Z*ZX*\2\3;
Y1:= 479 - (135 + Y*80 - Z*ZY*\2\3);
CUBE(X1, Y1, COLOR);
end;	\SHOWBLK



proc	SHOWCNT;	\Display the numbers of solutions and tries
begin
CURSOR(1, 29);   INTOUT(0, SOLVECNT);
CURSOR(30, 29);   RLOUT(0, COUNT);
\TEXT(0, "  ");   INTOUT(0, ZX);	\\Used for adjusting asthetic values
\TEXT(0, "  ");   INTOUT(0, ZY);
\TEXT(0, "  ");   INTOUT(0, ZS);
end;	\SHOWCNT



proc	SHOWPIECE(CELL, P2, COLOR);	\Display a puzzle piece
int	CELL, P2, COLOR;
begin
SHOWBLK(CELL, COLOR);
SHOWBLK(CELL+P2(0), COLOR);
SHOWBLK(CELL+P2(1), COLOR);
SHOWBLK(CELL+P2(2), COLOR);
SHOWBLK(CELL+P2(3), COLOR);

SHOWCNT;
DELAY(TIME-1);
end;	\SHOWPIECE



proc	SHOWBOX;	\Display the entire puzzle BOX
int	X, Y, Z, C;
begin
C:= 0;
for X:= 0, BOXW do
    for Y:= 0, BOXH do
	for Z:= 0, BOXD do
		begin
		case of X=BOXW, Y=BOXH, Z=BOXD: []
		other	SHOWBLK(C, BOX(C));
		C:= C +1;
		end;
SHOWCNT;
end;	\SHOWBOX



proc	DOKEY;		\Handle keyboard commands
int	CH;
begin
CH:= GETKEY;
case CH of
  ^S:	[ZS:= ZS +1;   CHOUT(0, FF);   SHOWBOX];
  ^X:	[ZS:= ZS -1;   CHOUT(0, FF);   SHOWBOX];
  ^A:	[ZY:= ZY -1;   CHOUT(0, FF);   SHOWBOX];
  ^Z:	[ZY:= ZY +1;   CHOUT(0, FF);   SHOWBOX];
  ^.:	[ZX:= ZX +1;   CHOUT(0, FF);   SHOWBOX];
  ^/:	[ZX:= ZX -1;   CHOUT(0, FF);   SHOWBOX];
  CR:	  ;		\Just continue and don't beep
  SP:	  SOUNDFL:= ~SOUNDFL;
  ^p, ^P: PAUSEFL:= ~PAUSEFL;
  ^S-CTRL: CH:= GETKEY;
  ^C-CTRL, ESC:  EXIT
other	begin
	if CH>=^0 & CH<=^9 then
		begin
		if TIME =0 then SHOWBOX;  \Display is not current if TIME = 0
		TIME:= CH - ^0;
		if TIME >4 then TIME:= (TIME-2)*(TIME-2);
		if TIME >=49 then TIME:= $7FFF;
		end
	else CHOUT(0, BEL);		\Illegal command
	end;
end;	\DOKEY

\----------------------------------------------------------------------

proc	PUTPIECE(CELL);	\Put piece into box at CELL location
addr	CELL;	\Current cell position
int	P,	\Current piece
	OR,	\Current piece's orientation
	P1,	\Array: Contains all orientations for the current piece
	P2;	\Array: Contains positions of cells for the oriented piece
addr	C;	\Scratch (cell/color)
begin
for P:= 0, 11 do			\For all available pieces...
    if AVAIL(P) then
	begin
	P1:= PIECE(P);
	for OR:= 0, OMAX(P) do		\For all orientations of the piece...
		begin			\Most frequently executed code is right
		P2:= P1(OR);		\ here
		case of			\If any cell is occupied then forget it
		  CELL(P2(0)), CELL(P2(1)), CELL(P2(2)), CELL(P2(3)): []
		other	begin			\Piece fits
			C:= P + 2;		\Put piece into box
			CELL(0):= C;		\Color:= piece value +2
			CELL(P2(0)):= C;	\(Color can't be 0 which is vacant)
			CELL(P2(1)):= C;
			CELL(P2(2)):= C;
			CELL(P2(3)):= C;

			AVAIL(P):= false;	\Mark piece unavailable
			COUNT:= COUNT +1.0;

			if CHKKEY then DOKEY;

			if TIME \#0\ then
				begin
				SHOWPIECE(CELL-BOX, P2, C);
				if CHKKEY then DOKEY;
				end;

			C:= CELL +1;		\Scan for next empty cell
			while C(0) \#0\ do C:= C +1;

			if C # CELLEND then	\Try to put next piece into box
				PUTPIECE(C)
			else	begin		\Solved !!
				SOLVECNT:= SOLVECNT +1;
				if SOUNDFL then CHOUT(0, BEL);
				SHOWBOX;
				if PAUSEFL then
					begin
					CURSOR(57, 29);
					TEXT(0, " HIT ENTER TO CONTINUE");
					OPENI(0); \(Messages must be same length)
					DOKEY;
					CURSOR(57, 29);	\Restore original message
					TEXT(0, "KEYS 0-9 CONTROL SPEED");
					end;
				end;

			CELL(0):= 0;		\Remove piece from box
			CELL(P2(0)):= 0;
			CELL(P2(1)):= 0;
			CELL(P2(2)):= 0;
			CELL(P2(3)):= 0;

			AVAIL(P):= true;	\Mark piece available
			COUNT:= COUNT +1.0;

			if TIME \#0\ then
				begin
				SHOWPIECE(CELL-BOX, P2, 0);
				if CHKKEY then DOKEY;
				end;
			end;
		end;
	end;
end;	\PUTPIECE

\----------------------------------------------------------------------

proc	INIT;		\Initialize
int	I, X, Y, Z, C;
begin
FORMAT(1,0);
TIME:= 2;		\Ten moves per second
PAUSEFL:= true;
SOUNDFL:= true;
COUNT:= 0.0;
SOLVECNT:= 0;
ZX:=9;   ZY:=6;   ZS:= 29;

\Set up empty box with borders:
C:= 0;
for X:= 0, BOXW do
    for Y:= 0, BOXH do
    	for Z:= 0, BOXD do
		begin
		case of X=BOXW, Y=BOXH, Z=BOXD:
			BOX(C):= ^#	\Border
		other	BOX(C):= 0;	\Empty
		C:= C +1;
		end;

BOX(C-1):= 0;		\Last corner is used to detect when box is full
CELLEND:= BOX +C -1;	\Note that no piece can occupy this vacant cell

for I:= 0, 11 do
	AVAIL(I):= true;

\Maximum orientations for each piece (-1):
\	0   1   2   3   4   5   6   7   8   9  10  11
OMAX:= [2,  0,  2, 11, 11, 11, 11, 15, 15, 15, 23, 23];

\This defines the pieces and all their orientations. Each piece consists of
\ five cells. The first cell scanned is the origin (=0). The other cells are
\ offsets from this origin.
PIECE:=
    [ [ [ 3,  4,  5, 8],	\0  -#--
	[16, 20, 24, 40],	\   ###-
	[19, 20, 21, 40] ],	\   -#--

      [ [20, 40, 60, 80] ],	\1  #####

      [ [ 4,  6,  7,  8],	\2  ###-
	[ 4,  8, 28, 48],	\   #---
	[ 1,  2, 22, 42] ],	\   #---
	\Other possible orientations are eliminated to elimate
	\ rotations and reflections in the final solutions)

      [ [ 1,  5,  6, 10],	\3  #---
	[ 4, 16, 20, 36],	\   ##--
	[19, 20, 38, 39],	\   --##
	[ 1,  3,  4,  7],
	[20, 24, 44, 48],
	[20, 21, 41, 42],
	[ 4,  5,  9, 10],
	[16, 20, 32, 36],
	[ 1, 19, 20, 39],
	[ 3,  4,  6,  7],
	[ 4, 24, 28, 48],
	[ 1, 21, 22, 42] ],

      [ [ 4,  7,  8,  9],	\4  ###-
	[12, 16, 20, 40],	\   -#--
	[ 1,  2, 21, 41],	\   -#--
	[ 1,  2,  5,  9],
	[20, 24, 28, 40],
	[20, 39, 40, 41],
	[ 2,  3,  4,  8],
	[ 4,  8, 24, 44],
	[18, 19, 20, 40],
	[ 4,  5,  6,  8],
	[20, 36, 40, 44],
	[20, 21, 22, 40] ],

      [ [ 1,  5,  9, 10],	\5  ##--
	[12, 16, 20, 32],	\   -#--
	[ 1, 20, 39, 40],	\   -##-
	[ 1,  4,  7,  8],
	[20, 24, 28, 48],
	[ 1, 21, 41, 42],
	[ 4,  5,  6, 10],
	[ 4, 20, 36, 40],
	[18, 19, 20, 38],
	[ 2,  3,  4,  6],
	[ 4, 24, 44, 48],
	[20, 21, 22, 42] ],

      [ [ 2,  4,  5,  6],	\6  ###-
	[ 4, 24, 40, 44],	\   #-#-
	[ 1,  2, 20, 22],
	[ 1,  5,  8,  9],
	[ 4,  8, 20, 28],
	[ 1, 21, 40, 41],
	[ 1,  4,  8,  9],
	[ 8, 20, 24, 28],
	[ 1, 20, 40, 41],
	[ 1,  2,  4,  6],
	[ 4, 20, 40, 44],
	[ 2, 20, 21, 22] ],

      [ [ 4,  8, 12, 20],	\7  -#--
	[ 1,  4,  8, 12],	\   -#--
	[20, 40, 60, 61],	\   -#--
	[ 4,  8, 12, 32],	\   ##--
	[ 4,  8, 12, 13],
	[ 1, 20, 40, 60],
	[ 8, 12, 16, 20],
	[ 4,  8, 11, 12],
	[ 1, 21, 41, 61],
	[20, 24, 28, 32],
	[ 1,  5,  9, 13],
	[20, 40, 59, 60],
	[ 4, 24, 44, 64],
	[20, 40, 56, 60],
	[ 4, 20, 40, 60],
	[20, 40, 60, 64] ],

      [ [ 4,  8, 12, 24],	\8  -#--
	[ 4,  5,  8, 12],	\   -#--
	[20, 40, 41, 60],	\   ##--
	[ 4,  8, 12, 28],	\   -#--
	[ 4,  8,  9, 12],
	[20, 21, 40, 60],
	[12, 16, 20, 24],
	[ 4,  7,  8, 12],
	[19, 20, 40, 60],
	[16, 20, 24, 28],
	[ 3,  4,  8, 12],
	[20, 39, 40, 60],
	[16, 20, 40, 60],
	[20, 36, 40, 60],
	[20, 24, 40, 60],
	[20, 40, 44, 60] ],

      [ [ 4,  8, 16, 20],	\9  ---#
	[ 3,  4,  7, 11],	\   ---#
	[20, 40, 41, 61],	\   --##
	[ 4,  8, 28, 32],	\   --#-
	[ 4,  8,  9, 13],
	[19, 20, 39, 59],
	[ 4, 12, 16, 20],
	[ 4,  7,  8, 11],
	[20, 21, 41, 61],
	[ 4, 24, 28, 32],
	[ 4,  5,  9, 13],
	[20, 39, 40, 59],
	[20, 24, 44, 64],
	[20, 36, 40, 56],
	[16, 20, 36, 56],
	[20, 40, 44, 64] ],

      [ [ 3,  4,  8,  9],	\10 ##--
	[ 1, 19, 20, 40],	\   -##-
	[ 4, 16, 20, 40],	\   -#--
	[ 3,  4,  5,  7],	\   ----
	[20, 21, 22, 41],
	[20, 24, 28, 44],
	[ 1,  5,  6,  9],
	[20, 21, 39, 40],
	[20, 24, 36, 40],
	[ 2,  3,  4,  7],
	[19, 20, 21, 41],
	[16, 20, 24, 44],
	[ 4,  5,  7,  8],
	[ 1, 21, 22, 41],
	[ 4, 24, 28, 44],
	[ 4,  5,  6,  9],
	[19, 20, 21, 39],
	[16, 20, 24, 36],
	[ 1,  3,  4,  8],
	[19, 20, 40, 41],
	[16, 20, 40, 44],
	[ 3,  4,  5,  9],
	[18, 19, 20, 39],
	[12, 16, 20, 36] ],

      [ [ 1,  3,  4,  5],	\11 ###-
	[ 4, 20, 24, 44],	\   ##--
	[ 1,  2, 21, 22],	\   ----
	[ 1,  4,  5,  9],	\   ----
	[ 4,  8, 20, 24],
	[19, 20, 39, 40],
	[ 1,  2,  4,  5],
	[20, 24, 40, 44],
	[ 1, 20, 21, 22],
	[ 4,  5,  8,  9],
	[ 4, 16, 20, 24],
	[ 1, 20, 21, 40],
	[ 1,  4,  5,  6],
	[16, 20, 36, 40],
	[ 1,  2, 20, 21],
	[ 3,  4,  7,  8],
	[ 4,  8, 24, 28],
	[ 1, 20, 21, 41],
	[ 1,  2,  5,  6],
	[ 4, 20, 24, 40],
	[ 1, 19, 20, 21],
	[ 1,  4,  5,  8],
	[ 4, 20, 24, 28],
	[20, 21, 40, 41] ] ];
end;	\INIT

\----------------------------------------------------------------------

begin	\MAIN
BOX:= RESERVE((BOXW+1) *(BOXH+1) *(BOXD+1));
AVAIL:= RESERVE(12);

CPUREG:= GETREG;
MODE0:= GETVID;
SETVID($12);	\VGA: 640x480x16 colors

TEXT(0, "		       3 - D   P E N T O M I N O E S

			      by Lenny Boreal





			 Keyboard controls:

			   0       Maximum speed

			   9       Single step

			   Space   Sound off and on

			   P       Pause off and on

			   Esc     Exit");

CURSOR(25, 28);   TEXT(0, "Press ENTER to begin...");

OPENI(0);
if CHIN(0) then;
CHOUT(0, FF);
TEXT(0, "		       3 - D   P E N T O M I N O E S");
INIT;
CURSOR(57, 29);   TEXT(0, "KEYS 0-9 CONTROL SPEED");

PUTPIECE(BOX);	\Put first piece into corner

SHOWBOX;
CURSOR(29, 10);   TEXT(0, " ALL SOLUTIONS FOUND ");
CURSOR(29, 12);   TEXT(0, " PRESS ENTER TO EXIT ");
CHOUT(0, BEL);   DELAY(2);   CHOUT(0, BEL);   DELAY(2);   CHOUT(0, BEL);
OPENI(0);
if GETKEY then;
EXIT;
end;	\MAIN
