\PI.XPL		DEC-13-93
\PI CALCULATOR
\BY RICHARD WILLMORE
\OCT-18-86, MODIFIED FOR AMIGA AND MAC II.
\NOV-07-90, MODIFIED FOR PC.
\DEC-13-93, SPEEDED UP AND MODIFIED BY L.B.

\This uses Machin's formula to calculate pi to a large number of places.
\
\	Pi = 16*atan(1/5) - 4*Atan(1/239)
\
\The Taylor's series for the arctangent is:
\
\	Atan(X) = X - X^3/3 + X^5/5 - X^7/7 + ...
\
\This converges very quickly when X is small. When X = 1/5 for example,
\ more than one decimal digit of accuracy is added for each term
\ calculated.


code	REM=2, RESERVE=3, CHIN=7, CHOUT=8, CRLF=9,
	INTIN=10, INTOUT=11, TEXT=12, OPENO=14, CLOSE=15;

int	PLACES,	\Number of places requested
	SIZE,	\Number of places necessary to calculate to prevent rounding errors
	DEV;	\Output device number



func	STRDIV(DIVISOR, STR);	\STR:= STR / DIVISOR
int	DIVISOR;		\Returns 'true' if STR equals 0
char	STR;
int	I, J;
begin
loop	begin			\Skip leading zeros
	for J:= 0, SIZE-1 do
		if STR(J) then quit;
	return true;
	end;

I:= 0/1;			\Clear remainder
for I:= J, SIZE-1 do
	STR(I):= (10 *REM(0) + STR(I)) /DIVISOR;
return false;
end;	\STRDIV



proc	STRADD(A, B);		\B:= B + A
char	A, B;
int	SUM, CARRY, I;
begin
CARRY:= 0;
for I:= -(SIZE-1), 0 do
	begin
	SUM:= B(-I) + A(-I) + CARRY;
	if SUM >= 10 then
		[SUM:= SUM -10;   CARRY:= 1]
	else 	CARRY:= 0;
	B(-I):= SUM;
	end;
end;	\STRADD



proc	STRSUB(A, B);		\B:= B - A
char	A, B;
int	DIF, BORROW, I;
begin
BORROW:= 0;
for I:= -(SIZE-1), 0 do
	begin
	DIF:= B(-I) - A(-I) - BORROW;
	if DIF < 0 then
		[DIF:= DIF +10;   BORROW:= 1]
	else	BORROW:= 0;
	B(-I):= DIF;
	end
end;	\STRSUB

\----------------------------------------------------------------------

proc	CALCPI;		\Calculate and output pi to specified number of places
char	TERM,		\Array of digits: A term in the Taylor's series
	RESULT,		\Array of digits: The cumulative sum of terms
	POWER;		\Array of digits: Portion of TERM that is raised to the EXP
int	I,
	ADDTERM,	\Flag to add or subtract terms
	DONE,		\Flag: Done with calculation (TERM = 0)
	EXP;		\Exponent, which is also the divisor
begin
POWER:= RESERVE(SIZE);
TERM:= RESERVE(SIZE);
RESULT:= RESERVE(SIZE);

\16*Atan(1/5) = 16/5^1/1 - 16/5^3/3 + 16/5^5/5 - 16/5^7/7 + ...

for I:= 0, SIZE-1 do
	[POWER(I):= 0;   RESULT(I):= 0];
POWER(0):= 16;				\POWER:= 16/5
STRDIV(5, POWER);
ADDTERM:= true;
EXP:= 1;
repeat	begin				\RESULT:= RESULT +/- POWER /EXP
	for I:= 0, SIZE-1 do TERM(I):= POWER(I);
	STRDIV(EXP, TERM);
	if ADDTERM then
		STRADD(TERM, RESULT)
	else	STRSUB(TERM, RESULT);

	EXP:= EXP +2;
	ADDTERM:= ~ADDTERM;
	DONE:= STRDIV(25, POWER);	\POWER:= POWER/(5^2)
	end;				\Constant is already squared
until DONE;				\Done when POWER = 0

\-4*Atan(1/239) = -4/239^1/1 + 4/239^3/3 - 4/239^5/5 + 4/239^7/7 - ...

for I:= 0, SIZE-1 do POWER(I):= 0;
POWER(0):= 4;
STRDIV(239, POWER);
ADDTERM:= false;
EXP:= 1;
repeat	begin				\RESULT:= RESULT +/- POWER /EXP
	for I:= 0, SIZE-1 do TERM(I):= POWER(I);
	STRDIV(EXP, TERM);
	if ADDTERM then
		STRADD(TERM, RESULT)
	else	STRSUB(TERM, RESULT);

	EXP:= EXP +2;
	ADDTERM:= ~ADDTERM;
	STRDIV(239, POWER);		\POWER:= POWER / (239^2)
	DONE:= STRDIV(239, POWER);
	end;
until DONE;				\Done when POWER = 0

TEXT(DEV, "THE VALUE OF PI TO ");   INTOUT(DEV, PLACES);
TEXT(DEV, " PLACES IS: ");   CRLF(DEV);

for I:= 0, PLACES do POWER(I):= 0;	\Round RESULT
POWER(PLACES+1):= 5;
STRADD(POWER, RESULT);

INTOUT(DEV, RESULT(0));			\Output RESULT
CHOUT(DEV, ^.);
for I:= 1, PLACES do INTOUT(DEV, RESULT(I));
CRLF(DEV);
end;	\CALCPI

\----------------------------------------------------------------------

begin	\MAIN
\TEXT(0, "OUTPUT DEVICE (0-7)? ");
\DEV:= INTIN(0);
DEV:= 0;
OPENO(DEV);

loop	begin
	TEXT(0, "HOW MANY PLACES WOULD YOU LIKE? ");
	PLACES:= INTIN(0);
	if PLACES <= 0 then quit;
	SIZE:= PLACES + PLACES /10 + 2;
	CALCPI;
	CRLF(DEV);
	end;
CLOSE(DEV);
end;	\MAIN
