\PICSIM.XPL	OCT-31-97
\Simulator/Debugger for the PIC 16C74 Microcontroller
\ by Loren Blaney

\REVISIONS:
\1.1, AUG-27-96, Use correct file register names for upper bank in
\ disassembly. Fix cycle counter at 1000. Show correct value for INDF
\ register in hex dump. Fix stack overflow message. Dump shows register
\ names and a dot next to registers that have changed. Any keystroke stops
\ go (G) command, preventing infinite loops. Simulate a power-on reset by
\ setting RAM to random values. Set special registers to proper values for
\ reset (R) command. Display RP0 status. Align tabs for STATUS,RP0. Revised
\ HELP summary.
\1.2, SEP-05-96, Fix DC status for ADDWF and ADDLW.
\1.3, OCT-31-97, Fix SLEEP status. Fix cycle count for DECFSZ and when PCL
\ is modified.

\NOTE:
\Most of the special registers, such as the timers and interrupt
\ controller, are not simulated.


code	RAN=1,		REM=2,		SWAP=4,		EXTEND=5,
	RESTART=6,	CHIN=7,		CHOUTX=8,	TEXTX=12,
	OPENI=13,	OPENO=14,	CLOSE=15,	RERUN=19,
	FSET=24,	FOPEN=29,	CHKKEY=33,	SOFTINT=34,
	GETREG=35,	BLIT=36;

def	ROMSIZE=$1000,	\Size of instruction memory (14-bit words) (power of 2)
	RAMSIZE=$100,	\Size of file register space including hardware registers
	STACKSIZE=8;	\Maximum depth of subroutine stack

def	ROMMASK=ROMSIZE-1;	\Note that ROMSIZE must be a power of 2

def	INDF=$0, PCL=$2, STATUS=$3, FSR=$4, PCLATH=$A, INTCON=$B, \File registers
	OPTION=$81, TRISA=$85, TRISB=$86, TRISC=$87, TRISD=$88, TRISE=$89;

def	BEL=$07, TAB=$09, LF=$0A, FF=$0C, CR=$0D, ESC=$1B;	\Ascii characters
def	TV=0, KB=0, PR=2;

char	RAM(RAMSIZE),	\File registers, including hardware registers
	RAM0(RAMSIZE),	\Copy of RAM at last dump, used to hilite changes
	MAP(RAMSIZE),	\Handles duplicate addresses for PCL, FSR, etc.
	DONE(ROMSIZE),	\'true' if corresponding PC location was executed
	LAB(ROMSIZE);	\'true' if corresponding PC location has label
int	ROM(ROMSIZE+1),	\Code space (last loc is for clearing break point)
	STACK(STACKSIZE); \Subroutine return address stack

int	PC,		\Program counter (PCLACH:PCL)
	CY0, CY1, CY2,	\Cycle counter (enough for 32000,000,000 cycles)
	WR,		\Working register (accumulator)
	ZF, CF, DC,	\Flags: Zero, Carry, Digit Carry
	SP,		\Stack pointer for STACK
	WDT,		\Watchdog timer
	WDTPRE,		\Watchdog prescaler
	CMD,		\Command from keyboard
	BPADDR,		\Break point address
	BPCODE,		\Contents of location before break point was set
	LISTADDR,	\List address
	TRACEMODE,	\Flag: Show state of PIC and wait for Enter key
	PRINTFL,	\Flag: Send output to printer as well as console
	LASTCHAR,	\Last character typed on the keyboard
	BACKFLAG;	\Flag: Used to reread last character struck on keyboard

\DOS file I/O variables:
int	CPUREG,		\Pointer to system info
	PSPSEG,		\Address of PSP segment (for file I/O)
	DATASEG,	\Address of data segment (heap & stack)
	INHAND;		\Input file handle
def	NAMESIZE=40;	\Maximum number of characters in a file name
char	INFILE(NAMESIZE); \Input file name

\=============================== DOS FILE I/O =================================

proc	DOSOPEN;	\Routine to open MS-DOS files
int	INX;		\Index into TAIL
char	TAIL, TLEN(80);



proc	COPYNAME(FROM, TO, EXT);
\Copy file name from FROM to TO and replace any existing extention with EXT
char	FROM, TO,	\Strings are terminated by 0 and must be <= NAMESIZE long
	EXT;		\Extension started with "." and terminated with an
			\ unused char with MSB set
int	I, J, CH;
begin
I:= 0;
loop	begin
	CH:= FROM(I);
	if CH=^. ! CH=0 then quit;
	TO(I):= CH;
	if I < NAMESIZE-1 then I:= I +1;
	end;
for J:= 0, 3 do
	begin
	TO(I):= EXT(J);
	if I < NAMESIZE-1 then I:= I +1;
	end;
TO(I):= 0;
end;	\COPYNAME



procc	GETNAME(NAME);
\Inputs: INX, TAIL, NAMESIZE
char	NAME;
int	I, CH;
begin
while TAIL(INX) =SP do INX:= INX +1;
I:= 0;
repeat	CH:= TAIL(INX);
	INX:= INX +1;
	if CH=SP ! CH=CR then CH:= 0;	\terminate NAME with 0
	NAME(I):= CH;
	if I < NAMESIZE-1 then I:= I +1;
until CH =0;
end;	\GETNAME



begin	\DOSOPEN
TAIL:= TLEN +1;
BLIT(PSPSEG, $80, DATASEG, TLEN, 79);
if TLEN(0)>79 then TLEN(0):= 79;

INX:= 0;
GETNAME(INFILE);			\fatal error if no file specified
if INFILE(0) =0 then
	[TEXTX(0, "No input file specified
");   exit];

COPYNAME(INFILE, INFILE, ".HEX ");
INHAND:= FOPEN(INFILE, 0);		\get input handle
FSET(INHAND, ^I);
OPENI(3);
end;	\DOSOPEN

\==============================================================================

func	GETKEY;		\Get character from keyboard (with backup)
begin
if BACKFLAG then
	begin
	BACKFLAG:= false;
	return LASTCHAR;
	end
else	begin
	LASTCHAR:= CHIN(KB);
	return LASTCHAR;
	end;
end;	\GETKEY



proc	BACKUP;		\Used to reread the last character from the keyboard
begin
BACKFLAG:= true;
end;	\BACKUP



func	GETHEX;		\Get a hex value from the keyboard
int	CH, HEX, FLAG, DIGIT;
begin
HEX:= 0;
FLAG:= false;
loop	begin
	loop	begin
		CH:= GETKEY;
		if CH>=^a & CH<=^z then CH:= CH -$20;	\convert to uppercase

		if CH>=^0 & CH<=^9 then DIGIT:= CH -^0	\convert to hex digit
		else if CH>=^A & CH<=^F then DIGIT:= CH -^A +10
		else quit;

		HEX:= HEX <<4 ! DIGIT;
		FLAG:= true;				\hex digit encountered
		end;
	if FLAG then quit;				\we must have some hex
	end;
return	HEX;
end;	\GETHEX

\------------------------------------------------------------------------------

proc	CHOUT(DEV, CH);	\Output to the console and possibly the printer
int	DEV;
int	CH;
begin
CHOUTX(DEV, CH);
if PRINTFL & DEV#PR then
	CHOUTX(PR, CH);
end;	\CHOUT



proc	CRLF(DEV);	\Output a CR & LF to the console and possibly the printer
int	DEV;
begin
CHOUT(DEV, CR);
CHOUT(DEV, LF);
end;	\CRLF



proc	TEXT(DEV, STR);	\Output text string to device and possibly to the printer
int	DEV;
char	STR;
int	I, CH;
begin
I:= 0;
loop	begin
	CH:= STR(I);
	if CH>=$80 then quit;
	I:= I +1;
	CHOUT(DEV, CH);
	end;
CHOUT(DEV, CH & $7F);
end;	\TEXT



proc	HEX1OUT(DEV, N); \Output ascii hex byte to device
int	DEV, N;
char	HEXDIGIT; \Array: ascii hex DIGITS (0 - F)
begin
HEXDIGIT:= "0123456789ABCDEF ";
CHOUT(DEV, HEXDIGIT(N & $0F));
end;	\HEX1OUT



proc	HEX2OUT(DEV, N); \Output ascii hex byte to device
int	DEV, N;
char	HEXDIGIT; 	\Array: ascii hex DIGITS (0 - F)
begin
HEXDIGIT:= "0123456789ABCDEF ";
CHOUT(DEV, HEXDIGIT(N >>4 & $0F));
CHOUT(DEV, HEXDIGIT(N & $0F));
end;	\HEX2OUT



proc	HEX4OUT(DEV, N); \Output ascii hex word (4 digits) to device
int	DEV, N;
char	HEXDIGIT; 	\Array: ascii hex DIGITS (0 - F)
begin
HEX2OUT(DEV, N >>8);
HEX2OUT(DEV, N);
end;	\HEX4OUT



proc	DOTAB(COL);	\Move cursor to column COL on the monitor
int	COL;
int	I;
begin
I:= 0;
repeat	begin
	CHOUT(TV, ^ );
	CPUREG(0):= $0300;	\function $03
	CPUREG(1):= $0000;
	SOFTINT($10);
	I:= I +1;		\prevent potiential infinite loop
	end;
until	(CPUREG(3)&$00FF)>=COL ! I>=COL;
end;	\DOTAB

\------------------------------------------------------------------------------

proc	ERROR(STR);	\Display error message and enter trace mode
char	STR;
begin
CHOUT(TV, BEL);
TEXT(TV, "ERROR - ");
TEXT(TV, STR);
TEXT(TV, ". Press Enter to continue.");
OPENI(KB);
if GETKEY then;
TRACEMODE:= true;
end;	\ERROR



proc	ILLEGAL(OP);	\Display error message for an illegal opcode
int	OP;
begin
CHOUT(TV, BEL);
TEXT(TV, "Illegal opcode (");
HEX4OUT(TV, OP);
TEXT(TV, "). Press Enter to continue.");
OPENI(KB);
if GETKEY then;
TRACEMODE:= true;
end;	\ILLEGAL

\------------------------------------------------------------------------------

func	HEXBYTE;	\Read in 2 chars and return the value of the hex byte
int	CH, HEX;	\ (Optimized for speed)
begin
loop	begin
	CH:= CHIN(3);
	if CH <= ^9 then
		if CH >= ^0 then [HEX:= (CH -^0) <<4; quit];
	if CH >= ^A then
		if CH <= ^F then [HEX:= (CH -$37) <<4; quit];
	ERROR("HEX CHARACTER EXPECTED");   exit;
	end;

CH:= CHIN(3);
if CH <= ^9 then
	if CH >= ^0 then return HEX + CH -^0;
if CH >= ^A then
	if CH <= ^F then return HEX + CH -$37;
ERROR("HEX CHARACTER EXPECTED");   exit;
end;	\HEXBYTE



proc	LOAD;		\Load the .HEX file output by MPASM
int	RECLEN,	\record length
	ADDR,
	TYPE,	\record type: 00 for data, 01 for end-of-file
	CKSUM,
	N,
	DATA;
char	MEM;	\facilitates loading bytes into ROM array
begin
\RECORD FORMAT:
\:BBAAAATTHHHH...HHHHCC
\BB = number of data bytes (HH) in record
\AAAA = address to load into, high byte first
\TT = record type: 00 for data, 01 for end-of-file
\HH = data, low byte, high byte
\CC = checksum, two's complement of the sum of all preceding bytes in the record

MEM:= ROM;
OPENI(3);
loop	begin
	repeat N:= CHIN(3) until N =^:;

	RECLEN:= HEXBYTE;
	CKSUM:= RECLEN;

	ADDR:= HEXBYTE;
	CKSUM:= CKSUM + ADDR;
	ADDR:= SWAP(ADDR) ! HEXBYTE;
	CKSUM:= CKSUM + ADDR;

	TYPE:= HEXBYTE;
	CKSUM:= CKSUM + TYPE;

	for N:= 1, RECLEN do
		begin
		DATA:= HEXBYTE;
		if ADDR >= ROMSIZE*2 then	\(ROMSIZE in bytes)
			[ERROR("FILE TOO BIG");   quit];
		MEM(ADDR):= DATA;
		CKSUM:= CKSUM + DATA;
		ADDR:= ADDR + 1;
		end;
	CKSUM:= CKSUM + HEXBYTE;
	if (CKSUM & $FF) # $00 then [ERROR("CHECKSUM");   quit];
	if TYPE = $01 then quit;
	end;
end;	\LOAD



proc	SETLAB;		\Set array indicating which instructions have labels
\WARNING: Only works for first page--first 2K
\WARNING: ROMSIZE better be >= 2K
int	I;
begin
for I:= 0, ROMSIZE-1 do LAB(I):= false;
for I:= 0, ROMSIZE-1 do
	begin
	case ROM(I) & $3800 of
	 $2000:	LAB(ROM(I) & $07FF):= true;	\CALL
	 $2800:	LAB(ROM(I) & $07FF):= true	\GOTO
	other	[];
	end;
end;	\SETLAB

\------------------------------------------------------------------------------

proc	DISASM(PC);	\Disassemble an instruction
int	PC;	\Address of instruction
int	TBL, I, OP, FR, D, B, OPCODE, REGS0, REGS1;


	proc	SHOWFR;	\Display file register by name or number
	begin
	case of
	  FR<=$1F:	TEXT(TV, REGS0(FR));	\16C74
	\  FR<=$0B:	TEXT(TV, REGS0(FR));	\\16C84
	  FR>=$80 & FR<=$9F:
			TEXT(TV, REGS1(FR-$80))
	other	[CHOUT(TV, ^0);   HEX2OUT(TV, FR);   CHOUT(TV, ^h)];
	end;	\SHOWFR


\OP NAME               OP NAME                OP NAME
\
\00 MOVWF              02 SUBWF               10 BCF
\00 NOP                03 DECF                14 BSF
\00 CLRWDT             04 IORWF               18 BTFSC
\00 RETFIE             05 ANDWF               1C BTFSS
\00 RETURN             06 XORWF               
\00 SLEEP              07 ADDWF               20 CALL
\                      08 MOVF                28 GOTO
\01 CLR F              09 COMF                
\01 CLR W              0A INCF                30 MOVLW
\                      0B DECFSZ              34 RETLW
\                      0C RRF                 38 IORLW
\                      0D RLF                 39 ANDLW
\                      0E SWAPF               3A XORLW
\                      0F INCFSZ              3C SUBLW
\                                             3E ADDLW
begin	\DISASM
TBL:= [	"???",   "CLR",   "SUBWF", "DECF",  "IORWF", "ANDWF", "XORWF", "ADDWF",
	"MOVF",  "COMF",  "INCF",  "DECFSZ","RRF",   "RLF",   "SWAPF", "INCFSZ",
	"BCF",   "BCF",   "BCF",   "BCF",   "BSF",   "BSF",   "BSF",   "BSF",
	"BTFSC", "BTFSC", "BTFSC", "BTFSC", "BTFSS", "BTFSS", "BTFSS", "BTFSS",
	"CALL",  "CALL",  "CALL",  "CALL",  "CALL",  "CALL",  "CALL",  "CALL",
	"GOTO",  "GOTO",  "GOTO",  "GOTO",  "GOTO",  "GOTO",  "GOTO",  "GOTO",
	"MOVLW", "movlw", "movlw", "movlw", "RETLW", "retlw", "retlw", "retlw",
	"IORLW", "ANDLW", "XORLW", "???",   "SUBLW", "sublw", "ADDLW", "addlw" ];

\16C74:
REGS0:=["INDF",   "TMR0",   "PCL",    "STATUS", "FSR",    "PORTA",  "PORTB",  "PORTC",
	"PORTD",  "PORTE",  "PCLATH", "INTCON", "PIR1",   "PIR2",   "TMR1L",  "TMR1H",
	"T1CON",  "TMR2",   "T2CON",  "SSPBUF", "SSPCON", "CCPR1L", "CCPR1H", "CCP1CON",
	"RCSTA",  "TXREG",  "RCREG",  "CCPR2L", "CCPR2H", "CCP2CON","ADRES", "ADCON0"];
REGS1:=["INDF",   "OPTION", "PCL",    "STATUS", "FSR",    "TRISA",  "TRISB",  "TRISC",
	"TRISD",  "TRISE",  "PCLATH", "INTCON", "PIE1",   "PIE2",   "PCON",   "???",
	"???",    "???",    "PR2",    "SSPADD", "SSPSTAT","???",    "???",    "???",
	"TXSTA",  "SPBRG",  "???",    "???",    "???",    "???",    "???",    "ADCON1"];

\16C84:
\REGS0:=["INDF",   "TMR0",   "PCL",    "STATUS", "FSR",    "PORTA",  "PORTB",  "???",
\	"EEDATA", "EEADR",  "PCLATH", "INTCON"];
\REGS1:=["INDF",   "OPTION", "PCL",    "STATUS",  "FSR",    "TRISA",  "TRISB",  "???",
\	"EECON1", "EECON2", "PCLATH", "INTCON"];

HEX4OUT(TV, PC);
TEXT(TV, ": ");

OPCODE:= ROM(PC);	\get opcode
if OPCODE=$0003\BREAK\ & PC=BPADDR then
	[OPCODE:= BPCODE; HEX4OUT(TV, OPCODE); TEXT(TV, " <bp>")]
else 	[HEX4OUT(TV, OPCODE); if LAB(PC) then TEXT(TV, " lab")];
CHOUT(TV, TAB);

FR:= OPCODE & $007F;
if RAM(STATUS) & $20 then FR:= FR ! $80;
D:= OPCODE & $0080;
B:= (OPCODE >> 7) & $07;

OP:= OPCODE >> 8;
if OP > 0 then TEXT(TV, TBL(OP&$3F))
else	begin
	if D then
		begin
		TEXT(TV, "MOVWF	");
		SHOWFR;
		end
	else	begin
		case OPCODE & $007F of
		  $00:	TEXT(TV, "NOP");
		  $01:	TEXT(TV, "INPUT");
		  $02:	TEXT(TV, "OUTPUT");
		  $03:	TEXT(TV, "BREAK");	\(in case of DATA 0003)
		  $20,$40,$60:
			TEXT(TV, "nop");
		  $08:	TEXT(TV, "RETURN");
		  $09:	TEXT(TV, "RETFIE");
		  $62:	TEXT(TV, "OPTION");
		  $63:	TEXT(TV, "SLEEP");
		  $64:	TEXT(TV, "CLRWDT");
		  $65:	TEXT(TV, "TRISA");
		  $66:	TEXT(TV, "TRISB");
		  $67:	TEXT(TV, "TRISC")
		other	TEXT(TV, "???");
		CHOUT(TV, TAB);
		end;
	end;

case of
  OP=1:							\CLRF CLRW
	begin
	if D then [TEXT(TV, "F	"); SHOWFR]
	else TEXT(TV, "W	");
	end;
  OP>=2 & OP<=$F:					\SUBWF	INDF,W
	begin
	CHOUT(TV, TAB);
	SHOWFR;
	if D=0 then TEXT(TV, ",W");
	end;

  OP>=$10 & OP<=$1F:					\BCF	FR,B
	begin
	CHOUT(TV, TAB);
	SHOWFR;
	CHOUT(TV, ^,);
	if (FR & $7F) = STATUS then
		begin \ 0     1     2     3     4     5      6      7
		TBL:= ["C", "DC", "Z", "PD", "TO", "RP0", "RP1", "IRP"];
		TEXT(TV, TBL(B));
		end
	else	HEX1OUT(TV, B);
	end;

  OP>=$20 & OP<=$2F:					\CALL	0234
	begin
	CHOUT(TV, TAB);
	HEX4OUT(TV, OPCODE & $07FF);
	end;

  OP>=$30 & OP<=$3F:					\MOVLW	03Ah
	begin
	CHOUT(TV, TAB);
	CHOUT(TV, ^0);
	HEX2OUT(TV, OPCODE & $00FF);
	CHOUT(TV, ^h);
	end
other	[];
end;	\DISASM

\------------------------------------------------------------------------------

proc	SHOWFLAG(N, F);	\E.G: Z=1
int	N, F;
begin
CHOUT(TV, N);
TEXT(TV, "=");
CHOUT(TV, if F then ^1 else ^0);
TEXT(TV, " ");
end;	\SHOWFLAG



proc	SHOWSTATE;	\Show contents of all registers and next instruction
int	I;
begin
TEXT(TV, "W=");
HEX2OUT(TV, WR);

TEXT(TV, "  SP=");
HEX2OUT(TV, SP);

TEXT(TV, "  RP");
SHOWFLAG(^0, RAM(STATUS) & $20);
SHOWFLAG(^Z, ZF);
SHOWFLAG(^H, DC);
SHOWFLAG(^C, CF);
CHOUT(TV, TAB);
DISASM(PC&ROMMASK);

DOTAB(67);
CHOUT(TV, if DONE(PC) then ^* else ^ );
CHOUT(TV, ^ );
end;	\SHOWSTATE



func	GETADDR(ADDR);	\Get an address or use the default address
int	ADDR;	\Default address
int	CH;
begin
loop	begin
	loop	begin
		CH:= GETKEY;
		if CH =CR then quit;	\Use existing ADDR
		case of CH>=^0 & CH<=^9, CH>=^A & CH<=^F, CH>=^a & CH<=^f:
			begin
			BACKUP;
			ADDR:= GETHEX;
			quit;
			end
		other;			\Ignore other characters (such as spaces)
		end;

	if ADDR>=$0000 & ADDR<=ROMSIZE then quit;
	CHOUT(TV, BEL);			\Invalid address
	end;

return ADDR;
end;	\GETADDR

\------------------------------------------------------------------------------

proc	DUMP;		\Dump 256 bytes of RAM starting at 00
\00:  00 11 22 33   44 55 66 77   00 11 22 33   44 55 66 77   ................
int	I, J, CH, DUMPADDR;
begin
RAM(STATUS):= RAM(STATUS) & $F8  !  ZF & $04  !  DC & $02  !  CF & $01;
RAM(PCL):= PC;
RAM(INDF):= RAM(RAM(MAP(FSR)));

TEXT(TV,							\16C74
"      I  T  P  S    F  P  P  P    P  P  P  I    P  P  T  T
      N  M  C  T    S  O  O  O    O  O  C  N    I  I  M  M
      D  R  L  A    R  R  R  R    R  R  L  T    R  R  R  R
      F  0     T       T  T  T    T  T  A  C    1  2  1  1
               U       A  B  C    D  E  T  O          L  H
               S                        H  N
");

DUMPADDR:= 0;
for I:= 0, RAMSIZE>>4 -1 do
	begin
	HEX2OUT(TV, DUMPADDR);
	TEXT(TV, ":  ");
	for J:= 0, $F do
		begin
		HEX2OUT(TV, RAM(MAP(DUMPADDR)));
		if RAM(MAP(DUMPADDR)) = RAM0(MAP(DUMPADDR)) then CHOUT(TV, ^ )
		else CHOUT(TV, ^.);
		if (J&3) = 3 then TEXT(TV, "  ");
		DUMPADDR:= DUMPADDR + 1;
		end;
	DUMPADDR:= DUMPADDR - $10;
	TEXT(TV, "  ");
	for J:= 0, $F do
		begin
		CH:= RAM(MAP(DUMPADDR)) & $7F;
		CHOUT(TV, if CH>=$20 & CH<=$7E then CH else ^.);
		DUMPADDR:= DUMPADDR +1;
		end;
	CRLF(TV);
	if I = 7 then
		TEXT(TV, "      0  1  2  3    4  5  6  7    8  9  A  B    C  D  E  F
");
	end;
for I:= 0, RAMSIZE-1 do RAM0(I):= RAM(I);
RAM(INDF):= $00;
end;	\DUMP



proc	ALTERLOC;	\Alter contents of specified location
char	ADDR;
int	CH, I;
begin
\S aa xx xx xx ...
ADDR:= GETHEX;
loop	begin
	BACKUP;				\Re-read character that terminated number
	CH:= GETKEY;
	if CH =CR then quit;
	case of CH>=^0 & CH<=^9, CH>=^A & CH<=^F, CH>=^a & CH<=^f:
		begin
		BACKUP;
		I:= GETHEX;
		if ADDR>=$00 & ADDR<RAMSIZE then \valid address
			RAM(MAP(ADDR)):= I	\(beware of infinite loop)
		else	CHOUT(TV, BEL);		\invalid address
		ADDR:= ADDR +1;
		end
	other	CH:= GETKEY;		\Ignore other characters (such as spaces)
	end;
ZF:= (RAM(STATUS) & $04) # 0;
DC:= (RAM(STATUS) & $02) # 0;
CF:= (RAM(STATUS) & $01) # 0;
end;	\ALTERLOC



proc	SETBREAK;	\Move break point to specified location
begin
ROM(BPADDR):= BPCODE;		\Restore contents of previous break point
BPADDR:= GETADDR(ROMSIZE);	\Default address is an unused location
BPCODE:= ROM(BPADDR);		\Save contents
ROM(BPADDR):= $0003;		\Set break point code (an unused opcode)
end;	\SETBREAK



proc	LIST;		\Disassemble 12 lines of code starting at $aaaa
int	I;
begin
LISTADDR:= GETADDR(LISTADDR);
for I:= 0, 11 do
	begin
	DISASM(LISTADDR);
	DOTAB(35);
	if DONE(LISTADDR) then CHOUT(TV, ^*);

	LISTADDR:= LISTADDR + 1;
	CRLF(TV);
	end;
end;	\LIST



proc	SHOWCYCLES;	\Display Oscillator cycle count
int	FL;

	proc	NUMOUT(N);
	int	N;
	int	M, D;
	begin
	M:= 100;
	repeat	begin
		D:= N/M;
		N:= REM(0);
		if D \#0\ then FL:= true;
		if FL then CHOUT(TV, D+^0);
		M:= M /10;
		end;
	until M = 0;
	end;	\NUMOUT

begin	\SHOWCYCLES
TEXT(TV, "Instruction cycles: ");
FL:= false;
NUMOUT(CY2);
if FL then CHOUT(TV, ^,);
NUMOUT(CY1);
if FL then CHOUT(TV, ^,);
NUMOUT(CY0);
if ~FL then CHOUT(TV, ^0);
CRLF(TV);
end;	\SHOWCYCLES



proc	SHOWSTACK;	\Show the subroutine return stack
int	I;
begin
for I:= 0, STACKSIZE-1 do
	begin
	HEX2OUT(TV, I);
	TEXT(TV, ": ");
	HEX4OUT(TV, STACK(I));
	CRLF(TV);
	end;
end;	\SHOWSTACK;



proc	HELP;
begin
TEXT(TV, "	Enter  - Execute next instruction (single step)
	G aaaa - Go to address aaaaH (or current address) and begin executing
	T aaaa - Go to address aaaaH (or current address) and begin Tracing
	D      - Dump RAM contents - display all file registers
	S      - Show Subroutine return Stack
	B aaaa - Move Break point to address aaaaH (if no aaaa then clear it)
	L aaaa - List assembly code starting at aaaaH (or current list address)
	I      - Display Instruction cycle count
	P      - Turn printer on/off
	R      - Simulate a hardware Reset (MCLR_ - also clears break point)
	Ctrl-C - Exit simulator program

	A aa xx xx xx ... - Alter RAM locations starting at aaH and set to xxH
	W xx   - Set W register to xxH
	Z x    - Set or clear Zero status bit (x = 1 or 0)
	H x    - Set or clear digit (Half) carry status bit (x = 1 or 0)
	C x    - Set or clear Carry status bit (x = 1 or 0)
");
end;	\HELP

\------------------------------------------------------------------------------

proc	BUMPCY;		\Increment cycle counter
begin
CY0:= CY0 +1;
if CY0 >= 1000 then
	begin
	CY0:= 0;
	CY1:= CY1 +1;
	if CY1 >= 1000 then
		begin
		CY1:= 0;
		CY2:= CY2 +1;
		if CY2 >= 1000 then CY2:= 0;	\wrap at 999,999,999
		end;
	end;
end;	\BUMPCY



proc	EXECUTE;	\Execute PIC instructions beginning at PC
int	FR,	\file register address
	D,	\destination select; d=0 store result in W, else store in FR
	T,	\temporary
	B,	\bit
	LIT,	\literal value
	OPCODE;	\instruction code (14 bits)



proc	CLASS00;
begin
case OPCODE & $007F of
  $00,$20,$40,$60:
	begin							\NOP
	end;
  $01:	begin							\INPUT
	WR:= CHIN(1);
	if WR = ESC then TRACEMODE:= true;
	end;
  $02:	begin							\OUTPUT
	CHOUT(TV, WR);
	end;
  $03:	begin							\BREAK POINT
	end;	\ = NOP (in case of DATA 0003)
  $08:	begin							\RETURN
	SP:= SP -1;
	if SP < 0 then
		begin
		CRLF(TV);   TEXT(TV, " *** STACK UNDERFLOW *** ");   CRLF(TV);
		SP:= STACKSIZE + SP;
		end;
	PC:= STACK(SP);
	BUMPCY;
	end;
  $09:	begin							\RETFIE
	SP:= SP -1;
	if SP < 0 then
		begin
		CRLF(TV);   TEXT(TV, " *** STACK UNDERFLOW *** ");   CRLF(TV);
		SP:= STACKSIZE + SP;
		end;
	PC:= STACK(SP);
	RAM(INTCON):= RAM(INTCON) ! $80;	\Enable GIE
	BUMPCY;
	end;
  $62:	begin							\OPTION
	RAM(OPTION):= WR;
	end;
  $63:	begin							\SLEEP
	WDT:= 0;
	WDTPRE:= 0;
	RAM(STATUS):= RAM(STATUS) & $F7 ! $10;	\ TO_:=1; PD_:=0
	TRACEMODE:= true;
	end;
  $64:	begin							\CLRWDT
	WDT:= 0;
	WDTPRE:= 0;
	RAM(STATUS):= RAM(STATUS) ! $18;	\ TO_:=1; PD_:=1
	end;
  $65:	begin							\TRIS A
	RAM(TRISA):= WR;
	end;
  $66:	begin							\TRIS B
	RAM(TRISB):= WR;
	end;
  $67:	begin							\TRIS C
	RAM(TRISC):= WR;
	end
other	ILLEGAL(OPCODE);
end;	\CLASS00



proc	CLASS0;
begin
FR:= OPCODE & $007F;
D:=  OPCODE & $0080;	\destination select; d=0 store into W else store into FR

if FR = INDF then FR:= RAM(MAP(FSR));
if FR = STATUS then
	RAM(STATUS):= RAM(STATUS) & $F8  !  ZF & $04  !  DC & $02  !  CF & $01
else if FR = PCL then RAM(PCL):= PC;
if RAM(STATUS) & $20 then FR:= FR ! $80;

case (OPCODE & $0F00) >> 8 of
  $0:	if D then RAM(MAP(FR)):= WR				\MOVWF
	else CLASS00;
  $1:	begin							\CLRF/CLRW
	if D then RAM(MAP(FR)):= 0 else WR:= 0;
	ZF:= true;
	end;
  $2:	begin							\SUBWF
	T:= (RAM(MAP(FR)) ! $FF00) - (WR & $00FF);
	ZF:= (T & $FF) = $00;
	CF:= (T & $100) = $100;
	DC:= ( (RAM(MAP(FR)) ! $FFF0) - (WR & $0F) & $10 ) = $10;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $3:	begin							\DECF
	T:= RAM(MAP(FR)) - 1;
	ZF:= (T & $FF) = $00;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $4:	begin							\IORWF
	T:= WR ! RAM(MAP(FR));
	ZF:= (T & $FF) = $00;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $5:	begin							\ANDWF
	T:= WR & RAM(MAP(FR));
	ZF:= (T & $FF) = $00;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $6:	begin							\XORWF
	T:= WR | RAM(MAP(FR));
	ZF:= (T & $FF) = $00;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $7:	begin							\ADDWF
	T:= (WR & $FF) + RAM(MAP(FR));
	ZF:= (T & $FF) = $00;
	CF:= T >= $100;
	DC:= (WR & $0F) + (RAM(MAP(FR)) & $0F) >= $10;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $8:	begin							\MOVF
	T:= RAM(MAP(FR));
	ZF:= (T & $FF) = $00;
	if D then \RAM(MAP(FR)):= T\ else WR:= T;
	end;
  $9:	begin							\COMF
	T:= ~RAM(MAP(FR));
	ZF:= (T & $FF) = $00;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $A:	begin							\INCF
	T:= RAM(MAP(FR)) + 1;
	ZF:= (T & $FF) = $00;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $B:	begin							\DECFSZ
	T:= RAM(MAP(FR)) - 1;
	if (T & $FF) = $00 then [PC:= PC +1;   BUMPCY];
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $C:	begin							\RRF
	T:= RAM(MAP(FR)) ! CF & $0100;
	CF:= (T & $01) = $01;
	if D then RAM(MAP(FR)):= T>>1 else WR:= T>>1;
	end;
  $D:	begin							\RLF
	T:= RAM(MAP(FR)) << 1;
	if CF then T:= T ! $01;
	CF:= T >= $100;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $E:	begin							\SWAPF
	T:= RAM(MAP(FR))<<4 ! RAM(MAP(FR))>>4;
	if D then RAM(MAP(FR)):= T else WR:= T;
	end;
  $F:	begin							\INCFSZ
	T:= RAM(MAP(FR)) + 1;
	if (T & $FF) = $00 then [PC:= PC +1;   BUMPCY];
	if D then RAM(MAP(FR)):= T else WR:= T;
	end
other	ERROR("CLASS0");

if (FR & $7F) = PCL then
    if D then
	begin
	PC:= RAM(PCL) ! RAM(PCLATH)<<8;
	BUMPCY;
	end;
if (FR & $7F) = STATUS then
    if D then
	begin
	ZF:= (RAM(STATUS) & $04) # 0;
	DC:= (RAM(STATUS) & $02) # 0;
	CF:= (RAM(STATUS) & $01) # 0;
	end;
end;	\CLASS0



proc	CLASS1;
begin
FR:= OPCODE & $007F;
if FR = INDF then FR:= RAM(MAP(FSR));
if FR = STATUS then
	RAM(STATUS):= RAM(STATUS) & $F8  !  ZF & $04  !  DC & $02  !  CF & $01
else if FR = PCL then RAM(PCL):= PC;
if RAM(STATUS) & $20 then FR:= FR ! $80;
B:= (OPCODE & $0380) >> 7;

case (OPCODE & $0C00) >> 10 of
  $0:	begin
	RAM(MAP(FR)):= RAM(MAP(FR)) & ~(1<<B);			\BCF
	if (FR & $7F) = PCL then
		begin
		PC:= RAM(PCL) ! RAM(PCLATH)<<8;
		BUMPCY;
		end;
	end;
  $1:	begin
	RAM(MAP(FR)):= RAM(MAP(FR)) ! 1<<B;			\BSF
	if (FR & $7F) = PCL then
		begin
		PC:= RAM(PCL) ! RAM(PCLATH)<<8;
		BUMPCY;
		end;
	end;
  $2:	begin
	if RAM(MAP(FR)) & 1<<B then []				\BTFSC
	else [PC:= PC +1;   BUMPCY];
	end;
  $3:	begin
	if RAM(MAP(FR)) & 1<<B then [PC:= PC +1;   BUMPCY];	\BTFSS
	end
other	ERROR("CLASS1");
if (FR & $7F) = STATUS then
	begin
	ZF:= (RAM(STATUS) & $04) # 0;
	DC:= (RAM(STATUS) & $02) # 0;
	CF:= (RAM(STATUS) & $01) # 0;
	end;
end;	\CLASS1



proc	CLASS2;
begin
case (OPCODE & $0800) >> 11 of
  $0:	begin							\CALL
	if SP >= STACKSIZE then
		begin
		CRLF(TV);   TEXT(TV, " *** STACK OVERFLOW *** ");   CRLF(TV);
		SP:= SP - STACKSIZE;
		end;
	STACK(SP):= PC;
	SP:= SP +1;
	PC:= OPCODE & $07FF ! RAM(PCLATH)<<8;
	BUMPCY;
	end;
  $1:	begin							\GOTO
	PC:= OPCODE & $07FF ! RAM(PCLATH)<<8;
	BUMPCY;
	end
other	ERROR("CLASS2");
end;	\CLASS2



proc	CLASS3;
begin
LIT:= OPCODE & $00FF;
case (OPCODE & $0F00) >> 8 of
  $0,$1,$2,$3:
	begin							\MOVLW
	WR:= LIT;
	end;
  $4,$5,$6,$7:
	begin							\RETLW
	SP:= SP -1;
	if SP < 0 then
		begin
		CRLF(TV);   TEXT(TV, " *** STACK UNDERFLOW *** ");   CRLF(TV);
		SP:= STACKSIZE + SP;
		end;
	PC:= STACK(SP);
	WR:= LIT;
	BUMPCY;
	end;
  $8:	begin							\IORLW
	WR:= WR ! LIT;
	ZF:= (WR & $FF) = $00;
	end;
  $9:	begin							\ANDLW
	WR:= WR & LIT;
	ZF:= (WR & $FF) = $00;
	end;
  $A:	begin							\XORLW
	WR:= WR | LIT;
	ZF:= (WR & $FF) = $00;
	end;
  $B:	ILLEGAL(OPCODE);					\??? secret opcode ?
  $C,$D:
	begin							\SUBLW
	DC:= ( (LIT ! $FFF0) - (WR & $0F) & $10 ) = $10;
	WR:= (LIT ! $FF00) - (WR & $00FF);
	ZF:= (WR & $FF) = $00;
	CF:= (WR & $100) = $100;
	end;
  $E,$F:
	begin							\ADDLW
	DC:= (WR & $0F) + (LIT & $0F) >= $10;
	WR:= (WR & $FF) + LIT;
	ZF:= (WR & $FF) = $00;
	CF:= WR >= $100;
	end
other	ERROR("CLASS3");
end;	\CLASS3



begin	\EXECUTE
PC:= PC & ROMMASK;
OPCODE:= ROM(PC);	\Fetch instruction from ROM

if OPCODE = $0003\BREAK\ then
	if TRACEMODE then
		[if (PC) = BPADDR then OPCODE:= BPCODE]	\do op at break point
	else [TRACEMODE:= true;   return];

DONE(PC):= true;
PC:= PC +1;
BUMPCY;

if CY0 < 10 then	\(for speed)
	if CHKKEY then TRACEMODE:= true;

case (OPCODE & $3000) >> 12 of
  0:	CLASS0;
  1:	CLASS1;
  2:	CLASS2;
  3:	CLASS3
other	ERROR("EXECUTE");
end;	\EXECUTE



proc	DOGO;		\Go command
begin
PC:= GETADDR(PC);
TRACEMODE:= true;	\execute past possible break point
EXECUTE;
TRACEMODE:= false;
repeat EXECUTE until TRACEMODE;
SHOWSTATE;
end;	\DOGO

\------------------------------------------------------------------------------

proc	INIT;		\Initialize program
int	I;
begin
SP:= 0;
PC:= 0;
CY0:= 0;   CY1:= 0;   CY2:= 0;
WR:= 0;
WDT:= 0;   WDTPRE:= 0;

TRACEMODE:= true;
PRINTFL:= false;
LISTADDR:= 0;

\Set break point address where it won't hurt anything
BPADDR:= ROMSIZE;   BPCODE:= 0;

for I:= 0, STACKSIZE-1 do STACK(I):= 0;		\(makes it easy to see amt used)
for I:= 0, ROMSIZE-1 do ROM(I):= $3FFF;		\unburned
for I:= 0, ROMSIZE-1 do DONE(I):= false;	\unexecuted

for I:= 0, RAMSIZE-1 do MAP(I):= I;		\some registers have two addresses
MAP($80):= $00;
MAP($82):= $02;
MAP($83):= $03;
MAP($84):= $04;
MAP($8A):= $0A;
MAP($8B):= $0B;

if ~RERUN then					\power up with random garbage
	for I:= 0, RAMSIZE-1 do RAM(I):= RAN(256);

RAM($00):= $00;
\RAM($02) is handled by PC
RAM($03):= RAM($03) & $1F ! $18; \ T0_=1 PD_=1; ZF, DC, & CF are stuffed in later
ZF:= (RAM($03) & $04) # 0;
DC:= (RAM($03) & $02) # 0;
CF:= (RAM($03) & $01) # 0;
RAM($05):= RAM($05) & $3F;
RAM($09):= RAM($09) & $07;
RAM($0A):= $00;
RAM($0B):= RAM($0B) & $01;
RAM($0C):= $00;
RAM($0D):= $00;
if ~RERUN then RAM($10):= $00
else RAM($10):= RAM($10) & $3F;
RAM($11):= $00;
RAM($12):= $00;
RAM($14):= $00;
RAM($17):= $00;
RAM($18):= RAM($18) & $01;
RAM($19):= $00;
RAM($1A):= $00;
RAM($1D):= $00;
RAM($1F):= $00;

RAM($81):= $FF;
RAM($85):= $3F;
RAM($86):= $FF;
RAM($87):= $FF;
RAM($88):= $FF;
RAM($89):= $07;
RAM($8C):= $00;
RAM($8D):= $00;
if ~RERUN then RAM($8E):= $00
else RAM($8E):= RAM($8E) & $02;
for I:= $8F, $9F do RAM(I):= $00;
RAM($92):= $FF;
RAM($98):= $02;

for I:= 0, RAMSIZE-1 do RAM0(I):= RAM(I);
end;	\INIT

\------------------------------------------------------------------------------

begin	\MAIN
CPUREG:= GETREG;
PSPSEG:= CPUREG(11);
DATASEG:= CPUREG(12);

if RERUN then OPENI(3) else DOSOPEN;

OPENI(KB);
BACKFLAG:= false;

INIT;
LOAD;
SETLAB;

if ~RERUN then
	begin
	TEXT(TV,"
-- PIC SIM, V1.3 --

Type: ^"? <Enter>^" for help.
");
	end;
CRLF(TV);	\(for reset)

SHOWSTATE;
loop	begin
	TEXT(TV, ">");
	CMD:= GETKEY;
	if PRINTFL then CRLF(PR);
	if CMD>=^a & CMD<=^z then CMD:= CMD & $DF;	\convert to uppercase
	case CMD of
	  CR:	[TRACEMODE:= true;   EXECUTE;   SHOWSTATE];
	  ^G:	DOGO;
	  ^T:	[PC:= GETADDR(PC);   TRACEMODE:= true;   EXECUTE;   SHOWSTATE];
	  ^D:	DUMP;
	  ^S:	SHOWSTACK;
	  ^B:	SETBREAK;
	  ^L:	LIST;
	  ^I:	SHOWCYCLES;
	  ^P:	[PRINTFL:= ~PRINTFL;
		if PRINTFL then OPENO(PR) else CLOSE(PR)];
	  ^R:	RESTART;
	  ^A:	ALTERLOC;
	  ^W:	WR:= GETHEX;
	  ^Z:	ZF:= GETHEX # 0;
	  ^H:	DC:= GETHEX # 0;
	  ^C:	CF:= GETHEX # 0;
	  ^?:	HELP
	other	[CHOUT(TV, BEL);   CRLF(TV)];

	if CMD # CR then OPENI(KB);
	end;
end;	\MAIN
