
include C:\CXPL\CODESI.XPL;
include C:\CXPL\XPLDEFS.XPL;


int X,O,N;

\MUSIC CONSTANTS

integer
	CHROMATIC;	\Strings of note names
int
	BOTTOM;			\Bottom octave for musical scale

define	\Define note values
	NOTE_C,NOTE_CS,NOTE_D,NOTE_DS,NOTE_E,NOTE_F,
	NOTE_FS,NOTE_G,NOTE_GS,NOTE_A,NOTE_AS,NOTE_B;


\------------------------------------------------------------------------

define
	SB_MPUBASE=$330,	\MPU-401 base port address
	SB_MIXPORT=$220,	\Mixer port address	
	SB_BASE=$220,		\Sound blaster base port address

	MIX_REGISTER=SB_MIXPORT+$04,	\MIX Register port
	MIX_DATA=SB_MIXPORT+$05,	\MIX Data port

	DSP_RESET=SB_BASE+$06,		\DSP Reset port
	DSP_READ=SB_BASE+$0A,		\DSP Read data
	DSP_WRITE=SB_BASE+$0C,		\DSP Write data
	DSP_WSTATUS=SB_BASE+$0C,	\DSP Write status
	DSP_RSTATUS=SB_BASE+$0E,	\DSP Read status

	FM_STATUS=SB_BASE+$08,		\FM Status port
	FM_REGISTER=SB_BASE+$08,	\FM Register port
	FM_DATA=SB_BASE+$09;		\FM Data port


integer
	REG_ARRAY;		\Array of FM control register values


define	\Register index values
	FMR_20_A, FMR_20_B,
	FMR_40_A, FMR_40_B,
	FMR_60_A, FMR_60_B,
	FMR_80_A, FMR_80_B,
	FMR_E0_A, FMR_E0_B,
	FMR_C0, FMR_01, FMR_08, FMR_BD,
	MAXREG;


integer
	ModulateOffset,		\Modulator offsets for each channel
	CarrierOffset;		\Carrier offsets for each channel

integer
	ChanNote,		\Array of notes for each channel
	ChanOctave,		\Array of octaves for each channel
	ChanTime;		\Array of times for each channel to be on

define	MAXCHAN=9;		\Maximum number of channels

define	IntSize=2;

\------------------- Routines to read and write DSP ports --------------

procedure ReadDSP(Port);
\Read from specified port
int Port,C;
begin
for C:=0,1600 do if PIN(DSP_RSTATUS,0) & $80 then return PIN(Port,0); 
return 0;
end;


procedure WriteDSP(Port,Value);
\Read from specified port
int Port,Value,C;
begin
for C:=0,1600 do if PIN(DSP_WSTATUS,0) & $80 then [POUT(Value,Port,0); return;];
POUT(Value,Port,0);
end;

\------------------------------------------------------------------------


procedure ResetDSP;
\Reset the DSP chip
int C;
begin
POUT(1,DSP_RESET,0);
for C:=0,256 do;
POUT(0,DSP_RESET,0);
return ReadDSP(DSP_READ,0) = $AA;
end;


procedure real DSPVersion;
\Get DSP version number
begin
WriteDSP(DSP_WRITE,$E1);
return FLOAT(ReadDSP(DSP_READ,0)) +FLOAT(ReadDSP(DSP_READ,0))/100.0;
end;

\------------------- Mixer Control Routines ----------------------------


procedure SetMIXReg(Reg,Value);
\Set specified register REG to Value
int Reg,Value;int C,T;
begin
POUT(Reg,MIX_REGISTER,0);
POUT(Value,MIX_DATA,0);
end;



\-------------------- FM Chip Control Routines -------------------------

procedure GetFMType;
\Get FM chip type
return PIN($388,0);


procedure SetFMRegister(Reg,Value);
\Set specified register REG to Value
int Reg,Value;int C,T;
begin
POUT(Reg,FM_REGISTER,0);
for C:=1,6 do T:=PIN(FM_STATUS,0);
POUT(Value,FM_DATA,0);
for C:=1,35 do T:=PIN(FM_STATUS,0);
end;


procedure ResetFM;
\Reset all FM register
int R;
begin
for R:=-$F5,-2 do SetFMRegister(-R,0);
SetFMRegister($04,$06);
SetFMRegister($01,$20);
end;


procedure GotFM;
\Test for FM sound chip
int T1,T2,D;
begin
SetFMRegister($04,$60);
SetFMRegister($04,$80);
T1:=PIN(FM_STATUS,0);
SetFMRegister($02,$FF);
SetFMRegister($04,$21);
for D:=1,80 do;
T2:=PIN(FM_STATUS,0);
SetFMRegister($04,$60);
SetFMRegister($04,$80);
return ((T1 & $E0)=0) & ((T2 & $E0)=$C0);  
end;


\------------------- Music Routines ------------------------------------

procedure SETNOTES;
begin
CHROMATIC:=["C    ","C#/Db","D    ","D#/Eb","E    ","F    ",
            "F#/Gb","G    ","G#/Ab","A    ","A#/Bb","B    "];

BOTTOM:=[$156,$16B,$181,$198,$1B0,$1CA,$1E5,$202,$220,$241,$263,$287,$2A3];
end;


\------------------ Routine to control the volume ----------------------

procedure SetVolume;
int Master,FMVolume,C;
begin
Master:=4; FMVolume:=4;
CHOUT(0,FF);
TEXT(0,"
************** SET VOLUME ****************
M = Master Volume   F = FM Volume Q = Quit");
loop
	begin
	CURSOR(0,5);
	TEXT(0,"Master = ");
	INTOUT(0,Master);
	TEXT(0," FM = ");
	INTOUT(0,FMVolume);

	C:=CHIN(1);
	case C of
	^M: Master:=(Master+1) & 7;	
	^m: Master:=(Master-1) & 7;	
	^F: FMVolume:=(FMVolume+1) & 7;	
	^f: FMVolume:=(FMVolume-1) & 7;
	^Q,^q: quit
	other;	
	end;
Master:=Master<<1;
Master:=Master<<4 ! Master;
SetMIXReg($22,Master);

FMVolume:=FMVolume<<1;
FMVolume:=FMVolume<<4 ! FMVolume;
SetMIXReg($26,FMVolume);
end;



\------------------- FM Synthysis Routines -----------------------------



procedure FMDefaults;
\Set sound quality defaults
begin
REG_ARRAY(FMR_01):=$20;
REG_ARRAY(FMR_08):=0;
REG_ARRAY(FMR_BD):=0;

REG_ARRAY(FMR_20_A):=$01; REG_ARRAY(FMR_20_B):=$01;
REG_ARRAY(FMR_40_A):=$10; REG_ARRAY(FMR_40_B):=$00;
REG_ARRAY(FMR_60_A):=$F0; REG_ARRAY(FMR_60_B):=$F0;
REG_ARRAY(FMR_80_A):=$77; REG_ARRAY(FMR_80_B):=$77;
REG_ARRAY(FMR_E0_A):=0; REG_ARRAY(FMR_E0_B):=0;
REG_ARRAY(FMR_C0):=0;
end;



procedure ZeroChanArray;
\Zero out channel array
int Chan;
for Chan:=0,MAXCHAN-1 do ChanNote(Chan):=-1;



procedure SetTimbre(CHAN,ARRAY);
\Set sound quality for a channel
int CHAN,ARRAY;
int MO,CO;
begin
MO:=ModulateOffset(CHAN);
CO:=CarrierOffset(CHAN);
SetFMRegister($01,ARRAY(FMR_01));
SetFMRegister($08,ARRAY(FMR_08));
SetFMRegister($BD,ARRAY(FMR_BD));

SetFMRegister($20+MO,ARRAY(FMR_20_A));
SetFMRegister($20+CO,ARRAY(FMR_20_B));

SetFMRegister($40+MO,ARRAY(FMR_40_A));
SetFMRegister($40+CO,ARRAY(FMR_40_B));

SetFMRegister($60+MO,ARRAY(FMR_60_A));
SetFMRegister($60+CO,ARRAY(FMR_60_B));

SetFMRegister($80+MO,ARRAY(FMR_80_A));
SetFMRegister($80+CO,ARRAY(FMR_80_B));

SetFMRegister($E0+MO,ARRAY(FMR_E0_A));
SetFMRegister($E0+CO,ARRAY(FMR_E0_B));

SetFMRegister($C0,ARRAY(FMR_C0));
SetFMRegister($40+CO,ARRAY(FMR_40_B)+5);
end;




procedure SetChannel(Chan,Note,Octave,Array);
\Set sound channel
int Chan,Note,Octave,Array;
int F1,F2;
begin
SetTimbre(Chan,Array);
F1:=BOTTOM(Note);
F2:=SWAP(F1) & $03; F2:=F2 ! Octave <<2; F2:=F2 ! $20;
SetFMRegister($A0+Chan,F1);
SetFMRegister($B0+Chan,F2);
return F1;
end;


procedure ChanOff(Chan,Note,Octave);
\Turn specified channel off
int Chan,Note,Octave;
int F1,F2;
begin
F1:=BOTTOM(Note);
F2:=SWAP(F1) & $03; F2:=F2 ! Octave <<2;
SetFMRegister($A0+Chan,F1);
SetFMRegister($B0+Chan,F2);
end;




procedure PlayChord;
\Play chord in array
int Chan,F;
begin
for Chan:=0,MAXCHAN-1 do [INTOUT(0,ChanNote(Chan)); TEXT(0," ");];
CRLF(0);

for Chan:=0,MAXCHAN-1 do
if ChanNote(Chan)#-1 then
	begin
	F:=SetChannel(Chan,ChanNote(Chan),ChanOctave(Chan),REG_ARRAY);
	INTOUT(0,F);TEXT(0," ");
	end;
CRLF(0);
SOUND(0,ChanTime(0),5);
for Chan:=0,MAXCHAN-1 do
  if ChanNote(Chan)#-1 then [ChanOff(Chan,ChanNote(Chan),ChanOctave(Chan)); ChanNote(Chan):=-1;];
end;



\-----------------------------------------------------------------------

procedure SetRegister(RegValues);
\Set timbre register array
\Values in the follow order: 20A, 20B, 40A, 40B, 60A, 60B, 80A, 80B, E0A, E0B, C0, 01, 08, BD
int RegValues,P;
for P:=0,MAXREG-1 do REG_ARRAY(P):=RegValues(P);


procedure COINFALL;
int Reg;
begin
Reg:=[$04,$05,$00,$0B,$78,$FA,$FF,$DF,$03,$00,$01,$20,$00,$00];
SetRegister(Reg);
ChanNote(0):=NOTE_E; ChanOctave(0):=4;ChanTime(0):=2; PLAYCHORD;
ChanNote(0):=NOTE_D; ChanOctave(0):=4;ChanTime(0):=2; PLAYCHORD;
ChanNote(0):=NOTE_C; ChanOctave(0):=4;ChanTime(0):=2; PLAYCHORD;
end;


procedure FIVECOIN;
int Reg;
begin
Reg:=[$05,$01,$4E,$80,$DA,$F0,$15,$13,$00,$00,$0A,$20,$00,$00];
SetRegister(Reg);
ChanNote(0):=NOTE_C; ChanOctave(0):=5;ChanTime(0):=2; PLAYCHORD;
ChanNote(0):=NOTE_E; ChanOctave(0):=5;ChanTime(0):=2; PLAYCHORD;
ChanNote(0):=NOTE_G; ChanOctave(0):=5;ChanTime(0):=2; PLAYCHORD;
end;


procedure COINPAY;
int Reg;
begin
Reg:=[$05,$01,$4E,$80,$DA,$F0,$15,$13,$00,$00,$0A,$20,$00,$00];
SetRegister(Reg);
ChanNote(0):=NOTE_C; ChanOctave(0):=6;ChanTime(0):=1; PLAYCHORD;
end;


procedure DEALCARD;
int Reg;
begin
Reg:=[$1F,$14,$C0,$80,$00,$90,$00,$0F,$03,$02,$0E,$20,$00,$00];
SetRegister(Reg);
ChanNote(0):=NOTE_C; ChanOctave(0):=4;ChanTime(0):=5; PLAYCHORD;
end;



procedure WINNER;
int Reg;
begin
Reg:=[$31,$21,$94,$00,$33,$A0,$73,$97,$00,$00,$0E,$20,$00,$00];
SetRegister(Reg);

\G7 Chord
ChanNote(0):=NOTE_G; ChanOctave(0):=5;ChanTime(0):=10;
ChanNote(1):=NOTE_B; ChanOctave(1):=5;
ChanNote(2):=NOTE_D; ChanOctave(2):=5;
ChanNote(3):=NOTE_F; ChanOctave(3):=5;
PLAYCHORD;

\C Chord
ChanNote(0):=NOTE_C; ChanOctave(0):=5;ChanTime(0):=10;
ChanNote(1):=NOTE_E; ChanOctave(1):=5;
ChanNote(2):=NOTE_G; ChanOctave(2):=5;
PLAYCHORD;
end;



procedure LOSER;
int Reg;
begin
Reg:=[$53,$00,$85,$00,$3F,$5F,$06,$07,$01,$00,$06,$20,$00,$00];
SetRegister(Reg);
ChanNote(0):=NOTE_C; ChanOctave(0):=2;ChanTime(0):=10; PLAYCHORD;
end;




procedure MENU;
int COM;
begin
CHOUT(0,FF);
TEXT(0,"

********** TEST POKER SOUNDS **********

C = COIN FALL
F = FIVE COIN SOUND
D = DEAL CARD
P = COIN PAYOFF
W = WINNER
L = LOSER

V = SET VOLUME

Q = QUIT

CHOOSE ONE: ");
COM:=CHIN(1);
case COM of
^C,^c: COINFALL;
^F,^f: FIVECOIN;
^D,^d: DEALCARD;
^P,^p: COINPAY;
^W,^w: WINNER;
^L,^l: LOSER;
^V,^v: SetVolume;
^Q,^q: return true
other;
return false;
end;


begin
REG_ARRAY:=RESERVE(MAXREG*2);
SETNOTES;
ModulateOffset:=[$00,$01,$02,$08,$09,$0A,$10,$11,$12];
CarrierOffset:= [$03,$04,$05,$0B,$0C,$0D,$13,$14,$15];
ChanNote:=RESERVE(MAXCHAN*IntSize);
ChanOctave:=RESERVE(MAXCHAN*IntSize);
ChanTime:=RESERVE(MAXCHAN*IntSize);
ZeroChanArray;
FMDefaults;

if GotFM then TEXT(0,"FM Chip Installed") else TEXT(0,"No FM Chip Installed");
CRLF(0);

if ResetDSP then TEXT(0,"DSP Reset") else TEXT(0,"Unable to Reset DSP");
CRLF(0);

FORMAT(5,2);
TEXT(0,"Version Number = ");RLOUT(0,DSPVersion);CRLF(0);
TEXT(0,"FM TYPE= ");HEXOUT(0,GetFMType);CRLF(0);
ResetFM;
TEXT(0,"Hit Any Key");
X:=CHIN(1);

repeat until MENU;
ResetFM;
end;
