\PRINTFIL.XPL	JAN-31-93
\This program takes text containing special control characters formats it
\ and sends it to an output file called PRINTOUT.TXT.
\ by Loren Blaney

\REVISIONS:
\JUL-30-92, Version for Okidata 380 based on ITOHPNT version 1.4 (JUN-28-84).
\JAN-30-93, Converted PRINTOKI to print to a file.

\In addition to the normal CR, LF, FF, and TAB, this program
\ recognizes the following control character commands:
\	CTRL-C		Center line
\ All the following control characters toggle (on/off):
\	CTRL-SHIFT-N	Scoop up a page header
\
\When centering lines, the CTRL-C can be anywhere on the line.
\ Leading spaces and tabs are ignored. If a title (big print)
\ is centered it is assumed that the entire line is in title
\ font.
\If you want page headers, then you must answer "Y" to the
\ "PRINT PAGE NUMBERS" question. However, page numbers less
\ than or equal to 1 will not be printed.

code	REM=2,		RESERVE=3,	CHIN=7,		CHOUT=8,
	CRLF=9,		INTIN=10,	INTOUT=11,	TEXT=12,
	OPENI=13,	OPENO=14,	CLOSE=15,	FSET=24,
	FOPEN=29,	FCLOSE=32,	GETREG=35,	BLIT=36;

define	EOF=$1A, CR=$0D, FF=$0C, LF=$0A, TAB=$09, SP=$20,
	BS=$08, ESC=$1B, ETX=$03, US=$1F, EOL=LF,
	SO=$0E, SI=$0F,
	CHPITCH=15,	\NUMBER OF DOT SPACES WIDE A CHAR IS (ELITE ONLY)
	\CONTROL CHARACTERS:
	CTRLB=$02, CTRLC=$03, CTRLG=$07,
	CTRLN=$0E, CTRLT=$14, CTRLY=$19, CTRLSN=$1E,
	NPITCH=15,	\NOMINAL PITCH, I.E. DOTS PER CHAR (ELITE ONLY)
	LINESIZ=200;	\BUFFER SIZE FOR THE INPUT LINE

int	CPUREG,	\REGISTER ARRAY FROM GETREG
	INHAND, OUTHAND,	\FILE HANDLES
	CHAR,	\CHARACTER
	CHPOS,	\POSITION ON A LINE (IN CHARACTERS)
	LNPOS,	\LINE NUMBER ON THE PAGE
	PRDEV,	\PRINTER DEVICE NUMBER
	II,	\SCRATCH
	LINELEN,\INPUT LINE INDEX COUNTER
	NLET,	\NUMBER OF LETTERS IN INPUT LINE
	NSPS,	\TOTAL NUMBER OF SPACES IN INPUT LINE
	NSPS2,	\NUMBER OF SPACES IN INPUT LINE AFTER THE 1ST PRINTABLE CHAR
	SPCNT,	\SPACE COUNTER FOR "ISSUE" PROCEDURE
	PAGENO,	\PAGE NUMBER COUNTER
	LPITCH,	\LETTER (PRINTABLE CHARACTER) PITCH
	SPITCH,	\SPACE PITCH
	ADDS,	\EXTRA PITCHES TO BE DISTRIBUTED IN SPACES
	CPITCH,	\PITCH FOR CENTERED LINE (NOMINAL OR TITLE)
	PITCH,	\PITCH FOR CURRENT CHARACTER
	CENTER,	\BOOLEAN--CENTER LINE
	FILL,	\BOOLEAN--FILL LINES WITH "PITCHES" TO RIGHT JUSTIFY MARGINS
	BOLD,	\BOOLEAN--DO BOLD FACE PRINT
	UNDERLINE, \BOOLEAN--DO UNDERLINE
	TITLE,	\BOOLEAN--DO BOLD UNDERLINE (FOR TITLES, ETC.)
	HEADER,	\BOOLEAN--SCOOP UP A PAGE HEADER
	HEADLEN,\LENGTH IN CHARS OF HEADER BUFFER
	HEADMAX;\MAXIMUM LENGTH OF "HEADLEN"

char	CMDTAIL,\COPY OF COMMAND TAIL
	LINE,	\THE INPUT LINE
	HEADBUF;\PAGE HEADER CHARACTER BUFFER

int	OWN;	\CONSTANT ARRAY FOR 'OWN' TYPE VARIABLES
define \OWN\	\ELEMENTS OF CONSTANT ARRAY, I.E. OPTIONS:
	CHMAX,	\MAXIMUM LINE LENGTH IN CHARACTERS
	PAGES,	\BOOLEAN--PAGINATE (DO PAGE NUMBERS AND HEADERS)
	LEFTLIM,\LEFT INDENT (START AT 1)
	LNMAX,	\NUMBER OF LINES TO USE PER PAGE
	STPAGE,	\PAGE NUMBER TO START
	LOLIM,	\LOW PAGE LIMIT; DON'T PRINT PAGES BELOW THIS
	HILIM,	\HIGH PAGE LIMIT; DON'T PRINT BEYOND THIS PAGE
	FONT,	\TRUE = PICA, FALSE = ELITE
	BAKBAK;	\BOOLEAN--FOR PAGE NOS. WHEN PAGES ARE BACK TO BACK



func	LEN(I);	\RETURNS NUMBER OF DIGITS IN INTEGER "I"
int	I;
int	J;
begin
for J:= 1, 5 do
	[I:= I/10;
	if I=0 then return J];
end;	\LEN



proc	SEND(CH);
int	CH;
int	P;
begin
CHOUT(PRDEV, CH);
CHOUT(0, CH);		\MONITOR ON THE CONSOLE (PRDEV MAY BE = NULL)
end;	\SEND



proc	ISSUE(CHAR);	\SEND A CHARACTER TO BE PRINTED
int	CHAR;
begin
if CHAR=SP then		\DISTRIBUTE ANY EXTRA PITCHES IN THE
	begin		\ FIRST "ADDS" SPACES IN THE LINE
	SPCNT:= SPCNT+1;
	if SPCNT<=ADDS then PITCH:= SPITCH+1 else PITCH:= SPITCH;
	end
else PITCH:= LPITCH;
SEND(CHAR);
if CHAR>=SP then CHPOS:= CHPOS+1;
if TITLE then SEND(SP);	\DOUBLE SPACE TITLES
end;	\ISSUE



proc	DOCR;		\DO A CARRIAGE RETURN
int	I;
begin
SEND(CR);
for I:= 1, OWN(LEFTLIM) do SEND(SP);
PITCH:= NPITCH;		\START NEW LINE WITH NOMINAL PITCH
CHPOS:= 0;   SPCNT:= 0;
end;	\DOCR



proc	DOHEAD;		\DO A PAGE HEADER
int	I;
begin
if OWN(PAGES) then
	begin
	PAGENO:= PAGENO+1;
	PRDEV:= if PAGENO<OWN(LOLIM) ! PAGENO>OWN(HILIM) then 7
		else 3;
	DOCR;   	\PAGINATE--PRINT PAGE NO. AND HEADER:
	if OWN(BAKBAK) & (PAGENO&1)#1 \EVEN\ then
		begin	\PRINT LEFT-HAND PAGE
		if PAGENO>1 then	\START NUMBERING AT PAGE 2
			INTOUT(PRDEV, PAGENO);
		for I:= 1, 5 do SEND(SP);
		for I:= 0, HEADLEN-1 do SEND(HEADBUF(I));
		end
	else
		begin	\PRINT NORMAL OR RIGHT-HAND PAGE
		for I:= 1, OWN(CHMAX) -HEADLEN -5 -LEN(PAGENO) do
			SEND(SP);
		for I:= 0, HEADLEN-1 do SEND(HEADBUF(I));
		for I:= 1, 5 do SEND(SP);
		if PAGENO>1 then	\START NUMBERING AT PAGE 2
			INTOUT(PRDEV, PAGENO);
		end;
	for I:= 1, 4 do [SEND(CR);   SEND(LF)];
	for I:= 1, OWN(LEFTLIM) do SEND(SP);
	CHPOS:= 0;   SPCNT:= 0;
	LNPOS:= LNPOS+5;
	end;
end;	\DOHEAD



proc	DOFF;		\DO A FORM FEED
begin
LNPOS:= 0;
\DOCR;
SEND(FF);	\(THE ITOH FF MUST BE THE FIRST THING ON A LINE)
DOHEAD;
end;	\DOFF



proc	DOLF;		\DO A LINE FEED
int	I;
begin
if LNPOS<OWN(LNMAX)-1 then
	begin
	ISSUE(LF);
	LNPOS:= LNPOS+1;
	end
else DOFF;
end;	\DOLF



proc	DOTAB;		\DO A TAB (EVERY 8 COLUMNS)
begin
repeat ISSUE(SP)
until (CHPOS & 7) = 0
end;	\DOTAB

\----------------------------------------------------------------------

proc	LINEIN;
\READ IN A LINE AND OUTPUT THE PITCH FACTORS: LPITCH, SPITCH, ADDS
int	HASTAB,REMAIN,LNS;



proc	COUNTER;
\COUNTS THE NUMBER OF SPACES (NSPS & NSPS2) AND THE NUMBER OF PRINTABLE
\ CHARS (NLET) IN A LINE. LINES CONTAINING TABS ARE FLAGGED (HASTAB) AND
\ ARE NOT FILLED.
int	I;
begin
NSPS:= 0;   NSPS2:= 0;   NLET:= 0;   HASTAB:= false;
for I:= 0, LINELEN-1 do
	case LINE(I) of
	TAB: HASTAB:= true;
	SP:  [NSPS:= NSPS+1;   if NLET>0 then NSPS2:= NSPS2+1]
	else if LINE(I)>SP then NLET:= NLET+1;
end;	\COUNTER



proc	GETCH;
begin
repeat CHAR:= CHIN(3);
	if CHAR=CTRLG then FILL:= not FILL;
	if CHAR=CTRLC then CENTER:= true;
	if CHAR=CTRLN then CPITCH:= 2*NPITCH;	\TITLE HAS DOUBLE-WIDE CHARS
until CHAR#CTRLG & CHAR#CTRLC;
end;	\GETCH



begin	\LINEIN
LINELEN:= 0;
CENTER:= false;		\MOST LINES ARE NOT CENTERED
CPITCH:= NPITCH;	\CENTERED LINE PITCH (NOMINAL OR TITLE)
repeat	GETCH;		\READ LINE INTO BUFFER
	LINE(LINELEN):= CHAR;
	LINELEN:= LINELEN+1;
until CHAR=EOF ! CHAR=EOL ! LINELEN>=LINESIZ;
			\CHOP OFF END OF LINES THAT ARE WAY TOO LONG
while CHAR#EOF & CHAR#EOL do GETCH;
COUNTER;		\NOW COMPUTE PITCHES
if	CENTER
!	not FILL
!	OWN(CHMAX)-(NLET+NSPS) >= OWN(CHMAX)/7	\CAN ONLY STRETCH SO FAR
!	HASTAB		\CONTAINS A TAB
then	begin
	LPITCH:= NPITCH;
	SPITCH:= NPITCH;
	ADDS:= 0;
	end
else	begin
		\COMPUTE TOTAL NUMBER OF PITCHES ON A LINE
	LNS:= OWN(CHMAX)*NPITCH;
		\LETTER PITCH = AVE NO. OF PITCHES USED BY
		\ LETTERS & SPACES
	LPITCH:= LNS/(NSPS+NLET);
		\REMAIN = TOTAL PITCHES USED BY SPACES
	REMAIN:= LNS-(LPITCH*NLET);
	SPITCH:= if NSPS#0 then REMAIN/NSPS else NPITCH;
		\EXTRA PITCHES TO BE DISTRIBUTED IN THE SPACES
	ADDS:= REMAIN-SPITCH*NSPS;
	end;
end;	\LINEIN

\----------------------------------------------------------------------

proc	LINEOUT;	\OUTPUT A LINE TO FILE
int	I,J,POSN;



proc	PUT(CH);	\HANDLE CONTROL CHARS OR OUTPUT A PRINTABLE CHAR
int	CH;
begin
case CH of
	TAB:	DOTAB;
	LF:	DOLF;
	CR:	DOCR;
	FF:	DOFF;
	CTRLT:	;	\STRIP OUT INDEX TAGS
	CTRLB:	[BOLD:= not BOLD];
	CTRLN:	begin
		TITLE:= not TITLE;
		end;
	CTRLY:	[UNDERLINE:= not UNDERLINE];
	CTRLSN:	[HEADER:= not HEADER;   if HEADER then HEADLEN:= 0]
other ISSUE(CH);
if HEADER then if CH>=$20 then
	begin
	HEADBUF(HEADLEN):= CH;
	if HEADLEN<HEADMAX-1 then HEADLEN:= HEADLEN+1;
	end;
end;	\PUT



begin	\LINEOUT
if CENTER then		\CENTER THE LINE OF TEXT
	begin		\(DON'T COUNT LEADING SPACES OR TABS)
	POSN:= (OWN(CHMAX)*NPITCH -
		(NLET + NSPS2)*CPITCH) / (2*NPITCH);
		   \SEND CHARS OUT, UP TO FIRST PRINTABLE LETTER
	loop for J:= 0, LINELEN-1 do
		begin	\IGNORE LEADING SPACES AND TABS
			\ BUT PUT OUT ANY CTRL CHARS LIKE FF
		if LINE(J)>$20 ! LINE(J)=CTRLN then quit;
		if LINE(J)#SP & LINE(J)#TAB then PUT(LINE(J));
		end;
	for I:= 1, POSN do SEND(SP);	\MOVE FROM LEFT MARGIN
			\SEND OUT THE REST OF THE LINE
	for I:= J, LINELEN-1 do PUT(LINE(I));
	end
else for I:= 0, LINELEN-1 do PUT(LINE(I));
end;	\LINEOUT

\----------------------------------------------------------------------

begin	\MAIN
	\ARRAY FOR DEFAULT VALUES
OWN:= [73, true, 5, 59, -2, -10, 10000, false, true];

CPUREG:= GETREG;		\Get DOS info: PSP and data segment
CMDTAIL:= RESERVE($80);		\Get copy of command tail
BLIT(CPUREG(11), $81, CPUREG(12), CMDTAIL, $7F);

INHAND:= FOPEN(CMDTAIL, 0);	\Open file name for input
FSET(INHAND, ^I);
OPENI(3);

OUTHAND:= FOPEN("PRINTOUT.TXT", 1);	\Open file name for output
FSET(OUTHAND, ^O);
OPENO(3);

TEXT(0, "-- TEXT FORMATTER TO FILE, V1.0 --

");
TEXT(0, "PRINT PAGE NUMBERS: ");		\SHOW DEFAULT SETTINGS
TEXT(0, if OWN(PAGES) then "YES" else "NO");   CRLF(0);
if OWN(PAGES) then
	begin
	TEXT(0, "STARTING PAGE: ");
	INTOUT(0, OWN(STPAGE));   CRLF(0);
	TEXT(0, "LOW, HIGH PAGE LIMITS: ");
	INTOUT(0, OWN(LOLIM));   CHOUT(0, ^,);   INTOUT(0, OWN(HILIM));
	CRLF(0);
	TEXT(0, "BACK-TO-BACK PAGES: ");
	TEXT(0, if OWN(BAKBAK) then "YES" else "NO");   CRLF(0);
	end;
TEXT(0, "LINE LENGTH: ");   INTOUT(0, OWN(CHMAX));   CRLF(0);
TEXT(0, "LEFT INDENTATION: ");   INTOUT(0, OWN(LEFTLIM));   CRLF(0);
TEXT(0, "LINES TO USE PER PAGE: ");   INTOUT(0, OWN(LNMAX));   CRLF(0);
TEXT(0, "DEFAULTS OK? ");

OPENI(0);
if (CHIN(0) ! $20) =^n then
	begin				\GET NEW DEFAULTS
	CRLF(0);
	TEXT(0, "PRINT PAGE NUMBERS? ");
	OPENI(0);   OWN(PAGES):= (CHIN(0) ! $20) =^y;

	if OWN(PAGES) then
		begin
		OPENI(0);   TEXT(0, "STARTING PAGE? ");
		OWN(STPAGE):= INTIN(0);
		OPENI(0);   TEXT(0, "LOW, HIGH PAGE LIMITS? ");
		OWN(LOLIM):= INTIN(0);   OWN(HILIM):= INTIN(0);
		TEXT(0, "BACK-TO-BACK PAGES? ");
		OPENI(0);   OWN(BAKBAK):= (CHIN(0) ! $20) =^y;
		end;
	OPENI(0);   TEXT(0, "LINE LENGTH? ");   OWN(CHMAX):= INTIN(0);
	OPENI(0);   TEXT(0, "LEFT INDENTATION? ");
	OWN(LEFTLIM):= INTIN(0);
	OPENI(0);   TEXT(0, "LINES TO USE PER PAGE? ");
	OWN(LNMAX):= INTIN(0);
	OPENI(0);   TEXT(0, "READY? ");   II:= CHIN(0);
	end;
LINE:= RESERVE(LINESIZ);
HEADMAX:= OWN(CHMAX)-10;
HEADBUF:= RESERVE(HEADMAX);
FILL:= true;   TITLE:= false;   BOLD:= false;
UNDERLINE:= false;   HEADER:= false;
LNPOS:= 0;   CHPOS:= 0;   SPCNT:= 0;   HEADLEN:= 0;
PITCH:= NPITCH;
PAGENO:= OWN(STPAGE)-1;
PRDEV:= 3;

DOHEAD;
repeat LINEIN; LINEOUT until CHAR=EOF;

SEND(FF);

CHOUT(3, EOF);
CLOSE(3);
FCLOSE(OUTHAND);
end;	\MAIN
