\RAY.XPL	FEB-03-91
\A not so simple Ray Tracer
\ by Loren Blaney
\
\This is based on an article in Byte Magazine, December 1990, page 263.
\
\Vectors are the key to understanding this program.
\
\A right handed, 3-D coordinate system is used with the origin (O) at the
\ upper-left corner of the display screen.
\
\				  Z (into screen)
\			        /
\			      /
\			    /
\			  /
\		      O +------------------- X
\			|
\			|
\			|
\			|		V
\			|	      /
\			|	    /
\			Y	  /
\			     P0 *
\
\The viewer is centered in front of the screen and back some distance, which
\ is negative Z. Objects are behind the screen with positive Z coordinates.
\
\A light ray is represented by a starting position (P0x, P0y, P0z), and a
\ direction vector (Vx, Vy, Vz). Rays are traced backwards, from the eye to
\ the objects. This makes the program much faster by eliminating rays that
\ the viewer never sees. This works because the path that a ray traces is
\ the same whether it is coming or going.
\
\Colors are scaled by 16.

include	\CXPL\CODESI.XPL;

def	TVW=640, TVH=480, TVD=8; \Screen width and height (also see SETVID in MAIN)
			  \WARNING: SAVEIMAGE must have (640, 480, 8)
def	INTSIZE=2;	\Number of bytes in an integer
def	SIZE0=16;	\Initial size of "blocks" (must be power of 2 [1..32])
def	BIG=1E20, SMALL=0.01;		\Relative to size of our objects
def	BEL=$07, CR=$0D, ESC=$1B, SP=$20, CTRL=$40;	\Ascii characters

def	\SURF\	NONE, FILTER, MIRROR, GLASS;

int	CPUREG,		\Pointer to CPU registers. For calls to BIOS routines
	MODE0,		\Video mode when this program started
	OBJECTS,	\Number of objects (spheres)
	SPSURF,		\Array: surface type of each sphere
	SPSURF2,	\Array: surface properties for each surface type
	REDRAW,		\Flag: redraw the image (because the view has changed)
	COLORS,		\The maximum number of colors allowed
	PALRED, PALGRN, PALBLU,	\Array: RGB values for each color (*16)
	PALCNT,		\Array: Counter for each color
	II;		\Scratch for main
real	SPCENTER,	\2-D array: center coordinates for each sphere
	SPRADIUS,	\Array: radius of each sphere
	SP2RADIUS,	\Array: radius squared of each sphere
	EYE,		\Array: coordinates of viewer
	GAIN,		\Magnification of screen image
	XOFFSET,	\Screen image position (horizontal and vertical)
	YOFFSET;

\For DOS I/O:
int	PSPSEG,		\Address of PSP segment (for file I/O)
	DATASEG,	\Address of data segment (heap & stack)
	OUTHAND;	\Output file handle
addr	OUTFILE,	\Output file name
	OUTTEMP,	\Temporary output file name (.$$$)
	OUTBACK;	\Backup output file name (.BAK)
def	NAMESIZE=40;	\Maximum number of characters in a file name

\----------------------- FUNDAMENTAL ROUTINES -------------------------

func	GETKEY;		\Get character from keyboard.
int	CH;		\This is a low-level routine with no echo,
begin			\ no ctrl-C, and no cursor.
CPUREG(0):= 0;		\Requires: CPUREG:= GETREG;   Function = $00
SOFTINT($16);		\Call BIOS routine
CH:= CPUREG(0) & $FF;
if CH = 0 then
	begin			\Handle special keys
	case CPUREG(0)>>8 of	\Arrows:
	  72:	CH:= $0B;	\ up
	  80:	CH:= $0A;	\ down
	  75:	CH:= $08;	\ left
	  77:	CH:= $15	\ right
	other	CH:= -(CPUREG(0)>>8);
	end;
return CH;
end;	\GETKEY



func	GETVID;		\Returns the current video mode
begin			\Requires: CPUREG:= GETREG;
CPUREG(0):= $0F00;	\Function $0F
SOFTINT($10);		\Call BIOS routine
return CPUREG(0) & $FF;
end;	\GETVID



proc	PALETTE(N, R, G, B);	\Set color register N
int	N,		\Register to set
	R, G, B;	\Red, Green, and Blue [0..255]
begin			\Requires: CPUREG:= GETREG;
CPUREG(0):= $1010;		\AX function $10, subfunction $10
CPUREG(1):= N;			\BX select color register
CPUREG(2):= G<<8 ! (B&$FF);	\CX green<<8 blue
CPUREG(3):= R<<8;		\DX red<<8
SOFTINT($10);			\Call BIOS routine
end;	\PALETTE



proc	FIXPALETTE;	\Make palette registers correspond to DAC registers
int	N;
begin			\Requires: CPUREG:= GETREG;
for N:= 0, 15 do
	begin
	CPUREG(0):= $1000;	\AX function $10, subfunction $00
	CPUREG(1):= N<<8!N;	\BX DAC and palette registers
	SOFTINT($10);		\Call BIOS routine
	end;
end;	\FIXPALETTE

\---------------------------- I/O COMMANDS ------------------------------

proc	DOSOPEN;	\Routine to open MS-DOS files
int	INX;		\index into TAIL
addr	TAIL, TLEN, STR;



proc	COPYNAME(FROM, TO, EXT);
\Copy file name from FROM to TO and replace any existing extention with EXT
addr	FROM, TO,	\Strings are terminated by 0 and must be <= NAMESIZE long
	EXT;		\Extension started with "." and terminated with an
			\ unused char with MSB set
int	I, J, CH;
begin
I:= 0;
loop	begin
	CH:= FROM(I);
	if CH=^. ! CH=0 then quit;
	TO(I):= CH;
	if I < NAMESIZE-1 then I:= I +1;
	end;
for J:= 0, 3 do
	begin
	TO(I):= EXT(J);
	if I < NAMESIZE-1 then I:= I +1;
	end;
TO(I):= 0;
end;	\COPYNAME



procc	GETNAME(NAME);
\Inputs: INX, TAIL, NAMESIZE
addr	NAME;
int	I, CH;
begin
while TAIL(INX) =SP do INX:= INX +1;
I:= 0;
repeat	CH:= TAIL(INX);
	INX:= INX +1;
	if CH=SP ! CH=CR then CH:= 0;	\terminate NAME with 0
	NAME(I):= CH;
	if I < NAMESIZE-1 then I:= I +1;
until CH =0;
end;	\GETNAME



begin	\DOSOPEN
TLEN:= RESERVE(80);			\get command tail from PSP
TAIL:= TLEN +1;
BLIT(PSPSEG, $80, DATASEG, TLEN, 79);
if TLEN(0)>79 then TLEN(0):= 79;

STR:= "RAYIMAGE.XXX ";
INX:= 0;
GETNAME(OUTFILE);
if OUTFILE(0) =0 then			\if no outfile then use 
	for INX:= 0, 11 do
		OUTFILE(INX):= STR(INX);

COPYNAME(OUTFILE, OUTFILE, ".IFF ");	\set default extension
COPYNAME(OUTFILE, OUTTEMP, ".$$$ ");	\set up other output file names
COPYNAME(OUTFILE, OUTBACK, ".BAK ");

OUTHAND:= FOPEN(OUTTEMP, 1);
FSET(OUTHAND, ^O);
OPENO(3);
end;	\DOSOPEN



proc	DOSCLOSE;	\Close MS-DOS files
begin
CLOSE(3);
FCLOSE(OUTHAND);

CPUREG(\AX\0):= $4100;			\DEL OUTFILE.BAK
CPUREG(\DS\9):= DATASEG;
CPUREG(\DX\3):= OUTBACK;
SOFTINT($21);				\(.BAK file might not exist)

CPUREG(\AX\0):= $5600;			\REN OUTFILE.EXT OUTFILE.BAK
CPUREG(\DS\9):= DATASEG;
CPUREG(\DX\3):= OUTFILE;
CPUREG(\ES\11):= DATASEG;
CPUREG(\DI\4):= OUTBACK;
SOFTINT($21);

CPUREG(\AX\0):= $5600;			\REN OUTFILE.$$$ OUTFILE.EXT
CPUREG(\DS\9):= DATASEG;
CPUREG(\DX\3):= OUTTEMP;
CPUREG(\ES\11):= DATASEG;
CPUREG(\DI\4):= OUTFILE;
SOFTINT($21);
end;	\DOSCLOSE

\----------------------------------------------------------------------

proc	SAVEIMAGE;	\Save current 640x480x256 color image as an .IFF file



proc	WORDOUT(W);	\Output a 16-bit long word to disk
int	W;
begin
CHOUT(3, W>>8);
CHOUT(3, W);
end;	\WORDOUT



proc	LONGOUT(H, L);	\Output a 32-bit long word to disk
int	H, L;
begin
WORDOUT(H);
WORDOUT(L);
end;	\LONGOUT



proc	DOBODY;		\Write the body chunk to disk
addr	LINE;
int	X, Y, Z, PLANE, BIT, BYTE;
\Inputs: TVW, TVH, TVD			MUST BE (640, 480, 8)
\TVW must be an even multiple of 8
begin
LINE:= RESERVE(TVW);

TEXT(3, "BODY");
\TVW/8 *TVH *TVD
LONGOUT($0004, $B000);

for Y:= 0, TVH-1 do
	begin
	for X:= 0, TVW-1 do
		LINE(X):= READPIX(X, Y);

	PLANE:= $01;
	for Z:= 0, TVD-1 do
		begin
		X:= 0;
		repeat	BIT:= $80;   BYTE:= 0;
			repeat	if LINE(X) & PLANE then BYTE:= BYTE ! BIT;
				X:= X +1;
				BIT:= BIT >>1;
			until BIT = 0;
			CHOUT(3, BYTE);
		until X >= TVW;
		PLANE:= PLANE <<1;
		end;
	end;
end;	\DOBODY



proc	DOCMAP;		\Write the color map chunk to disk
int	N, R, G, B;
begin
TEXT(3, "CMAP");
LONGOUT($0000, 1<<TVD *3);

for N:= 0, 1<<TVD-1 do
	begin
	CPUREG(0):= $1015;		\AX function $10, subfunction $15
	CPUREG(1):= N;			\BX select color register
	SOFTINT($10);
	G:= CPUREG(2) >>8;
	B:= CPUREG(2) &$FF;
	R:= CPUREG(3) >>8;

	CHOUT(3, R <<2);
	CHOUT(3, G <<2);
	CHOUT(3, B <<2);
	end;
end;	\DOCMAP



proc	DOBMHD;		\Write the bit map header chunk to disk
begin
TEXT(3, "BMHD");
LONGOUT($0000, $0014);

WORDOUT(TVW);		\WIDTH
WORDOUT(TVH);		\HEIGHT
WORDOUT(0);		\XOFFSET
WORDOUT(0);		\YOFFSET
CHOUT(3, TVD);		\DEPTH
CHOUT(3, 0);		\MASKING = NONE
CHOUT(3, 0);		\COMPRESSION = NONE
CHOUT(3, 0);		\PAD1
WORDOUT(0);		\TRANSPARENT (NOT USED)
CHOUT(3, 1);		\XASPECT
CHOUT(3, 1);		\YASPECT
WORDOUT(TVW);		\PAGEWIDTH
WORDOUT(TVH);		\PAGEHEIGHT
end;	\DOBMHD



begin	\SAVEIMAGE
DOSOPEN;

TEXT(3, "FORM");
LONGOUT($0004, $B330);

TEXT(3, "ILBM");
DOBMHD;
DOCMAP;
DOBODY;

DOSCLOSE;
CHOUT(0, BEL);	\Done at last!
end;	\SAVEIMAGE

\----------------------------------------------------------------------

proc	DOKEY;		\Handle keyboard commands
int	CH;
begin
CH:= GETKEY;
REDRAW:= true;		\Anticipate that a redraw will be necessary
case CH of
  ^.:	GAIN:= GAIN *1.1;			\Zoom in a little
  ^,:	GAIN:= GAIN /1.1;			\Zoom out a little
  ^>:	GAIN:= GAIN *2.0;			\Zoom in to double size
  ^<:	GAIN:= GAIN /2.0;			\Zoom out to half size
  $15:	XOFFSET:= XOFFSET -FLOAT(SIZE0);	\Shift to the left
  $08:	XOFFSET:= XOFFSET +FLOAT(SIZE0);	\Shift to the right
  $0A:	YOFFSET:= YOFFSET -FLOAT(SIZE0);	\Shift up
  $0B:	YOFFSET:= YOFFSET +FLOAT(SIZE0);	\Shift down (Byte, byte, byte!)
  ^C-CTRL, ESC:  [SETVID(MODE0);   exit];	\Restore video mode and exit
  ^S-CTRL: [CH:= GETKEY;   REDRAW:=false];	\Stop/start program
  ^Q-CTRL: [SAVEIMAGE;   REDRAW:=false]		\Save image to disk
other	[CHOUT(0, BEL);   REDRAW:= false];	\Illegal command -- beep
end;	\DOKEY



proc	BLOCK(X0, Y0, C, SIZE);	\Plot a square block
int	X0, Y0,	\Coordinates of upper-left corner of block
	C,	\Color of the block
	SIZE;	\Number of pixels on a side (of the block)
int	X, Y1;
begin
if SIZE = 1 then
	POINT(X0, Y0, C)	  \(For speed)
else	begin
	Y1:= Y0 +SIZE -1;				\ Y0 ---|||--
	for X:= X0, X0+SIZE-1 do			\       |||
		[MOVE(X, Y0);   LINE(X, Y1, C)];	\       |||
	end;
end;	\BLOCK



proc	PLOT(X, Y, COLOR, SIZE);	\Plot a square block
int	X, Y,	\Coordinates of upper-left corner of block
	COLOR,	\Vector: color of the block
	SIZE;	\Number of pixels on a side (of the block)
\Other inputs: COLORS, PALRED, PALGRN, PALBLU
int	LEAST, N, SN, R, G, B;
begin
R:= COLOR(0)>>4;   G:= COLOR(1)>>4;   B:= COLOR(2)>>4;
for N:= 0, COLORS-1 do
	begin	\Find the palette number N that matches the COLOR
	if R=PALRED(N) then
	  if G=PALGRN(N) then
	    if B=PALBLU(N) then
		[BLOCK(X, Y, N, SIZE);   \PALCNT(N):= PALCNT(N) +1;\   return];
	end;

\No palette matches COLOR so set up a new palette:
LEAST:= $7FFF;		\16-bits are not enough!
for N:= 1, COLORS-1 do	\Don't change color 0, the border
	begin
	if PALCNT(N) < LEAST then [LEAST:= PALCNT(N);   SN:= N];
	end;
if LEAST # 0 then CHOUT(0, BEL); \Warning: existing color being replaced
PALRED(SN):= R;
PALGRN(SN):= G;
PALBLU(SN):= B;
PALETTE(SN, R, G, B);
PALCNT(SN):= 1;
BLOCK(X, Y, SN, SIZE);
end;	\PLOT

\------------------------- VECTOR ROUTINES ----------------------------

func real INTPLANE(RP, RV, Y0);	\Intersection of a ray and the X-Z plane.
\Returns the distance from RP to the point of intersection. The coordinates
\ of this point of intersection can be obtained by: P = RP + RV *DIST.
\Returns 0.0 if there is no intersection.
real	RP,	\Array: Ray point
	RV,	\Array: Ray vector
	Y0;	\Y position of X-Z plane
real	DIST;	\Distance along RV from RP to point of intersection
begin
\DIST = (Y0 - P0y) /Vy
if RLABS(RV(1)) < SMALL then return 0.0;	\Avoid divide by 0.0 error
DIST:= (Y0 - RP(1)) /RV(1);

if DIST > 0.0 then return DIST;	\The usual case
return 0.0;			\The plane is behind the ray
end;	\INTPLANE



func real INTSPHERE(RP, RV, S, R2);	\Intersection of a ray and a sphere.
\Returns the distance from RP to the point of intersection. The coordinates
\ of this point of intersection can be obtained by: P = RP + RV *DIST.
\Returns 0.0 if there is no intersection.
real	RP,	\Array: Ray point
	RV,	\Array: Ray vector
	S,	\Array: Sphere center
	R2;	\Radius of sphere squared
real	B, C, D, G0, G1, G2,		\Intermediate variables
	DIST, DIST1, DIST2;	\Distances to points of intersection
begin
G0:= RP(0) -S(0);			\G:= P0 - S;
G1:= RP(1) -S(1);
G2:= RP(2) -S(2);

B:= RV(0)*G0 + RV(1)*G1 + RV(2)*G2;	\B:= RV.G

C:= G0*G0 + G1*G1 + G2*G2 - R2;

D:= B *B - C;
if D <= 0.0 then return 0.0;		\Ray does not intersect sphere

D:= SQRT(D);
DIST1:= +D -B;
DIST2:= -D -B;

DIST:= if DIST1 < DIST2 then DIST1 else DIST2;	\Get the closer point
if DIST > SMALL then return DIST;		\The usual case
if DIST1 > SMALL then return DIST1;	\The ray starts inside the sphere
if DIST2 > SMALL then return DIST2;
return 0.0;				\The sphere is behind the ray
end;	\INTSPHERE



proc	REFLECT(I, N, R);		\Reflect a ray
real	I, N, R;	\Unit vectors: Incident, Normal, Reflected
real	D;

\		                 N
\		                 |        R
		          \      |      /
		            \    |    /
		              \ I|  /
		                \|/
\		        ---------*----------

begin
D:= I(0)*N(0) + I(1)*N(1) + I(2)*N(2);	\R = I - 2 *(I.N) *N
D:= D +D;

R(0):= I(0) - N(0) *D;
R(1):= I(1) - N(1) *D;
R(2):= I(2) - N(2) *D;
end;	\REFLECT



proc	REFRACT(I, N, R, N1, N2);	\Refract a ray
real	I, N, R;	\Unit vectors: Incident, Normal, Refracted
real	N1, N2;		\Indexes of refraction
real	C, D, E;

\		                N
\		                |
		         \      |
		           \    |
		             \ I|
		               \|       N1
\		       ---------*----------
		                 \      N2
		                  \
		                   \
		                    \
\		                     R
begin
E:= N2 /N1;
C:= -(I(0)*N(0) + I(1)*N(1) + I(2)*N(2));	\C = -I.N

D:= C*C +E*E -1.0;	\R:= ((C - SQRT(C*C +E*E -1.0)) *N + I) /E

if D < 0.0 then		\Total internal reflection; REFLECT(I, N, R)
	begin				\R = I - 2 *(I.N) *N
	C:= C +C;			\R = I + 2 *C *N
	R(0):= I(0) + N(0) *C;
	R(1):= I(1) + N(1) *C;
	R(2):= I(2) + N(2) *C;
	return
	end;

D:= C - SQRT(D);

R(0):= (D *N(0) + I(0)) /E;		\R:= (D *N + I) /E
R(1):= (D *N(1) + I(1)) /E;
R(2):= (D *N(2) + I(2)) /E;
end;	\REFRACT

\----------------------------- THE HEART ------------------------------

proc	TRACERAYS;	\Determine the correct color for each pixel and plot it
real	RAYV,		\Array: Ray vector
	RAV,		\Array: Ray unit vector
	LEN, LEN1, LEN2, \Length of ray vector (RAYV)
	DRX;		\Delta Ray vector X
int	X, Y,		\Pixel coordinates on the screen
	SIZE,		\Block size (resolution control)
	COLOR,		\Vector: Pixel color
	LEVEL,		\Level of recursion
	XF, YF,		\Flags used to plot smaller and smaller blocks
	FIRSTTIME;


proc	TRACE(RP, RV);	\Trace ray from point RP in direction RV.
\Set COLOR to the color of the closest object intersected by this ray.
\COLOR is left unchanged if no object is intersected.
real	RP, RV;		\Ray point, Ray vector
real	RP2, RV2,	\Ray point, Ray vector for reflected ray, if any
	RP3, RV3,	\Ray point, Ray vector for refracted ray, if any
	NORM,		\Vector normal (perpendicular) to sphere at point
			\ of intersection
	DIST, DISTMIN,	\Distance to point of intersection, Closest distance
	SPC, SPR;	\Array: closest sphere center, Closest sphere radius
int	OBJ, OBJMIN,	\Object, Closest object
	SURF;


	proc	CHECKER(D);	\Make checker-board pattern for X-Z plane
	\Inputs RP and RV
	real	D;	\Distance along RV from RP to point of intersection
	real	PX, PZ;
	int	IPX, IPZ, F;
	def	CHECKSIZE= 1000.0 \18.75\;	\Size of checker squares
	begin
	\Px = P0x + Vx *D		Find point of intersection
	PX:= RP(0) + RV(0) *D;		PZ:= RP(2) + RV(2) *D;
	IPX:= FIX(PX /CHECKSIZE - 0.5);	IPZ:= FIX(PZ /CHECKSIZE - 0.5);
	\COLOR:= if (IPX&$0F)= $0F ! (IPZ&$0F)=$0F then RED else BLACK;
	\COLOR:= if IPX &IPZ &$0F then BLACK else RED;
	\COLOR:= if (IPX&$03)= $03 ! (IPZ&$03)=$03 then BLUE else WHITE;

	F:= (IPX&$03)= $03 ! (IPZ&$03)=$03;
	COLOR(0):= if F then 0 else $300;
	COLOR(1):= if F then $200 else $300;
	COLOR(2):= if F then $300 else $300;
	end;	\CHECKER


begin	\TRACE
if LEVEL >= 100 then return;	\Limit the recursion
LEVEL:= LEVEL +1;		\Next level of recursion down

SURF:= NONE;
DISTMIN:= BIG;			\Infinitely far away (more or less)
for OBJ:= 0, OBJECTS-1 do	\For all objects see if ray intersects them
	begin
	DIST:= INTSPHERE(RP, RV, SPCENTER(OBJ), SP2RADIUS(OBJ));
	if DIST > SMALL then	\(Prevent infinite reflections due to rounding errors)
	    if DIST < DISTMIN then	\Ray intersects -- save closest point
		begin
		DISTMIN:= DIST;
		SURF:= SPSURF(OBJ);	\Save surface type of closest point
		OBJMIN:= OBJ;
		end;
	end;

if SURF = FILTER then
	begin
	COLOR(0):= SPSURF2(OBJMIN, 0);
	COLOR(1):= SPSURF2(OBJMIN, 1);
	COLOR(2):= SPSURF2(OBJMIN, 2);
	end;

if DISTMIN = BIG then
	begin		\Ray didn't hit sphere. See if it hits the bottom plane.
	DIST:= INTPLANE(RP, RV, 3500.0 \340.0\); \Y distance to bottom plane
	if DIST > SMALL then	\(Prevent infinite reflections due to rounding errors)
	    if DIST < DISTMIN then \Ray intersects -- save closest point
		begin
		DISTMIN:= DIST;
		CHECKER(DIST);	\Get color of this closest point
		SURF:= FILTER;
		end;
	end;

if SURF = MIRROR then	\Reflect ray and trace it
	begin		\(This does not check for reflections off planes yet)
	RP2:= RLRES(3);   RV2:= RLRES(3);   NORM:= RLRES(3);

	\RP2 = RP + RV *DISTMIN		Point where ray intersects sphere
	RP2(0):= RP(0) + RV(0) *DISTMIN;
	RP2(1):= RP(1) + RV(1) *DISTMIN;
	RP2(2):= RP(2) + RV(2) *DISTMIN;

	SPC:= SPCENTER(OBJMIN);   SPR:= SPRADIUS(OBJMIN);

	\NORM:= (RP2 - C) /R	Normal to sphere at point of intersection
	NORM(0):= (RP2(0) - SPC(0)) /SPR;
	NORM(1):= (RP2(1) - SPC(1)) /SPR;
	NORM(2):= (RP2(2) - SPC(2)) /SPR;

	REFLECT(RV, NORM, RV2);		\(Incident, Normal, Reflected)

	\Background color for reflections (gray)
	COLOR(0):= $80;   COLOR(1):= $100;   COLOR(2):= $200;

	TRACE(RP2, RV2);		\Trace reflected ray

	COLOR(0):= COLOR(0) - SPSURF2(OBJMIN, 0);
	COLOR(1):= COLOR(1) - SPSURF2(OBJMIN, 1);
	COLOR(2):= COLOR(2) - SPSURF2(OBJMIN, 2);
	if COLOR(0) < 0 then COLOR(0):= 0;
	if COLOR(1) < 0 then COLOR(1):= 0;
	if COLOR(2) < 0 then COLOR(2):= 0;
	end

else if SURF = GLASS then	\Refract ray and trace it
	begin		\(This does not check for refractions in planes yet)
	RP2:= RLRES(3);   RV2:= RLRES(3);   NORM:= RLRES(3);
	RP3:= RLRES(3);   RV3:= RLRES(3);

	\RP2 = RP + RV *DISTMIN		Point where ray intersects sphere
	RP2(0):= RP(0) + RV(0) *DISTMIN;
	RP2(1):= RP(1) + RV(1) *DISTMIN;
	RP2(2):= RP(2) + RV(2) *DISTMIN;

	SPC:= SPCENTER(OBJMIN);   SPR:= SPRADIUS(OBJMIN);

	\NORM:= (RP2 - C) /R	Normal to sphere at point of intersection
	NORM(0):= (RP2(0) - SPC(0)) /SPR;
	NORM(1):= (RP2(1) - SPC(1)) /SPR;
	NORM(2):= (RP2(2) - SPC(2)) /SPR;

	REFRACT(RV, NORM, RV2, 1.0, 1.5);	\(Incident, Normal, Refracted)

	DIST:= INTSPHERE(RP2, RV2, SPC, SPR*SPR);

	\RP3 = RP2 + RV2 *DIST		Point where ray intersects sphere
	RP3(0):= RP2(0) + RV2(0) *DIST;
	RP3(1):= RP2(1) + RV2(1) *DIST;
	RP3(2):= RP2(2) + RV2(2) *DIST;

	\NORM:= -(RP3 - C) /R	Normal to sphere at point of intersection
	NORM(0):= (SPC(0) - RP3(0)) /SPR;
	NORM(1):= (SPC(1) - RP3(1)) /SPR;
	NORM(2):= (SPC(2) - RP3(2)) /SPR;

	REFRACT(RV2, NORM, RV3, 1.5, 1.0);	\(Incident, Normal, Refracted)

	\Background color for refractions (cyan)
	COLOR(0):= $00;   COLOR(1):= $300;   COLOR(2):= $300;

	TRACE(RP3, RV3);		\Trace refracted ray

	COLOR(0):= COLOR(0) - SPSURF2(OBJMIN, 0);
	COLOR(1):= COLOR(1) - SPSURF2(OBJMIN, 1);
	COLOR(2):= COLOR(2) - SPSURF2(OBJMIN, 2);
	if COLOR(0) < 0 then COLOR(0):= 0;
	if COLOR(1) < 0 then COLOR(1):= 0;
	if COLOR(2) < 0 then COLOR(2):= 0;
	end;
end;	\TRACE



begin	\TRACERAYS
RAYV:= RLRES(3);   RAV:= RLRES(3);
COLOR:= RESERVE(3*INTSIZE);
FIRSTTIME:= true;
SIZE:= SIZE0;

RAYV(2):= \0.0\ - EYE(2);
LEN2:= RAYV(2)*RAYV(2);

loop	begin
	YF:= false;
	Y:= 0;
	while Y < TVH do
		begin
		if KEYHIT then DOKEY;			\Check for key commands
		if REDRAW then return;

		RAYV(1):= (FLOAT(Y) - EYE(1) + YOFFSET) /GAIN;
		LEN1:= RAYV(1)*RAYV(1) + LEN2;

		RAYV(0):= (XOFFSET - EYE(0)) /GAIN;
		DRX:= FLOAT(SIZE) /GAIN;

		XF:= false;
		X:= 0;
		while X < TVW do
			begin
			if XF ! YF ! FIRSTTIME then	\(Resolution control)
				begin	\Project ray from viewer to pixel X,Y
				\Make it a unit vector
				LEN:= SQRT(RAYV(0)*RAYV(0) + LEN1);
				RAV(0):= RAYV(0) /LEN;
				RAV(1):= RAYV(1) /LEN;
				RAV(2):= RAYV(2) /LEN;

				\Background color (black):
				COLOR(0):= 0;   COLOR(1):= 0;   COLOR(2):= 0;

				LEVEL:= 0;
				TRACE(EYE, RAV);	\Trace ray and get COLOR

				PLOT(X, Y, COLOR, SIZE); \Plot COLOR
				end;
			X:= X +SIZE;			\Skip SIZE pixels
			XF:= ~XF;
			RAYV(0):= RAYV(0) + DRX;
			end;
		Y:= Y +SIZE;				\Skip SIZE pixels
		YF:= ~YF;
		end;
	SIZE:= SIZE >>1;				\Chop block size in half
	if SIZE = 0 then quit;
	FIRSTTIME:= false;
	end;
end;	\TRACERAYS

\----------------------------------------------------------------------

proc	INITIALIZE;
begin
MODE0:= GETVID;		\Save current video mode so it can be restored upon exit
SETVID($62);		\Set VGA mode (also see def TVW, TVH)

COLORS:=256;

\FIXPALETTE;
PALETTE(0, 0, 0, 0);		\Border color = black



\Make some objects (spheres):
OBJECTS:= 5;				\Number of objects (spheres)

SPSURF:= [GLASS, GLASS, GLASS,
	GLASS, GLASS, GLASS,
	GLASS, GLASS, GLASS,
	GLASS, GLASS, GLASS,
	GLASS, GLASS, GLASS,
	GLASS, GLASS, GLASS];		\Color of each sphere

SPRADIUS:= [300.0, 100.0, 100.0,
	    100.0, 100.0, 100.0,
	    100.0, 100.0, 100.0,
	    100.0, 100.0, 100.0,
	    100.0, 100.0, 100.0,
	    100.0, 100.0, 100.0];	\Radii

SPCENTER:=[[450.0, 240.0, 900.0],	\Centers (X,Y,Z)

	   [300.0, 240.0, 000.0],
	   [600.0, 240.0, 000.0],

	   [000.0, 240.0, 300.0],
	   [300.0, 240.0, 300.0],
	   [600.0, 240.0, 300.0],

	   [000.0, 240.0, 600.0],
	   [300.0, 240.0, 600.0],
	   [600.0, 240.0, 600.0],

	   [-300.0, 240.0, 900.0],
	   [000.0, 240.0, 900.0],
	   [300.0, 240.0, 900.0],

	   [-300.0, 240.0, 1200.0],
	   [000.0, 240.0, 1200.0],
	   [300.0, 240.0, 1200.0],
	   [600.0, 240.0, 1200.0],

	   [-600.0, 240.0, 1500.0],
	   [600.0, 240.0, 1500.0],

	   [1600.0, -4000.0, -4000.0]];

EYE:= [640.0, -400.0, -2000.0];		\Coordinates of viewer (X,Y,Z)

GAIN:= 0.80;   XOFFSET:= 100.0;   YOFFSET:= -200.0;




\Make some objects (spheres):
OBJECTS:= 3;				\Number of objects (spheres)

SPSURF:= [FILTER,
	GLASS,
	MIRROR];

SPSURF2:= [
	[$3F0, $200, $100],		\COLOR RGB (63*16)
	[$280, $40, $40],
	[$180, $00, $100] ];

SPRADIUS:= [200.0, 300.0, 600.0];	\Radii
SPCENTER:=[[400.0, 700.0, 1200.0],	\Centers (X,Y,Z)
	   [-100.0, 400.0, 1640.0],
	   [700.0, 170.0, 2000.0]];

EYE:= [320.0, 240.0, -1500.0];		\Coordinates of viewer (X,Y,Z)

GAIN:= 1.00;   XOFFSET:= -30.0;   YOFFSET:= 90.0;
end;	\INITIALIZE

\----------------------------------------------------------------------

begin	\MAIN
CPUREG:= GETREG;	\Get registers so we can call BIOS routines
PSPSEG:= CPUREG(11);
DATASEG:= CPUREG(12);

OUTFILE:= RESERVE(NAMESIZE);
OUTTEMP:= RESERVE(NAMESIZE);
OUTBACK:= RESERVE(NAMESIZE);

INITIALIZE;

PALRED:= RESERVE(COLORS*INTSIZE);
PALGRN:= RESERVE(COLORS*INTSIZE);
PALBLU:= RESERVE(COLORS*INTSIZE);
PALCNT:= RESERVE(COLORS*INTSIZE);

for II:= 0, COLORS-1 do
	begin
	PALRED(II):= 0;
	PALGRN(II):= 0;
	PALBLU(II):= 0;
	PALCNT(II):= 0;
	end;

SP2RADIUS:= RLRES(OBJECTS);	\Set up squares of radii (for speed)
for II:= 0, OBJECTS-1 do
	SP2RADIUS(II):= SPRADIUS(II) *SPRADIUS(II);

loop	begin
	loop	begin
		REDRAW:= false;		\This flag is set to true by DOKEY
		TRACERAYS;
		if ~REDRAW then quit;
		end;
	CHOUT(0, BEL);	\Done at last!
	repeat DOKEY until REDRAW;	\Wait for key press to exit or redraw
	end;
end;	\MAIN
