\ROLL.XPL	NOV-07-93

\This displays 640x480 .LBM files. (InterLeaved BitMap image in Interchange
\ File Format). Reference: ROM Kernel Reference Manual, IFF, p I-28.

\REVISION HISTORY:
\DEC-03-89, Original on Amiga
\OCT-24-93, Converted to run on PC.
\NOV-07-93, Roll and unroll image.

inc	C:\CXPL\CODESI;		\Include code definitions for intrinsic routines
ext	CopyLine;		\Copy a scan line from Raster to screen

def	INTSIZE= 2;		\Number of bytes in an integer
def	TV=0, KB=0, PR=2;					\Device numbers
def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0a, FF=$0c,	\Control chars
	CR=$0d, EOF=$1a, Esc=$1b, Sp=$20, Ctrl=$40;

int	Cpureg;			\Address of CPU register array (from Getreg)
def	AX, BX, CX, DX, DI, SI, BP, CF, CS, DS, SS, ES;		\Getreg registers

def	Pi = 3.14159265;

def	RADIUS=30, RADIUS2=RADIUS*2;	\Radius of roller

def	FO=$464F, RM=$524D, IL=$494C, BM=$424D, HD=$4844, CM=$434D, AP=$4150,
	BO=$424F, DY=$4459;

int	HAVEBMHD,	\Flag: Bit map header chunk has been read in
	RECHDH,		\Record header
	RECHDL,		\Record header, 4-character name
	WIDTH, HEIGHT, DEPTH,	\Image dimensions (pixels)
	COMPRESSION,	\Compression algorithm (0= none, 1=ByteRun1)
	VIDMODE,	\Initial video mode
	II;		\Scratch for Main

seg addr Screen;	\Array: Graphic display screen
seg addr RASTER;	\Array: Image in memory
int	Roller;		\Array: Used to make image of roller

\======================================================================

proc	WaitKey;		\Wait for a keystroke
begin
Openi(1);
repeat until ChkKey;
end;	\WaitKey



proc	WaitVB;			 \Wait for beginning of next vertical blank
begin				 \WARNING: even $3ba might not work for monochrome
\while Pin($3da, 0) & $08 do;	 \\Wait for no vertical blank
repeat until Pin($3da, 0) & $08; \Wait for vertical blank
end;	\WaitVB;



func	GetVid;			\Returns the current video mode
begin
Cpureg(AX):= $0f00;		\Function $0f
Softint($10);			\Call BIOS routine
return Cpureg(AX) & $ff;
end;	\GetVid



proc	SetColReg(N, R, G, B);	\Set VGA color register (DAC) N
int	N,	 \Register to set (0..255)
	R, G, B; \Red, Green, and Blue (only low 6 bits are used)
begin		 \ 2**6 = 64;  64 *64 *64 = 262144 possible colors
Cpureg(AX):= $1010;		\Function $10, subfunction $10
Cpureg(BX):= N;			\Select color register
Cpureg(CX):= G<<8 ! B&$ff;	\Green<<8 blue
Cpureg(DX):= R<<8;		\Red<<8
Softint($10);			\Call BIOS routine
end;	\SetColReg



proc	SetPalReg(N, C);	\Set palette register
int	N,	\Register to set (0..15)
	C;	\Color for EGA bits: 0 0 r g b R G B  (64 colors)
begin		\                    0 0 0 I 0 R G B  (16 colors)
Cpureg(AX):= $1000;
Cpureg(BX):= C<<8 ! N&$ff;
Softint($10);			\Call BIOS
end;	\SetPalReg



proc	AlignPalette;		\Make palette regs correspond to VGA regs (DAC)
int	N;
begin
for N:= 0, 15 do
	begin
	Cpureg(AX):= $1000;	\Function $10, subfunction $00
	Cpureg(BX):= N<<8 !N;	\DAC and palette registers
	Softint($10);		\Call BIOS routine
	end;
end;	\AlignPalette



proc	SetBorder(C);		\Set border color register
int	C;	\Color for EGA bits: 0 0 r g b R G B  (64 colors)
begin		\                    0 0 0 I 0 R G B  (16 colors)
Cpureg(AX):= $1001;
Cpureg(BX):= C<<8;
Softint($10);
end;	\SetBorder

\======================================================================

proc	FATAL(STR);		\Display error message and abort
addr	STR;
begin
TEXT(TV,"
NOPE - ");
CHOUT(TV, BEL);
TEXT(TV, STR);
CRLF(TV);

WaitKey;
Openi(1);
SETVID(VIDMODE);
exit;
end;	\FATAL



proc	SHOWSTATS;		\Show information about image
begin
if COMPRESSION = 0 then
	[Cursor(34, 23);
	TEXT(6, "UNCOMPRESSED")];
Cursor(35, 24);
Attrib($01);
Intout(6, WIDTH);
Chout(6, ^x);
Intout(6, HEIGHT);
Chout(6, ^x);
Intout(6, DEPTH);
end;	\SHOWSTATS



func	HEXWORD;		\Read in 2 bytes and return value of the word
return SWAP(CHIN(3)) + CHIN(3);

\----------------------------------------------------------------------

proc	DOFORM;			\Read in form type (must be ILBM)
int	CKSIZE;
begin
CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;
if HEXWORD#IL ! HEXWORD#BM then FATAL("^"ILBM^" MISSING");
end;	\DOFORM

\----------------------------------------------------------------------

proc	DOBMHD;			\Read in BitMap Header chunk
int	CKSIZE,
	XOFFSET, YOFFSET,
	MASKING,	\These parameters are read in but are not used
	PAD1,
	TRANSPARENT,
	XASPECT, YASPECT,
	PAGEWIDTH, PAGEHEIGHT;
begin
CKSIZE:= HEXWORD;	\Chunk size in bytes (must be even)
CKSIZE:= HEXWORD;

WIDTH:= HEXWORD;	\Raster width and height in pixels
HEIGHT:= HEXWORD;
XOFFSET:= HEXWORD;	\Offset (in pixels) for this image
YOFFSET:= HEXWORD;	\ (not used)
DEPTH:= CHIN(3);	\Number of bit planes (can be 0)
MASKING:= CHIN(3);	\Masking technique (0=no mask, is assumed)
COMPRESSION:= CHIN(3);	\Compression algorithm (0=none, 1=byte run)
PAD1:= CHIN(3);		\Unused by IFF standard
TRANSPARENT:= HEXWORD;	\Transparent color (only used w masking)
XASPECT:= CHIN(3);	\Aspect ratio (1:1 is assumed here)
YASPECT:= CHIN(3);
PAGEWIDTH:= HEXWORD;	\Source page size (not used here)
PAGEHEIGHT:= HEXWORD;

HAVEBMHD:= true;	\Indicate that BMHD chunk has been read in
end;	\DOBMHD

\----------------------------------------------------------------------

proc	DOCMAP;		\Color Map, set up color registers
int	CKSIZE,		\Number of bytes in CMAP field
	COLREGS,	\Number of color registers to use
	N,
	RED, GREEN, BLUE;
begin
CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;
COLREGS:= CKSIZE /3;
if COLREGS >256 then COLREGS:= 256;	\Limit maximum number of color registers

for N:= 0, COLREGS-1 do
	begin
	RED:= CHIN(3)>>2;   GREEN:= CHIN(3)>>2;   BLUE:= CHIN(3)>>2;
	SetColReg(N, RED, GREEN, BLUE);
	end;

\Eat the remainder. Must end on an even-byte boundary.
for N:= CKSIZE/3 *3, CKSIZE-1 do
	RED:= CHIN(3);
end;	\DOCMAP

\======================================================================

proc	DrawRoller(Y);		\Draw the roller
\Inputs: RADIUS, Roller, Height
int	Y;	\Top edge
int	J, K, Yi;
begin
K:= RADIUS2 + Fix(Pi *Float(RADIUS));

for J:= 0, RADIUS2-1 do
	begin
	Yi:= Y + Roller(J);
	if Yi >= Height then
		Yi:= Y + (if J <= RADIUS then J else K-Roller(J));
	if Yi < Height then CopyLine(Yi, Y+J, Raster, Depth);
	end;
end;	\DrawRoller



proc	MakeRoller;		\Make Roller, an image of the roller
\Inputs: RADIUS
\Outputs: Roller
def	Pi2 = Pi/2.0, Pi32 = Pi2*3.0;
int	J, K, Yi;
real	R, Ang;
begin
R:= Float(RADIUS);
for J:= 0, RADIUS2-1 do
	begin
	K:= RADIUS -J;
	Ang:= Acos(Float(K) / R);
	Yi:= Fix((Pi32 + 1.0 - Ang) *R);
	Roller(J):= Yi;
	end;
end;	\MakeRoller



proc	UnrollImage;		\Unroll image from screen
int	Y;
begin
for Y:= -(Height-1), RADIUS2 do
	begin
	DrawRoller(-Y);
	Move(0, RADIUS2-Y);   Line(Width-1, RADIUS2-Y, 0);
	Y:= Y +1;
	Move(0, RADIUS2-Y);   Line(Width-1, RADIUS2-Y, 0);
	Y:= Y +1;
	Move(0, RADIUS2-Y);   Line(Width-1, RADIUS2-Y, 0);
	Y:= Y +1;
	Move(0, RADIUS2-Y);   Line(Width-1, RADIUS2-Y, 0);
	WaitVB;
	end;
end;	\UnrollImage



proc	RollImage;		\Roll image onto screen
int	Y;
begin
for Y:= -RADIUS2, Height-1 do
	begin
	DrawRoller(Y);
	WaitVB;
	CopyLine(Y, Y, Raster, Depth);
	Y:= Y +1;
	CopyLine(Y, Y, Raster, Depth);
	Y:= Y +1;
	CopyLine(Y, Y, Raster, Depth);
	Y:= Y +1;
	CopyLine(Y, Y, Raster, Depth);
	end;
end;	\RollImage

\----------------------------------------------------------------------

proc	DOBODY;		\Read in image raster data, display it then exit
int	WID,		\Image WIDTH in bytes (and rounded up to next word)
	Y,		\Vertical scan line in pixels (0..HEIGHT-1)
	D,		\Bit plane depth (0..DEPTH-1)
	CKSIZE;
int	J,		\Index into RASTER
	X,		\Horizontal position (0..WIDTH-1)
	X1;


	proc	DOLINE;		\Read a compressed line for one bit plane
	int	N,		\Byte from input file
		M,		\Byte to replicate
		END;		\Last byte in RASTER for line (+1)
	begin
	END:= J +WID;
	repeat	N:= CHIN(3);
		if N <= $7F then
			begin			\Copy next N+1 bytes literally
			X:= J;   X1:= J +N +1;
			repeat	RASTER(D, X):= CHIN(3);
				X:= X +1;
			until X = X1;
			J:= X;
			end
		else if N = $80 then []		\Ignore
		else	begin			\N is in the range: $81..$FF
			\Replicate the next byte -EXTEND(N)+1 times
			M:= CHIN(3);		\Read next byte
			X:= J;   X1:= J -EXTEND(N) +1;
			repeat	RASTER(D, X):= M;
				X:= X +1;
			until X = X1;
			J:= X;
			end;
	until J = END;
	end;	\DOLINE


begin	\DOBODY
if ~HAVEBMHD then FATAL("^"BMHD^" MISSING");
SHOWSTATS;

CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;

WID:= WIDTH /8;		\WIDTH in bytes
for Y:= 0, HEIGHT-1 do
	begin
	for D:= 0, DEPTH-1 do
		begin
		J:= Y *WID;
		if COMPRESSION then
			DOLINE
		else	begin		\Read in a line of pixels
			X:= J;   X1:= J +WID;
			repeat	RASTER(D, X):= CHIN(3);
				X:= X +1;
			until X = X1;
			end;
		end;
	end;

Roller:= Reserve(RADIUS2*Intsize);
MakeRoller;
RollImage;

WAITKEY;
Openi(1);
UnrollImage;

SETVID(VIDMODE);
exit;
end;	\DOBODY

\======================================================================

proc	EATCHUNK;	\Read in and dispose of a chunk
int	CKSIZE, N, I;
begin
CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;
for N:= 0, CKSIZE-1 do
	I:= CHIN(3);
end;	\EATCHUNK

\----------------------------------------------------------------------

proc	INIT;		\Initialize
int	HANDIN;				\File handles
char	CMDTAIL;			\Copy of command tail
begin
VIDMODE:= GETVID;
HAVEBMHD:= false;

CMDTAIL:= RESERVE($80);			\Get copy of command tail
BLIT(CPUREG(11), $81, CPUREG(12), CMDTAIL, $7F);

HANDIN:= FOPEN(CMDTAIL, 0);		\Open first file name for input
FSET(HANDIN, ^I);
OPENI(3);
end;	\INIT

\----------------------------------------------------------------------

begin	\MAIN
CPUREG:= GETREG;

Screen:= RESERVE(INTSIZE);
Screen(0):= $A000;

RASTER:= RESERVE(4*INTSIZE);
for II:= 0, 3 do
	RASTER(II):= Malloc(80/16*480);

INIT;

SETVID($12);		\ASSUME 640x480x16
ALIGNPALETTE;

if HEXWORD#FO ! HEXWORD#RM then FATAL("FILE DID NOT START WITH ^"FORM^"");
DOFORM;

\WARNING: If BODY is missing then this will read beyond the end of the input file
loop	begin
	RECHDH:= HEXWORD;	\Read chunk header
	RECHDL:= HEXWORD;
	     if RECHDH=FO & RECHDL=RM then DOFORM
	else if RECHDH=BM & RECHDL=HD then DOBMHD
	else if RECHDH=CM & RECHDL=AP then DOCMAP
	else if RECHDH=BO & RECHDL=DY then DOBODY   \(Never returns)
	else EATCHUNK;		\Ignore any unrecognized chunks
	end;
end;	\MAIN
