	PAGE	55,120

CSEG	SEGMENT BYTE PUBLIC 'CODE'

	ASSUME	CS:CSEG,DS:DSEG,ES:CSEG,SS:CSEG

;ASSEMBLY LANGUAGE SUBROUTINE TO COUNT LINES IN THE DOCUMENT

	PUBLIC	LCOUNT

; LCOUNT(LSTART,LEND) = Count Carriage Returns between LSTART and LEND
; Return the number of Carriage Returns on the stack.

;DEFINE STACK FRAME LOCATIONS OF ARGUMENTS

LSTART	EQU	[BP+6]
LEND	EQU	[BP+4]

LCOUNT	PROC	FAR
	CLD			;WE WILL BE INCREMENTING
	MOV	BP,SP		;GET STACK FRAME
	MOV	DX,0		;DX HOLDS THE COUNT
	MOV	DI,LSTART	;POINT TO THE START OF THE BUFFER
	MOV	CX,LEND		;CALCULATE THE SIZE
	SUB	CX,DI
	MOV	AL,CR		;LOOKING FOR CARRIAGE RETURNS
	JCXZ	LCDONE		;EXIT IF ZERO LENGTH

LCONT1:
REPNE	SCASB
	JNE	LCDONE		;EXIT IF DONE
	INC	DX		;COUNT LINES
	JMP	LCONT1		;LOOP
LCDONE:	MOV	LSTART,DX	;PUT RESULT ON STACK
	RET	2
LCOUNT	ENDP
	PAGE

;ASSEMBLY LANGUAGE ROUTINE TO HANDLE SCREEN UPDATES FOR XED
;Information is passed to the subroutine via a data structure.
;A pointer to the data structure is passed on the stack.

	PUBLIC	SCREEN

;SCREEN(struc WINDOW) =  UPDATE WHOLE SCREEN

;STRUCTURE WHICH HOLDS WINDOW PARAMETER FOR SCREEN UPDATE

WINDOW	STRUC
SCRSEG	DW	?		;SEGMENT OF SCREEN
BSTART	DW	?		;START OF TEXT BUFFER
BEND	DW	?		;END OF THE TEXT BUFFER
PTR1	DW	?		;FIRST TEXT POINTER (CINX)
PTR2	DW	?		;SECOND TEXT POINTER (C2INX)
COLLUM	DW	?		;CURRENT COLLUMN OF CURSOR
LWIND	DW	?		;WINDOW LIMITS FOR SCREEN
RWIND	DW	?
SHWFLG	DW	?		;FLAGS SHOW ALL CONTROL CHARACTERS
SCREND	DW	?		;SCREEN SIZE
LINLEN	DW	?		;NUMBER CHARACTER PER LINE
LBEFOR	DW	?		;LINES BEFORE CURSOR
CURATT	DW	?		;CURSOR ATTRIBUTE
WINDOW	ENDS


SCREEN	PROC	FAR
	MOV	BP,SP		;GET STACK FRAME BASE
	MOV	BP,[BP+4]	;GET POINTER TO STRUCTURE
	PUSH	DI		;SAVE DI
	PUSH	ES		;SAVE ES

	CALL	BACKUP		;BACKUP TO START
	CALL	SETWND		;SETUP WINDOW
	CALL	DISBEF		;DISPLAY CHARACTERS BEFORE CURSOR
	CALL	DISAFT		;DISPLAY CHARACTERS AFTER CURSOR
	CALL	DISCUR		;DISPLAY CURSOR

	POP	ES		;RESTORE REGISTER AND EXIT
	POP	DI
	RET	2
SCREEN	ENDP
	PAGE

;BACKUP X LINES OR TO BEGINING OF TEXT
;ALSO LOCATE CURSOR POSITION

BACKUP:	STD			;WE WILL BE DECREMENTING
	MOV	LNGLIN,0	;NO LONG LINE YET
	MOV	DI,DS:[BP].PTR1	;GET PTR1
	MOV	FRSTCR,DI	;SET FIRST CR POSITION
	MOV	CX,DI
	SUB CX,DS:[BP].BSTART	;CALCULATE LENGTH
	MOV	BEFLEN,CX	;SAVE THIS VALUE FOR LATER
	JCXZ	BACKU4		;EXIT IF AT BEGINING

	MOV	BX,[BP].LBEFOR	;NUMBER OF LINES BEFORE THE CURSOR
	MOV	AL,CR		;LOOK FOR RETURNS
	DEC	DI		;PRE- DECREMENT

;BACKUP TO FIRST CR AND SAVE THE POSITION

REPNE	SCASB			;SCAN FOR RETURN
	JCXZ	BACKU5		;EXIT IF AT BEGINING
	MOV	FRSTCR,DI	;SAVE LOC OF FIRST CR
	ADD	FRSTCR,2	;POINT PAST CR
	JMP	BACKU2		;ENTER COMMON CODE
	PAGE

;BACKUP THE REST OF LINES

BACKU1:
REPNE	SCASB			;SCAN FOR RETURN
	JCXZ	BACKU3		;EXIT IF AT BEGINING

BACKU2:	DEC	BX		;COUNT LINES
	JNZ	BACKU1		;LOOP UNTIL ALL LINES DONE
	INC	DI		;POINT PAST CR
BACKU3:	INC	DI
BACKU4:	MOV	SI,DI		;USE AS FROM INDEX
	RET

;HANDLE TEXT BEFORE CURSOR WITH NO CARRIAGE RETURNS

BACKU5:	INC	DI		;POINT PAST START
	MOV	SI,DS:[BP].PTR1	;GET CURSOR POSITION
	SUB	SI,BLIMIT	;BACKUP TO LIMIT
	CMP	SI,DI		;BEFORE BEGINING OF TEXT?
	JA	BACKU6		;SKIP IF NOT
	MOV	SI,DI		;THEN USE DI
	MOV	FRSTCR,SI	;SET FIRST CR
	RET

BACKU6:	MOV	LNGLIN,0FFFFH	;FLAG THIS IS A LONG LINE
	MOV	FRSTCR,SI	;SET FIRST CR
	RET
	PAGE

;SET UP WINDOW FOR HORIZONAL SCROLLING

SETWND:	CLD			;SET INCREMENT MODE
	MOV	DX,0		;START WITH ZERO CURSOR POSITION
	MOV	DI,FRSTCR	;POINT TO FIRST CR
	MOV	CX,DS:[BP].PTR1	;CACULATE CHARS FROM CR TO CURSOR
	SUB	CX,DI
	JCXZ	SETWD3		;SKIP IF CURSOR AT START OF LINE

;DO WE HANDLE TABS OR NOT?

	CMP [BP].SHWFLG,0	;SHOW CONTROLS?
	JE	SETWD0		;NO, THEN HANDLE TABS
	MOV	DX,CX		;GET COLLUMN COUNT
	JMP	SETWD3		;ENTER COMMON CODE

;LOCATE COLLUMN COUNT TAKING INTO ACCOUNT TABS

SETWD0:	MOV	AL,TAB		;LOOK FOR TABS
SETWD1:	SCASB
	JNE	SETWD2		;SKIP NOT TAB
	OR	DX,7		;COMPENSATE FOR TAB
SETWD2:	INC	DX		;COUNT CHAR POSITIONS
	LOOP	SETWD1

;FIND COLLUMN NUMBER AND CURSOR FOR SHORT OR LONG LINES

SETWD3:	MOV	AX,DX		;GET COLLUMNS AFTER LAST CR
	CMP	LNGLIN,0	;LONG LINE?
	JE	SETWD4		;SKIP IF NOT
	MOV	AX,BEFLEN	;USE CHAR BEFORE CURSOR AS COLLUMN
SETWD4:	MOV DS:[BP].COLLUM,AX	;USE THIS AS COLLUMN
	CMP DX,DS:[BP].LINLEN	;CURSOR ON SCREEN?
	JAE	MOVWND		;SKIP IF NOT
	MOV DS:[BP].LWIND,0		;RESET WINDOW
	MOV	AX,DS:[BP].LINLEN
	MOV DS:[BP].RWIND,AX
	RET

;MOVE WINDOW SO THAT CURSOR IS ON SCREEN

MOVWND:	ADD	DX,RGARDE	;ADD IN GAURD BAND
	MOV DS:[BP].RWIND,DX
	SUB DX,DS:[BP].LINLEN	;CALCULATE LEFT WINDW
	MOV DS:[BP].LWIND,DX
	RET

	PAGE

;DISPLAY TEXT BEFORE THE CURSOR

DISBEF:	MOV	BX,0			;ZERO COLLUMN COUNTER
	MOV	AX,DS:[BP].SCRSEG	;POINT ES TO SCREEN
	MOV	ES,AX
	MOV	DI,SCRSTR	;POINT TO UPPER LEFT
	MOV	CX,DS:[BP].PTR1	;GET CURSOR POINTER
	SUB	CX,SI		;CALCULATE CHARS BEFORE CURSOR
	JCXZ	DISBFX		;EXIT IF AT START OF TEXT
	CLD			;SET INCREMENT MODE
	MOV	AH,07H		;SET SCREEN ATTRIBUTE
;	CALL	VDWAIT		;WAIT FOR VIDEO
DISBF1:	LODSB			;GET A CHARACTER FROM BUFFER
	CALL	CHOUT		;OUTPUT CHARACTER
	LOOP	DISBF1		;LOOP UNTIL DONE
DISBFX:	MOV	CURADD,DI	;SAVE THE ADDRESS OF THE CURSOR
	RET

;DISPLAY CHARACTERS AFTER THE CURSOR

DISAFT:	CLD			;SET INCREMENT MODE
	MOV	SI,DS:[BP].PTR2	;POINT TO CURSOR
	MOV	CX,DS:[BP].BEND	;CALCULATE NUMBER AFTER
	SUB	CX,SI
	JCXZ	DISAF2		;SKIP IF NONE
;	CALL	VDWAIT		;WAIT FOR VIDEO
	MOV	AH,07H		;SET SCREEN ATTRIBUTE
DISAF1:	LODSB			;GET A CHARACTER FROM BUFFER
	CALL	CHOUT		;OUTPUT CHARACTER
	CMP DI,DS:[BP].SCREND	;AT END OF SCREEN?
	JAE	DISAFX		;SKIP IF SO
	LOOP	DISAF1		;LOOP UNTIL DONE

DISAF2:	MOV	AL,SPACE	;BLANK REST OF SCREEN
	MOV	AH,07H		;SET ATTRIBUTE
	MOV CX,DS:[BP].SCREND	;CALCULATE NUMBER OF BLANKS
	SUB	CX,DI
	SHR	CX,1
	JBE	DISAFX		;EXIT IF DONE
REP	STOSW			;NOW BLANK IT
DISAFX:	RET

;DISPLAY CUROR

DISCUR:	MOV AH,BYTE PTR DS:[BP].CURATT	;SET CURSOR ATTRIBUTE
	MOV	DI,CURADD	;GET THE CURSOR ADDRESS
	MOV	AL,ES:[DI]	;GET THE CHARACTER
	MOV	ES:[DI],AX	;NOW WRITE WITH NEW ATTRIBUTE
	RET
	PAGE

;ROUTINE TO HANDLE GENERAL CHARACTER OUTPUT

CHOUT:	CMP	AL,SPACE	;CONTROL CHARS?
	JA	LOWOUT		;SKIP IF NOT
	JE	SPAOUT		;HANDLE SPACES
	CMP	AL,CR		;TEST FOR RETURN
	JE	CROUT		;THEN BLANK LINE AND EXIT
	CMP	AL,TAB		;IS IT A TAB
	JE	TABOUT		;THEN HANDLE IT
CHOUT1:	PUSH	AX		;SAVE CHAR AND ATTRIBUTE
	MOV	AH,070H		;INVERT CHARACTER
	OR	AL,040H		;MAKE ALPHABETIC
	CALL	LOWOUT		;DISPLAY IT
	POP	AX		;RESTORE ATTRIBUTE
	RET


;DISPLAY SPACE CHARACTER

SPAOUT:	CMP [BP].SHWFLG,0	;SHOW CONTROLS?
	JE	LOWOUT		;NO, THEN ENTER
	MOV	AL,'_'		;CONVERT TO DASH
	JMP	LOWOUT		;AND OUTPUT IT

      	
;HANDLE TAB CHARACTER

TABOUT:	CMP [BP].SHWFLG,0	;SHOW CONTROLS?
	JNE	CHOUT1		;YES, THEN HANDLE AS CONTROL
TABHND:	MOV	AL,SPACE	;GET A SPACE

TABHD1:	CALL	LOWOUT		;DISPLAY SPACE
	TEST	BX,07H		;TEST MODULO 8
	JNZ	TABHD1		;SKIP IF DONE
	RET
	PAGE

;HANDLE CARRIAGE RETURNS

CROUT:	CMP [BP].SHWFLG,0	;SHOW CONTROLS?
	JE	BLINE		;NO, THEN DO NORMAL CR

	CMP BX, DS:[BP].LWIND	;OFF SCREEN TO THE LEFT?
	JB	CROUT1		;EXIT IF IT IS
	CMP BX, DS:[BP].RWIND	;OFF SCREEN TO THE RIGHT?
	JAE	CROUT1		;EXIT IF IT IS

	PUSH	AX		;SAVE ATTRIBUTE AND CHARACTER
	MOV	AX,0704DH	;DISPLAY INVERTED M
	STOSW			;PUT ON SCREEN
	POP	AX
CROUT1:	INC	BX		;COUNT CHARACTERS

;BLANK TO THE END OF A  LINE

BLINE:	XCHG	CX,DX		;SAVE TOTAL COUNT
	CMP BX, DS:[BP].LWIND	;OFF LEFT SIDE?
	JAE	BLIN2		;SKIP IF NOT
	MOV CX,DS:[BP].LINLEN	;THEN BLANK WHOLE LINE
	JMP	BLIN3		;ENTER COMMON CODE

BLIN2:	MOV CX, DS:[BP].RWIND	;GET RIGHT WINDOW
	SUB	CX,BX		;CALCULATE SPACE TO RIGHT EDGE
	JB	BLIN1		;SKIP IF PAST RIGHT EDGE
BLIN3:	MOV	AL,SPACE	;GET A SPACE
REP	STOSW			;STORE IT

BLIN1:	MOV	BX,0		;RESTART AT LEFT SIDE
	XCHG	CX,DX		;RESTORE CX
	RET
	PAGE

;LOWEST LEVEL OUTPUT ROUTINE

LOWOUT:	CMP BX, DS:[BP].LWIND	;OFF SCREEN TO THE LEFT?
	JB	LOWOT1		;EXIT IF IT IS
	CMP BX, DS:[BP].RWIND	;OFF SCREEN TO THE RIGHT?
	JAE	LOWOT2		;EXIT IF IT IS
	STOSW			;DISPLAY CHAR
LOWOT1:	INC	BX		;ADVANCE TO NEXT COLLUMN
	RET

;SCAN AHEAD IF OFF SCREEN RIGHT FOR SPEED

LOWOT2:	PUSH	ES		;SAVE SCREEN SEGMENT
	MOV	DX,DS		;POINT TO DATA SEGMENT
	MOV	ES,DX
	XCHG	DI,SI		;PUT INDEX IN CORRECT REGISTER
	MOV	AL,CR		;SCAN FOR RETURN
REPNE	SCASB			;SCAN
	JCXZ	LOWOT3		;SKIP IF AT THE END
	JMP	LOWOT4

LOWOT3:	INC	CX		;FIX CX SO CALLING LOOP WILL TERMINATE
LOWOT4:	MOV	BX,0		;ZERO COLLUMN COUNTER
	XCHG	DI,SI		;RESTORE REGISTERS
	POP	ES		;RESTORE SCREEN SEGMENT
	RET

	PAGE

;ROUTINE TO WAIT FOR VIDEO REFRESH

VDWAIT:	MOV	DX,VRTREF	;POINT TO VIDEO PORT
VDWAT1:	IN	AL,DX		;WAIT FOR VERTICAL REFRESH
	TEST	AL,1
	JZ	VDWAT1
	RET

;SOME CONSTANTS

SCRSTR	EQU	0		;STARTING MEMORY LOCATION OF SCREEN
RGARDE	EQU	4		;GUARD BAND FOR RIGHT SIDE
BLIMIT	EQU	500		;BACKUP LIMIT

VRTREF	EQU	03DAH		;REGISTER FOR VERTICAL REFRESH

CR	EQU	0DH
SPACE	EQU	20H
TAB	EQU	09H

DSEG	SEGMENT BYTE PUBLIC 'DATA'
	public	BEFLEN,FRSTCR
;SOME VARIABLES

BEFLEN	DW	0		;NUMBER CHARS BEFORE CURSOR
LNGLIN	DW	0		;FLAG LONG LINE BEFORE CURSOR
CURADD	DW	0		;SCREEN ADDRESS OF CURSOR
FRSTCR	DW	0		;ADDRESS OF FIRST CR BACK FROM CURSOR
DSEG	ENDS

CSEG	ENDS
	END
