\SCROLBAR.XPL	DEC-13-97
\Soft buttons and scroll bars

inc	\CXPL\CODESI;

def	ROMSize=$800;	\Size of instruction memory

def	ExitX=0, ExitY=0,		\Exit button: []
	StackX=17, StackY=1,
	ListWinX=1, ListWinY=14;	\Code listing window

int	Cpureg,
	CursorShape,
	HaveMouse,	\A mouse driver is installed and the mouse works
	;
int	Stack(8),	\Subroutine return address stack
	II,
	KnobX, KnobY,	\Knob position on scroll bar (character coordinates)
	WinHeight,	\Height of the listing window in lines
	ROMSizeEven,	\ROMSize rounded down to even multiple of ScrollBarHeight
	ScreenHeight,	\Number of lines of text on the screen
	ScreenWidth,	\Number of character columms on the screen
	ScrollBarHeight,\Characters in scroll bar (not including arrows)
	WinPtr;		\ROM address where listing window starts (can be < 0)

def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0A, FF=$0C,	\Control chars
	CR=$0D, Up=$18, Dn=$19, EOF=$1A, Esc=$1B, Space=$20, Ctrl=$40;
def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D,	\Key scan codes
	PageUp=$49, PageDn=$51, Home=$47, End=$4F, Func=$3A,
	AltFunc=$67, AltX=$2D;
def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,	\Attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \ EGA palette
def	Flashing=$80;	\(Windows doesn't do flashing)

\===============================================================================

func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES);\Call software interrupt (BIOS)
int	Int, AX, BX, CX, DX, BP, DS, ES; \(Unused arguments need not be passed)
begin
Cpureg:= Getreg;
Cpureg(0):= AX;
Cpureg(1):= BX;
Cpureg(2):= CX;
Cpureg(3):= DX;
Cpureg(6):= BP;
Cpureg(9):= DS;
Cpureg(11):= ES;
Softint(Int);
return Cpureg(0);		\Return AX register
end;	\CallInt



proc	SetBkgndColor(X, Y, C);	\Set background color for text
int	X, Y, C;	\Text coordinates (X,Y) and background color (C)
int	AC;		\Attribute (high byte) and character (low byte)
begin
\Set cursor position
CallInt($10, $0200, $0000, 0, Y<<8!X);

\Get attribute and character at cursor
AC:= CallInt($10, $0800, $0000);

\Change background color and write attribute back at cursor
CallInt($10,   $0900 ! AC&$00FF,   AC>>8 & $0F ! C<<4,   1);
end;	\SetBkgndColor



proc	Backlight(X0, Y0, X1, Y1, C);  \"Hilight" but don't change foregnd color
int	X0, Y0, X1, Y1, C;
int	X, Y;
begin
if X0 > X1 then
	[X:= X0;   X0:= X1;   X1:= X];
if Y0 > Y1 then
	[Y:= Y0;   Y0:= Y1;   Y1:= Y];

for Y:= Y0, Y1 do
    for X:= X0, X1 do
	SetBkgndColor(X, Y, C);
end;	\Backlight

\============================== MOUSE ROUTINES =================================

func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\Pointer's set to center of screen but is hidden
CallInt($21, $3533);		\Make sure mouse vector $33 points to something
HaveMouse:= false;
if Cpureg(1)=0 & Cpureg(11)=0 then return false;
HaveMouse:= CallInt($33, $0000); \Reset mouse and get status
if HaveMouse then HaveMouse:= true; \(Beware of 'not' operator in 32-bit XPL)
return HaveMouse;		\Return 'false' if failure
end;	\OpenMouse



proc	ShowMouse(On);		\Turn mouse pointer on or off
\The video mode should be set before calling this routine.
\This counts the number of times the pointer is turned on or off. If the
\ pointer is turned off twice, it must be turned on twice before it
\ actually goes on. The pointer should be turned off before drawing over
\ it and before your program exits. Setting the video mode will also turn
\ off the pointer.
int	On;	\Flag: True = pointer on; False = pointer off
if HaveMouse then
	CallInt($33, if On then $0001 else $0002);



func	GetMousePosition(N);	\Return position of specified mouse coordinate
int	N;	\0 = X coordinate; 1 = Y coordinate
\For video modes $0-$E and $13 the maximum coordinates are 639x199, minus
\ the size of the pointer. For modes $F-$12 the coordinates are the same as
\ the pixels. For 80-column text modes divide the mouse coordinates by 8 to
\ get the character cursor position.
begin
if ~HaveMouse then return 0;
CallInt($33, $0003);
return if N then Cpureg(3) else Cpureg(2);
end;	\GetMousePosition



func	GetMouseButton(N);	\Return 'true' if specified mouse button is down
int	N;	\Button number: 0 = left; 1 = right (or middle)
begin
if ~HaveMouse then return false;
CallInt($33, $0003);
return if N then (Cpureg(1)&2)=2 else (Cpureg(1)&1)=1;
end;	\GetMouseButton



proc	MoveMouse(X, Y);	\Move mouse pointer to X,Y
int	X, Y;
if HaveMouse then
	CallInt($33, $0004, 0, X, Y);

\===============================================================================

proc	ShowScrollBar(X0, Y0, X1, Y1, Atb, RelPosn);
int	X0, Y0, X1, Y1, Atb;
real	RelPosn;	\Relative position [0.0..1.0]
\Outputs: KnobX, KnobY.  ??
\Also do horizontal case
def	Knob=$FE, Bar=$DB, Up=$1E, Dn=$1F, Lt=$11, Rt=$10;	\Ascii chars
int	X, Y;
begin
if ~HaveMouse then return;

if X0 > X1 then
	[X:= X0;   X0:= X1;   X1:= X;   RelPosn:= 1.0 - RelPosn];
if Y0 > Y1 then
	[Y:= Y0;   Y0:= Y1;   Y1:= Y;   RelPosn:= 1.0 - RelPosn];

Cursor(X0, Y0);
Attrib(Atb);
if X0 = X1 then
	begin				\Vertical bar
	KnobX:= X0;
	Y:= Fix( RelPosn * Float((Y1-1) - (Y0-1)) );
	KnobY:= Y0+1 + Y;

	ShowMouse(false);
	Chout(6, Up);			\Up arrow
	Cursor(X1, Y1);
	Chout(6, Dn);			\Down arrow
	for Y:= Y0+1, Y1-1 do
		begin
		Cursor(X0, Y);
		Chout(6, if Y = KnobY then Knob else Bar);
		end;
	end
else	begin				\Horizontal bar
	KnobY:= Y0;
	X:= Fix( RelPosn * Float((X1-1) - (X0-1)) );
	KnobX:= X0+1 + X;

	ShowMouse(false);
	Chout(6, Lt);			\Left arrow
	Cursor(X1, Y1);
	Chout(6, Rt);			\Right arrow
	for X:= X0+1, X1-1 do
		begin
		Cursor(X, Y0);
		Chout(6, if X = KnobX then Knob else Bar);
		end;
	end;

ShowMouse(true);
end;	\ShowScrollBar



func	VerifyExit;	\Show message and return true if verified
begin
ShowMouse(false);
Attrib(Blue<<4 ! Yellow);
SetWind((ScreenWidth-15)>>1, 5, (ScreenWidth+15)>>1-1, 5+2, 0, true);
Text(6, "
  Exit (Y/N)?");

SetWind(0, 0, ScreenWidth-1, ScreenHeight-1, $0102, false);	\Turn off scroll
ShowMouse(true);

Openi(0);
repeat until Chkkey;				\Wait for keystroke
if (GetKey ! $20) = ^y then return true;	\(don't echo, especially Ctrl-C)

DisplayRegs;
ShowCode(LineCursor);	\(outputs PCLine)
PointOut;		\(inputs PCLine)
return false;
end;	\VerifyExit



proc	Exit;		\Make a clean exit
begin
ShowMouse(false);
Chout(0, FF);
CallInt($10, $0100, 0, CursorShape);	\Restore flashing cursor
exit;
end;	\Exit



func	ButtonTrigger(ID);	\Returns 'true' when autorepeat button triggers
int	ID;	\Button identifier (don't change buttons in middle of a repeat)
int	Timer, SID;
begin
Timer:= [0];	\Own (static) variables
SID:= [0];	\Save initial button ID (usually is X & Y screen coordinates)

if ButtonState = 0 then SID(0):= ID;	\First-time call for this button

if ID = SID(0) then		\If it's the button we started with...
	begin
	case ButtonState of
	  0:	begin
		Timer(0):= 9;	\About 1/2 second delay before autorepeat starts
		ButtonState:= 1;
		return true;	\Trigger on initial press
		end;
	  1:	begin		\Subsequent calls come here and check timer
		Sound(0, 1, 1);	\Delay 1/18 second
		Timer(0):= Timer(0) - 1;
		if Timer(0) <= 0 then
			begin
			ButtonState:= 2;
			return true;	\Trigger button when timer times out
			end;
		end
	  other	begin
		Timer(0):= 2;	\About 0.1 second delay for autorepeat rate
		ButtonState:= 1;
		end
	end;
return false;
end;	\ButtonTrigger

\------------------------------- TEST ROUTINE ----------------------------------

proc	ShowCode(Loc);
int	Loc;
real	RelPosn;
begin
\Display some code in a window--just show number for demo
Cursor(10, 40);
Hexout(6, Loc);

\Display scroll bar with knob position ratioed to Loc
\KnobYOffset/ScrollBarHeight = Loc/ROMSizeEven
RelPosn:= Float(Loc) / Float(ROMSizeEven);
ShowScrollBar(79, 5, 79, 5+ScrollBarHeight, Blue<<4!White, RelPosn);
end;	\ShowCode



proc	UpdateDisplay;
int	I;
begin
for I:= 0, 7 do
	begin
	Cursor(StackX, StackY+I);
	Hexout(6, Stack(I));
	end;

Cursor();
Hexout();

Cursor();
Hexout();
end;	\UpdateDisplay



proc	DoCmd;
int	I, T, B, X, Y,
	Inc, But0, But1,
	YRel, OldWinPtr,
	ButtonState;	\Used by mouse button autorepeat
begin
ButtonState:= 0;
loop begin
But0:= GetMouseButton(0);		\Get copies of mouse buttons for
But1:= GetMouseButton(1);		\ consistent readings
if But0 ! But1 then			\A button is pressed
	begin
	X:= GetMousePosition(0) >> 3;	\Get mouse position in text coordinates
	Y:= GetMousePosition(1) >> 3;

	case of				\See if a soft button is pressed
	X>=StackX & X<=StackX+3 & Y>=StackY & Y<=StackY+7:		\STACK
	    if ButtonTrigger(Y<<8!X) then
		begin
		Inc:= 1 << ((3+StackX-X)<<2);	\$1000, $100, $10 or $1
		if But1 then Inc:= -Inc;
		Stack(Y-StackY):= Stack(Y-StackY) + Inc;
		UpdateDisplay;
		end;

	X>=ExitX & X<=ExitX+2 & Y=ExitY:				\[]
	    if ButtonTrigger(Y<<8!X) then
		begin
		repeat	begin
			B:= GetMousePosition(1)>>3 = ExitY &
			    GetMousePosition(0)>>3 >= ExitX &
			    GetMousePosition(0)>>3 <= ExitX+2;
			ShowMouse(false);
			Backlight(ExitX, ExitY, ExitX+2, ExitY,
				if B then Cyan else White);
			ShowMouse(true);
			Sound(0, 1, 1);	\Delay to reduce flicker
			end;
		until ~GetMouseButton(0) & ~GetMouseButton(1);
		if B then Exit;
		end;

	\Scroll bar:
	X=KnobX & Y=ListWinY:					\Up arrow
	    if ButtonTrigger(Y<<8!X) then
		begin
		WinPtr:= WinPtr - 1;
		if WinPtr < -1 then WinPtr:= -1;
		ShowCode(WinPtr+2);
		UpdateDisplay;
		end;
	X=KnobX & Y=ScreenHeight-2:				\Down arrow
	    if ButtonTrigger(Y<<8!X) then
		begin
		WinPtr:= WinPtr + 1;
		if WinPtr >= ROMSize then WinPtr:= ROMSize;
		ShowCode(WinPtr+2);
		UpdateDisplay;
		end;
	X=KnobX & Y>=ListWinY+1 & Y<KnobY:			\Up page
	    if ButtonTrigger(Y<<8!X) then
		begin
		WinPtr:= WinPtr - WinHeight;
		if WinPtr < -1 then WinPtr:= -1;
		ShowCode(WinPtr+2);
		UpdateDisplay;
		end;
	X=KnobX & Y>KnobY & Y<=ScreenHeight-3:			\Down page
	    if ButtonTrigger(Y<<8!X) then
		begin
		WinPtr:= WinPtr + WinHeight;
		if WinPtr >= ROMSize then WinPtr:= ROMSize;
		ShowCode(WinPtr+2);
		UpdateDisplay;
		end;
	X=KnobX & Y=KnobY:					\Knob
	    if ButtonTrigger(Y<<8!X) then
		begin
		OldWinPtr:= -2;		\(used to reduce flicker)
		loop	begin
			\WP / ROMSizeEven = YRel / ScrollBarHeight
			YRel:= Y - (ListWinY+1);
			WinPtr:= Fix( Float(YRel) * Float(ROMSizeEven) /
				Float(ScrollBarHeight-1) ) - 1;
			if WinPtr # OldWinPtr then	\Only redraw code window
				begin			\ if it changes to reduce
				ShowCode(WinPtr+2);	\ flicker
				OldWinPtr:= WinPtr;
				end;
			if ~GetMouseButton(0) & ~GetMouseButton(1) then quit;
			Sound(0, 1, 1);		\Delay to reduce cursor flicker
			Y:= GetMousePosition(1) >> 3;
			if Y < ListWinY+1 then Y:= ListWinY+1;
			if Y > ScreenHeight-3 then Y:= ScreenHeight-3;
			end;
		end;

\F1=Help F2=Bkpt F3=Clear F4=Here F5=Time F6=Go F7=Step F8=Over F9=Run F10=MClr
\                         F4=Jump       F6=Slow F7=Back F8=Return     F10=Reset
	Y = ScreenHeight-1:
		begin
		\X coordinates of "Fn" (See ShowPrompts)
		T:= if AltKey then
			\F0   1   2   3   4   5   6   7   8   9  10  11
			[ 0,  1,  9, 17, 26, 34, 40, 48, 56, 66, 70, 80]
		else	[ 0,  1,  9, 17, 26, 34, 42, 48, 56, 64, 71, 80];
		for I:= 1, 10 do
		    begin
		    if X >= T(I)  &  X <= T(I+1)-2 then
			if ButtonTrigger(I+^F) then
				begin
				repeat	begin
					B:= GetMousePosition(1)>>3 = Y &
					     GetMousePosition(0)>>3 >= T(I) &
					     GetMousePosition(0)>>3 <= T(I+1)-2;
					ShowMouse(false);
					Backlight(T(I), Y, T(I+1)-2, Y, 
					  if B then Cyan else White);
					ShowMouse(true);
					Sound(0, 1, 1);	\Delay to reduce flicker
					end;
				until ~GetMouseButton(0) & ~GetMouseButton(1);

				ShowMouse(false);	\Unbacklight button
				Backlight(T(I), Y, T(I+1)-2, Y, White);
				ShowMouse(true);

				\If still on original button then fake keystroke
				if B then CallInt($16, $0500, 0,
				 ((if AltKey then AltFunc else Func) + I) << 8);
				end;
		    end;
		ButtonTrigger(Y<<8!X);	\Don't trigger soft button if cursor
		end		\ slides onto it while mouse button is down
	other	[];
	end;
end;	\loop
end;	\DoCmd

\-------------------------------------------------------------------------------

begin	\Main
for II:= 0, 7 do Stack(II):= 0;
ScrollBarHeight:= 15;
ROMSizeEven:= ROMSize / ScrollBarHeight * ScrollBarHeight;
DoCmd;
end;	\Main
