\SHOW12.XPL	OCT-24-93

\This displays 640x480x16 .LBM files. (InterLeaved BitMap image in Interchange
\ File Format). Reference: ROM Kernel Reference Manual, IFF, p I-28.

\REVISION HISTORY:
\DEC-03-89, Original on Amiga
\OCT-24-93, Converted to run on PC.

inc	C:\CXPL\CODESI;		\Include code definitions for intrinsic routines

def	INTSIZE= 2;	\Number of bytes in an integer
def	TV=0, KB=0, PR=2;					\Device numbers
def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0a, FF=$0c,	\Control chars
	CR=$0d, EOF=$1a, Esc=$1b, Sp=$20, Ctrl=$40;

int	Cpureg;			\Address of CPU register array (from Getreg)
def	AX, BX, CX, DX, DI, SI, BP, CF, CS, DS, SS, ES;		\Getreg registers

def	FO=$464F, RM=$524D, IL=$494C, BM=$424D, HD=$4844, CM=$434D, AP=$4150,
	BO=$424F, DY=$4459;

int	HAVEBMHD,	\Flag: Bit map header chunk has been read in
	RECHDH,		\Record header
	RECHDL,		\Record header, 4-character name
	WIDTH, HEIGHT, DEPTH,	\Image dimensions (pixels)
	COMPRESSION,	\Compression algorithm (0= none, 1=ByteRun1)
	VIDMODE,	\Initial video mode
	;
seg addr RASTER;	\Display screen

\======================================================================

proc	WaitKey;		\Wait for a keystroke
begin
Openi(1);
repeat until ChkKey;
end;	\WaitKey



func	GetVid;			\Returns the current video mode
begin
Cpureg(AX):= $0f00;		\Function $0f
Softint($10);			\Call BIOS routine
return Cpureg(AX) & $ff;
end;	\GetVid



proc	SetColReg(N, R, G, B);	\Set VGA color register (DAC) N
int	N,	 \Register to set (0..255)
	R, G, B; \Red, Green, and Blue (only low 6 bits are used)
begin		 \ 2**6 = 64;  64 *64 *64 = 262144 possible colors
Cpureg(AX):= $1010;		\Function $10, subfunction $10
Cpureg(BX):= N;			\Select color register
Cpureg(CX):= G<<8 ! B&$ff;	\Green<<8 blue
Cpureg(DX):= R<<8;		\Red<<8
Softint($10);			\Call BIOS routine
end;	\SetColReg



proc	SetPalReg(N, C);	\Set palette register
int	N,	\Register to set (0..15)
	C;	\Color for EGA bits: 0 0 r g b R G B  (64 colors)
begin		\                    0 0 0 I 0 R G B  (16 colors)
Cpureg(AX):= $1000;
Cpureg(BX):= C<<8 ! N&$ff;
Softint($10);			\Call BIOS
end;	\SetPalReg



proc	AlignPalette;		\Make palette regs correspond to VGA regs (DAC)
int	N;
begin
for N:= 0, 15 do
	begin
	Cpureg(AX):= $1000;	\Function $10, subfunction $00
	Cpureg(BX):= N<<8 !N;	\DAC and palette registers
	Softint($10);		\Call BIOS routine
	end;
end;	\AlignPalette



proc	SetBorder(C);		\Set border color register
int	C;	\Color for EGA bits: 0 0 r g b R G B  (64 colors)
begin		\                    0 0 0 I 0 R G B  (16 colors)
Cpureg(AX):= $1001;
Cpureg(BX):= C<<8;
Softint($10);
end;	\SetBorder

\======================================================================

proc	FATAL(STR);	\Display error message and abort
addr	STR;
begin
OPENO(TV);		\Back to text mode
TEXT(TV,"
NOPE - ");
CHOUT(TV, BEL);
TEXT(TV, STR);
CRLF(TV);

WaitKey;
SETVID(VIDMODE);
exit;
end;	\FATAL



proc	SHOWSTATS;
begin
INTOUT(TV, WIDTH);
TEXT(TV, "x");
INTOUT(TV, HEIGHT);
TEXT(TV, "x");
INTOUT(TV, DEPTH);
CRLF(TV);
TEXT(TV, "
Press ^"Enter^" to display it.

After picture is displayed, press ^"Enter^" to return to DOS.
");
WAITKEY;
end;	\SHOWSTATS



func	HEXWORD;	\Read in 2 bytes and return the value of the word
return SWAP(CHIN(3)) + CHIN(3);

\----------------------------------------------------------------------

proc	DOFORM;		\Read in form type (must be ILBM)
int	CKSIZE;
begin
CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;
if HEXWORD#IL ! HEXWORD#BM then FATAL("^"ILBM^" MISSING");
end;	\DOFORM

\----------------------------------------------------------------------

proc	DOBMHD;		\Read in BitMap Header chunk
int	CKSIZE,
	XOFFSET, YOFFSET,
	MASKING,	\These parameters are read in but are not used
	PAD1,
	TRANSPARENT,
	XASPECT, YASPECT,
	PAGEWIDTH, PAGEHEIGHT;
begin
CKSIZE:= HEXWORD;	\Chunk size in bytes (must be even)
CKSIZE:= HEXWORD;

WIDTH:= HEXWORD;	\Raster width and height in pixels
HEIGHT:= HEXWORD;
XOFFSET:= HEXWORD;	\Offset (in pixels) for this image
YOFFSET:= HEXWORD;	\ (not used)
DEPTH:= CHIN(3);	\Number of bit planes (can be 0)
MASKING:= CHIN(3);	\Masking technique (0=no mask, is assumed)
COMPRESSION:= CHIN(3);	\Compression algorithm (0=none, 1=byte run)
PAD1:= CHIN(3);		\Unused by IFF standard
TRANSPARENT:= HEXWORD;	\Transparent color (only used w masking)
XASPECT:= CHIN(3);	\Aspect ratio (1:1 is assumed here)
YASPECT:= CHIN(3);
PAGEWIDTH:= HEXWORD;	\Source page size (not used here)
PAGEHEIGHT:= HEXWORD;

HAVEBMHD:= true;	\Indicate that BMHD chunk has been read in
end;	\DOBMHD

\----------------------------------------------------------------------

proc	DOCMAP;		\Color Map, set up color registers
int	CKSIZE,		\Number of bytes in CMAP field
	COLREGS,	\Number of color registers to use
	N,
	RED, GREEN, BLUE;
begin
CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;
COLREGS:= CKSIZE /3;
if COLREGS >256 then COLREGS:= 256;	\Limit maximum number of color registers

for N:= 0, COLREGS-1 do
	begin
	RED:= CHIN(3)>>2;   GREEN:= CHIN(3)>>2;   BLUE:= CHIN(3)>>2;
	SetColReg(N, RED, GREEN, BLUE);
	end;

\Eat the remainder. Must end on an even-byte boundary.
for N:= CKSIZE/3 *3, CKSIZE-1 do
	RED:= CHIN(3);
end;	\DOCMAP

\----------------------------------------------------------------------

proc	DOBODY;		\Image raster data
int	WID,		\Image WIDTH in bytes (and rounded up to next word)
	Y,		\Vertical scan line in pixels (0..HEIGHT-1)
	D,		\Bit plane depth (0..DEPTH-1)
	CKSIZE;
int	J,		\Index into RASTER
	X,		\Horizontal position (0..WIDTH-1)
	X1;



proc	DOLINE;		\Read a single line for one bit plane (compressed data)
int	N,		\Byte from input file
	M,		\Byte to replicate
	END;		\Last byte in RASTER for line (+1)
begin
END:= J +WID;
repeat	N:= CHIN(3);
	if N <= $7F then
		begin			\Copy the next N+1 bytes literally
		X:= J;   X1:= J +N +1;
		repeat	RASTER(0, X):= CHIN(3);
			X:= X +1;
		until X = X1;
		J:= X;
		end
	else if N = $80 then []		\Ignore
	else	begin			\N is in the range: $81..$FF
		\Replicate the next byte -EXTEND(N)+1 times
		M:= CHIN(3);		\Read next byte
		X:= J;   X1:= J -EXTEND(N) +1;
		repeat	RASTER(0, X):= M;
			X:= X +1;
		until X = X1;
		J:= X;
		end;
until J = END;
end;	\DOLINE



begin	\DOBODY
if ~HAVEBMHD then FATAL("^"BMHD^" MISSING");
\SHOWSTATS;

CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;

WID:= WIDTH /8;		\WIDTH in bytes
for Y:= 0, HEIGHT-1 do
	begin
	for D:= 0, DEPTH-1 do
		begin
		POUT(2, $3C4, 0);	\Enable appropriate plane
		POUT(1<<D, $3C5, 0);
		J:= Y *WID;
		if COMPRESSION then
			DOLINE
		else	begin		\Read in a line of pixels
			X:= J;   X1:= J +WID;
			repeat	RASTER(0, X):= CHIN(3);
				X:= X +1;
			until X = X1;
			end;
		end;
	end;
WAITKEY;
SETVID(VIDMODE);
exit;
end;	\DOBODY

\----------------------------------------------------------------------

proc	EATCHUNK;	\Read in and dispose of a chunk
int	CKSIZE, N, I;
begin
CKSIZE:= HEXWORD;
CKSIZE:= HEXWORD;
for N:= 0, CKSIZE-1 do
	I:= CHIN(3);
end;	\EATCHUNK

\----------------------------------------------------------------------

proc	INIT;		\Initialize
int	HANDIN;				\File handles
char	CMDTAIL;			\Copy of command tail
begin
VIDMODE:= GETVID;
HAVEBMHD:= false;

CMDTAIL:= RESERVE($80);			\Get copy of command tail
BLIT(CPUREG(11), $81, CPUREG(12), CMDTAIL, $7F);

HANDIN:= FOPEN(CMDTAIL, 0);		\Open first file name for input
FSET(HANDIN, ^I);
OPENI(3);
end;	\INIT

\----------------------------------------------------------------------

begin	\MAIN
CPUREG:= GETREG;

RASTER:= RESERVE(INTSIZE);
RASTER(0):= $A000;

INIT;

SETVID($12);		\ASSUME 640x480x16
ALIGNPALETTE;

if HEXWORD#FO ! HEXWORD#RM then FATAL("FILE DID NOT START WITH ^"FORM^"");
DOFORM;

\WARNING: If BODY is missing then this will read beyond the end of the input file
loop	begin
	RECHDH:= HEXWORD;	\Read chunk header
	RECHDL:= HEXWORD;
	     if RECHDH=FO & RECHDL=RM then DOFORM
	else if RECHDH=BM & RECHDL=HD then DOBMHD
	else if RECHDH=CM & RECHDL=AP then DOCMAP
	else if RECHDH=BO & RECHDL=DY then DOBODY   \(Never returns)
	else EATCHUNK;		\Ignore any unrecognized chunks
	end;
end;	\MAIN
