\SHOWILBM.XPL	OCT-24-93
\This program displays pictures. It reads in an ILBM IFF (InterLeaved
\ BitMap image in Interchange File Format).
\ Reference: ROM Kernel Reference Manual, IFF, p I-28.

\REVISION HISTORY:
\JUL-22-88, Original
\SEP-23-88, Added CAMG chunk, which is ignored.
\SEP-29-89, Reads binary (rather than ascii hex).
\NOV-27-89, Display HAM and other images of almost any dimensions
\DEC-03-89, Add CRNG chunk to cycle colors
\OCT-24-93, Converted to run on PC.

code	ABS=0,		RAN=1,		REM=2,		RESERVE=3,
	SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
	CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
	TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
	CHKKEY=33;

	BITMAP=108,	VIEW=110,	PEEK_W=113,
	POKE_W=114,	DEVINFO=122,	GETKEY=125,	

def	TV= 0, KB= 0;
def	BEL= $07, FF= $0C, CR= $0D, EOF= $1A, SP= $20, CTRL=$40;
def	INTSIZE= 2;	\Number of bytes in an integer


def	FORM= $464F524D, ILBM= $494C424D, BMHD= $424D4844, CMAP= $434D4150,
	BODY= $424F4459, CAMG= $43414D47, CRNG= $43524E47;

def	V_HIRES= $8000, V_HAM= $0800, V_LACE= $0004;	\View mode bits

def	CRNGS=10;				\Maximum number of CRNG chunks
def	ACTIVEBIT= $0001, REVBIT= $0002;	\Bits in CYCFLAGS

int	HAVEBMHD,	\Flag: Bit map header chunk has been read in
	RECHD,		\Record header, 4-character name
	WIDTH, HEIGHT, DEPTH,	\Image dimensions (pixels)
	VIEWMODES,	\View mode flag bits (HIRES, HAM, LACE)
	COMPRESSION,	\Compression algorithm (0= none, 1=ByteRun1)
	CYCINX,		\Index into CYCRATE, CYCFLAGS, CYCLOW, and CYCHIGH
	COLORTBL,	\Array: Contains copy of color registers
	CYCRATE,	\Array: Cycle rate (60/sec = $4000)
	CYCFLAGS,	\Array: Cycle flags: bit 0 = active, bit 1 = reverse
	CYCLOW, CYCHIGH; \Arrays: Range of color registers to cycle

seg addr RASTER;	\Display screen

\----------------------------------------------------------------------

proc	FATAL(STR);	\Display error message and abort
addr	STR;
begin
OPENO(TV);		\Back to text mode
TEXT(TV,"
NOPE - ");
CHOUT(TV, BEL);
TEXT(TV, STR);
CRLF(TV);
exit;
end;	\FATAL



proc	WAITVB;		\Wait for vertical blank on the video screen
begin
end;	\WAITVB



proc	PALETTE(N, COLOR);
int	N, COLOR;
int	R, G, B;
begin
R:= COLOR>>12 & $3F;
G:= COLOR>>6  & $3F;
B:= COLOR     & $3F;
SetColReg(N, R, G, B);
end;	\PALETTE

\----------------------------------------------------------------------

func	HEXWORD;	\Read in 4 chars and return the value of the hex word
return SWAP(CHIN(3)) + CHIN(3);



func	HEXLONG;  \Read in 8 chars and return the value of the hex long word
return HEXWORD <<16 + HEXWORD;

\----------------------------------------------------------------------

proc	SHOWSTATS;
begin
if COMPRESSION then TEXT(TV, "COMPRESSED, ");
TEXT(TV, if VIEWMODES & V_HIRES then "HI-RES" else "LO-RES");
if VIEWMODES & V_LACE then TEXT(TV, ", INTERLACED");
if VIEWMODES & V_HAM then TEXT(TV, ", HAM");
TEXT(TV, " PICTURE ");

INTOUT(TV, WIDTH);
TEXT(TV, "x");
INTOUT(TV, HEIGHT);
TEXT(TV, "x");
INTOUT(TV, DEPTH);
CRLF(TV);
TEXT(TV, "
Press ^"Enter^" to display it.

After picture is displayed, press ^"Enter^" to return to DOS.
");
WAITKEY;
end;	\SHOWSTATS

\----------------------------------------------------------------------

proc	DOFORM;		\Read in form type (must be ILBM)
int	CKSIZE;
begin
CKSIZE:= HEXLONG;
if HEXLONG # ILBM then FATAL("^"ILBM^" MISSING");
end;	\DOFORM

\----------------------------------------------------------------------

proc	DOBMHD;		\Read in BitMap Header chunk
int	CKSIZE,
	XOFFSET, YOFFSET,
	MASKING,	\These parameters are read in but are not used
	PAD1,
	TRANSPARENT,
	XASPECT, YASPECT,
	PAGEWIDTH, PAGEHEIGHT;
begin
CKSIZE:= HEXLONG;	\Chunk size in bytes (must be even)

WIDTH:= HEXWORD;	\Raster width and height in pixels
HEIGHT:= HEXWORD;
XOFFSET:= HEXWORD;	\Offset (in pixels) for this image
YOFFSET:= HEXWORD;	\ (not used)
DEPTH:= CHIN(3);	\Number of bit planes (can be 0)
MASKING:= CHIN(3);	\Masking technique (0=no mask, is assumed)
COMPRESSION:= CHIN(3);	\Compression algorithm (0=none, 1=byte run)
PAD1:= CHIN(3);		\Unused by IFF standard
TRANSPARENT:= HEXWORD;	\Transparent color (only used w masking)
XASPECT:= CHIN(3);	\Aspect ratio (1:1 is assumed here)
YASPECT:= CHIN(3);
PAGEWIDTH:= HEXWORD;	\Source page size (not used here)
PAGEHEIGHT:= HEXWORD;

HAVEBMHD:= true;	\Indicate that BMHD chunk has been read in

\Set view modes to default values:
VIEWMODES:= 0;
if WIDTH > 384 then VIEWMODES:= VIEWMODES ! V_HIRES;
if HEIGHT > 240 then VIEWMODES:= VIEWMODES ! V_LACE;
end;	\DOBMHD

\----------------------------------------------------------------------

proc	DOCMAP;		\Color Map, set up COLORTBL
int	CKSIZE,		\Number of bytes in CMAP field
	COLREGS,	\Number of color registers to use
	N,
	RED, GREEN, BLUE;
begin
CKSIZE:= HEXLONG;
COLREGS:= CKSIZE /3;
if COLREGS >256 then COLREGS:= 256;	\Limit maximum number of color registers

for N:= 0, COLREGS-1 do
	begin
	RED:= CHIN(3)>>2;   GREEN:= CHIN(3)>>2;   BLUE:= CHIN(3)>>2;
	COLORTBL(N):= RED<<12 ! GREEN<<6 ! BLUE;
	end;

\Eat the remainder. Must end on an even-byte boundary.
for N:= CKSIZE/3 *3, CKSIZE-1 do
	RED:= CHIN(3);
end;	\DOCMAP

\----------------------------------------------------------------------

proc	DOCAMG;		\Read in Commodore-Amiga chunk, set view mode
int	CKSIZE, N, I;
begin
CKSIZE:= HEXLONG;
CKSIZE:= HEXLONG;
end;	\DOCAMG

\----------------------------------------------------------------------

proc	DOCRNG;		\Read in a color range
int	CKSIZE, PAD1;
begin
CKSIZE:= HEXLONG;

if CYCINX >=CRNGS then FATAL("TOO MANY ^"CRNG^" CHUNKS");

PAD1:= HEXWORD;		\Unused by IFF standard
CYCRATE(CYCINX):= HEXWORD;
CYCFLAGS(CYCINX):= HEXWORD;
CYCLOW(CYCINX):= CHIN(3);
CYCHIGH(CYCINX):= CHIN(3);
CYCINX:= CYCINX +1;
end;	\DOCRNG

\----------------------------------------------------------------------

proc	SHOWCYCLE;	\Show image and handle color cycling if enabled
int	TIMER,		\Array: Count down timers, one for each range
	TIMEMAX,	\Array: Delay between cycles in 60ths of a second
	TABFLAG,	\Flag: Tab control to turn cycling on and off
	J;


	proc	CYCLE(LOW, HIGH, REV);	\Cycle the colors
	int	LOW, HIGH,	\Range of color registers to cycle
		REV;		\Flag: cycle in reverse direction
	int	TEMP, N;
	begin
	if REV then
		begin
		TEMP:= COLORTBL(LOW);
		for N:= LOW, HIGH-1 do		\Move colors down
			COLORTBL(N):= COLORTBL(N+1);
		COLORTBL(HIGH):= TEMP;		\Wrap bottom color to top
		end
	else	begin
		TEMP:= COLORTBL(HIGH);
		for N:= -HIGH, -(LOW+1) do	\Move colors up
			COLORTBL(-N):= COLORTBL(-N-1);
		COLORTBL(LOW):= TEMP;		\Wrap top color to bottom
		end;
	end;	\CYCLE


begin	\SHOWCYCLE
TIMER:= RESERVE(CRNGS *INTSIZE);
TIMEMAX:= RESERVE(CRNGS *INTSIZE);

TABFLAG:= false;
OPENI(0);

for J:= 0, CRNGS-1 do			\Set up timers
	begin
	if CYCRATE(J) <=36 then CYCFLAGS(J):= CYCFLAGS(J) & ~ACTIVEBIT
	else TIMEMAX(J):= $4000 /CYCRATE(J);		\(for DPaintII)

	TIMER(J):= TIMEMAX(J);
	end;

loop	begin				\Poll loop
	if TABFLAG then
		begin
		for J:= 0, CRNGS-1 do
		    if TIMER(J)=0 & (CYCFLAGS(J) &ACTIVEBIT) then
			begin
			CYCLE(CYCLOW(J), CYCHIGH(J), CYCFLAGS(J)&REVBIT);
			TIMER(J):= TIMEMAX(J);
			end;
		WAITVB;
		for J:= 0, 256-1 do			\Set colors
			PALETTE(J, COLORTBL(J));
		for J:= 0, CRNGS-1 do			\Tick timers
			if TIMER(J) >0 then TIMER(J):= TIMER(J) -1;
		end;

	if CHKKEY then			\Check keyboard for commands
		begin
		case GETKEY of
		  ^I-CTRL:	TABFLAG:= ~TABFLAG
		other	[OPENO(TV);   CHOUT(TV, FF);   exit];
		end;
	end;
end;	\SHOWCYCLE

\----------------------------------------------------------------------

proc	DOBODY;		\Image raster data
int	WID,		\Image WIDTH in bytes (and rounded up to next word)
	H, W,		\Adjusted HEIGHT and WIDTH
	M,		\Modulo
	Y,		\Vertical scan line in pixels (0..HEIGHT-1)
	D,		\Bit plane depth (0..DEPTH-1)
	PLANESIZE,	\Size of bit plane in bytes
	CKSIZE;
int	J,		\Index into RASTER
	X;		\Horizontal position (0..WIDTH-1)



proc	DOLINE;		\Read a single line for one bit plane (compressed data)
int	N,		\Byte from input file
	M,		\Byte to replicate
	END;		\Last byte in RASTER for line (+1)
begin
END:= J +WID;
repeat	N:= CHIN(3);
	if N <=$7F then
		begin			\Copy the next N+1 bytes literally
		for X:= J, J+N do
			RASTER(0, X):= CHIN(3);
		J:= X;
		end
	else if N =$80 then []		\Ignore
	else	begin			\N is in the range: $81..$FF
		\Replicate the next byte -EXTEND(N)+1 times
		M:= CHIN(3);		\Read next byte
		for X:= J, J-EXTEND(N) do
			RASTER(0, X):= M;
		J:= X;
		end;
until J >= END;
end;	\DOLINE



begin	\DOBODY
if ~HAVEBMHD then FATAL("^"BMHD^" MISSING");
SHOWSTATS;

WID:= WIDTH /8;		\WIDTH in bytes
SETVID($12);		\ASSUME 640x480x16
for J:= 0, 256-1 do			\Set colors
	PALETTE(J, COLORTBL(J));

CKSIZE:= HEXLONG;
PLANESIZE:= WID *HEIGHT;
for Y:= 0, HEIGHT-1 do
	begin
	for D:= 0, DEPTH-1 do
		begin
		POUT(2, $3C4, 0);	\ENABLE APPROPRIATE PLANE
		POUT(1<<D, $3C5, 0);
		J:= Y *WID;
		if COMPRESSION then
			DOLINE
		else	for X:= J, J+WID-1 do	\Read in a line of pixels
				RASTER(0, X):= CHIN(3);
		end;
	end;

SHOWCYCLE;		\Display the image and handle color cycling if any
end;	\DOBODY

\----------------------------------------------------------------------

proc	EATCHUNK;	\Read in and dispose of a chunk
int	CKSIZE, N, I;
begin
CKSIZE:= HEXLONG;
for N:= 0, CKSIZE-1 do
	I:= CHIN(3);
end;	\EATCHUNK

\----------------------------------------------------------------------

proc	INIT;		\Initialize
int	J;
begin
HAVEBMHD:= false;
CYCINX:= 0;
for J:= 0, CRNGS-1 do		\Tick timers
	begin
	CYCRATE(J):= 0;
	CYCFLAGS(J):= 0;
	CYCLOW(J):= 0;
	CYCHIGH(J):= 0;
	end;

OPENI(3);
end;	\INIT

\----------------------------------------------------------------------

begin	\MAIN
RASTER:= RESERVE(INTSIZE);
RASTER(0):= $A000;

CYCRATE:= RESERVE(CRNGS *INTSIZE);
CYCFLAGS:= RESERVE(CRNGS *INTSIZE);
CYCLOW:= RESERVE(CRNGS *INTSIZE);
CYCHIGH:= RESERVE(CRNGS *INTSIZE);
COLORTBL:= RESERVE(256 *INTSIZE);

TEXT(TV,"-- SHOWILBM, V2.0 --

");
INIT;

if HEXLONG # FORM then FATAL("FILE DID NOT START WITH ^"FORM^"");
DOFORM;

\WARNING: If BODY is missing then this will read beyond the end of the input file
loop	begin
	RECHD:= HEXLONG;	\Read chunk header
	case RECHD of		\Handle chunk
	  FORM:	DOFORM;
	  BMHD:	DOBMHD;
	  CMAP:	DOCMAP;
	  CAMG:	DOCAMG;
	  CRNG:	DOCRNG;
	  BODY:	DOBODY		\(Never returns)
	other	EATCHUNK;	\Ignore any unrecognized chunks
	end;
end;	\MAIN
