\SHOWPCX2.XPL	DEC-09-95	Loren Blaney
\Show a 640x480x8 .PCX file using VESA graphics.

int	DataSeg,	\Our data segment address
	OrigVidMode;	\Original video mode (to be restored upon exit)
char	Palette(256*3);	\256 color registers; each color is 3 bytes (RGB)

inc	\CXPL\STDLIB;

\----------------------------------------------------------------------

proc	LoadPalette(Pal);
char	Pal;
int	I;
begin
I:= Chin(3);				\Skip $0C, start-of-palette byte
for I:= 0, 256*3-1 do
	Pal(I):= Chin(3) >> 2;
end;	\LoadPalette

\----------------------------------------------------------------------

proc	ShowImage;
\Read in a 640x480x8 .PCX file, decompress it, and display it
seg char Screen(1);	\Video screen memory at $A000
int	I,		\Scratch
	J,		\Index into Screen ($A000)
	Byte,		\Byte read from PCX file
	Cnt,		\Number of bytes to repeat (for compressed images)
	Info(128),	\SVGA mode info returned by function 1 (=256 bytes)
	X, Y,		\Screen coordinates

	WinSize,	\Window Size is the amount of video RAM mapped into
			\ $A000 at one time, in KB (usually = 64)
	WS1024,		\WinSize * 1024; WinSize in bytes
	GranSize,	\Window positioning granularity size, in KB. The smallest
			\ step size that video RAM can be mapped into $A000
	Page;		\Particular chunk of video RAM that is currently accessable
			\ at $A000. Each Page contains WinSize KB of video RAM.
			\Logically, GranSize should equal WinSize, but sometimes
			\ it does not. The number of GranSize steps needed to
			\ get to the next Page is WinSize/GranSize. To access a
			\ particular Page, pass WinSize/GranSize*Page to function 5.


	proc	NextPage;	\Map next WinSize chunk of video memory at $A000
	begin
	Page:= Page +1;		\Next Page
	CallInt($10, $4F05, 0, 0, WinSize/GranSize*Page);	\VESA Function 5
	J:= 0;			\Reset $A000 index
	end;	\NextPage


begin	\ShowImage
Screen(0):= $A000;

\Get necessary information for mode $101
Cpureg(0\AX\):= $4F01;
Cpureg(2\CX\):= $101;
Cpureg(4\DI\):= Info;			\256-byte array (=128 integers)
Cpureg(11\ES\):= DataSeg;
Softint($10);
WinSize:= Info(3);
WS1024:= WinSize * 1024;		\(64*1024 overflow is ok)
GranSize:= Info(2);

for I:= 0, 128-1 do J:= Chin(3);	\Skip .PCX header information

Page:= 0;
J:= 0;
for Y:= 0, 480-1 do
    for X:= 0, 640-1 do
	begin
	Byte:= Chin(3);			\Get byte from .PCX file
	if Byte < $C0 then
		begin
		Screen(0,J):= Byte;
		J:= J +1;
		if J = WS1024 then NextPage;
		end
	else	begin			\Decompress multiple copies
		Cnt:= Byte & $3F;
		Byte:= Chin(3);
		for I:= 1, Cnt do
			begin
			Screen(0,J):= Byte;
			J:= J +1;
			if J = WS1024 then NextPage;
			end;
		X:= X + Cnt - 1;
		end;
	end;
end;	\ShowImage

\----------------------------------------------------------------------

proc	SetVesaMode(M);		\Set SVGA mode M
int	M;
begin
CallInt($10, $4F02, M);
if (Cpureg(0) & $FF00) # 0 then
	[Text(0,"Error setting VESA mode: "); Hexout(0, Cpureg(0)); exit];
end;	\SetVesaMode

\----------------------------------------------------------------------

proc	DosOpen; \Open input file (with extension) specified on command line
int	Handle;				\File handle
char	CmdTail($80);			\Copy of command tail
begin
Blit(CpuReg(11), $81, DataSeg, CmdTail, $7F);	\Get copy of command tail

Handle:= Fopen(CmdTail, 0);		\Open file name for input
Fset(Handle, ^I);
Openi(3);
end;	\DosOpen

\======================================================================

begin	\Main
Cpureg:= Getreg;
DataSeg:= Cpureg(12);
DosOpen;		\Open .PCX file specified on command line

OrigVidMode:= GetVid;
SetVesaMode($101);	\Set VESA mode 101: 640x480x8

ShowImage;		\Show .PCX image
LoadPalette(Palette);

\Set block of 256 color registers starting with register 0
CallInt($10, $1012, 0, 256, Palette, 0, 0, DataSeg);

WaitKey;
SetVid(OrigVidMode);
end;	\Main
