\SND.XPL	AUG-08-00
\Make some noise from the Adlib / Sound Blaster's FM chip

inc	\CXPL\CODESI;

def	MaxChan = 9;			\Maximum number of channels

def	SB_BASE = $220,			\Sound Blaster base port address

	DSP_RESET = SB_BASE+$06,	\DSP Reset port
	DSP_READ = SB_BASE+$0A,		\DSP Read data
	DSP_WRITE = SB_BASE+$0C,	\DSP Write data
	DSP_WSTATUS = SB_BASE+$0C,	\DSP Write status
	DSP_RSTATUS = SB_BASE+$0E,	\DSP Read status

	FM_STATUS = SB_BASE+$08,	\FM Status port (read)
	FM_REGISTER = SB_BASE+$08,	\FM Register select port (write)
	FM_DATA = SB_BASE+$09;		\FM Data port (write only)

def	\RegArray index values
	FM_20_A, FM_20_B,
	FM_40_A, FM_40_B,
	FM_60_A, FM_60_B,
	FM_80_A, FM_80_B,
	FM_E0_A, FM_E0_B,
	FM_C0, FM_01, FM_08, FM_BD,
	MaxReg;

def	\Define note values
	NoteC, NoteCS, NoteD, NoteDS, NoteE, NoteF,
	NoteFS, NoteG, NoteGS, NoteA, NoteAS, NoteB;

int	Bottom,			\Bottom octave for musical scale
	ModulateOffset,		\Modulator offsets for each channel
	CarrierOffset,		\Carrier offsets for each channel

	ChanNote(MaxChan),	\Array of notes for each channel
	ChanOctave(MaxChan),	\Array of octaves for each channel
	ChanTime(MaxChan),	\Array of times for each channel to be on

	RegArray(MaxReg),	\Array of FM control register values
	II;			\Scratch index

\-------------------------------------------------------------------------------

proc	ReadDSP(Port);			\Read from specified port
int	Port;
int	C;
begin
for C:= 0, 1600 do if Pin(DSP_RSTATUS, 0) & $80 then return Pin(Port, 0);
return 0;
end;	\ReadDSP



proc	ResetDSP;			\Reset DSP chip (Required for Windows)
int	C;
begin
Pout(1, DSP_RESET, 0);
for C:= 0, 256 do;
Pout(0, DSP_RESET, 0);
return ReadDSP(DSP_READ, 0) = $AA;
end;	\ResetDSP

\-------------------------- FM Chip Control Routines ---------------------------

proc	SetFMReg(Reg, Value);		\Set specified register to "Value"
int	Reg, Value;
int	C, T;
begin
Pout(Reg, FM_REGISTER, 0);		\Select register
for C:= 1, 6 do T:= Pin(FM_STATUS, 0);	\Delay at least 3.3us

Pout(Value, FM_DATA, 0);		\Write data to selected register
for C:= 1, 35 do T:= Pin(FM_STATUS, 0);	\Delay at least 23us
end;	\SetFMReg



proc	ResetFM;			\Reset all FM register
int	R;
begin
for R:= -$F5, -2 do SetFMReg(-R, 0);
SetFMReg($04, $06);
SetFMReg($01, $20);
end;	\ResetFM



proc	GotFM;				\Return 'true' if FM chip is detected
int	T1, T2, D;
begin
SetFMReg($04, $60);
SetFMReg($04, $80);			\Reset
T1:= Pin(FM_STATUS, 0);			\High 3 bits should be 0's

SetFMReg($02, $FF);			\Set timer 1 for 80us
SetFMReg($04, $21);			\Enable "IRQ"
for D:= 1, 80 do;			\Delay at least 80us
T2:= Pin(FM_STATUS, 0);			\Bits 6 & 7 should be set
SetFMReg($04, $60);
SetFMReg($04, $80);

return (T1 & $E0)=$00 & (T2 & $E0)=$C0;
end;	\GotFM

\--------------------------- FM Synthesis Routines -----------------------------

proc	SetFMDefaults;			\Set sound quality defaults
begin
RegArray(FM_01):= $20;
RegArray(FM_08):= 0;
RegArray(FM_BD):= 0;

RegArray(FM_20_A):= $01;   RegArray(FM_20_B):= $01;
RegArray(FM_40_A):= $10;   RegArray(FM_40_B):= $00;
RegArray(FM_60_A):= $F0;   RegArray(FM_60_B):= $F0;
RegArray(FM_80_A):= $77;   RegArray(FM_80_B):= $77;
RegArray(FM_E0_A):= 0;     RegArray(FM_E0_B):= 0;
RegArray(FM_C0):= 0;
end;	\SetFMDefaults



proc	SetTimbre(CHAN, ARRAY);		\Set sound quality for a channel
int	CHAN, ARRAY;
int	MO, CO;
begin
MO:= ModulateOffset(CHAN);
CO:= CarrierOffset(CHAN);

SetFMReg($01, ARRAY(FM_01));
SetFMReg($08, ARRAY(FM_08));
SetFMReg($BD, ARRAY(FM_BD));

SetFMReg($20+MO, ARRAY(FM_20_A));
SetFMReg($20+CO, ARRAY(FM_20_B));

SetFMReg($40+MO, ARRAY(FM_40_A));
SetFMReg($40+CO, ARRAY(FM_40_B));

SetFMReg($60+MO, ARRAY(FM_60_A));
SetFMReg($60+CO, ARRAY(FM_60_B));

SetFMReg($80+MO, ARRAY(FM_80_A));
SetFMReg($80+CO, ARRAY(FM_80_B));

SetFMReg($E0+MO, ARRAY(FM_E0_A));
SetFMReg($E0+CO, ARRAY(FM_E0_B));

SetFMReg($C0, ARRAY(FM_C0));
SetFMReg($40+CO, ARRAY(FM_40_B)+5);
end;	\SetTimbre



proc	SetChannel(Chan, Note, Octave, Array);	\Set sound channel
int	Chan, Note, Octave, Array;
int	F1, F2;
begin
SetTimbre(Chan, Array);
F1:= Bottom(Note);
F2:= SWAP(F1) & $03;   F2:= F2 ! Octave<<2;   F2:= F2 ! $20;
SetFMReg($A0+Chan, F1);
SetFMReg($B0+Chan, F2);
return F1;
end;	\SetChannel



proc	ChanOff(Chan, Note, Octave);	\Turn specified channel off
int	Chan, Note, Octave;
int	F1, F2;
begin
F1:= Bottom(Note);
F2:= SWAP(F1) & $03;   F2:= F2 ! Octave<<2;
SetFMReg($A0+Chan, F1);
SetFMReg($B0+Chan, F2);
end;	\ChanOff



proc	PlayChord;			\Play chord in array
int	Chan, F;
begin
for Chan:= 0, MaxChan-1 do
	[Intout(0, ChanNote(Chan));   Chout(0, ^ )];
Crlf(0);

for Chan:= 0, MaxChan-1 do
    if ChanNote(Chan) # -1 then
	begin
	F:= SetChannel(Chan, ChanNote(Chan), ChanOctave(Chan), RegArray);
	Intout(0, F);
	Chout(0, ^ );
	end;
Crlf(0);

Sound(0, ChanTime(0), 5);		\Delay

for Chan:= 0, MaxChan-1 do
    if ChanNote(Chan)#-1 then
	[ChanOff(Chan, ChanNote(Chan), ChanOctave(Chan));
	ChanNote(Chan):= -1];
end;	\PlayChord



proc	SetReg(RegValues);
\Set timbre register array values in the follow order:
\ 20A, 20B, 40A, 40B, 60A, 60B, 80A, 80B, E0A, E0B, C0, 01, 08, BD
int	RegValues;
int	I;
begin
for I:= 0, MaxReg-1 do RegArray(I):= RegValues(I);
end;	\SetReg



proc	FiveCoin;			\Make "five coin" sound
int	Reg;
begin
Reg:= [$05, $01, $4E, $80, $DA, $F0, $15, $13, $00, $00, $0A, $20, $00, $00];
SetReg(Reg);
ChanNote(0):= NoteC;   ChanOctave(0):= 5;   ChanTime(0):= 2;   PlayChord;
ChanNote(0):= NoteE;   ChanOctave(0):= 5;   ChanTime(0):= 2;   PlayChord;
ChanNote(0):= NoteG;   ChanOctave(0):= 5;   ChanTime(0):= 2;   PlayChord;
end;	\FiveCoin



proc	WINNER;				\Make "winner" sound
int	Reg;
begin
Reg:= [$31, $21, $94, $00, $33, $A0, $73, $97, $00, $00, $0E, $20, $00, $00];
SetReg(Reg);

\G7 Chord
ChanNote(0):= NoteG;   ChanOctave(0):= 5;   ChanTime(0):= 10;
ChanNote(1):= NoteB;   ChanOctave(1):= 5;
ChanNote(2):= NoteD;   ChanOctave(2):= 5;
ChanNote(3):= NoteF;   ChanOctave(3):= 5;
PlayChord;

\C Chord
ChanNote(0):= NoteC;   ChanOctave(0):= 5;   ChanTime(0):= 10;
ChanNote(1):= NoteE;   ChanOctave(1):= 5;
ChanNote(2):= NoteG;   ChanOctave(2):= 5;
PlayChord;
end;	\Winner



begin	\Main
Bottom:= [$156, $16B, $181, $198, $1B0, $1CA, $1E5, $202, $220, $241, $263, $287, $2A3];
ModulateOffset:= [$00, $01, $02, $08, $09, $0A, $10, $11, $12];
CarrierOffset:=  [$03, $04, $05, $0B, $0C, $0D, $13, $14, $15];

for II:= 0, MaxChan-1 do ChanNote(II):= -1;

SetFMDefaults;

Text(0, if GotFM then "FM chip installed" else "No FM chip installed");
Crlf(0);

Text(0, if ResetDSP then "DSP reset" else "Unable to reset DSP");
Crlf(0);

ResetFM;
FiveCoin;
Winner;
ResetFM;
end;	\Main
