\PROGRAM TO READ SERIAL DATA FROM DMO AND DISPLAY FOURIER TRANSFORM

include \CXPL\CODESI.XPL;


\FOURIER TRANFORM VARIABLES

define NPOINTS=512;		\NUMBER OF POINT IN THE TRANSFORM

real
	SINTAB,			\SINE TABLE
	FOURX,FOURY,			\ARRAY OF FOURIER INPUTS
	MAGNITUDE,		\MAGNITUDE OF SPECTRUM
	PHASE;			\PHASE OF SPECTRUM

\GRAPHICS WINDOW VARIABLES

define
	COMY=20,COMX=2,		\COMMAND LINE
	ACTY=21,ACTX=2,		\ACTIVITY LINE LOCATION
	STATY=23,STATX=2,	\STATUS LINE
	MAXWIND=5;		\MAXIMUM NUMBER OF WINDOWS

integer
	VIDMOD,			\CURRENT VIDEO MODE
	ACTWIND,		\CURRENT ACTIVE WINDOW
	MAXX,MAXY,ORX,ORY;	\ARRAY OF CORNERS FOR WINDOWS
real
	XSCALE,YSCALE,		\ARRAY OF SCALES FOR WINDOWS
	XOFF,YOFF;		\ARRAY OF OFFSETS FOR WINDOWS


integer				\LINE COLORS FOR GRAPHICS
	CURCOL,			\COLOR OF MEASUREMENT CURSOR
	PLOTCOL,		\COLOR FOR PLOT
	GRIDCOL,		\COLOR OF GRID LINES
	BOXCOL,			\COLOR OF BOX
	TXTCOL1,		\TEXT COLORS
	TXTCOL2;

define				\EGA COLOR SET

	BLACK=0,BLUE=1,GREEN=2,CYAN=3,
	RED=4,MAGEN=5,BROWN=6,WHITE=7,
	GREY=8,LTBLUE=9,LTGREEN=10,LTCYAN=11,
	LTRED=12,LTMAGEN=13,YELLOW=14,LTWHITE=15;

define
	LEFTARROW=75,		\LEFT ARROW SCAN CODE
	RIGHTARROW=77;		\RIGHT ARROW SCAN CODE

integer
	OPMODE;			\OPERATING MODE

\VARIABLES FOR OSCILLOSCOPE INTERFACE

real
	FREQTAB;		\TABLE OF FREQUENCY VS. TIMEBASE


\VARIABLES FOR SERIAL COMMUNICATION

char
	STATSTR,	\STATUS STRING FROM SCOPE
	WAVESTR;	\STRING TO HOLD WAVE FROM SCOPE


integer 
	T,		\GLOBAL TEMPORARY

	COM,		\CURRENT COMMUNICATION CHANNEL
	BASE,		\BASE PORT ADDRESS OF SERIAL PORT
	MODCON,		\SERIAL PORT MODEM CONTROL REGISTER
	LINCON,		\SERIAL PORT LINE CONTROL REGISTER

	CPUREG;		\CPU REGISTER FOR DOS CALLS


integer SPARAM;		\PARAMETERS FOR EXTERNAL SERIAL INPUT ROUTINE
define
	SPORT,		\LOCATION OF STATUS PORT
	DPORT,		\LOCATION OF DATA PORT
	ARRAY,		\POINTER TO THE ARRAY
	CHRLEN,		\NUMBER OF CHARACTER TO READ

	STCLEN;		\LENGTH OF THE STRUCTURE !!MUST BE LAST!!


external SERIAL;	\EXTERNAL ASSEMBLY LANGUAGE ROUTINE
			\TO READ A BLOCK OF DATA


\******************* LOW LEVEL SUPPORT ROUTINES *****************


procedure BLINE(Y);
\BLANK THE SPECIFIED LINE
integer Y;
begin
CURSOR(2,Y);
TEXT(6,"                                                                            ");
end;


\******** SERIAL COMMUNICATION ROUTINES *********

procedure SINITIAL;
\INITIALIZE SERIAL PORT
integer TEMP;
begin
\SET TO 9600 BAUD 8N1
CPUREG(0):=$00E3;;
CPUREG(3):=COM;
SOFTINT($14);

\SET PORT BASE ADDRESS
case COM of
	0: BASE:=$3F8;
	1: BASE:=$2F8
else;
SPARAM(DPORT):=BASE;
SPARAM(SPORT):=BASE+5;
MODCON:=BASE+4;
LINCON:=BASE+3;

\SELECT PRIMARY REGISTER SET
TEMP:=PIN(LINCON,0) & $7F;
POUT(TEMP,LINCON,0);
\TURN ON RTS AND DTR
POUT($03,MODCON,0);
end;


procedure SEROUT(CHAR);
\WRITE ONE SERIAL CHARACTER
integer CHAR;
begin
repeat until (PIN(SPARAM(SPORT),0) & $20) # 0;
POUT(CHAR,SPARAM(DPORT),0);
end;



procedure SERTEST;
\TEST FOR SERIAL CHARACTER WAITING
return (PIN(SPARAM(SPORT),0) & 1) # 0;



procedure SERIN;
\READ ONE CHARACTER FROM PORT
begin
repeat until (PIN(SPARAM(SPORT),0) & 1) # 0;
return PIN(SPARAM(DPORT),0) & $7F;
end;


\************ GRAPHICS WINDOW CODE *************


procedure WINDOW(X1,Y1,X2,Y2);
\OPEN A WINDOW
\ X1Y1= LOWER LEFT CORNER
\ X2Y2= UPPER RIGHT CORNER
integer X1,Y1,X2,Y2;
begin
MOVE(X1,Y1);
LINE(X2,Y1,1);LINE(X2,Y2,GRIDCOL);
LINE(X1,Y2,1);LINE(X1,Y1,GRIDCOL);
MAXX(ACTWIND):=ABS(X2-X1);MAXY(ACTWIND):=ABS(Y2-Y1);
ORX(ACTWIND):=X1;ORY(ACTWIND):=Y1;
end;



procedure OPOINT(X,Y,C);
\PLACE POINT, REFERENCED TO ORIGIN
integer X,Y,C;
POINT(ORX(ACTWIND)+X,ORY(ACTWIND)-Y,C);



procedure OLINE(X,Y,C);
\DRAW LINE, REFERRENCE TO ORIGIN
integer X,Y,C;
LINE(ORX(ACTWIND)+X,ORY(ACTWIND)-Y,C);



procedure OMOVE(X,Y);
\MOVE, REFERENCED TO ORIGIN
integer X,Y;
MOVE(ORX(ACTWIND)+X,ORY(ACTWIND)-Y);



procedure RLINE(X,Y,C);
\DRAW SCALED AND OFFSET LINE
real X,Y;
integer C;
OLINE(FIX((X-XOFF(ACTWIND))*XSCALE(ACTWIND)),
	FIX((Y-YOFF(ACTWIND))*YSCALE(ACTWIND)),C);



procedure RPOINT(X,Y,C);
\PLACE SCALED AND OFFSET POINT
real X,Y;
integer C;
OPOINT(FIX((X-XOFF(ACTWIND))*XSCALE(ACTWIND)),
	FIX((Y-YOFF(ACTWIND))*YSCALE(ACTWIND)),C);



procedure RMOVE(X,Y);
\MOVE SCALED AND OFFSET
real X,Y;
OMOVE(FIX((X-XOFF(ACTWIND))*XSCALE(ACTWIND)),
	FIX((Y-YOFF(ACTWIND))*YSCALE(ACTWIND)));


procedure SYMBOL(T,X,Y);
\PLACE A SYMBOL IN WINDOW
\SYMBOL TYPE IN T
real X,Y;
integer T,X1,Y1,SHAPE;
begin
SHAPE:=[
[[2,2],[2,-2],[-2,-2],[-2,2],[2,2]], \SQUARE
[[0,2],[2,0],[0,-2],[-2,0],[0,2]], \DIAMOND
[[2,2],[-2,-2],[2,-2],[-2,2],[2,2]] \X
];
if T>=3 then T:=0;

X1:=FIX((X-XOFF(ACTWIND))*XSCALE(ACTWIND));
Y1:=FIX((Y-YOFF(ACTWIND))*YSCALE(ACTWIND));
OMOVE(X1+SHAPE(T,0,0),Y1+SHAPE(T,0,1));
OLINE(X1+SHAPE(T,1,0),Y1+SHAPE(T,1,1),TXTCOL1);
OLINE(X1+SHAPE(T,2,0),Y1+SHAPE(T,2,1),TXTCOL1);
OLINE(X1+SHAPE(T,3,0),Y1+SHAPE(T,3,1),TXTCOL1);
OLINE(X1+SHAPE(T,4,0),Y1+SHAPE(T,4,1),TXTCOL1);
end;


procedure VGRID(X,R1,R2);
\DRAW VERTICAL GRID
\NUMBER OF LINES IN X
\START AND END OF SCALE IN R1 AND R2
integer X;
real R1,R2,DL;
integer O,I;
begin
XSCALE(ACTWIND):=FLOAT(MAXX(ACTWIND))/(R2-R1);
XOFF(ACTWIND):=R1;

O:=MAXX(ACTWIND)/X;
for I:=1,X do
	begin
	OMOVE(I*O,0);
	OLINE(I*O,MAXY(ACTWIND),GRIDCOL);
	end;


FORMAT(2,0);
DL:=(R2-R1)/FLOAT(X);
for I:=0,X do
	begin
	CURSOR((I*O+ORX(ACTWIND))/8-2,(ORY(ACTWIND)/8)+1);
	RLOUT(0,R1);
	R1:=R1+DL;
	end;
end;




procedure HGRID(Y,R1,R2);
\DRAW VERTICAL GRID
\NUMBER OF LINES IN X
\START AND END OF SCALE IN R1 AND R2
integer Y;
real R1,R2,DL;
integer O,I;
begin
YSCALE(ACTWIND):=FLOAT(MAXY(ACTWIND))/(R2-R1);
YOFF(ACTWIND):=R1;

O:=MAXY(ACTWIND)/Y;
for I:=1,Y do
	begin
	OMOVE(0,O*I);
	OLINE(MAXX(ACTWIND),O*I,GRIDCOL);
	end;

FORMAT(2,0);
DL:=(R2-R1)/FLOAT(Y);
for I:=0,Y do
	begin
	CURSOR((ORX(ACTWIND)/8)-4,(ORY(ACTWIND)-(O*I))/8);
	RLOUT(0,R1);
	R1:=R1+DL;
	end;
end;



procedure SETWND;
\SETUP THE WINDOWS
begin
SETVID(VIDMOD);

ATTRIB(TXTCOL1);
CURSOR(38,0);TEXT(6,"Spectrum");

ACTWIND:=0;
WINDOW(32,82,638,10);
VGRID(8,0.,256.);
HGRID(6,-60.,0.);

CURSOR(33,12);TEXT(6,"Input Wave Form");
ACTWIND:=1;
WINDOW(32,140,638,106);
VGRID(4,0.,512.);
HGRID(2,-128.,128.0);

CURSOR(COMX,COMY);
TEXT(6,"ESC=EXIT    R=RESET DISPLAY    M=MEASURE  D=DISPLAY");

MOVE(0,152);
LINE(639,152,BOXCOL);LINE(639,199,BOXCOL);
LINE(0,199,BOXCOL);LINE(0,152,BOXCOL);
ATTRIB(TXTCOL2);
end;


procedure VEDSETUP;
\SETUP VIDEO PARAMETERS
integer C;
begin
TEXT(0,"EGA COMPATABLE? Y/N");
C:=CHIN(1);
if C=^Y ! C=^y then 
	begin
	VIDMOD:=$0E;
	CURCOL:=LTCYAN;
	PLOTCOL:=YELLOW;
	GRIDCOL:=BLUE;
	BOXCOL:=GREEN;
	TXTCOL1:=RED;
	TXTCOL2:=MAGEN;
	end
 else 
	begin
	VIDMOD:=6;
	CURCOL:=1;PLOTCOL:=1;
	GRIDCOL:=1;BOXCOL:=1;
	TXTCOL1:=1;TXTCOL2:=1;
	end;
end;


\************ FOURIER TRANFORM CODE ************



proc PLOT(PNTS,N,F);
real PNTS;
int N,F,I,C;
begin
C:=PLOTCOL;if F then C:=C ! $80;

RMOVE(0.0,PNTS(0));
for I:=1,N-1 do RLINE(FLOAT(I),PNTS(I),C);
end;


proc SETSIN;
integer I;
for I:=0,NPOINTS do
  SINTAB(I):=SIN(6.283185308*FLOAT(I)/FLOAT(NPOINTS));


proc real SINT(I);
integerI;
begin
if I<0 then return SINT(I+NPOINTS)
else if I>NPOINTS then return SINT(I-NPOINTS)
else return SINTAB(I);
end;

proc real COST(I);
integer I;
return SINT(I-NPOINTS/4);


proc  FFT(REAL,IMAG);
real  REAL,IMAG,TR,TI,WR,WI,MIN,MAX;
integer MR,M,L,NN,CHAR,STEP,A;
integer  I,J;
begin
MR:=0;
NN:=NPOINTS-1;
for M:=1,NN do
	begin
	L:=NPOINTS;
	repeat L:=L/2 until MR+L<=NN;
	MR:=REM(MR/L)+L;
	if MR>M then
		begin
		TR:=REAL(M);
		REAL(M):=REAL(MR);
		REAL(MR):=TR;
		TI:=IMAG(M);
		IMAG(M):=IMAG(MR);
		IMAG(MR):=TI;
		end;
	end;

L:=1;
while L<NPOINTS do
	begin
	CURSOR(ACTX,ACTY);
	TEXT(6,"FOURER: ");
	INTOUT(6,L);
	STEP:=2*L;
	for M:=0,L-1 do
		begin
		A:=(NPOINTS/STEP)*(-M);
		WR:=-COST(A);
		WI:=SINT(A);
		I:=M;
		repeat
			begin
			J:=I+L;
			TR:=(WR*REAL(J)-WI*IMAG(J));
			TI:=(WR*IMAG(J)+WI*REAL(J));
			REAL(J):=REAL(I)-TR;
			IMAG(J):=IMAG(I)-TI;
			REAL(I):=REAL(I)+TR;
			IMAG(I):=IMAG(I)+TI;
			I:=I+STEP
			end
		until I>=NPOINTS-1;
		end;
	L:=STEP;
	end;

for I:=0,NPOINTS/2 do
	begin
	TR:=SQRT(REAL(I)*REAL(I)+IMAG(I)*IMAG(I))/FLOAT(NPOINTS);
	if TR<.25 then TR:=.25;
	MAGNITUDE(I):=20.*LOG(TR/128.);
	\TR:=if REAL(I)=0.0 then 1.0E-300 else REAL(I);
	\PHASE(I):=ATAN2(IMAG(I),TR);
	end;
end;


\******* ROUTINES TO READ THE SCOPE *********



procedure HERTZ(D,F);
int D;real F;
begin
FORMAT(6,3);
case of
	F>=1E6: [RLOUT(D,F/1E6);TEXT(D," MHZ");];
	F>=1E3: [RLOUT(D,F/1E3);TEXT(D," KHZ");]
else  [RLOUT(D,F);TEXT(D," HZ ");];
end;



procedure READSTATUS;
\READ STATUS FROM SCOPE
integer I,T;
real B;
begin
SPARAM(ARRAY):=STATSTR;
SPARAM(CHRLEN):=13;
CURSOR(ACTX,ACTY);
TEXT(6,"STATUS IN  ");
SEROUT(^Y);
SERIAL(SPARAM);
CURSOR(ACTX,ACTY);
TEXT(6,"           ");
if SPARAM(CHRLEN)#13 then return;

FORMAT(6,2);
CURSOR(STATX,STATY);
TEXT(6,"RANGE: ");
B:=FREQTAB(STATSTR(9));
HERTZ(0,B);
TEXT(0," - ");
HERTZ(0,256. * B);
end;



procedure HANDSHAKE;
\HANDLE HANDSHAKING AND TIMEOUTS
integer T;
seg char CLOCK;

	procedure SAMPLE;
	\WAIT FOR SCOPE TO SAMPLE LONG WAVE FORM
	begin
	CURSOR(ACTX,ACTY);
	TEXT(6,"SAMPLING   ");
	loop
		begin
		if SERTEST then if SERIN=^r then return;
		if CHKKEY then return;
		end;
	end;


begin
CLOCK:=RESERVE(2);
CLOCK(0):=0;
CURSOR(ACTX,ACTY);
TEXT(6,"HANDSHAKE  ");
SEROUT(^W);
T:=CLOCK(0,$46C);T:=(T+$0040) & $FF;
loop
	begin
	if SERTEST then
	case SERIN of
		^r : quit;
		^s : [SAMPLE;quit;]
	else SEROUT(^W);
	if CLOCK(0,$46C)=T then quit;
	end;
end;



procedure READWAVE;
\READ A WAVE FORM FROM SCOPE
integer I,O,T,START;
begin
READSTATUS;

SPARAM(ARRAY):=WAVESTR;
SPARAM(CHRLEN):=512;

HANDSHAKE;

CURSOR(ACTX,ACTY);
TEXT(6,"WAVE IN    ");
SEROUT(^r);
SERIAL(SPARAM);
SEROUT($20);SEROUT($20);

for I:=0,NPOINTS-1 do FOURX(I):=FLOAT(WAVESTR(I)-128);
for I:=0,NPOINTS-1 do FOURY(I):=0.0;
ACTWIND:=1;
PLOT(FOURX,NPOINTS,OPMODE);
end;


\******** INTERFACE ROUTINES *************


procedure MEASURE;
\MEASURE SPECTRUM POINTS
integer CX,CC,OX;
real L;

	procedure UPDATE;
	\UPDATE CURSOR LINE
	begin
	RMOVE(FLOAT(OX),0.);RLINE(FLOAT(OX),-60.,CC);
	RMOVE(FLOAT(CX),0.);RLINE(FLOAT(CX),-60.,CC);
	OX:=CX;
	end;

begin
CC:=CURCOL ! $80;
FORMAT(5,2);
CX:=0;OX:=0;
RMOVE(FLOAT(CX),0.);RLINE(FLOAT(CX),-60.,CC);
loop
	begin
	CURSOR(ACTX,ACTY);
	L:=if CX=0 then .5 else FLOAT(CX);
	FORMAT(3,0);
	TEXT(6,"X: ");RLOUT(0,FLOAT(CX));
	TEXT(6,"  FREQ:");HERTZ(0,L * FREQTAB(STATSTR(9)));
	TEXT(6,"  AMP:");RLOUT(0,MAGNITUDE(CX));
	case CHIN(1) of
		$1B: quit;
		0:	begin
			case CHIN(1) of
			LEFTARROW : if CX>0 then [CX:=CX-1;UPDATE];
			RIGHTARROW: if CX<256 then [CX:=CX+1;UPDATE;]
			other;
			end		
	other;
	end;
RMOVE(FLOAT(CX),0.);RLINE(FLOAT(CX),-60.,CC);
BLINE(ACTY);
end;



procedure DECODE;
\DECODE USER COMMANDS;
begin
case CHIN(1) of
$1B: return true;
^D,^d: OPMODE:=not OPMODE;
^R,^r: SETWND;
^M,^m: MEASURE
other;
return false;
end;



procedure INITIALIZE;
\INITIALIZE VARIABLES ETC.
begin
\TABLE OF FREQUENCIE VS. TIME DIVISION OF SCOPE

FREQTAB:=[
	10E6,\10ns\	5E6,\20ns\	2E6,\50ns\
	1E6,\100ns\	500E3,\200ns\	200E3,\500ns\
	100E3,\1us\	50E3,\2us\	20E3,\5us\
	10E3,\10us\	5E3,\20us\	2E3,\50us\
	1E3,\100us\	500.,\200us\	200.,\500us\
	100.,\1ms\	50.,\2ms\	20.,\5ms\
	10.,\10ms\	5.,\20ms\	2.,\50ms\
	1.,\100ms\	500E-3,\200ms\	200E-3,\500ms\
	100E-3,\1s\	50E-3,\2s\	20E-3,\5s\
	10E-3,\10s\	5E-3];\20s\
end;



begin
CPUREG:=GETREG;
SPARAM:=RESERVE(STCLEN*2);
WAVESTR:=RESERVE(1024);
STATSTR:=RESERVE(16);
COM:=1;			\SELECT COM2


MAXX:=RESERVE(MAXWIND*2);MAXY:=RESERVE(MAXWIND*2);
ORX:=RESERVE(MAXWIND*2);ORY:=RESERVE(MAXWIND*2);

XSCALE:=RLRES(MAXWIND);YSCALE:=RLRES(MAXWIND);
XOFF:=RLRES(MAXWIND);YOFF:=RLRES(MAXWIND);

SINTAB:=RLRES(NPOINTS+1);
MAGNITUDE:=RLRES((NPOINTS/2)+1);
PHASE:=RLRES((NPOINTS/2)+1);
FOURX:=RLRES(NPOINTS);
FOURY:=RLRES(NPOINTS);

OPMODE:=false;
INITIALIZE;
SINITIAL;
VEDSETUP;
SETWND;
SETSIN;

loop
	begin
	if CHKKEY then if DECODE then quit;
	READWAVE;
	FFT(FOURX,FOURY);CHOUT(0,7);
	ACTWIND:=0;
	PLOT(MAGNITUDE,NPOINTS/2+1,OPMODE);
	end;
SETVID(3);
end;
