\SSS.XPL	FEB-04-94
\Simple VGA Screen Saver by Loren Blaney

\Note: Unless 32-bit XPL is used, the three CWD instructions in the
\ compiled SSS.ASM file must be removed by hand.

inc	C:\CXPL\CODESI;	\Include code definitions for intrinsic routines

def	AX, BX, CX, DX, DI, SI, BP, CF, CS, DS, SS, ES;	\Getreg registers
int	CpuReg;		\Address of CPU register array (from Getreg)

def	N=7;		\Index of last spline knot
int	P0, P1,		\(N,2) Arrays: Positions of spline knots
			\ P1 draws, P0 follows later and erases
	Dir0, Dir1,	\(N,2) Arrays: Amount to move knots for each step
	Color,		\Color register (0..15)
	R, G, B,	\Red, Green, Blue intensities (0..63)
	DR, DG, DB,	\Amount to adjust intensity for each step (delta color)
	VMode,		\Initial video mode
	II, JJ;		\Scratch for main
def	X, Y;		\Subscripts for X and Y directions



proc	Park;			\Park hard drive heads
begin
\Park first drive (C)
CpuReg(AX):= $1100;		\Recalibrate
CpuReg(DX):= $0080;		\Select first drive
Softint($13);
CpuReg(AX):= $0800;		\Get drive parameters
Softint($13);
CpuReg(AX):= $0C00;		\Seek
CpuReg(DX):= $0080;
Softint($13);

\Park second drive (D) (in case there is one)
CpuReg(AX):= $1100;		\Recalibrate
CpuReg(DX):= $0081;		\Select second drive
Softint($13);
CpuReg(AX):= $0800;		\Get drive parameters
Softint($13);
CpuReg(AX):= $0C00;		\Seek
CpuReg(DX):= $0081;
Softint($13);
end;	\Park



proc	WaitVB;			\Wait for beginning of next vertical blank
begin				\WARINING: Requires color display
while Pin($3DA, 0) & $08 do;	 \Wait for no vertical blank
repeat until Pin($3DA, 0) & $08; \Wait for vertical blank
end;	\WaitVB;

\----------------------------------------------------------------------

func	Getvid;			\Returns the current video mode
begin
CpuReg(AX):= $0F00;		\Function $0F
Softint($10);			\Call BIOS routine
return CpuReg(AX) & $FF;
end;	\Getvid



proc	FixPalette;		\Make palette registers correspond to DAC
int	N;
begin
for N:= 0, 15 do
	begin
	CpuReg(AX):= $1000;	\Function $10, subfunction $00
	CpuReg(BX):= N<<8!N;	\DAC and palette registers
	Softint($10);		\Call BIOS routine
	end;
end;	\FixPalette



proc	SetColReg(N, R, G, B);	\Set color register N
int	N,			\Register to set
	R, G, B;		\Red, Green, and Blue [0..255]
begin
CpuReg(AX):= $1010;		\Function $10, subfunction $10
CpuReg(BX):= N;			\Select color register
CpuReg(CX):= G<<8 ! (B&$FF);	\Green<<8 blue
CpuReg(DX):= R<<8;		\Red<<8
Softint($10);			\Call BIOS routine
end;	\SetColReg



proc	NewColor(N);		\Change color register N
int	N;
begin
R:= R +DR;
if R<0 ! R>63 then [DR:= -DR;   R:= R +DR];
G:= G +DG;
if G<0 ! G>63 then [DG:= -DG;   G:= G +DG];
B:= B +DB;
if B<0 ! B>63 then [DB:= -DB;   B:= B +DB];

SetColReg(N, R, G, B);
end;	\NewColor

\----------------------------------------------------------------------

proc	DrawCurve(P, Dir, Col);	\Draw a closed curve using a Bezier spline
int	P,		\Pointer to array of knot positions
	Dir,		\Pointer to array of amounts to move knots
	Col;		\Color register (0..15)
int	Px, Py;		\Local copy of array knot positions
int	U;		\Fractional step position for spline
int	I, J, K, L;
def	Scale=64;	\Used to calculate fractional U value
def	Segs=64;	\Number of line segments used to approximate curve
begin
Px:= Reserve((N+1)*4);
Py:= Reserve((N+1)*4);

Col:= Col ! $80;	\Draw with XOR color

P(7,X):= P(0,X);	\Make end point same as begin point for a closed loop
P(7,Y):= P(0,Y);

\Remove the kink in the loop by projecting point 6 as a mirror image of point 1
\P6  =  P0 -(P1-P0)  =  2*P0 - P1
P(6,X):= 2*P(0,X) - P(1,X);
P(6,Y):= 2*P(0,Y) - P(1,Y);

\Draw the spline
Move(P(0,X), P(0,Y));
for K:= 1, Segs do
	begin
	U:= K *Scale /Segs;	\Fractional part of distance between knots

	for I:= 0, N do
		[Px(I):= P(I,X)<<3;
		 Py(I):= P(I,Y)<<3];

	for J:= -(N-1), 0 do
	    for I:= 0, -J do
		[Px(I):= Px(I) + (U *(Px(I+1) - Px(I))) /Scale;
		 Py(I):= Py(I) + (U *(Py(I+1) - Py(I))) /Scale];

	I:= Px(0) >>3;   J:= Py(0) >>3;
	Line(I, J, Col);	\Draw the segment
	Point(I, J, Col);	\Fix black dot caused by XOR & overlapping lines
	end;

\Move the knots
for I:= 0, N-2 do		\The last 2 knots are handled elsewhere
	begin
	P(I,X):= P(I,X) + Dir(I,X);		\Move knots
	if P(I,X)>=500 ! P(I,X)<140 then	\Reverse direction near edges
		[Dir(I,X):= -Dir(I,X);
		P(I,X):= P(I,X) + Dir(I,X)];

	P(I,Y):= P(I,Y) + Dir(I,Y);
	if P(I,Y)>=380 ! P(I,Y)<100 then
		[Dir(I,Y):= -Dir(I,Y);
		P(I,Y):= P(I,Y) + Dir(I,Y)];
	end;
end;	\DrawCurve

\----------------------------------------------------------------------

begin	\Main
CpuReg:= Getreg;

P0:= Reserve((N+1)*4);			\Set up the arrays
P1:= Reserve((N+1)*4);
Dir0:= Reserve((N+1)*4);
Dir1:= Reserve((N+1)*4);
for II:= 0, N do			\For all knots...
	begin
	P0(II):= Reserve(2*4);
	P0(II,X):= 320;			\Start in center
	P0(II,Y):= 240;

	P1(II):= Reserve(2*4);		\Make both arrays the same
	P1(II,X):= P0(II,X);
	P1(II,Y):= P0(II,Y);

	Dir0(II):= Reserve(2*4);	\Set up random speeds and directions
	Dir0(II,X):= RAN(25)-12;
	Dir0(II,Y):= RAN(25)-12;

	Dir1(II):= Reserve(2*4);	\Make 2nd array the same as first
	Dir1(II,X):= Dir0(II,X);
	Dir1(II,Y):= Dir0(II,Y);
	end;

Chout(0, $0C);				\Clear screen with a form feed
Rawtext(0, "


				   
				   imple
				   creen
				   aver
				   




				Loren Blaney

	





			  Parking hard drive heads...  ");

Park;

for II:= 1, 200 do [WaitVB;   if Chkkey then II:= 30000];	\Pause

VMode:= Getvid;				\Save video mode
Setvid($12);				\640x480x16
FixPalette;

R:= 0;   G:= 0;   B:= 0;		\Set up the colors
DR:= 2;  DG:= 1;  DB:= 3;
for Color:= 1, 15 do NewColor(Color);	\Prevent XOR glitches
Color:= 1;

Openi(0);
II:= 1;
loop	begin
	if II > 15*4 then 		\Follow 15*4 steps later with black
		DrawCurve(P0,Dir0,Color) \XOR with same color to make black
	else	II:= II +1;
	NewColor(Color);		\Change colors
	DrawCurve(P1,Dir1,Color);	\Draw new colored curve
	Color:= (Color+1) & $0F;	\Next palette register
	if Color=0 then Color:= 1;	\Don't use black
	if Chkkey then quit;
	for JJ:= 1, 6 do WaitVB;	\Don't go too fast
	end;
Setvid(VMode);				\Restore original video mode
end;	\Main
