\STUDENT.XPL	AUG-27-90
\By Lenny Boreal
\
\Artificial intelligence is better than having no intelligence at all.
\
\This program learns by reading. You can give it reading assignments then
\ judge its level of comprehension by asking it questions.
\
\In this program the fundamental unit of data is the English word.
\ Sentences are handled as linked lists of words.

code	ABS=0,		RAN=1,		REM=2,		RESERVE=3,
	SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
	CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
	TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
	ABORT=16,	TRAP=17,	FREE=18,	RERUN=19,
	GETHP=20,	SETHP=21,	GETERR=22,	CURSOR=23,
	SCAN=24,	SETRUN=25,	HEXIN=26,	HEXOUT=27,
	CHAIN=28,	OPENF=29,	FWRITE=30,	FREAD=31,
	\TESTPT=32,\	FGET=33,	\FASAVE=34,\	FSAVE=35,
	BLIT=36,	BUTTON=37,	JOYSTICK=38,	\SOUND=39,\
	CLEAR=40,	POINT=41,	LINE=42,	MOVE=43,
	\SCREEN=44,\	\BLOCK=45,\	FIX=50,		BACKUP=64,
	WAITVB=107,
	BITMAP=108,	BITMAP2=109,	VIEW=110,	PALETTE=111,
	CARRY=112,	PEEK_W=113,	POKE_W=114,	PEEK_L=115,
	POKE_L=116,	SWAP_W=117,	EXT_L=118,	CURSOR1=119,
	BUTES1=120,	SHOCUR1=121,	DEVINFO=122,	UNTINFO=123,
	BUTES=124,	GETKEY=125,	KEYHIT=126,	SHOCUR=127;

def	BEL=$07, TAB=$09, FF=$0C, CR=$0D, EOF=$1A, ESC=$1B, SP=$20;
def	TV=0, INDEV=3;
def	INTSIZE=4;	\Number of bytes in an integer
def	INF=$7FFFFFFF,	\Largest possible signed integer
	NIL=0;		\Terminator for linked lists (an impossible address)

int	KNOWLEDGE,	\Lists of sentences which state facts (the ball is blue)
	SRAM,		\Used to keep the garbage from getting too deep
	EOFFOUND,	\Flag: End-of-file read
	;
addr	TOPMEM,		\Top of usable memory (heap) space +1
	RAM;		\Pointer to allocated heap space



func	ALLOCATE(AMOUNT);
\This reserves memory, but, unlike the RESERVE intrinsic, this doesn't
\ relinquish the space when the calling procedure returns.
int	AMOUNT;
addr	TEMP;
begin
TEMP:= RAM;
RAM:= RAM + AMOUNT;			\Reserve bytes
if RAM >= TOPMEM then
	begin
	CHOUT(TV, BEL);
	CRLF(TV);   TEXT(TV, "I NEED MORE BRAIN CELLS (I.E. RAM)");   CRLF(TV);
	exit;
	end;
return TEMP;
end;	\ALLOCATE



func	INSERT(A, L);	\Insert atom A at the head of the linked list L
int	A, L;		\This returns the new head of the list, thus the
int	L2, I;		\ form of the call is: L:= INSERT(A, L);
begin
L2:= ALLOCATE(2*INTSIZE);
\The 68000 insists on having pointers to words at even addresses
if L2 & 1 then [L2:= L2 +1;   I:= ALLOCATE(1)];

L2(0):= A;
L2(1):= L;
return L2;
end;	\INSERT



func	REVERSE(L);	\Reverse the order of the linked list L
int	L;		\This returns the new head of the list, thus the
int	S, L0;		\ form of the call is: L:= REVERSE(L);
begin
L0:= NIL;
while L # NIL do
	begin
	S:= L(1);	\Save the pointer to the next cell
	L(1):= L0;	\Point to the previous cell
	L0:= L;		\Save pointer to the current cell
	L:= S;		\Point to the next cell
	end;
return	L0;
end;	\REVERSE



proc	PRINT(DEV, L);	\Print the linked list L on the specified device.
int	DEV, L;
begin
while L # NIL do
	begin
	TEXT(DEV, L(0));
	CHOUT(DEV, SP);
	L:= L(1);
	end;
end;	\PRINT



proc	STRCOPY(A, B);	\Copy string A into B (including the terminator)
addr	A, B;
int	I;
begin
for I:= 0, INF do
	begin
	B(I):= A(I);
	if A(I) = 0 then return;
	end;
end;	\STRCOPY



func	STREQU(A, B);	\Returns 'true' if string A is equal to B.
\Note that strings must be the same length to be equal.
addr	A, B;
int	I;
begin
for I:= 0, INF do
	begin
	if A(I) # B(I) then return false;
	if A(I) = 0 then return true;
	end;
end;	\STREQU



func	MEMBER(A, L);	\Returns 'true' if atom A is in linked list L
int	A, L;
begin
loop	begin
	if L = NIL then return false;
	if STREQU(A, L(0)) then return true;
	L:= L(1);
	end;
end;	\MEMBER



func	SEARCH(L1, L2);	\Search for list L1 in L2. Returns 'true' if found.
int	L1, L2;
int	L10, L20;
begin
if L1=NIL ! L2=NIL then return false;
L10:= L1;   L20:= L2;
loop	begin
	if STREQU(L1(0), L2(0)) then
		begin			\words match
		L1:= L1(1);		\get next atom
		L2:= L2(1);
		if L1 = NIL then return true;
		if L2 = NIL then return false;
		end
	else	begin			\words don't match
		L20:= L20(1);
		if L20 = NIL then return false;
		L1:= L10;
		L2:= L20;
		end;
	end;
end;	\SEARCH



func	SIZEOF(L);	\Returns the number of items in a linked list
int	L;
int	I;
begin
for I:= 0, INF do
	begin
	if L = NIL then return I;
	L:= L(1);
	end;
end;	\SIZEOF



func	COPYOF(L);	\Make a copy of the linked list L
\WARNING: This does not copy the atoms, only the list of pointers.
\This does not, however, cause any problem when throwing out the garbage.
int	L;
int	L2;
begin
L2:= NIL;
while L # NIL do
	begin
	L2:= INSERT(L(0), L2);
	L:= L(1);
	end;
return REVERSE(L2);
end;	\COPYOF

\----------------------------------------------------------------------

func	GETSEN(DEV);
\Read in a sentence and return a pointer to its list of words.
int	DEV;
int	CH, L, L2;


	func	GETWORD;
	int	I;
	addr	WD, WD2;
	begin
	WD:= RESERVE(80);
	repeat CH:= CHIN(DEV); until CH#SP & CH#CR & CH#TAB & CH#FF;
	I:= 0;
	loop	begin
		if CH>=^a & CH<=^z then CH:= CH & $DF;	\Convert to uppercase
	\	if (CH<^A ! CH>^Z) & (CH<^0 ! CH>^9) then quit;
		case CH of
		  SP, CR, ^., ^?, ^!, TAB, FF, EOF: quit
		other;
	\	if CH=SP ! CH=^, ! CH=^. ! CH=CR then quit;
		WD(I):= CH;
		if I < 80-1 then I:= I +1;
		CH:= CHIN(DEV);
		end;
	WD(I):= 0;		\String termination
	WD2:= ALLOCATE(I+1);
	STRCOPY(WD, WD2);
	return WD2;
	end;	\GETWORD


begin	\GETSEN
L:= NIL;
loop	begin
	L:= INSERT(GETWORD, L);
	if CH=^. ! CH=^? ! CH=^! ! CH=EOF then quit;
	end;
if CH=EOF then EOFFOUND:= true;
return REVERSE(L);
end;	\GETSEN



proc	DUMP(DEV, L);	\Print a list of lists on the specified device
int	DEV, L;
begin
if L # NIL then
	begin		\Use recursion to output the lists in reverse order
	DUMP(DEV, L(1));
	PRINT(DEV, L(0));
	CRLF(DEV);
	end;
end;	\DUMP

\----------------------------------------------------------------------

proc	READ;		\Read a text file and add any (new) facts to KNOWLEDGE
int	SENTENCE,	\A list of words
	;
begin
OPENI(INDEV);
EOFFOUND:= false;
TEXT(TV, "READING...
");
while ~EOFFOUND do
	begin
	SRAM:= RAM;
	SENTENCE:= GETSEN(INDEV);	\Read in a sentence from the input device

	\If the sentence contains "is", "are" etc. then save the sentence in KNOWLEDGE
	case of
	    MEMBER("IS", SENTENCE),
	    MEMBER("ARE", SENTENCE),
	    MEMBER("WAS", SENTENCE),
	    MEMBER("WERE", SENTENCE),
	    MEMBER("HAS", SENTENCE),
	    MEMBER("HAVE", SENTENCE):
		KNOWLEDGE:= INSERT(SENTENCE, KNOWLEDGE)
	other RAM:= SRAM;	\Throw out the garbage before it piles up
	end;

KNOWLEDGE:= REVERSE(KNOWLEDGE);
CRLF(TV);
DUMP(TV, KNOWLEDGE);		\Print all saved sentences

TEXT(TV, "
THAT WAS VERY IN-TER-ES-TING!  I NOW KNOW ");
INTOUT(TV, SIZEOF(KNOWLEDGE));
TEXT(TV, " FACTS.
");
end;	\READ

\----------------------------------------------------------------------

proc	QUESTION;	\Question-answer session
int	SENTENCE,	\A list of words
	NEWSEN,		\Copy of SENTENCE
	L,
	;
begin
SRAM:= RAM;
SENTENCE:= GETSEN(0);	\For example: What is blue?

\Assume first word is "WHAT", "WHICH", "WHO", "WHEN", "WHY", "HOW", ETC.

\Wack off everything up to and including the "IS", etc. e.g: blue?
NEWSEN:= COPYOF(SENTENCE);
L:= NEWSEN;
loop	begin
	if L = NIL then quit;
	case of
	    STREQU(L(0), "IS"),
	    STREQU(L(0), "ARE"),
	    STREQU(L(0), "WAS"),
	    STREQU(L(0), "WERE"),
	    STREQU(L(0), "HAS"),
	    STREQU(L(0), "HAVE"):
		begin
		NEWSEN:= L(1);
		quit;
		end
	other;
	L:= L(1);
	end;

\Search KNOWLEDGE for string following "IS" etc.
L:= KNOWLEDGE;
loop	begin
	if L = NIL then
		begin
		TEXT(TV, "I DON'T KNOW. TELL ME ");	\What is blue?
		PRINT(TV, SENTENCE);   CHOUT(TV, ^?);   CRLF(TV);
		RAM:= SRAM;		\Discard question (garbage)

		SENTENCE:= GETSEN(0);	\For example: The ball is blue.
		KNOWLEDGE:= INSERT(SENTENCE, KNOWLEDGE);
		quit;
		end;

	if SEARCH(NEWSEN, L(0)) then	\If found then print KNOWLEDGE item
		begin
		PRINT(TV, L(0));   CHOUT(TV, ^.);   CRLF(TV);
		RAM:= SRAM;		\Discard question
		quit;			\(Might have multiple answers
		end;			\ but don't say "I DON'T KNOW")

	L:= L(1);
	end;
end;	\QUESTION

\----------------------------------------------------------------------

begin	\MAIN
RAM:= RESERVE(FREE-4000);		\Reserve memory KNOWLEDGE, etc.
TOPMEM:= RESERVE(0);			\Pointer to top (end) of available memory

TEXT(TV,"
-- STUDENT, V0.0 --
");

KNOWLEDGE:= NIL;

READ;

TEXT(TV, "
GO AHEAD, ASK ME A QUESTION ABOUT THE STUFF I READ.
");
loop	QUESTION;
end;	\MAIN
KNOWLEDGE:= NIL;

READ;

TEXT(TV, "
GO AHEAD, ASK ME A QUESTION ABOUT THE STUFF I READ.
");
loop	QUESTION;
en