\STUDENT.XPL	AUG-25-90
\In the LISP tradition, words are atoms and sentences are lists.

code	ABS=0,		RAN=1,		REM=2,		RESERVE=3,
	SWAP=4,		EXTEND=5,	RESTART=6,	CHIN=7,
	CHOUT=8,	CRLF=9,		INTIN=10,	INTOUT=11,
	TEXT=12,	OPENI=13,	OPENO=14,	CLOSE=15,
	ABORT=16,	TRAP=17,	FREE=18,	RERUN=19,
	GETHP=20,	SETHP=21,	GETERR=22,	CURSOR=23,
	SCAN=24,	SETRUN=25,	HEXIN=26,	HEXOUT=27,
	CHAIN=28,	OPENF=29,	WRITE=30,	READ=31,
	\TESTPT=32,\	FGET=33,	\FASAVE=34,\	FSAVE=35,
	BLIT=36,	BUTTON=37,	JOYSTICK=38,	\SOUND=39,\
	CLEAR=40,	POINT=41,	LINE=42,	MOVE=43,
	\SCREEN=44,\	\BLOCK=45,\	FIX=50,		BACKUP=64,
	WAITVB=107,
	BITMAP=108,	BITMAP2=109,	VIEW=110,	PALETTE=111,
	CARRY=112,	PEEK_W=113,	POKE_W=114,	PEEK_L=115,
	POKE_L=116,	SWAP_W=117,	EXT_L=118,	CURSOR1=119,
	BUTES1=120,	SHOCUR1=121,	DEVINFO=122,	UNTINFO=123,
	BUTES=124,	GETKEY=125,	KEYHIT=126,	SHOCUR=127;

def	BEL=$07, CR=$0D, EOF=$1A, ESC=$1B, SP=$20;
def	TV=0;
def	INTSIZE=4;
def	INF=$7FFFFFFF, NIL=0;

int	SENTENCE,
	DATA,
	SRAM,
	;
addr	TOPMEM,		\Top of usable memory (heap) space +1
	RAM;		\Pointer to allocated heap space



func	ALLOCATE(AMOUNT);
\This reserves memory, but, unlike the RESERVE intrinsic, this doesn't
\ relinquish the space when the calling procedure returns.
int	AMOUNT;
addr	TEMP;
begin
TEMP:= RAM;
RAM:= RAM + AMOUNT;			\Reserve bytes
if RAM >= TOPMEM then
	begin
	CHOUT(TV, BEL);
	CRLF(TV);   TEXT(TV, "OUT OF MEMORY");   CRLF(TV);
	exit;
	end;
return TEMP;
end;	\ALLOCATE



func	CAR(L);
int	L;
begin
if L = NIL then return NIL;
return L(0);
end;	\CAR



func	CDR(L);
int	L;
begin
if L = NIL then return NIL;
return L(1);
end;	\CDR



func	CONS(A, L);	\Insert atom A at the head of the linked list L
int	A, L;		\This returns the new head of the list, thus the
int	L2, I;		\ form of the call is: L:= CONS(A, L);
begin
L2:= ALLOCATE(2*INTSIZE);
\The 68000 requires even addresses for pointers to words
if L2 & 1 then [L2:= L2 +1;   I:= ALLOCATE(1)];

L2(0):= A;
L2(1):= L;
return L2;
end;	\CONS



proc	PRINT(DEV, L);	\Print the linked list L on the specified device
int	DEV, L;		\This prints the last atom first
begin
if L = NIL then
	CRLF(DEV)
else	begin
	PRINT(DEV, CDR(L));
	TEXT(DEV, CAR(L));
	CHOUT(DEV, SP);
	end;
end;	\PRINT



proc	STRCOPY(A, B);	\Copy string A into B (includes string terminator)
addr	A, B;
int	I;
begin
for I:= 0, INF do
	[B(I):= A(I);
	if A(I) = 0 then return];
end;	\STRCOPY



func	STREQ(A, B);	\Returns 'true' if string A is equal to B.
\Note that strings must be the same length to be equal.
addr	A, B;
int	I;
begin
for I:= 0, INF do
	[if A(I) # B(I) then return false;
	if A(I) = 0 then return true];
end;	\STREQ



func	MEMBER(A, L);	\Returns 'true' if atom A is in linked list L
int	A, L;
begin
loop	[if L = NIL then return false;
	if STREQ(A, L(0)) then return true;
	L:= L(1)];
end;	\MEMBER



func	SIZEOF(L);	\Returns the number of items in a linked list
int	L;
int	I;
begin
for I:= 0, INF do
	[if L = NIL then return I;
	L:= L(1)];
end;	\SIZEOF



func	GETSEN(DEV);
\Read in a sentence and return a pointer to its list of words.
\Currently, sentences must be terminated with a period.
int	DEV;
int	CH, L, L2;


	func	GETWORD;
	int	I;
	addr	WD, WD2;
	begin
	WD:= RESERVE(80);
	repeat CH:= CHIN(DEV); until CH#SP & CH#CR;
	I:= 0;
	loop	begin
		if CH>=^a & CH<=^z then CH:= CH & $DF;	\UPPERCASE
		if CH<^A ! CH>^Z then quit;
	\	if CH=SP ! CH=^, ! CH=^. ! CH=CR then quit;
		WD(I):= CH;
		if I < 80-1 then I:= I +1;
		CH:= CHIN(DEV);
		end;
	WD(I):= 0;		\TERMINATE STRING
	WD2:= ALLOCATE(I+1);
	STRCOPY(WD, WD2);
	return WD2;
	end;	\GETWORD


begin	\GETSEN
L:= NIL;
loop	begin
	L:= CONS(GETWORD, L);
	if CH=^. then quit;
	end;
return L;
end;	\GETSEN



proc	DUMP(DEV, L);	\Print a list of lists on the specified device
int	DEV, L;		\This prints the last list first
begin
if L # NIL then
	begin
	DUMP(DEV, CDR(L));
	PRINT(DEV, CAR(L));
	end;
end;	\DUMP



begin	\MAIN
RAM:= RESERVE(FREE-2000);		\Reserve memory for tree & lists
TOPMEM:= RESERVE(0);			\Pointer to top of usable memory

TEXT(TV,"-- STUDENT, V0.0 --

TYPE IN A SENTENCE ENDING WITH A PERIOD.
");

DATA:= NIL;
loop	begin
	SRAM:= RAM;
	SENTENCE:= GETSEN(0);	\Read in a sentence of the input file
	PRINT(TV, SENTENCE);

	INTOUT(0, SIZEOF(SENTENCE));   CRLF(0);
	if SIZEOF(SENTENCE) <= 1 then quit;

	\If the sentence contains "is" or "are" then save the sentence
	if MEMBER("IS", SENTENCE) ! MEMBER("ARE", SENTENCE) then
		DATA:= CONS(SENTENCE, DATA)
	else RAM:= SRAM;	\Discard the garbage
	end;

\Print all saved sentences
CRLF(TV);
DUMP(TV, DATA);
end;	\MAIN
