\TM2.XPL	FEB-06-96
\Texture Mapping Program by Loren Blaney

\A right-hand coordinate system is used:
\
\			       |   Z  (into screen)
\			       |  /
\			       | /
\			       |/
\		       --------+-------- X
\			      /|
\			     / |
\			    /  |
\			       |
\			       Y
\

include	\CXPL\CODESI;

int	VCurX, VCurY;	\Character cursor coordinates in pixels
int	VLineX, VLineY;	\Line coordinates (VMove and end of last line drawn)

def	VWidth = 640,	\VESA display dimensions in pixels
	VHeight = 480,
	VPoints = 11;	\Character point size (height in pixels) (see font table)

def	IntSize= 4;	\Number of bytes in an integer (2 or 4)

ext	BestColor(Palette, R, G, B),	\Find closest color in Palette
	GetTime,	\Get 32-bits of system timer counts
	HLine,		\Draw horizontal line
	VShow(ScreenBuf, GranShift);	\Display 640x480x8 image at ScreenBuf

char	ImageData,	\Array holding copy of PCX image (the texture map)
	Palette,	\256 color registers. Each color is 3 bytes (RGB).
	ScreenBuf,	\Array holding copy of screen
	VESAInfo;	\SVGA mode information returned by function 1 (=256 bytes)

int	CLUT,		\Array: Color Look-Up Table: 64 intensities of 256 colors
	Cpureg,		\Address of CPU register array (from Getreg)
	DataSeg,	\Our data segment address
	GranShift,	\Bits to shift to compensate for VESA granularity
	HaveMouse,	\A mouse driver is installed and mouse works
	Lambert,	\Counter: Lambert lighting vs. colored facets vs. hilights
	OrigVidMode,	\Original video mode (to be restored upon exit)
	Solid,		\Flag: do solid filled polygons as opposed to lines
	Texturize,	\Flag: do texture fill of polygons
	PSPSeg,		\Our PSP segment address
	II;		\Scratch for Main

real	Ambient;	\Ambient light level

def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0A, FF=$0C,	\Control chars
	CR=$0D, EOF=$1A, Esc=$1B, Sp=$20, Ctrl=$40;

def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\Scan codes

def	Pi = 3.14159265358979323846;

\-------------------------------------------------------------------------------

func real Rad(X);		\Convert degrees to radians
real	X;
return X /57.2957795130823;



func	WExtend(N);		\Sign extend 16-bit word to 32-bit double word
int	N;
return if N & $8000 then N ! $FFFF0000 else N & $0000FFFF;



func	ILn2(N);	\Integer log to base 2
int	N;		\ (this truncates, e.g: ILn2(31)=4)
int	B, I;
begin
B:= $8000;
for I:= -15, 0 do
	begin
	if N & B then return -I;
	B:= B>>1;
	end;
return -1;		\error
end;	\ILn2



proc	WaitVB;		 	\Wait for beginning of next vertical blank
int	SR;
begin
SR:= Peek($40, $63) + $306;	\Port address of status register
if (SR & $39A) # $39A then return;
while Pin(SR, 0) & $08 do;	\Wait for no vertical blank
repeat until Pin(SR, 0) & $08;	\Wait for vertical blank
end;	\WaitVB;



proc	Beep;			\A not-too-obnoxious beep
begin
Sound(false, 1, 1000);
Sound(true, 1, 3000);
end;	\Beep



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(Unused arguments need not be passed)
begin
Cpureg:= Getreg;
Cpureg(0):= AX;
Cpureg(1):= BX;
Cpureg(2):= CX;
Cpureg(3):= DX;
Cpureg(6):= BP;
Cpureg(9):= DS;
Cpureg(11):= ES;
Softint(Int);
return WExtend(Cpureg(0));	\return AX register
end;	\CallInt



func	GetVid;			\Returns the current video mode
return CallInt($10, $0F00) & $FF; \Function $0F



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl-C, and no cursor.
SC:= CallInt($16, $0000);	\Function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\Return non-ascii chars as negative scan code
return Ch;
end;	\GetKey



proc	WaitKey;		\Wait for a keystroke
begin
Openi(1);
repeat until ChkKey;
end;	\WaitKey



proc	SetTimerMode(M);	\Set 8254 timer mode
int	M;
begin
\00110100
Pout($30+M<<1, $43, 0);
Pout($FF, $40, 0);
Pout($FF, $40, 0);
end;	\SetTimerMode

\--------------------------- MOUSE ROUTINES ---------------------------

func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\Pointer is set to center of screen but is hidden
CallInt($21, $3533);		\Make sure mouse vector ($33) points to something
HaveMouse:= false;
if WExtend(Cpureg(1))=0 & WExtend(Cpureg(11))=0 then return false;
HaveMouse:= CallInt($33, $0000); \Reset mouse and get status
if HaveMouse then HaveMouse:= true; \(Beware of 'not' operator in 32-bit XPL)
return HaveMouse;		\Return 'false' if failure
end;	\OpenMouse



proc	ShowMouse(On);		\Turn mouse pointer on or off
\The video mode should be set before calling this routine.
\This counts the number of times the pointer is turned on or off. If the
\ pointer is turned on twice, it must be turned off twice before it
\ actually goes off. The pointer should be turned off before drawing over
\ it and before your program exits. Setting the video mode will also turn
\ off the pointer.
int	On;	\Flag: True = pointer on; False = pointer off
if HaveMouse then
	CallInt($33, if On then $0001 else $0002);



func	GetMousePosition(N);	\Return position of specified mouse coordinate
int	N;	\0 = X coordinate; 1 = Y coordinate
\For video modes $0-$E and $13 the maximum coordinates are 639x199, minus
\ the size of the pointer. For modes $F-$12 the coordinates are the same as
\ the pixels. For 80-column text modes divide the mouse coordinates by 8 to
\ get the character cursor position.
begin
if ~HaveMouse then return 0;
CallInt($33, $0003);
return if N then WExtend(Cpureg(3)) else WExtend(Cpureg(2));
end;	\GetMousePosition



func	GetMouseButton(N);	\Return 'true' if specified mouse button is down
int	N;	\Button number: 0 = left; 1 = right (or middle)
begin
if ~HaveMouse then return false;
CallInt($33, $0003);
return if N then (Cpureg(1)&2)=2 else (Cpureg(1)&1)=1;
end;	\GetMouseButton



proc	MoveMouse(X, Y);	\Move mouse pointer to X,Y
int	X, Y;
if HaveMouse then
	CallInt($33, $0004, 0, X, Y);

\-------------------------------------------------------------------------------

proc	VPoint(X, Y, C);	\Plot a point in ScreenBuf
int	X, Y, C;		\Coordinates, color
begin
if X < 0 then return;		\clip
if X >= VWidth then return;
if Y < 0 then return;
if Y >= VHeight then return;

ScreenBuf(Y*VWidth+X):= C;
end;	\VPoint



proc	VLine(X1, Y1, C);	\Draw line from VLineX,VLineY to X1,Y1
\This plots lines from both ends. Lines with an odd number of pixels will
\ have their middle pixel plotted twice. This would be a problem if XORed
\ lines were used.
\WARNING: This does not clip lines and assumes that lines do not extend beyond
\ ScreenBuf.
int	X1, Y1, C;	\Coordinates of end-of-line, Color (0-255)
int	DX, DY,		\Delta X, Delta Y
	XI, YI,		\Plotting increments along X and Y axes
	S, SD,		\Step (along X or Y), Step Diagonal (along X and Y)
	D,		\Decision variable (move straight or diagonally?)
	E, NE,		\East, Northeast (straight or diagonal)
	I;
char	P, Q;		\Pointers to ends of lines, Q plots from opposite end
begin
DX:= X1 - VLineX;
if DX >= 0 then
	XI:= 1
else	[DX:= -DX;   XI:= -1];

DY:= Y1 - VLineY;
if DY >= 0 then
	YI:= VWidth
else	[DY:= -DY;   YI:= -VWidth];

SD:= XI + YI;
if DX >= DY then
	S:= XI
else	begin
	S:= YI;
	I:= DX;   DX:= DY;   DY:= I;
	end;

D:= 2*DY - DX;					\Set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

P:= ScreenBuf + VLineY*VWidth + VLineX;		\Set up pointers
Q:= ScreenBuf + Y1*VWidth + X1;

P(0):= C;			\Plot initial point
Q(0):= C;
for I:= 1, DX>>1 do
	begin
	if D < 0 then
		begin		\Move straight
		D:= D + E;
		P:= P + S;
		Q:= Q - S;
		end
	else	begin		\Move diagonally
		D:= D + NE;
		P:= P + SD;
		Q:= Q - SD;
		end;
	P(0):= C;		\Plot point
	Q(0):= C;
	end;

VLineX:= X1;   VLineY:= Y1;	\End of current line is start of next line
end;	\VLine



proc	VMove(X, Y);		\Move pen to the beginning of a line
int	X, Y;
begin
VLineX:= X;
VLineY:= Y;
end;	\VMove

\-------------------------------------------------------------------------------

proc	VPoly(XX, YY, A, C);	\Draw a (filled) polygon
int	XX, YY,	\Location to plot polygon (coordinates of first point)
	A,	\One-dimensional array of X,Y coordinates with first value
		\ giving the number of points (X,Y pairs)
	C;	\Fill color
int	N, BASE, P, MEMPTR, \MEM0\;
int	X0, Y0, X1, Y1;
int	I, J,
	Y,
	X00, Y00,	\Coordinates (absolute) of first vertex in polygon
	MIN, MAX,	\Maximum and minimum (absolute) Y coordinate of polygon
	HEIGHT;		\Height of polygon in pixels
def	NIL= $80000000;
def	SS=12;		\Scale shift factor (one pixel error when slope = 4096)


	proc	PLINE(X0, Y0, X1, Y1);
	\Build table containing coordinates of X points along line. This is a
	\ linked list header for each horizontal line through the polygon.
	int	X0, Y0, X1, Y1;
	int	X, Y, DX, DY, M, T, Q;
	begin
	if Y0 > Y1 then				\Swap ends
		begin
		T:= Y0;   Y0:= Y1;   Y1:= T;
		T:= X0;   X0:= X1;   X1:= T;
		end;

	DY:= Y1 - Y0;
	if DY = 0 then return;
	DX:= X1 - X0;
	M:= DX<<SS /DY;				\Scaled slope

	X:= X0<<SS;
	for Y:= Y0-MIN, Y1-MIN-1 do
		begin
		T:= X>>SS;			\Unscale X
		P:= BASE + Y*IntSize;
		loop	begin
			case of P(0)=NIL, P(0,1)>T:
				begin		\Insert item at front of list
				Q:= MEMPTR;
				MEMPTR:= MEMPTR +2*IntSize;
				Q(0):= P(0);
				Q(1):= T;
				P(0):= Q;
				quit;
				end
			other	P:= P(0);	\Follow pointers
			end;
		X:= X +M;
		end;
	end;	\PLINE


begin	\VPoly
\Find maximum and minimum (absolute) Y coordinate
N:= A(0);				\Get number of coordinates
MIN:= $7FFFFFFF;   MAX:= $80000000;
I:= 2;					\Index of first Y coordinate in A
for J:= 1, N do
	begin
	Y:= YY +A(I);
	if Y > MAX then MAX:= Y;
	if Y < MIN then MIN:= Y;
	I:= I +2;			\Next Y coordinate
	end;
HEIGHT:= MAX - MIN + 1;

\Set up array of list headers for each horizontal scan line in the polygon.
\ The list will contain the X coordinates (lowest value first) of the edges.
BASE:= Reserve(HEIGHT*IntSize);
for Y:= 0, HEIGHT-1 do BASE(Y):= NIL;

MEMPTR:= RESERVE(10000);		\Set up our own data space (or remove nested proc)
\MEM0:= MEMPTR;

I:= 1;					\Index of first coordinate in A
X0:= XX +A(I);   X00:= X0;   I:= I +1;	\Save first vertex of polygon in X00,Y00
Y0:= YY +A(I);   Y00:= Y0;   I:= I +1;
for J:= 2, N do				\Get rest of vertices
	begin
	X1:= XX +A(I);   I:= I +1;
	Y1:= YY +A(I);   I:= I +1;
	Pline(X0, Y0, X1, Y1);		\Draw edge of polygon between vertices
	X0:= X1;   Y0:= Y1;
	end;
Pline(X0, Y0, X00, Y00);		\Make sure the polygon is closed

for Y:= 0, HEIGHT-1 do			\Draw horizontal lines to fill polygon
	begin
	P:= BASE + Y*IntSize;
	loop	begin
		P:= P(0);
		if P = NIL then quit;
		X0:= P(1);
		P:= P(0);
		if P = NIL then quit;
		if X0#P(1) then HLine(ScreenBuf, X0, MIN+Y, P(1)-1, C);
		end;
	end;
end;	\VPoly

\-------------------------------------------------------------------------------

proc	VTPoly(XX, YY, A, TM, Intensity);	\Draw a texture-mapped polygon
int	XX, YY,	\Location to plot polygon (coordinates of first point)
	A,	\One-dimensional array of X,Y coordinates with first value
		\ giving the number of points (X,Y pairs)
	TM,
	Intensity; \Light intensity 0..63
int	N, Base, P, MemPtr;
int	X0, X1;
int	I, J,
	X, Y,
	I00,
	Min, Max,	\Maximum and minimum (absolute) Y coordinate of polygon
	Height;		\Height of polygon in pixels
char	CLUT_Intensity;	\Array: 256 bytes, color look up table for given intensity
def	Nil= $80000000;
def	SS=12;		\Scale shift factor (one pixel error when slope = 4096)

int	PY, QY, RY, SY,
	Tpx, Tpy, Tqx, Tqy, Trx, Try, Tsx, Tsy,
	C0, C1,
	Tx0, Ty0, Tx1, Ty1,
	Tdx, Tdy,
	Tx, Ty;
char	Texture;


	proc	PLine(I0, I1);
	\Build table containing coordinates of X points along line. This is a
	\ linked list header for each horizontal line through the polygon.
	int	I0, I1;
	int	X0, Y0, X1, Y1;
	int	X, Y, DX, DY, M, T, Q;
	begin
	X0:= A(I0) + XX;   Y0:= A(I0+1) + YY;
	X1:= A(I1) + XX;   Y1:= A(I1+1) + YY;

	if Y0 > Y1 then				\Swap ends
		begin
		T:= Y0;   Y0:= Y1;   Y1:= T;
		T:= X0;   X0:= X1;   X1:= T;
		T:= I0;   I0:= I1;   I1:= T;
		end;

	DY:= Y1 - Y0;
	if DY = 0 then return;
	DX:= X1 - X0;
	M:= DX<<SS /DY;				\Scaled slope

	X:= X0<<SS;
	for Y:= Y0-MIN, Y1-MIN-1 do
		begin
		T:= X>>SS;			\Unscale X
		P:= Base + Y*IntSize;
		loop	begin
			case of P(0)=Nil, P(0,1)>T:
				begin		\Insert item at front of list
				Q:= MemPtr;
				MemPtr:= MemPtr + 12;	\3*IntSize;
				Q(0):= P(0);
				Q(1):= T;	\Unscaled X
				Q(2):= I0 + I1<<8; \Indicies of ends of line segment
				P(0):= Q;
				quit;
				end
			other	P:= P(0);	\Follow pointers
			end;
		X:= X +M;
		end;
	end;	\PLine


begin	\VTPoly
\Find maximum and minimum (absolute) Y coordinate
N:= A(0);				\Get number of coordinates
Min:= $7FFFFFFF;   Max:= $80000000;
I:= 2;					\Index of first Y coordinate in A
for J:= 1, N do
	begin
	Y:= YY +A(I);
	if Y > Max then Max:= Y;
	if Y < Min then Min:= Y;
	I:= I +2;			\Next Y coordinate
	end;
Height:= Max - Min + 1;

\Set up array of list headers for each horizontal scan line in the polygon.
\ The list will contain the X coordinates (lowest value first) of the edges,
\ along with the index to the coordinate for the start and end of each edge.
Base:= Reserve(Height*IntSize);
for Y:= 0, Height-1 do Base(Y):= Nil;

MemPtr:= Reserve(12000);		\Set up our own data space (or remove nested proc)

I:= 1;					\Index of first coordinate in A
I00:= I;				\Save index of first vertex of polygon
I:= I +2;
for J:= 2, N do				\Get rest of vertices
	begin
	I:= I +2;
	Pline(I-4, I-2);		\Draw edge of polygon between vertices
	end;
Pline(I-2, I00);			\Make sure the polygon is closed

\    0  1  2  3  4  5  6  7  8
\A:  N  PX PY QX QY RX RY SX SY
\TM: t  px py qx qy rx ry sx sy

CLUT_Intensity:= CLUT(Intensity); 	\Intensity range: 0..63 ADDRESS of 256 byte table
Texture:= TM(0);
for Y:= Min, Min+Height-1 do			\Draw horizontal lines to fill polygon
	begin
	P:= Base + (Y-Min)*IntSize;
	loop	begin
		P:= P(0);
		if P = Nil then quit;
		X0:= P(1);

		I:= P(2) & $00FF;
		SY:= A(I+1) + YY;
		Tsx:= TM(I);
		Tsy:= TM(I+1);

		I:= P(2)>>8;
		RY:= A(I+1) + YY;
		Trx:= TM(I);
		Try:= TM(I+1);

		P:= P(0);
		if P = Nil then quit;
		X1:= P(1);

		if X0 # X1 then
			begin
			I:= P(2) & $00FF;
			PY:= A(I+1) + YY;
			Tpx:= TM(I);
			Tpy:= TM(I+1);

			I:= P(2)>>8;
			QY:= A(I+1) + YY;
			Tqx:= TM(I);
			Tqy:= TM(I+1);

			C0:= (Y-RY)<<SS / (SY-RY);
			C1:= (Y-QY)<<SS / (PY-QY);

			Tx0:= C0 * (Tsx-Trx) + Trx<<SS;
			Ty0:= C0 * (Tsy-Try) + Try<<SS;
			Tx1:= C1 * (Tpx-Tqx) + Tqx<<SS;
			Ty1:= C1 * (Tpy-Tqy) + Tqy<<SS;

			Tdx:= (Tx1-Tx0) / (X1-X0);
			Tdy:= (Ty1-Ty0) / (X1-X0);

			Tx:= Tx0;	\Coordinates of start of line in texture map
			Ty:= Ty0;
			for X:= X0, X1-1 do	\Step one pixel at a time horizontally
				begin
				\ScreenBuf(Y*VWidth+X):= Texture(Tx>>SS+320*(Ty>>SS));
				ScreenBuf(Y*VWidth+X):= 
					CLUT_Intensity(Texture(Tx>>SS+320*(Ty>>SS)));
				Tx:= Tx + Tdx;	\Coordinates of next pixel in texture map
				Ty:= Ty + Tdy;
				end;
			end;
		end;
	end;
end;	\VTPoly

\-------------------------------------------------------------------------------

proc	VCrlf;			\Move cursor to the start of a new line
begin
VCurX:= 0;
VCurY:= VCurY + VPoints;
if VCurY+VPoints >= VHeight then VCurY:= 0;
end;	\VCrlf



proc	VCursor(X, Y);		\Set cursor coordinates (in character cells)
int	X, Y;
begin
VCurX:= X * 8;
VCurY:= Y * VPoints;
end;	\VCursor



proc	VChout(Ch, C);		\Plot character at current cursor location
int	Ch, C;	\character, color
\Also inputs: ScreenBuf, VPoints, VWidth, VHeight
int	I, J,	\indices
	B,	\byte from font table
	M,	\bit mask
	FT,
	;
char	A;
begin	\VChout
if Ch = $0D then [VCurX:= 0;   return];
if Ch = $0A then
	[VCurY:= VCurY + VPoints;
	if VCurY+VPoints >= VHeight then VCurY:= 0;
	return];

include	FONTTBL;

A:= ScreenBuf + VCurX + VCurY*VWidth;
for J:= 0, VPoints-1 do
	begin
	B:= FT(Ch*VPoints+J);			\Get byte from font table
	M:= $80;
	for I:= 0, 8-1 do
		begin				\Background is transparent
		if M & B then A(I):= C;
		M:= M >> 1;
		end;
	A:= A + VWidth;
	end;

VCurX:= VCurX + 8;				\Advance cursor
if VCurX+8 >= VWidth then VCrlf;
end;	\VChout



proc	VText(Str, C);		\Plot a string of characters at current cursor
char	Str;
int	C;
int	I, Ch;
begin
for I:= 0, 30000 do
	begin
	Ch:= Str(I);
	if Ch >= $80 then [Ch:= Ch&$7f;   I:= 30000];
	VChout(Ch, C);
	end;
end;	\VText



proc	VIntout(N, C);		\Plot the ascii representation of an integer
int	N, C;
int	Ch;
begin
Openo(8);
Intout(8, N);
Chout(8, ^,);
Openi(8);
loop	begin
	Ch:= Chin(8);
	if Ch = ^, then quit;
	VChout(Ch, C);
	end;
end;	\VIntout



proc	VRlout(N, C);		\Plot the ascii representation of a real
real	N;
int	C;
int	Ch;
begin
Openo(8);
Rlout(8, N);
Chout(8, ^,);
Openi(8);
loop	begin
	Ch:= Chin(8);
	if Ch = ^, then quit;
	VChout(Ch, C);
	end;
end;	\VRlout

\-------------------------------------------------------------------------------

proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
def	X, Y, Z;
begin
V3(X):= V1(Y)*V2(Z) - V1(Z)*V2(Y);
V3(Y):= V1(Z)*V2(X) - V1(X)*V2(Z);
V3(Z):= V1(X)*V2(Y) - V1(Y)*V2(X);
end;	\CrossProd



func real DotProd(V1, V2);	\Return the dot product of two 3D vectors
real	V1, V2;			\V1  V2
def	X, Y, Z;
return V1(X)*V2(X) + V1(Y)*V2(Y) + V1(Z)*V2(Z);



proc	Unitize(V);		\Convert a 3D vector to its unit vector
real	V;	\Vector to convert
real	M;	\Magnitude of vector
def	X, Y, Z;
begin
M:= Sqrt( V(X)*V(X) + V(Y)*V(Y) + V(Z)*V(Z) );
V(X):= V(X) / M;
V(Y):= V(Y) / M;
V(Z):= V(Z) / M;
end;	\Unitize



proc    Rotate(V, R, P, W); \3D rotate vector V
real    V,              \3D vector is 3-element array with X, Y and Z components
	R, P, W;        \Roll, Pitch and yaW (radians)
real    X, Y, Z, SW, SP, SR, CW, CP, CR, T;
begin
X:= V(0);   Y:= V(1);   Z:= V(2);	\get vector components
SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
\                               
\                     1   0   0 
\ [X Y Z]:= [X Y Z] * 0   CR  SR
\                     0  -SR  CR
\                               
T:= Y*CR - Z*SR;
Z:= Z*CR + Y*SR;
Y:= T;

\Rotate about Y axis (Pitch):
\                               
\                     CP  0  -SP
\ [X Y Z]:= [X Y Z] * 0   1   0 
\                     SP  0   CP
\                               
T:= X*CP + Z*SP;
Z:= Z*CP - X*SP;
X:= T;

\Rotate about Z axis (yaW):
\                                
\                      CW  SW  0 
\ [X Y Z]:= [X Y Z] * -SW  CW  0 
\                      0   0   1 
\                                
T:= X*CW - Y*SW;
Y:= Y*CW + X*SW;
X:= T;

V(0):= X;   V(1):= Y;   V(2):= Z;       \return values
end;    \Rotate

\===============================================================================

proc	MoveTorus;	\Display and move a torus
def	TColor=$FF;	\Text Color [0..$FF]
def	Max=20;		\Maximum values for Sectors and Sections
real	TorusR1,	\Distance from the center of the hole in the middle
			\ to the center of the eatable tube-shaped part.
	TorusR2,	\The distance from the center of the eatable tube-
			\ shaped part to the frosting on top.
	TorusX, TorusY, TorusZ,	\Arrays of coordinates of vertices (2-dim)
	TorusNX, TorusNY, TorusNZ, \Arrays of coordinates of normals (2-dim)
	TNX, TNY, TNZ;
real	Ref1, Ref2, Ref3,	\Arrays: 3 unit vectors forming reference frame
	Zoom,		\Zoom factor
	;
int	Sectors,	\Slice of torus (tube)
	Sections,	\Slice of a sector (patch)
	Facets,		\Number of facets
	ITorusX, ITorusY, ITorusZ,	\Arrays of coordinates of vertices (2-dim)
	ITorusZZ,	\Sorted Z coordinates (1-dimensional array)
	ITorusIJ,	\Array accompanying ITorusZZ containing I and J indices
	II;		\Scratch

int	Time, Time0;
real	FPS;

\-------------------------------------------------------------------------------

proc	DrawTorus;	\Draw the torus
int	I, J, K,	\Indices
	Color,
	Intensity,
	PA,		\Array: Polygon corners for a facet
	TM;
real	Light,		\Light vector. Unit vector specifies direction of light
	N,		\Normal vector
	DP, DP2, DP4;
def	CX=640/2, CY=480/2;	\Coordinates of center of the screen
def	X, Y, Z;
\Inputs: Sectors, Sections, ITorusX, ITorusY, Facets, ITorusIJ
begin
\DRAW POLYGONS:
if Solid then
begin
PA:= Reserve(IntSize + 4*2*IntSize);	\Space for a rectangular polygon
PA(0):= 4;				\4 corners

N:= Rlres(3);
Light:= [0.5, 1.0, 1.0];		\Light direction
Unitize(Light);

\Draw facets starting with the back-most one (the one with the greatest Z)
for K:= 0, Facets-1 do
	begin
	I:= ITorusIJ(K);			\extract indices
	J:= I>>8;
	I:= I & $00FF;

	PA(1):= ITorusX(I, J);		PA(2):= ITorusY(I, J);
	PA(3):= ITorusX(I+1, J);	PA(4):= ITorusY(I+1, J);
	PA(5):= ITorusX(I+1, J+1);	PA(6):= ITorusY(I+1, J+1);
	PA(7):= ITorusX(I, J+1);	PA(8):= ITorusY(I, J+1);

	\This determines if a facet is visible. The surface normal of a facet
	\ is the cross product of two vectors determined by adjacent edges. If
	\ the dot product of this normal and the viewing direction [0,0,1] is
	\ negative, the facet is visible. This eliminates about half the polygons.
	if (PA(3)-PA(1)) * (PA(8)-PA(2)) < (PA(7)-PA(1)) * (PA(4)-PA(2)) then
		begin
		if Lambert \#0\ then
			begin
			\VN
			\[0,0,1]N
			\if TNZ(I,J) < 0 then \\visible

			N(X):= TNX(I,J);
			N(Y):= TNY(I,J);
			N(Z):= TNZ(I,J);

			case Lambert of
			  1:	begin
			  	Color:= Fix((189.0-Ambient) * -DotProd(N, Light) +
					Ambient) + 64;		\189=3*64-1-2
				end;
			  2:	begin
				DP:= -DotProd(N, Light);
				DP2:= DP * DP;
				DP4:= DP2 * DP2;
				Color:= Fix((100.0-Ambient/2.0) * DP +
					(189.0 - 100.0) * DP4*DP4*DP4*DP4 +
					Ambient/2.0) + 64;
				end
			other	[];

			if Color < 64 then Color:= 64;	\fix some rounding glitches
			end
		else	Color:= Rem((I+J+J)/63)+1;

		if Texturize then
			begin
			N(X):= TNX(I,J);
			N(Y):= TNY(I,J);
			N(Z):= TNZ(I,J);

\			TM:= [0, 0,0, 0,199, 319,199, 319,0];

			if J & 1 then
				begin
				if I & 1 then
					TM:= [0, 0,100, 0,199, 159,199, 159,100]	\4
				else	TM:= [0, 0,0, 0,99, 159,99, 159,0];		\1
				end
			else	begin
				if I & 1 then
					TM:= [0, 160,100, 160,199, 319,199, 319,100]	\3
				else	TM:= [0, 160,0, 160,99, 319,99, 319,0];		\2
				end;

			TM(0):= ImageData;
		  	Intensity:= Fix((63.0-Ambient) * -DotProd(N, Light) + Ambient);
		  	if Intensity > 63 then Intensity:= 63;
		  	if Intensity < 0 then Intensity:= 0;

			VTPoly(CX, CY, PA, TM, Intensity);
			end
		else	VPoly(CX, CY, PA, Color);
		end;
	end;
end
else	begin
\DRAW LINES:
Color:= 3;
for J:= 0, Sections-1 do
	begin
	VMove(ITorusX(0,J)+CX, ITorusY(0,J)+CY);
	for I:= 1, Sectors do
		VLine(ITorusX(I,J)+CX, ITorusY(I,J)+CY, Color);
	end;

for I:= 0, Sectors-1 do
	begin
	VMove(ITorusX(I,0)+CX, ITorusY(I,0)+CY);
	for J:= 1, Sections do
		VLine(ITorusX(I,J)+CX, ITorusY(I,J)+CY, Color);
	end;
end;
end;	\DrawTorus

\-------------------------------------------------------------------------------

proc	Sort(A, IA, N);		\Sort arrays into descending Z coordinates
int	A,			\Address of array of Z coordinates
	IA,			\Address of array of I and J coordinates
	N;			\Number of elements in array (size)
int	J, Gap, I, JG, T;
begin
Gap:= N /2;
while Gap > 0 do
	begin
	for I:= Gap, N-1 do
		begin
		J:= I -Gap;
		loop	begin
			if J < 0 then quit;
			JG:= J +Gap;
			if A(J) >= A(JG) then quit;
			T:= A(J);    A(J):= A(JG);    A(JG):= T;   \Swap
			T:= IA(J);  IA(J):= IA(JG);  IA(JG):= T;
			J:= J -Gap;
			end;
		end;
		Gap:= Gap /2;
	end;
end;	\Sort



proc	DepthSort;
\Find the Z coordinate (*2) of the center of each facet. Save result in
\ array ITorusZZ. Array ITorusIJ saves I & J coordinates for corresponding
\ facet. Z increases going into the screen.
\Inputs: Sectors, Sections, ITorusZ
\Outputs: ITorusZZ, ITorusIJ, Facets
int	I, J;
begin
Facets:= 0;
for I:= 0, Sectors-1 do
    for J:= 0, Sections-1 do
	begin
	ITorusZZ(Facets):= ITorusZ(I,J) + ITorusZ(I+1,J+1);
	ITorusIJ(Facets):= I + J<<8;
	Facets:= Facets +1;
	end;
Sort(ITorusZZ, ITorusIJ, Facets);
end;	\DepthSort

\-------------------------------------------------------------------------------

proc	RotateTorus;
\Inputs: Sectors, Sections, TorusX, TorusY, TorusZ
\Outputs: ITorusX, ITorusY, ITorusZ
int	I, J,
	DX, DY, DZ;		\Translational motion
real	AngX, AngY, AngZ,	\Rotation angle about indicated axis (radians)
	Rxx, Rxy, Rxz,		\Rotation matrix
	Ryx, Ryy, Ryz,
	Rzx, Rzy, Rzz;
begin
if GetMouseButton(0) then
	begin
	CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Z axis
	AngZ:= Float(WExtend(Cpureg(2))) / 500.0;	\get angle (in radians) from mouse
	\Mouse Y motion zooms
	Zoom:= Zoom + Float(-WExtend(Cpureg(3))) / 2500.0;
	AngX:= 0.0;
	AngY:= 0.0;
	end
else	begin
	CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Y axis
	AngY:= Float(-WExtend(Cpureg(2))) / 500.0;
	\Mouse Y motion rotates object about X axis
	AngX:= Float(WExtend(Cpureg(3))) / 500.0;
	AngZ:= 0.0;
	end;

\STEP 1:
\Rotate reference frame according to changes in mouse position.
\Independently rotating P1 and P2 can cause them to drift relative to
\ each other, but the amount of drift is insignificant because reals are
\ calculated to 16-place accuracy.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2

\STEP 2:
\Rotate each vertex in the object so that it aligns with the reference frame:
\Translation is done at the same time for efficiency.
DX:= 0;   DY:= 0;   DZ:= 0;

\Replace subscripted variables for efficiency:
Rxx:= Ref1(0);   Rxy:= Ref1(1);   Rxz:= Ref1(2);
Ryx:= Ref2(0);   Ryy:= Ref2(1);   Ryz:= Ref2(2);
Rzx:= Ref3(0);   Rzy:= Ref3(1);   Rzz:= Ref3(2);

\This multiplies each point at its initial position (Po) by the rotation
\ matrix to get the rotated position.
\	                                                    
\	                                  Rxx  Rxy  Rxz 
\	 Px  Py  Pz  =  Pxo  Pyo  Pzo  *  Ryx  Ryy  Ryz 
\	                                  Rzx  Rzy  Rzz 
\	                                                    
\
for I:= 0, Sectors do
  for J:= 0, Sections do
	begin
	ITorusX(I,J):= Fix((TorusX(I,J) *Rxx + TorusY(I,J) *Ryx + TorusZ(I,J) *Rzx)*Zoom) + DX;
	ITorusY(I,J):= Fix((TorusX(I,J) *Rxy + TorusY(I,J) *Ryy + TorusZ(I,J) *Rzy)*Zoom) + DY;
	ITorusZ(I,J):= Fix( TorusX(I,J) *Rxz + TorusY(I,J) *Ryz + TorusZ(I,J) *Rzz) + DZ;

	if Lambert \#0\ then
	  if I<Sectors & J<Sections then
		begin
		TNX(I,J):= TorusNX(I,J) *Rxx + TorusNY(I,J) *Ryx + TorusNZ(I,J) *Rzx;
		TNY(I,J):= TorusNX(I,J) *Rxy + TorusNY(I,J) *Ryy + TorusNZ(I,J) *Rzy;
		TNZ(I,J):= TorusNX(I,J) *Rxz + TorusNY(I,J) *Ryz + TorusNZ(I,J) *Rzz;
		end;
	end;
end;	\RotateTorus

\-------------------------------------------------------------------------------

proc	CalcTorus;
\Find coordinates of vertices of facets that approximate a torus.
\Chop the surface of the torus up into polygons. To simplify later
\ calculations, the last coordinates duplicate the first coordinates.
\ Torus is initially in the X,Y plane (like a donut floating in hot grease).
\Inputs: Sectors, Sections, TorusR1, TorusR2
\Outputs: TorusX, TorusY, TorusZ, TorusNX, TorusNY, TorusNZ
int	I, J;
real	A,	\Angular increments about hole in center
	B,	\Angular increments about center of tube
	SA, CA, SB, CB;	\Sines and Cosines of A and B
real	V1, V2, V3;
def	Pi2 = Pi*2.0;
def	X, Y, Z;
begin
V1:= Rlres(3);
V2:= Rlres(3);
V3:= Rlres(3);

for I:= 0, Sectors do
	begin
	A:= Float(I) / Float(Sectors) * Pi2;
	SA:= Sin(A);   CA:= Cos(A);
	for J:= 0, Sections do
		begin
		B:= Float(J) / Float(Sections) * Pi2;
		SB:= Sin(B);   CB:= Cos(B);
		TorusX(I,J):= (TorusR1 + TorusR2*CB) * CA;
		TorusY(I,J):= (TorusR1 + TorusR2*CB) * SA;
		TorusZ(I,J):= TorusR2 * SB;
		end;
	end;

\Find the normal vectors for each facet. A normal vector is a vector that
\ is one unit long and is perpendicular to the facet. It points outward,
\ away from the inside of the donut (the inside of a facet is the side that
\ can never be seen no matter how the donut is rotated.
for I:= 0, Sectors-1 do
	for J:= 0, Sections-1 do
		begin
		V1(X):= TorusX(I+1,J) - TorusX(I,J);
		V2(X):= TorusX(I,J+1) - TorusX(I,J);
		V1(Y):= TorusY(I+1,J) - TorusY(I,J);
		V2(Y):= TorusY(I,J+1) - TorusY(I,J);
		V1(Z):= TorusZ(I+1,J) - TorusZ(I,J);
		V2(Z):= TorusZ(I,J+1) - TorusZ(I,J);
		CrossProd(V1, V2, V3);
		Unitize(V3);
		TorusNX(I,J):= V3(X);
		TorusNY(I,J):= V3(Y);
		TorusNZ(I,J):= V3(Z);
		end;
end;	\CalcTorus

\-------------------------------------------------------------------------------

begin	\MoveTorus
TorusX:= Rlres(Max);
for II:= 0, Max-1 do
	TorusX(II):= Rlres(Max);
TorusY:= Rlres(Max);
for II:= 0, Max-1 do
	TorusY(II):= Rlres(Max);
TorusZ:= Rlres(Max);
for II:= 0, Max-1 do
	TorusZ(II):= Rlres(Max);

TorusNX:= Rlres(Max);
for II:= 0, Max-1 do
	TorusNX(II):= Rlres(Max);
TorusNY:= Rlres(Max);
for II:= 0, Max-1 do
	TorusNY(II):= Rlres(Max);
TorusNZ:= Rlres(Max);
for II:= 0, Max-1 do
	TorusNZ(II):= Rlres(Max);

TNX:= Rlres(Max);
for II:= 0, Max-1 do
	TNX(II):= Rlres(Max);
TNY:= Rlres(Max);
for II:= 0, Max-1 do
	TNY(II):= Rlres(Max);
TNZ:= Rlres(Max);
for II:= 0, Max-1 do
	TNZ(II):= Rlres(Max);

ITorusX:= Reserve(Max*IntSize);
for II:= 0, Max-1 do
	ITorusX(II):= Reserve(Max*IntSize);
ITorusY:= Reserve(Max*IntSize);
for II:= 0, Max-1 do
	ITorusY(II):= Reserve(Max*IntSize);
ITorusZ:= Reserve(Max*IntSize);
for II:= 0, Max-1 do
	ITorusZ(II):= Reserve(Max*IntSize);

ITorusZZ:= Reserve(Max*Max*IntSize);
ITorusIJ:= Reserve(Max*Max*IntSize);

Ref1:= [1.0, 0.0, 0.0];		\unit vectors for reference frame
Ref2:= [0.0, 1.0, 0.0];		\Rotate procedure modifies these arrays
Ref3:= Rlres(3);		\calculated from Ref1 & Ref2 using cross product
Zoom:= 1.0;

TorusR1:= 100.0;
TorusR2:= 60.0;
Sectors:= 13;
Sections:= 13;

CalcTorus;

VCursor(0, 0);
VText("FPS:



 [ ] [ ]
 Sectors


 [ ] [ ]
Segments


  [S]
Solid/Wire


  [L]
Lighting


 [+] [-]
 Ambient


[Enter]
 Reset


 [Esc]
 Exit
", TColor);

\Arrows
VCursor(2,4);   VChout($1B,TColor);
VCursor(6,4);   VChout($1A,TColor);
VCursor(2,8);   VChout($18,TColor);
VCursor(6,8);   VChout($19,TColor);


Format(2,1);
Time:= GetTime;
loop	begin
	Time0:= Time;
	Time:= GetTime;
	FPS:= 1.19318E6 / Float(Time-Time0);

	VPoly(5*8, 0, [4, 0,0, 100,0, 100,12, 0,12], 0);	\clear background for FPS
	VCursor(5, 0);
	VRlout(FPS, TColor);

	RotateTorus;
	DepthSort;
	VPoly(320, 240, [4, -239,-239, 239,-239, 239,239, -239,239], 0);  \clear background
	DrawTorus;

	\WaitVB;	debug
	VShow(ScreenBuf,GranShift);

	if Chkkey then
		case GetKey of
		  ^S,^s: Solid:= ~Solid;
		  ^T,^t: Texturize:= ~Texturize;
		  ^L,^l: [Lambert:= Lambert +1;   if Lambert > 2 then Lambert:= 0];
		  ^-:	[Ambient:= Ambient - 3.0;
		  	if Ambient<0.0 then Ambient:= 0.0];
		  ^+:	[Ambient:= Ambient + 3.0;
		  	if Ambient > 189.0 then Ambient:= 189.0];
		  -LtArrow:
			if Sectors > 3 then [Sectors:= Sectors -1; CalcTorus];
		  -RtArrow:
			if Sectors < Max-1 then [Sectors:= Sectors +1; CalcTorus];
		  -UpArrow:
			if Sections < Max-1 then [Sections:= Sections +1; CalcTorus];
		  -DnArrow:
			if Sections > 3 then [Sections:= Sections -1; CalcTorus];
		  CR:	[Ref1(0):= 1.0;   Ref1(1):= 0.0;   Ref1(2):= 0.0;
			 Ref2(0):= 0.0;   Ref2(1):= 1.0;   Ref2(2):= 0.0;
			Zoom:= 1.0];
		  Esc:	quit
		other	Beep;
	end;
end;	\MoveTorus

\===============================================================================

proc	MakeCLUT;
int	CReg,	\Color Register number [0..255]
	R, G, B,	\Red, Green, Blue color component intensities [0..63]
	R0, G0, B0,	\Original red, green, blue intensities
	DR, DG, DB,	\Delta red, green, blue
	D,
	Min,
	I, J, J3, SJ;
char	CLUT_Intensity;	\Array: 256 bytes, color look up table for given intensity
\Inputs: Palette
\Outputs: CLUT
begin
\For all 256 colors...
for CReg:= 0, 255 do
	begin
	R0:= Palette(CReg*3+0);
	G0:= Palette(CReg*3+1);
	B0:= Palette(CReg*3+2);

	\Reduce the intensity in 64 steps
	for I:= -63, 0 do
		begin
		R:= (R0 * -I) >> 6;		\63/64 ????
		G:= (G0 * -I) >> 6;
		B:= (B0 * -I) >> 6;


		\For each reduced intensity color find the nearest existing color 
		SJ:= BestColor(Palette, R, G, B);

\		Min:= $7FFFFFFF;		\\(64^2 *3 = 12288)
\		for J:= 0, 255 do
\			begin
\			J3:= J+J+J;
\			DR:= Palette(J3) - R;
\			DG:= Palette(J3+1) - G;
\			DB:= Palette(J3+2) - B;
\
\			D:= DR*DR + DG*DG + DB*DB;
\			if D < Min then
\				begin
\				SJ:= J;		\\Remember
\				Min:= D;
\				end;
\			end;

		\ and store its color register number into the CLUT.
		CLUT_Intensity:= CLUT(-I); 	\Intensity range: 0..63
		CLUT_Intensity(Creg):= SJ;	\Remembered register number
		end;
	Chout(0, ^.);	\Show that something is happening......
	end;
end;	\MakeCLUT

\-------------------------------------------------------------------------------

proc	LoadPcx(Image);
\Read in a 320x200x256 .PCX file, decompress it, & put it into memory at "Image"
char	Image;	\Address of location to put image (can be $A0000-DataSeg)
int	I, J,	\Index into Image
	Byte,	\Byte read from PCX file
	Cnt;	\Number of bytes to repeat
begin
for I:= 0, 128-1 do J:= Chin(3);	\Skip header
J:= 0;
repeat	Byte:= Chin(3);			\Get byte from .PCX file
	if Byte < $C0 then
		begin
		Image(J):= Byte;
		J:= J +1;
		end
	else	begin			\Decompress multiple copies
		Cnt:= Byte & $3F;
		Byte:= Chin(3);
		for I:= 1, Cnt do
			begin
			Image(J):= Byte;
			J:= J +1;
			end;
		end;
until J = 320 *200;
end;	\LoadPcx

\-------------------------------------------------------------------------------

proc	LoadPcxPal(Pal);	\Read in a 256 color palette
\This is called immediately after LoadPcx when loading a normal .PCX file image.
char	Pal;	\Address of 256*3-byte array to hold palette
int	I;
begin
I:= Chin(3);				\Skip $0C, the start-of-palette byte
for I:= 0, 256*3-1 do
	Pal(I):= Chin(3) >> 2;
end;	\LoadPcxPal

\===============================================================================

func	Granularity(Mode);
\Granularity is the smallest increment (in KB) of SVGA video RAM that can
\ be mapped into segment $A000. Normally, it is 64 KB, but for some cards
\ it is less. To specify the second 64K page of video RAM for access the
\ "page" (used by VESA function $4F05) must be 1*64/Granularity. For
\ example, if granularity is 64, the second page is 1. If granularity is 4,
\ the second page is 1*64/4 = 16.
int	Mode;		\VESA mode (e.g: $101)
begin
\Get necessary information for mode $101
Cpureg(0\AX\):= $4F01;
Cpureg(2\CX\):= Mode;
Cpureg(4\DI\):= VESAInfo;
Cpureg(11\ES\):= DataSeg;
Softint($10);
return VESAInfo(4) + VESAInfo(5)<<8;
end;	\Granularity

\-------------------------------------------------------------------------------

proc	SetVesaMode(M);		\Set SVGA mode M
int	M;
begin
CallInt($10, $4F02, M);
if (Cpureg(0) & $FF00) # 0 then
	[Text(0,"Error setting VESA mode: "); Hexout(0, Cpureg(0)); exit];
end;	\SetVesaMode

\-------------------------------------------------------------------------------

proc	DosOpen; \Open input file (with extension) specified on command line
int	Handle;				\File handle
char	CmdTail;			\Copy of command tail
begin
CmdTail:= Reserve($80);			\Copy of command tail
Blit(PSPSeg, $81, DataSeg, CmdTail, $7F);	\Get copy of command tail
Handle:= Fopen(CmdTail, 0);		\Open file name for input
Fset(Handle, ^I);
Openi(3);
end;	\DosOpen

\===============================================================================

begin	\Main
CpuReg:= Getreg;
DataSeg:= Cpureg(12) & $FFFF;
PSPSeg:= Cpureg(11) & $FFFF;

\KLUDGE: Must be reserved before ScreenBuf and ImageData. Used by Granularity.
VESAInfo:= Reserve(256);

Palette:= Reserve(256*3);	\256 color registers. Each color is 3 bytes (RGB).
CLUT:=Reserve(64*IntSize);
for II:= 0, 64-1 do		\Color Look-Up Table: 64 intensities of 256 colors
	CLUT(II):= Reserve(256);
ImageData:= Reserve(320*200);

if ~OpenMouse then [Text(0, "Mouse not detected");   exit];

Text(0, "Loading PCX Image...
");
DosOpen;				\Open .PCX file specified on command line
LoadPCX(ImageData);
LoadPcxPal(Palette);

Text(0, "Generating CLUT...
");
MakeCLUT;

\Start Screen Buffer on a paragraph boundary
II:= Reserve(0);
II:= Reserve((16-II)&$f);
ScreenBuf:= Reserve(640*480);
\WARNING: LARGE ARRAY. DOS SEGMENT ADDRESSING CANNOT ACCESS ARRAYS RESERVED AFTER THIS POINT
for II:= 0, 640*480-1 do ScreenBuf(II):= 0;

Trapc(true);			\Ignore Ctrl+C if struck on keyboard

Solid:= true;
Texturize:= false;
Lambert:= 2;
Ambient:= 32.0;		\90.0;
OrigVidMode:= GetVid;
SetVesaMode($101);		\Set VESA mode 101: 640x480x8
GranShift:= ILn2(64/Granularity($101));
VCursor(0,0);
VLineX:= 0;   VLineY:= 0;
SetTimerMode(2);

\Set block of 256 color registers starting with register 0
CallInt($10, $1012, 0, 256, Palette, 0, 0, DataSeg);

MoveTorus;

SetVid(OrigVidMode);		\Restore original video mode (clears screen)
end;	\Main
