	page	59,136
;VGALINE.ASM	DEC-19-92
;VGA/EGA line drawing routines by Loren Blaney.
;Bresenham's algorithm is used for diagonal lines. This is based on code
; by Sutty and Blair in their book "Advanced Programmer's Guide to the
; EGA/VGA".

	.386
cseg	segment dword use16 public 'code'
	assume	cs:cseg, ds:dseg

	public	vline, vmove, vclear, vdraw, vshow

dseg	segment dword use16 public 'data'
vpage	dw	0		;page to draw to (0 or 1, nonzero is page 1)
x0	dw	0		;coordinates of begining of line
y0	dw	0
x1	dw	0		;coordinates of end of line
y1	dw	0
color	db	0		;COLOR (0..15); if bit 7 = 1 then XOR
linsiz	equ	044ah		;BIOS data: columns (bytes) per raster line
dseg	ends

dotdash	db	0		;dot-dash pattern (set bits plot pixels)

temp1	equ	[bp-2]		;temporary local variables on the stack
temp2	equ	[bp-4]
temp3	equ	[bp-6]

;----------------------------------------------------------------------
;Draw a line from the current position (X0, Y0) to X, Y. COLOR is 0 through
; 15. If bit 7 is set then the line will be XORed with what is already on
; the screen. The high byte of COLOR is used to draw dotted lines. A set
; bit in this byte corresponds to an undrawn pixel. There are only 4 colors
; because of page flipping.
; VLINE(X, Y, COLOR);
;
vline:	mov	bp,sp
	mov	ax,[bp]+4	;get arguments
	mov	color,al
	not	ah		;make set bits draw pixels
	mov	cs:dotdash,ah
	mov	ax,[bp]+8	;coordinates of end of line
	sar	ax,4		;divide by 16 (unzoom)
	add	ax,240		;move coordinate 0,0 to center of screen
	mov	y1,ax
	mov	ax,[bp]+12
	sar	ax,4
	add	ax,320
	mov	x1,ax

	pusha			;save registers
	push	es
	push	ds

	mov	bp,sp		;reserve space on stack for 3 local variables
	sub	sp,6

	mov	ax,x0		;make sure that x1 >= x0
	mov	cx,x1
	cmp	cx,ax
	jge	short vlin10
	mov	bx,y0		;swap x0,y0 with x1,y1
	mov	dx,y1
	mov	x0,cx
	mov	y0,dx
	mov	x1,ax
	mov	y1,bx
vlin10:
	sub	ax,ax		;get number of bytes per scan line
	mov	es,ax
	mov	ax,es:linsiz	;compute the offset (from a0000) to the byte
	mul	y0		; containing the pixel at the start of the line
	mov	bx,x0		; and save it on the stack
	mov	cl,3		;push y0*linsiz + x0/8
	shr	bx,cl
	add	ax,bx
	push	ax

;compute the bit mask for the first pixel and save it on the stack
	mov	cx,x0		;compute which bit (x mod 8) to modify
	and	cl,7		;  0  1  2  3  4  5  6  7 shift count
	mov	bx,80h		;  1  0  0  0  0  0  0  0  ---->
	shr	bx,cl		;(high byte is 00, see diagonal line below)
	push	bx		;save mask on stack

;set up registers for color
	mov	bl,color
	mov	dx,03ceh	;set up "rotate/function" register
	mov	al,3
	out	dx,al
	inc	dx
;	sub	al,al		;assume replace mode for ALU
;	test	bl,080h		;xor mode?
;	jz	short vlin20	;skip if not
;	mov	al,18h		;select ALU xor mode
	mov	al,10h		;select ALU OR mode
vlin20:	out	dx,al		;set ALU mode	

	dec	dx		;set "set/reset" register with COLOR
	sub	al,al
	mov	ah,bl
	and	ah,0fh		;mask off unused color bits
	out	dx,ax

	mov	al,1		;set "enable set/reset" register with 0fh
	mov	ah,0fh		; to enable all 4 planes of the set/reset
	out	dx,ax		; register

	mov	dx,03c4h	;set "map mask" register to enable the bit
	mov	al,2		; planes corresponding to the active page
	mov	ah,03h		;assume page 0
	cmp	vpage,0		;is it page 0?
	jz	short vlin25	;jump if not
	mov	ah,0ch		;select page 1
vlin25:	out	dx,ax

;determine if horizontal, vertical or diagonal line
	mov	ax,x0		;set up temp variables because ds goes away
	mov	temp1,ax
	mov	ax,es:linsiz	;number of bytes per raster line
	mov	temp3,ax

	mov	si,x1		;compute dx	       si
	sub	si,x0
	mov	di,y1		;compute dy	       di
	sub	di,y0

	mov	dx,0a000h	;point data segment to base of display
	mov	ds,dx		; memory
	mov	es,dx		;also point extra segment here (for movsb)

	jge	short vlin30	;jump if dy is positive
	neg	ax		;else negate linsiz
	mov	temp3,ax
	neg	di		;make dy positive
vlin30:
	test	si,si		;jump according to type of line
	jz	short vlin40	;jump if vertical line (dx = 0)
	test	di,di
	jz	short vlin50	;jump if horizontal line (dy = 0)
	jmp	vlin60		;must be diagonal line

;draw a vertical line
vlin40:	mov	cx,di		;set up counter cx with the number of pixels
	inc	cx		;add one to include first pixel
	mov	bx,temp3	;get +/- bytes per raster line

	pop	ax		;get bit mask for first pixel
	mov	ah,al
	mov	al,08h
	mov	dx,03ceh	;set "bit map" register
	out	dx,ax

	pop	di		;fetch offset to byte at beginning of line

	rol	cs:dotdash,1	;draw dotted lines
;	align	4		MASM 6.0 generates an illegal instruction!
vlin45:	jnc	short vlin47	;don't plot point if dotdash bit is clear
	xchg	al,[di]		;latch old data & modify unmasked bits
vlin47:	add	di,bx		;move to next scan line
	rol	cs:dotdash,1	;draw dotted lines
	dec	cx		;next pixel
	jne	short vlin45
	jmp	vlin80

;draw a horizontal line (from left to right)
vlin50:	mov	cx,si		;set cx counter with the number of pixels
	inc	cx		;add one for first pixel
	pop	si		;get bit mask for first pixel
	pop	di		;get offset to byte at beginning of line
	mov	dx,03ceh	;set up for selecting bit mask register

;draw pixels from the leading partial byte
	mov	ax,temp1	;x0
	and	ax,07h		;check for partial byte
	jz	short vlin54	;jump if none
	mov	bx,0ffh		;set up bit mask
	push	cx		;save number of pixels in line
	mov	cx,ax		;get number of bits to shift
	shr	bx,cl
	pop	cx		;restore number of pixels in line
	add	cx,ax		;adjust number of pixels in rest of line
	sub	cx,08h
	jge	short vlin52	;jump if line has at least one byte's worth
	neg	cx		; of pixels in it
	shr	bx,cl		;else set up bit mask for partial byte
	shl	bx,cl
	sub	cx,cx		;count = 0
vlin52:
	mov	al,08h		;set bit mask register with contents of bl
	mov	ah,bl
	and	ah,cs:dotdash	;draw dotted line
	out	dx,ax

	xchg	al,[di]		;latch old data & modify unmasked bits
	inc	di		;move offset to next byte to the right

;draw 8 pixels (= one byte) at a time for the middle of the line
vlin54:	mov	bx,cx		;check if any bytes to set
	cmp	bx,8
	jl	short vlin56	;jump if not
	shr	cx,3		;byte count = pixel count /8
	mov	ax,0ff08h	;set all the bits in the bit mask register
	and	ah,cs:dotdash	;draw dotted line
	out	dx,ax

	mov	si,di		;set up to read and write display latches
	cld
	rep movsb		;es:di++ <-- ds:si++  cx++

;draw the end of the line (partial byte)
vlin56:	and	bl,07h		;are there any pixels to plot?
	jz	short vlin59	;jump if not
	mov	ah,0ffh		;set up bit mask
	mov	cl,bl
	shr	ah,cl
	not	ah
	mov	al,08h		;set bit mask register
	and	ah,cs:dotdash	;draw dotted line
	out	dx,ax
	xchg	al,[di]		;latch old data & modify unmasked bits
vlin59:	jmp	vlin80

;Draw a diagonal line
;
; Octants:
;	 ------------+------------
;	            /|\
;	    3     /  |  \     0
;	        /    |    \
;	      /   2  |  1   \
;	             |
;
;
;determine which octant the line is in
vlin60:	cmp	si,di		;is dx <= dy
	jle	short vlin70	;jump if so -- octants 1 and 2

;compute constants for octant 0 and 3
	mov	cx,si		;set pixel counter to dx+1
	inc	cx		;include first pixel
	sal	di,1		;d1 = dy*2		di
	mov	bx,di		;d  = dy*2-dx		bx
	sub	bx,si
	neg	si		;d2 = dy*2-dx-dx	si
	add	si,bx

	mov	temp1,di	;d1
	mov	temp2,si	;d2
	mov	dx,03ceh	;set up to output to bit mask register
	mov	al,08h
	out	dx,al
	inc	dx
	pop	ax		;fetch bit mask for first pixel
	pop	di		;fetch address offset to first pixel

;draw line in octants 0 and 3
	mov	si,temp3	;bytes per raster line
	xchg	ah,al		;bit mask in ah, al = 0
	test	bx,bx		;test decision variable
;	align	4
vlin62:	jns	short vlin66	;jump based on decision variable, d
	or	al,ah		;combine bits
	ror	ah,1		;update bit mask
	jc	short vlin64	;jump if moving into next byte
	add	bx,temp1	;if d < 0 then d = d + d1
	loop	vlin62		;(don't alter status)

	and	al,cs:dotdash	;draw dotted line
	out	dx,al		;output final bit mask
	xchg	al,[di]		;latch old data & modify unmasked bits
	jmp	vlin80

;	align	4
vlin64:
	and	al,cs:dotdash	;draw dotted line
	out	dx,al		;output bit mask
	xchg	al,[di]		;latch old data & modify unmasked bits
	sub	al,al		;clear al
	inc	di		;point to next byte to the right
	add	bx,temp1	;if d < 0 then d = d + d1
	loop	vlin62		;(don't alter status)
	jmp	vlin80

;	align	4
vlin66:
	or	al,ah		;combine bits
	and	al,cs:dotdash	;draw dotted line
	out	dx,al		;output bit mask
	xchg	al,[di]		;latch old data & modify unmasked bits
	sub	al,al		;clear al
	ror	ah,1		;update bit mask
	adc	di,si		;move to next scan line and maybe next byte
	add	bx,temp2	;if d >= 0 then d = d + d2
	loop	vlin62		;(don't alter status)
	jmp	vlin80

;compute constants for octant 1 and 2
vlin70:	mov	cx,di		;set counter to dy+1
	inc	cx		;include first pixel
	sal	si,1		;d1 = dx * 2
	mov	bx,si		;d  = dx * 2 - dy
	sub	bx,di
	neg	di		;d2 = -dy + dx * 2 - dy
	add	di,bx

	mov	temp2,di	;d2
	mov	temp1,si	;d1
	mov	dx,03ceh	;set up to output to bit mask register
	mov	al,08h
	out	dx,al
	inc	dx
	pop	ax		;fetch bit mask for first pixel
	pop	di		;fetch address offset to first pixel
	out	dx,al		;output bit mask for first pixel

;draw a line in octants 1 and 2
	mov	si,temp3	;bytes per raster line
	test	bx,bx		;set status for decision variable, d
	rol	cs:dotdash,1	;draw dotted lines
;	align	4
vlin72:	jnc	short vlin73	;don't plot point if dotdash bit is clear
	xchg	ah,[di]		;latch old data & modify unmasked bits
vlin73:	jns	short vlin74
	add	di,si		;move down (or up) a scan line (y = y+1)
	add	bx,temp1	;if d < 0 then d = d + d1
	rol	cs:dotdash,1	;draw dotted lines
	loop	vlin72		;(don't alter status)
	jmp	vlin80

;	align	4
vlin74:
	ror	al,1		;update bit mask (x = x+1)
	adc	di,si		;move down a scan line and maybe right a byte
	out	dx,al		;set bit mask
	add	bx,temp2	;if d >= 0 then d = d + d2
	rol	cs:dotdash,1	;draw dotted lines
	loop	vlin72		;(don't alter status)

;restore plane enable and bit mask registers to BIOS defaults
vlin80:	mov	dx,03ceh	;enable all 8-bits in the bit mask register
	mov	ax,0ff08h
	out	dx,ax
	mov	ax,00003h	;make sure xor mode is off
	out	dx,ax
	mov	ax,00001h	;disable set/reset
	out	dx,ax

	mov	sp,bp		;discard local variables

	pop	ds		;restore registers
	pop	es
	popa

	mov	ax,[bp]+8	;set coordinates for beginning of next
	sar	ax,4		; line
	add	ax,240
	mov	y0,ax
	mov	ax,[bp]+12
	sar	ax,4
	add	ax,320
	mov	x0,ax
	retf	12		;drop 3 arguments

;----------------------------------------------------------------------
;Set the starting point of a line
; VMOVE(X, Y);
;
vmove:	mov	bp,sp
	mov	ax,[bp]+4
	sar	ax,4
	add	ax,240
	mov	y0,ax
	mov	ax,[bp]+8
	sar	ax,4
	add	ax,320
	mov	x0,ax
	retf	8

;----------------------------------------------------------------------
;Clear the specified page
; VCLEAR(PAGE);
;
vclear:	mov	ax,00302h	;assume page 0
	mov	bp,sp		;get argument
	cmp	word ptr [bp]+4,0
	jz	short vc10
	mov	ax,00c02h	;set page 1
vc10:	mov	dx,03c4h	;sequencer map mask register
	out	dx,ax

	mov	ax,0a000h	;clear
	mov	es,ax
	sub	di,di
	mov	ax,di
	mov	cx,19200	;640/16*480
	cld
	rep	stosw		;es:di++ <-- ax

	retf	4

;----------------------------------------------------------------------
;Select page for drawing onto
; VDRAW(PAGE);
;
vdraw:	mov	bp,sp		;get argument
	mov	ax,[bp]+4
	mov	vpage,ax
	retf	4

;----------------------------------------------------------------------
;Display specified page
; VSHOW(PAGE);
;
vshow:	mov	dx,03dah	;reset toggle to attribute address register
	in	al,dx

	mov	dx,03c0h	;color plane enable register
	mov	ax,0332h	;assume page 0
	mov	bp,sp		;get argument
	cmp	word ptr [bp]+4,0
	jz	short vs10
	mov	ax,0c32h	;page 1
vs10:	out	dx,al
	xchg	al,ah
	out	dx,al
	retf	4

cseg	ends
	end
