\XPLSHIFT.XPLV5D	19-JUL-2001		LOREN BLANEY
\This program converts upper case command words, in XPL source
\ files, to lower case.

\REVISIONS:
\8/20/81 VERSION 1.2
\19-JUL-2001, VERSION 2.0, MODIFIED TO RUN ON PC, L.B.

include	\CXPL\CODESI;

def	TV=0, KB=0, FILE=3;	\DEVICE NUMBERS
\ASCII CHARACTERS:
def	NULL=0, TAB=$9, LF=$0A, FF=$0C, CR=$0D, EOF=$1A, SP=$20;

int	CHAR,	\CURRENT CHARACTER
	CHARBAK,\PREVIOUS CHARACTER
	CHARFWD;\NEXT CHARACTER

int	CpuReg,		\address of CPU registers
	PspSeg,		\address of PSP segment (for file I/O)
	DataSeg;	\address of data segment (heap & stack)

\Global variables for DOS I/O routines:
int	InHandle,	\input file handle (= -1 if no input file)
	OutHandle;	\output file handle (= -1 if no output file)
char	InFileName,	\input file name (all file names are zero terminated)
	OutFileName,	\output file name
	TmpFileName,	\temporary output file name (.$$$)
	BakFileName;	\backup output file name (.BAK)
def	NameSize=80;	\maximum number of chars in a file name (incl path)

\============================== DOS I/O ROUTINES ===============================
\
\Routines to handle DOS file I/O almost as conveniently as Apex did.
\
\The call to DosOpen specifies the input and output extensions that will be
\ used, for example:
\
\	DosOpen("XPL", "I2L");		INPUT		OUTPUT
\	XPLNQ FROG			FROG.XPL	FROG.I2L
\	XPLNQ FROG TOAD			FROG.XPL	TOAD.I2L
\	XPLNQ FROG.XPL			FROG.XPL	FROG.I2L
\	XPLNQ FROG.PAS			FROG.XPL	FROG.I2L
\
\If "???" is specified by DosOpen then the extensions will be taken from
\ the command line (and they better be typed in):
\
\	DosOpen("???", "???");		INPUT		OUTPUT
\	ED PIG.TXT			PIG.TXT		PIG.TXT
\	COPY PIG.TXT DOG.DOC		PIG.TXT		DOG.DOC
\
\Where "   " is specified by DosOpen, the input or output file is not used:
\
\	DosOpen("???", "   ");		INPUT		OUTPUT
\	PRINT FROG.TXT			FROG.TXT	  --
\
\	DosOpen("   ", "MAP");		INPUT		OUTPUT
\	SAVE FROG			 --		FROG.MAP
\
\If a new output file replaces an existing file by the same name, the existing
\ file is renamed with a .BAK extension.
\
\GetCh in GetName can be modified to handle command-line switches. Currently,
\ all switches are stripped out and ignored.


func	DosOpen(InExt, OutExt);	\Open MS-DOS files. Returns: 0 if no errors;
\ 1 if error opening input file; 2 if error opening output file.
char	InExt, OutExt;	\input and output extensions
int	Inx;		\index into CmdLine
char	CmdLine;	\command line tail


	proc	CopyName(From, To, Ext);
	\Copy file name and replace any existing extension with "Ext".
	\If "Ext" = "???" then use extension from "From".
	char	From, To,	\strings are 0 terminated
		Ext;		\extension
	int	I, J, Ch;
	begin
	I:= 0;
	loop	begin				\copy file name
		Ch:= From(I);
		if Ch=^. ! Ch=0 then quit;
		To(I):= Ch;
		if I < NameSize-1 then I:= I +1;
		end;
	To(I):= ^.;				\copy extension
	if I < NameSize-1 then I:= I +1;
	for J:= 0, 2 do
		begin
		Ch:= Ext(J) & $7F;
		To(I):= if Ch=^? then From(I) else Ch;
		if I < NameSize-1 then I:= I +1;
		end;
	To(I):= 0;				\terminate string
	end;	\CopyName



	proc	GetName(Name);
	\Read file name from command line and return it in "Name".
	\The command line is terminated by a CR.
	char	Name;
	int	I, Ch;


		proc	GetCh;	\Get character, stripping out switches
		begin
		loop	begin
			Ch:= CmdLine(Inx);
			Inx:= Inx +1;
			if Ch # ^/ then quit;
			Ch:= CmdLine(Inx);	\ignore any one-char switches
			Inx:= Inx +1;
			if Ch = CR then quit;	\handle wise guys
			end;
		end;	\GetCh


	begin	\GetName
	repeat GetCh until Ch # SP;		\skip any leading spaces
	I:= 0;
	loop	begin
		case Ch of SP, CR, ^,: Ch:= 0 other; \terminate Name with 0
		Name(I):= Ch;
		if Ch = 0 then quit;
		if I < NameSize-1 then I:= I +1;
		GetCh;
		end;
	end;	\GetName


begin	\DosOpen
CmdLine:= Reserve($7F);			\get command tail from PSP
Blit(PspSeg, $81, DataSeg, CmdLine, $7F);
CmdLine($7E):= CR;			\make sure it's terminated with a CR

InHandle:= -1;   OutHandle:= -1;	\set to illegal values

Inx:= 0;
if InExt(0) # SP then
	begin				\get the input file name
	GetName(InFileName);
	CopyName(InFileName, InFileName, InExt);

	Trap(false);
	InHandle:= Fopen(InFileName, 0);
	Trap(true);
	if Geterr # 0 then		\file or path does not exist,
		[InHandle:= -1;   return 1]; \ SHARE conflict, etc.
	Fset(InHandle, ^I);
	end;

if OutExt(0) # SP then
	begin				\get the output file name
	if Inx > 0 then Inx:= Inx -1;	\backup in case of CR
	GetName(OutFileName);
	if OutFileName(0) =0 then	\if no outfile then use the infile
		CopyName(InFileName, OutFileName, OutExt)
	else	CopyName(OutFileName, OutFileName, OutExt);

	CopyName(OutFileName, TmpFileName, "$$$");
	CopyName(OutFileName, BakFileName, "BAK");

	Trap(false);
	OutHandle:= Fopen(TmpFileName, 1);
	Trap(true);
	if Geterr # 0 then		\path does not exist, file already exists
		[OutHandle:= -1;   return 2]; \ root directory full, etc.
	Fset(OutHandle, ^O);
	end;

return 0;				\no errors
end;	\DosOpen

\-------------------------------------------------------------------------------

func	DosClose;	\Close MS-DOS files. Returns: 0 if no errors;
\ 1 if error closing input file; 2 if error closing output file.
begin
if InHandle # -1 then
	begin
	Trap(false);
	Fclose(InHandle);
	Trap(true);
	if Geterr # 0 then return 1;
	end;

if OutHandle # -1 then
	begin
	Trap(false);
	Fclose(OutHandle);

	CpuReg(\AX\0):= $4100;		\DEL OutFileName.BAK
	CpuReg(\DS\9):= DataSeg;
	CpuReg(\DX\3):= BakFileName;
	Softint($21);			\(.BAK file might not exist)

	CpuReg(\AX\0):= $5600;		\REN OutFileName.EXT OutFileName.BAK
	CpuReg(\DS\9):= DataSeg;
	CpuReg(\DX\3):= OutFileName;
	CpuReg(\ES\11):= DataSeg;
	CpuReg(\DI\4):= BakFileName;
	Softint($21);

	CpuReg(\AX\0):= $5600;		\REN OutFileName.$$$ OutFileName.EXT
	CpuReg(\DS\9):= DataSeg;
	CpuReg(\DX\3):= TmpFileName;
	CpuReg(\ES\11):= DataSeg;
	CpuReg(\DI\4):= OutFileName;
	Softint($21);

	Trap(true);
	if Geterr # 0 then return 2;	\disk write error when flushing buffers
	end;

return 0;				\no errors
end;	\DosClose

\===============================================================================

proc	NEXTCHAR;
\OUTPUT THE CURRENT CHARACTER AND READ IN THE NEXT CHARACTER.
begin
if CHAR#NULL then CHOUT(FILE,$7F&CHAR);
CHARBAK:=CHAR;
CHAR:=CHARFWD;
if CHAR#EOF then CHARFWD:=CHIN(FILE);
if CHAR=^^ then CHARFWD:=CHARFWD!$80;	\FLAG META CHARS
end;	\NEXTCHAR



begin	\MAIN
TEXT(TV,"-- XPL CASE SHIFTER (FOR RUNNING ON THE PC), VERSION 2.0 --
");

CpuReg:= Getreg;
PspSeg:= CpuReg(11);
DataSeg:= CpuReg(12);

InFileName:= Reserve(NameSize);
OutFileName:= Reserve(NameSize);
TmpFileName:= Reserve(NameSize);
BakFileName:= Reserve(NameSize);

case DosOpen("XPL", "XPL") of
  1:	[Text(0, "Cannot open input file");  exit];
  2:	[Text(0, "Cannot open output file");  exit]
other;

OPENI(FILE); OPENO(FILE);
CHARFWD:=NULL; CHAR:=CHARFWD;
NEXTCHAR;
while CHAR#EOF do
	begin
	case CHAR of
	  ^":	repeat NEXTCHAR until CHAR=^" ! CHAR=EOF;
	  ^\:	repeat NEXTCHAR until CHAR=^\ ! CHAR=CR ! CHAR=EOF;
	  ^':	begin
		case CHARBAK of		\CHECK THE PREVIOUS CHAR
		  TAB, LF, FF, CR, SP, ^(, ^[, ^,, ^=: CHAR:=NULL
		other CHAR:=SP;
		repeat	NEXTCHAR;
			if CHAR>=^A & CHAR<=^Z then
				CHAR:=CHAR!$20;	\SHIFT LOW
		until CHAR=^' ! CHAR=EOF;
		case CHARFWD of		\CHECK THE FOLLOWING CHAR
		  TAB, LF, FF, CR, SP, ^), ^], ^;, ^,, ^:: CHAR:=NULL
		other CHAR:=SP;
		end
	other	[];	\NORMAL CHARACTER
	NEXTCHAR;
	end;
CLOSE(FILE);

case DosClose of
  1:	Text(0, "Cannot close input file");
  2:	Text(0, "Cannot close output file")
other;
end;	\MAIN
