\FULLER.XPL	DEC-13-92
\By Loren Blaney
\Inspired by FERRIS and Bucky.
\
\Revision History:
\JUN-04-89, Original on Amiga.
\DEC-13-92, Converted to run on PC.
\
\This program displays a 3-D image of a truncated icosahedron, commonly
\ known as a soccer ball. An efficient hidden-line algorithm, exploiting
\ symmetry, allows the image to be rotated in real time.
\
\A right-hand coordinate system is used. X increases to the right, Y
\ increases downward, and Z increases into the screen.
\
\			           Z
\			       |  /
\			       | /
\			       |/
\			-------+-------> X
\			      /|
\			     / |
\			    /  |
\			       V
\			       Y
\

include C:\CXPL\CODESI;	\Intrinsic routines

ext	VLINE, VMOVE, VCLEAR, VDRAW, VSHOW;	\External assembly routines

def	INTSIZE= 4;			\Number of bytes in an integer
def	BEL= $07, FF=$0C;		\ASCII control characters

int	PAGE,		\Video page (0 or 1)
	ICOS0,		\Array: Initial coordinates of icosahedron verticies.
	REF,		\Array: Coordinates of reference frame, REF(x/y/z, point)
	PT0,		\Array: Initial coordinates of verticies of trunc icos
	PT,		\Array: Rotated coordinates of trunc icos
	FACE1,		\Array: Given an edge (line) as an index, these give
	FACE2,		\ the two faces that intersect to form the edge
	FACEPT0,	\Array: Initial and rotated coordinates of the center
	FACEPT,		\ points for the faces of the truncated icosahedron
	SPEED,		\Array: Rotational speed (x, y, z)
	MORPH,		\Controls shape of object
	REM0, REM1, REM2, REM3, REM4, REM5, \Remainders from rotation calculation
	HIDSW,		\4-way switch which controls how hidden lines are shown
	HIDLINETYPE,	\Actual value used to draw hidden lines
	II;		\Scratch for MAIN
addr	LMOVE,		\Array of flags: DRAW must MOVE to the beginning of line
	LBEGIN,		\Array: Points at beginning of lines (edges) of trunc icos
	LEND;		\Array: Points at end of lines

\----------------------------------------------------------------------

proc	SETPALREG(N, C); \Set palette register
int	N,	\Register to set
	C;	\Color value (0..$3F)
int	CPUREG;
begin
CPUREG:= GETREG;
CPUREG(0):= $1000;
CPUREG(1):= C<<8 ! N;
SOFTINT($10);		\Call BIOS
end;	\SETPALREG



proc	SETBORDER(C);	\Set border color
int	C;	\Color value (0..$3F)
int	CPUREG;
begin
CPUREG:= GETREG;
CPUREG(0):= $1001;
CPUREG(1):= C<<8;
SOFTINT($10);
end;	\SETBORDER



proc	WAITVB;		\Wait for the beginning of the next vertical blank
begin
while PIN($3DA, 0) & $08 do;		\Wait for no vertical blank
repeat until PIN($3DA, 0) & $08;	\Wait for vertical blank
end;	\WAITVB;



func	GETKEY;		\Get character from keyboard
int	CPUREG, CH;	\This is a low-level routine with no echo,
begin			\ no ctrl-C, and no cursor.
CPUREG:= GETREG;
CPUREG(0):= 0;		\Function $00
SOFTINT($16);		\Call BIOS routine
CH:= CPUREG(0) & $FF;
if CH = 0 then
	begin			\Handle special keys
	case CPUREG(0)>>8 of	\Arrows:
	  72:	CH:= $0B;	\ up
	  80:	CH:= $0A;	\ down
	  75:	CH:= $08;	\ left
	  77:	CH:= $15	\ right
	other	CH:= -(CPUREG(0)>>8);
	end;
return CH;
end;	\GETKEY

\----------------------------------------------------------------------

proc	ROT(PX, PY, PZ, PX0, PY0, PZ0, UXX, UYX, UZX, UXY, UYY, UZY, UXZ, UYZ, UZZ, N);
\Rotate object about 3 axes
int	PX, PY, PZ,		\Point position after being rotated (output)
	PX0, PY0, PZ0,		\Point at initial position
	UXX, UYX, UZX, UXY, UYY, UZY, UXZ, UYZ, UZZ,	\Rotation matrix
	N;			\Number of points to rotate

\This multiplies each point at its initial position (Po) by the rotation
\ matrix to get the rotated position:
\	                                                    
\	                                  Uxx  Uxy  Uxz 
\	 Px  Py  Pz  =  Pxo  Pyo  Pzo  *  Uyx  Uyy  Uyz 
\	                                  Uzx  Uzy  Uzz 
\	                                                    
\
int	K;
begin
for K:= 0, N-1 do
	PX(K):= PX0(K) *UXX + PY0(K) *UYX + PZ0(K) *UZX;
for K:= 0, N-1 do
	PY(K):= PX0(K) *UXY + PY0(K) *UYY + PZ0(K) *UZY;
for K:= 0, N-1 do
	PZ(K):= PX0(K) *UXZ + PY0(K) *UYZ + PZ0(K) *UZZ;
end;	\ROT



proc	ROTATE;		\Rotate object about 3 axes
\Inputs: PT0, FACEPT0, REF, REMn, SPEED.
\Outputs: PT, FACEPT, REF, REMn.
int	UXX, UYX, UZX,	\Unit vectors for rotation matrix
	UXY, UYY, UZY,
	UXZ, UYZ, UZZ,
	D2, DX, DY, DZ; 
int	I;
begin
\Rotate reference frame:
for I:= 0, 2 do
	begin
	\Rotate about X axis:
	REF(1,I):= REF(1,I) + (REF(2,I) *SPEED(0) + REM0(I)) /$1000<<4;
	REM0(I):= REM(0);
	REF(2,I):= REF(2,I) - (REF(1,I) *SPEED(0) + REM1(I)) /$1000<<4;
	REM1(I):= REM(0);

	\Rotate about Y axis:
	REF(0,I):= REF(0,I) + (REF(2,I) *SPEED(1) + REM2(I)) /$1000<<4;
	REM2(I):= REM(0);
	REF(2,I):= REF(2,I) - (REF(0,I) *SPEED(1) + REM3(I)) /$1000<<4;
	REM3(I):= REM(0);

	\Rotate about Z axis:
	REF(1,I):= REF(1,I) + (REF(0,I) *SPEED(2) + REM4(I)) /$1000<<4;
	REM4(I):= REM(0);
	REF(0,I):= REF(0,I) - (REF(1,I) *SPEED(2) + REM5(I)) /$1000<<4;
	REM5(I):= REM(0);
	end;

\Because these three points are independantly rotated, errors can accumulate
\ and distort the reference frame. To prevent this, mathematical cross braces
\ are used to maintain the correct distances between these points.

\Adjust the distance between reference points 0 and 1:
DX:= REF(0,1) - REF(0,0);   DY:= REF(1,1) - REF(1,0);   DZ:= REF(2,1) - REF(2,0);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $0FFF<<16+$0004 then			\$0FFF0004 = $3FFE **2
	begin				\Lengthen line
	REF(0,1):= REF(0,1) + DX /$1000;
	REF(1,1):= REF(1,1) + DY /$1000;
	REF(2,1):= REF(2,1) + DZ /$1000;
	end;
if D2 > $1001<<16+$0004 then			\$10010004 = 4002 **2
	begin				\Shorten line
	REF(0,1):= REF(0,1) - DX /$1000;
	REF(1,1):= REF(1,1) - DY /$1000;
	REF(2,1):= REF(2,1) - DZ /$1000;
	end;

\Adjust the distance between reference points 0 and 2:
DX:= REF(0,2) - REF(0,0);   DY:= REF(1,2) - REF(1,0);   DZ:= REF(2,2) - REF(2,0);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $0FFF<<16+$0004 then			\$0FFF0004 = $3FFE **2
	begin				\Lengthen line
	REF(0,2):= REF(0,2) + DX /$1000;
	REF(1,2):= REF(1,2) + DY /$1000;
	REF(2,2):= REF(2,2) + DZ /$1000;
	end;
if D2 > $1001<<16+$0004 then			\$10010004 = 4002 **2
	begin				\Shorten line
	REF(0,2):= REF(0,2) - DX /$1000;
	REF(1,2):= REF(1,2) - DY /$1000;
	REF(2,2):= REF(2,2) - DZ /$1000;
	end;

\Adjust the distance between reference points 1 and 2:
DX:= REF(0,2) - REF(0,1);   DY:= REF(1,2) - REF(1,1);   DZ:= REF(2,2) - REF(2,1);
D2:= DX *DX + DY *DY + DZ *DZ;
if D2 < $1FFD<<16+$8B01 then			\$1FFD8B01 = $5A7F **2
	begin				\Lengthen line
	REF(0,2):= REF(0,2) + DX /$1000;
	REF(1,2):= REF(1,2) + DY /$1000;
	REF(2,2):= REF(2,2) + DZ /$1000;
	end;
if D2 > $2001<<16+$C919 then			\$2001C919 = $5A85 **2
	begin				\Shorten line
	REF(0,2):= REF(0,2) - DX /$1000;
	REF(1,2):= REF(1,2) - DY /$1000;
	REF(2,2):= REF(2,2) - DZ /$1000;
	end;

UXX:= REF(0,1) - REF(0,0);		\UX = REF(1) - REF(0)
UXY:= REF(1,1) - REF(1,0);
UXZ:= REF(2,1) - REF(2,0);

UYX:= REF(0,2) - REF(0,0);		\UY = REF(2) - REF(0)
UYY:= REF(1,2) - REF(1,0);
UYZ:= REF(2,2) - REF(2,0);

\Find UZ using cross product:
UZX:= (UXY *UYZ - UXZ *UYY) /$4000;	\UZ = UX X UY
UZY:= (UXZ *UYX - UXX *UYZ) /$4000;
UZZ:= (UXX *UYY - UXY *UYX) /$4000;

\Rotate all points to align with reference frame (unit vectors):
ROT(	PT(0), PT(1), PT(2),		\outputs
	PT0(0), PT0(1), PT0(2),		\inputs
	UXX, UYX, UZX,
	UXY, UYY, UZY,
	UXZ, UYZ, UZZ,
	60);

\Rotate face points to align with reference frame (unit vectors):
ROT(	FACEPT(0), FACEPT(1), FACEPT(2),		\outputs
	FACEPT0(0), FACEPT0(1), FACEPT0(2),		\inputs
	UXX, UYX, UZX,
	UXY, UYY, UZY,
	UXZ, UYZ, UZZ,
	32);
end;	\ROTATE

\----------------------------------------------------------------------

proc	DRAW;		\Draw truncated icosahedron on the bit map
\Inputs: PT, LMOVE, LBEGIN, LEND, FACEPT, FACE1, FACE2, HIDLINETYPE.
int	L, I;
begin
for L:= 0, 89 do			\For all lines (edges)...
	begin
	if LMOVE(L) then		\Must move to start of line
		begin
		I:= LBEGIN(L);		\Get index to start point
		VMOVE(PT(0,I)>>12, PT(1,I)>>12);
		end;
	I:= LEND(L);			\Get index to end point

	if FACEPT(2, FACE1(L)) <= 0 then
		VLINE(PT(0,I)>>12, PT(1,I)>>12, 5)
	else	begin
		if FACEPT(2, FACE2(L)) <= 0 then
			VLINE(PT(0,I)>>12, PT(1,I)>>12, 5)
		else	VLINE(PT(0,I)>>12, PT(1,I)>>12, HIDLINETYPE);
		\If neither face containing the edge (line) is visible (> 0)
		\ then draw a hidden line
		end;
	end;
end;	\DRAW

\----------------------------------------------------------------------

proc	SETUP2;		\Set up the verticies of the truncated icosahedron
\Inputs: ICOS0, MORPH.
\Outputs: PT0
int	TBL,
	I,
	A, B,		\Indicies for the verticies of an icosahedron
	C, D;		\Indicies for the truncated verticies
int	J, K;
begin
\From the verticies of the icosahedron (A & B) derive the verticies of the
\ truncated icosahedron (C & D): (See external documentation for a diagram
\ which shows the number assigned to each vertex.)
\	A	B	C	D
TBL:=	[0,	1,	1,	8,
	0,	2,	0,	5,
	0,	3,	4,	17,
	0,	4,	3,	14,
	0,	5,	2,	11,
	1,	2,	7,	6,	\--
	2,	3,	19,	18,
	3,	4,	16,	15,
	4,	5,	13,	12,
	5,	1,	10,	9,
	1,	6,	24,	23,	\--
	6,	2,	22,	21,
	2,	7,	20,	39,
	7,	3,	38,	37,
	3,	8,	36,	35,
	8,	4,	34,	33,	\--
	4,	9,	32,	31,
	9,	5,	30,	29,
	5,	10,	28,	27,
	10,	1,	26,	25,
	6,	7,	41,	40,	\--
	7,	8,	53,	52,
	8,	9,	50,	49,
	9,	10,	47,	46,
	10,	6,	44,	43,
	6,	11,	42,	56,	\--
	7,	11,	54,	55,
	8,	11,	51,	59,
	9,	11,	48,	58,
	10,	11,	45,	57];

J:= 0;
for I:= 0, 29 do	\For 30 pairs of verticies...
	begin
	A:= TBL(J);	J:= J +1;
	B:= TBL(J);	J:= J +1;
	C:= TBL(J);	J:= J +1;
	D:= TBL(J);	J:= J +1;
	for K:= 0, 2 do
		begin				\C:= (B-A) *MORPH/300 + A
		PT0(K,C):= (ICOS0(K,B) - ICOS0(K,A)) *MORPH /300 + ICOS0(K,A);
		PT0(K,D):= (ICOS0(K,A) - ICOS0(K,B)) *MORPH /300 + ICOS0(K,B);
		end;
	end;
end;	\SETUP2

\----------------------------------------------------------------------

proc	DOKEY;		\Process keyboard commands
\Inputs: SPEED, MORPH, HIDSW.
\Outputs: SPEED, MORPH, HIDSW, HILINETYPE.
int	I, LINETBL;
begin
case GETKEY of
  $1B:	[SETVID($03);   exit];				\Esc

  ^H, ^h:
	begin
	HIDSW:= (HIDSW +1) & $0003;
	LINETBL:= [$AA0A, $7F05, $0005, $FF05];
	HIDLINETYPE:= LINETBL(HIDSW);
	end;

  $20:	[SPEED(0):= 0;   SPEED(1):= 0;   SPEED(2):= 0];	\Spacebar

  $0B:	SPEED(0):= SPEED(0) +$40;			\Up arrow
  $0A:	SPEED(0):= SPEED(0) -$40;			\Down arrow

  $15:	SPEED(1):= SPEED(1) -$40;			\ -->
  $08:	SPEED(1):= SPEED(1) +$40;			\ <--

  ^.:	SPEED(2):= SPEED(2) +$40;			\ >
  ^,:	SPEED(2):= SPEED(2) -$40;			\ <

  ^>:	[MORPH:= MORPH +1;   SETUP2];			\ shift >
  ^<:	[MORPH:= MORPH -1;   SETUP2]			\ shift <

other	CHOUT(0, BEL);					\Oops
end;	\DOKEY

\----------------------------------------------------------------------

proc	SETUP;		\Set up the ref. points & verticies of truncated icos
\Inputs: ICOS0.
\Outputs: REF, LMOVE, LBEGIN, LEND, FACEPT0, FACE1, FACE2.
int	DODEC0,		\Array: Initial coordinates of dodecahedron verticies.
			\ (These are the centers of some of the truncated icos-
			\ ahedron faces and are used for hidden line removal)
	TBL,
	I, J, K,
	A, B,		\Indicies for the verticies of an icosahedron
	C, D;		\Indicies for the truncated verticies
def	F= false, T= true;
begin
DODEC0:= RESERVE(3 *INTSIZE);
for I:= 0, 2 do
	DODEC0(I):= RESERVE(20 *INTSIZE);	\20 points

\Reference points:
for J:= 0, 2 do		\For 3 points...
	for K:= 0, 2 do
		REF(K,J):= 0;
REF(0,1):= $4000;
REF(1,2):= $4000;

\From the verticies of the icosahedron, derive the verticies of the dodecahedron:
\This table contains 20 entries corresponding to verticies of a dodecahedron.
\Each entry consists of the surrounding verticies of the icosahedron
TBL:= [	0,	1,	2,
	0,	2,	3,
	0,	3,	4,
	0,	4,	5,
	0,	5,	1,
	10,	5,	1,	\--
	1,	10,	6,
	6,	1,	2,
	2,	6,	7,
	7,	2,	3,
	3,	7,	8,	\--
	8,	3,	4,
	4,	8,	9,
	9,	4,	5,
	5,	9,	10,
	11,	6,	7,	\--
	11,	7,	8,
	11,	8,	9,
	11,	9,	10,
	11,	10,	6 ];

J:= 0;
for D:= 0, 19 do	\For 20 dodecahedron verticies...
	begin
	A:= TBL(J);	J:= J +1;
	B:= TBL(J);	J:= J +1;
	C:= TBL(J);	J:= J +1;
	for K:= 0, 2 do
		\Center point = average X,Y,Z:   D:= (A+B+C) /3
		DODEC0(K,D):= (ICOS0(K,A) + ICOS0(K,B) + ICOS0(K,C)) /3;
	end;

\Table defining which verticies are connected with edges on trunc icos:
\(See external documentation for a diagram which shows the number assigned
\ to each vertex.)
TBL:=	[T, 0, 1,	F, 0, 2,	F, 0, 3,	F, 0, 4,	F, 0, 0,
	F, 0, 5,	F, 0, 6,	F, 0, 7,	F, 0, 8,	F, 0, 9,
	F, 0, 10,	F, 0, 11,	F, 0, 12,	F, 0, 13,	F, 0, 14,
	F, 0, 15,	F, 0, 16,	F, 0, 17,	F, 0, 18,	F, 0, 19,
	F, 0, 20,	F, 0, 21,	F, 0, 22,	F, 0, 23,	F, 0, 24,
	F, 0, 25,	F, 0, 26,	F, 0, 27,	F, 0, 28,	F, 0, 29,
	F, 0, 30,	F, 0, 31,	F, 0, 32,	F, 0, 33,	F, 0, 34,
	F, 0, 35,	F, 0, 36,	F, 0, 37,	F, 0, 38,	F, 0, 39,
	F, 0, 40,	F, 0, 41,	F, 0, 42,	F, 0, 43,	F, 0, 44,
	F, 0, 45,	F, 0, 46,	F, 0, 47,	F, 0, 48,	F, 0, 49,
	F, 0, 50,	F, 0, 51,	F, 0, 52,	F, 0, 53,	F, 0, 54,
	F, 0, 55,	F, 0, 56,	F, 0, 57,	F, 0, 58,	F, 0, 59,
	F, 0, 55,	T, 1, 8,	T, 2, 11,	T, 3, 14,	T, 4, 17,
	T, 5, 19,	T, 6, 21,	T, 7, 24,	T, 9, 25,	T, 10, 28,
	T, 12, 29,	T, 13, 32,	T, 15, 33,	T, 16, 36,	T, 18, 37,
	T, 39, 20,	T, 41, 22,	T, 43, 23,	T, 44, 26,	T, 46, 27,
	T, 47, 30,	T, 49, 31,	T, 50, 34,	T, 52, 35,	T, 53, 38,
	T, 54, 40,	T, 56, 42,	T, 57, 45,	T, 58, 48,	T, 59, 51];

J:= 0;
for I:= 0, 89 do			\For 90 edges...
	begin
	LMOVE(I):= TBL(J);	J:= J +1;
	LBEGIN(I):= TBL(J);	J:= J +1;
	LEND(I):= TBL(J);	J:= J +1;
	end;

\For each line above, there are two corresponding faces. These tables contain
\ indicies into the face point table, FACEPT. (See external documentation for
\ a diagram which shows the number assigned to each face.)
FACE1:=[0,	4,	3,	2,	1,	1,	22,	7,
	21,	21,	5,	25,	25,	13,	24,	2,
	11,	23,	23,	9,	22,	8,	8,	7,
	7,	21,	6,	5,	14,	25,	14,	29,
	12,	24,	11,	28,	10,	23,	10,	27,
	8,	15,	26,	26,	19,	30,	18,	14,
	29,	17,	17,	28,	16,	10,	27,	16,
	15,	19,	18,	17,	16,	4,	4,	3,
	2,	22,	22,	21,	5,	25,	25,	24,
	11,	23,	9,	8,	26,	6,	6,	14,
	29,	12,	28,	10,	27,	15,	19,	18,
	17,	16];

FACE2:=[20,	20,	20,	20,	20,	0,	0,	0,
	0,	4,	4,	4,	3,	3,	3,	24,
	2,	2,	1,	1,	9,	22,	7,	26,
	6,	6,	5,	30,	5,	14,	13,	13,
	13,	12,	12,	11,	11,	10,	9,	9,
	27,	8,	15,	19,	6,	19,	30,	18,
	18,	29,	12,	17,	28,	16,	16,	15,
	31,	31,	31,	31,	31,	0,	3,	2,
	1,	1,	7,	7,	21,	5,	13,	13,
	24,	11,	23,	9,	8,	26,	30,	30,
	14,	29,	12,	28,	10,	27,	15,	19,
	18,	17];

\The verticies of a dodecahedron and of an icosahedron are at the center
\ of the faces of the truncated icosahedron, and are used to remove hidden
\ lines.
for J:= 0, 19 do	\For 20 dodecahedron verticies...
	begin
	for K:= 0, 2 do
		FACEPT0(K,J):= DODEC0(K,J);
	end;

for J:= 0, 11 do	\For 12 icosahedron verticies...
	begin
	for K:= 0, 2 do
		FACEPT0(K,J+20):= ICOS0(K,J);
	end;
end;	\SETUP

\----------------------------------------------------------------------

proc	SETCOLORS;	\Set up the palette registers
int	TBL, J;
def	BLACK=$00, DBLUE=$08, BLUE=$01, WHITE=$07, GREEN=$02, RED=$04, LRED=$24;
begin
\TBL:=	[DBLUE,	WHITE,	GREEN,	RED,
\	WHITE,	0,	0,	WHITE,
\	GREEN,	0,	0,	0,
\	RED,	0,	0,	0];
TBL:=	[BLACK,	LRED,	BLUE,	LRED,
	LRED,	0,	0,	WHITE,
	BLUE,	0,	0,	0,
	LRED,	0,	0,	0];
for J:= 0, $F do SETPALREG(J, TBL(J));
\SETBORDER(DBLUE);
end;	\SETCOLORS

\----------------------------------------------------------------------

begin	\MAIN
SETVID($12);
SETCOLORS;

TEXT(0, "
                                  FULLER

   R. Buckminster Fuller obtained over 2,000 patents, wrote about twenty-
   five books, and was an extremely popular lecturer. In 1947 he designed
   the geodesic dome, a very efficient structure enclosing a large volume
   with a minimum of material. The recently discovered, large carbon
   molecules with geodetic form have been named fullerenes, or buckyballs,
   in his honor.

   Use the arrow keys and the ^"< >^" keys to rotate about the X, Y and Z
   axes. Press the spacebar to stop, the H key to change the hidden lines,
   and the Esc key to exit.

   If the ^"< >^" keys are pressed with the Shift key, they operate as a
   morph control. This changes the buckyball from a truncated icosahedron
   to an icosadodecahedron, and other nifty hedrons that transcend reality.

                                               -- Loren Blaney










   Press any key to start...");
OPENI(0);   OPENO(0);
if CHIN(1) then;

\			Edges	Vert's	Faces
\Icosahedron:		30	12	20
\Dodecahedron:		30	20	12
\Truncated icosahedron:	90	60	32

\Coordinates of an icosahedron (top row = X, middle = Y, bottom = Z):
\These were calculated by another program which used floating point and trig.
\(See external documentation for a diagram which shows the number assigned to
\ each vertex.)
\	     0     1     2     3     4     5     6     7     8     9    10    11
ICOS0:=[[    0,  916,  283, -741, -741,  283,  741, -283, -916, -283,  741,    0],
	[-1024, -458, -458, -458, -458, -458,  458,  458,  458,  458,  458, 1024],
	[    0,    0,  871,  538, -538, -871,  538,  871,    0, -871, -538,    0]];

REF:= RESERVE(3 *INTSIZE);		\3 dimensions
PT0:= RESERVE(3 *INTSIZE);
PT:= RESERVE(3 *INTSIZE);

FACEPT0:= RESERVE(3 *INTSIZE);
FACEPT:= RESERVE(3 *INTSIZE);
for II:= 0, 2 do
	begin
	REF(II):= RESERVE(3 *INTSIZE);		\3 points
	PT0(II):= RESERVE(60 *INTSIZE);		\60 points (verticies)
	PT(II):= RESERVE(60 *INTSIZE);
	FACEPT0(II):= RESERVE(32 *INTSIZE);	\32 points
	FACEPT(II):= RESERVE(32 *INTSIZE);	\32 points
	end;

LMOVE:= RESERVE(90);			\90 lines (edges)
LBEGIN:= RESERVE(90);
LEND:= RESERVE(90);

SPEED:= RESERVE(3 *INTSIZE);

REM0:= RESERVE(3 *INTSIZE);	REM1:= RESERVE(3 *INTSIZE);
REM2:= RESERVE(3 *INTSIZE);	REM3:= RESERVE(3 *INTSIZE);
REM4:= RESERVE(3 *INTSIZE);	REM5:= RESERVE(3 *INTSIZE);

HIDSW:= 0;   HIDLINETYPE:= $AA0A;
SETUP;
MORPH:= 100;
SETUP2;
for II:= 0, 2 do
	begin
	SPEED(II):= -$100;
	REM0(II):= 0;
	REM1(II):= 0;
	REM2(II):= 0;
	REM3(II):= 0;
	REM4(II):= 0;
	REM5(II):= 0;
	end;

ROTATE; 	\PT must be initialized for DRAW
PAGE:= 0;
loop	begin
	VCLEAR(PAGE);		\Erase page
	VDRAW(PAGE);		\Set page to draw on
	DRAW;			\Draw wheel
	ROTATE;			\Rotate wheel
	if CHKKEY then DOKEY;	\Handle keyboard commands
	WAITVB;			\Wait for vertical blank
	VSHOW(PAGE);		\Show page with wheel drawn in new position
	PAGE:= if PAGE = 0 then 1 else 0;	\Flip to other page
	end;			\Loop until Esc key is pressed
end;	\MAIN
